/* acl.c : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
       (might be the wrong person except for offline)

    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <ctype.h>
#include "tomsysdep.h"
#include "prototyp.h"
#include "user.h"
#include "tomtcl.h"


typedef struct acl_user_t
{
	char *name;
	struct acl_user_t *next;
} acl_user_t;

typedef struct acl_zeit_t
{
	UWORD from;
	UWORD to;
	struct acl_zeit_t *next;
} acl_zeit_t;

typedef struct acl_class_t
{
	BOOLEAN           level_oder;
	BOOLEAN           level[ULV_MAX+1];
	acl_user_t        *user;
	acl_zeit_t        *zeiten;
} acl_class_t;
	/* 26 (a..z) + 10 (0..9) + 1 (*) */
#define ANZ_KLASSEN	37

typedef struct acl_t
{
	char *filename;
	acl_class_t *klasse[ANZ_KLASSEN];
	struct acl_t *next;
} acl_t;

static acl_t *acl_anker;
static unsigned int acl_zahl;

static acl_t *acl_get_from_cache(const char *filename);
static void acl_free(acl_t *acl);
static acl_t * acl_load_cache(const char *filename);

static int acl_class(char c);

static int acl_class(char c)
{
	if (!c || !isascii(c))
		return -1;
	c=tolower(c);
	if (islower(c))
	{
		return (c-'a');
	}
	else if (c=='*')
		return 26;
	else if (isdigit(c))
		return (c-'0')+27;
	else
		return -1;
}

/*
 * Im Cache nach einer ACL suchen.
 */
static acl_t *
acl_get_from_cache(const char *filename)
{
	acl_t *l;
	l=acl_anker;
	while (l)
	{
		if (!strcmp(filename,l->filename))
			return l;
		l=l->next;
	}
	return NULL;
}

/*
 * Eine ACL laden und vorne in die Liste einhngen. Pointer darauf
 * zurckliefern.
 */
static acl_t *
acl_load_cache(const char *filename)
{
	char buf[1024];
	char realfname[PATH_MAX];
	acl_t *neu;
	FILE *f;
	int i;

	if (!do_interpret(filename,realfname,sizeof(realfname)))
		return FALSE;

	f=fopen(realfname,"r");
	if (!f)
		return FALSE;

	/* Speicherplatz fr die ACL besorgen, ACL-Namen sichern, 
	 * ACL vorne in die Liste einhngen
	 */
	neu=XMALLOC(sizeof(acl_t));
	for (i=0;i<ANZ_KLASSEN;i++)
		neu->klasse[i]=NULL;
	acl_zahl++;
	neu->filename=XSTRDUP(filename);
	neu->next=acl_anker;
	acl_anker=neu;
	while (fgets(buf,sizeof(buf),f))
	{
		int klasse;
		char kennchar;
		char *start;
		char *para;
		char *p=strpbrk(buf,"\r\n");
		if (p)
			*p=0;
		start=strtrim(buf);
		para=strpbrk(buf,"\t ");
		while (para && isspace((unsigned char) *para))
			*para++=0;

		klasse=acl_class(*start);
		if (klasse==-1)
		{
			continue;
		}

		start=para;
		para=strpbrk(start,"\t ");
		while (para && isspace((unsigned char) *para))
			*para++=0;
		kennchar=tolower(*start);
		/* "Type"-Kommando, oder type "ALL" */

		if (!neu->klasse[klasse])
		{
			neu->klasse[klasse]=XMALLOC(sizeof(acl_class_t));
			neu->klasse[klasse]->level_oder=TRUE;
			for (i=0;i<=ULV_MAX;i++)
				neu->klasse[klasse]->level[i]=FALSE;
			neu->klasse[klasse]->user=NULL;
			neu->klasse[klasse]->zeiten=NULL;
		}
		/* "User"-Kommando: User Uwe Ohse */
		if (kennchar=='u')
		{
			acl_user_t *u;
			char ubuf[1024];
			ub_canonical(para,ubuf);
			u=XMALLOC(sizeof(acl_user_t));
			u->name=XSTRDUP(ubuf);
			u->next=neu->klasse[klasse]->user;
			neu->klasse[klasse]->user=u;
		}
		/* "Zeit"-Kommando:  Zeit 19:30-21:05 */
		else if (kennchar=='z')
		{
			char *minus=strchr(para,'-');
			char *d1,*d2;
			acl_zeit_t *zeit;
			if (!minus)
				continue;
			*minus++=0;
			d1=strchr(para,':');
			if (!d1 && *para)
				continue;
			*d1++=0;
			d2=strchr(minus,':');
			if (!d2 && *minus)
				continue;
			if (d2)
				*d2++=0;
			zeit=XMALLOC(sizeof(acl_zeit_t));
			if (*para)
				zeit->from=60*strtol(para,NULL,10)+strtol(d1,NULL,10);
			else
				zeit->from=0;
			if (*minus)
				zeit->to=60*strtol(minus,NULL,10)+strtol(d2,NULL,10);
			else
				zeit->to=60*24;
			zeit->next=neu->klasse[klasse]->zeiten;
			neu->klasse[klasse]->zeiten=zeit;
		}
		/* Levelbeschrnkung? */
		else if (kennchar=='l')
		{
			/* 
			 * Syntax:
			 * level -2357-
			 * level 0-345-9
			 */
			char *s;
			int last=0;
			s=para;
			while (*s)
			{
				if (isdigit(*s))
				{
					last=*s-'0';
					neu->klasse[klasse]->level[last]=TRUE;
				}
				else if (*s=='-')
				{
					if (isdigit(s[1]))
					{
						int ende=s[1]-'0';
						for (i=last;i<=ende;i++)
							neu->klasse[klasse]->level[i]=TRUE;
						s++; /* ein Zeichen mehr verarbeitet */
					}
					else if (!s[1])
					{
						for (i=last;i<=ULV_MAX;i++)
							neu->klasse[klasse]->level[i]=TRUE;
					}
				}
				else if (*s=='*')
				{
					for (i=last;i<=ULV_MAX;i++)
						neu->klasse[klasse]->level[i]=TRUE;
				}
				else if (*s=='&')
				{
					neu->klasse[klasse]->level_oder=FALSE;
				}
				s++;
			}
		} /* auswahl des ersten Buchstabens */
	} /* while fgets() */
	fclose(f);
	return neu;
}


#define ERROR(str) do{ if (errortext) *errortext=str; } while(0)
BOOLEAN acl_check(const char *filename, const char *username, 
	UWORD userlevel, char typechar, const char **errortext)
{
	char ubuf[512];
	time_t now;
	struct tm *tm;
	UWORD vglzeit;
	acl_t *acl;
	acl_t *this,*last;
	acl_class_t *type;
	int klasse;
	int i;

	klasse=acl_class(typechar);
	if (klasse==-1)
	{
		ERROR("Typ ausserhalb des Wertebereiches a..z");
		return FALSE;
	}
	acl=acl_get_from_cache(filename);
	if (!acl)
	{
		acl=acl_load_cache(filename);
		if (!acl)
		{
			ERROR("kann Zugriffsliste nicht lesen");
			return FALSE;
		}
	}
	ub_canonical(username,ubuf);
	/* 
	 * Zeit auslesen.
	 */
	now=time(NULL);
	tm=localtime(&now);
	vglzeit=tm->tm_hour*60+tm->tm_min;

	/* 
	 * acl im Cache nach vorne ziehen 
	 */
	for (this=acl_anker,last=NULL;this;this=this->next)
	{
		if (this==acl)
		{
			if (this!=acl_anker)
			{
				/* aus der alten Position lschen */
				last->next=this->next;
				/* und vorne eintragen */
				this->next=acl_anker;
				acl_anker=this;
			}
			break;
		}
		last=this;
	}

	/*
	 * Cache verkleinern.
	 */
	if (acl_zahl>5)
	{
		this=acl_anker;
		last=NULL; 
		while (this)
		{
			if (!this->next)
			{
				last->next=NULL;
				acl_free(last);
				break;
			}
			last=this;
			this=this->next;
		}
	}

	/* zuerst die gewnschte Klasse, dann '*' */
	for (i=0;i<2;i++)
	{
		if (i==1)
			type=acl->klasse[acl_class('*')];
		else
			type=acl->klasse[klasse];
		if (!type)
			continue;
		/*
		 * Stimmen die Level?
		 */
		if (type->level_oder)
		{
			if (type->level[userlevel])
				return TRUE;
		}
		else
		{
			if (!type->level[userlevel])
			{
				ERROR("fr ihre Zugriffsklasse nicht gestattet");
				return FALSE;
			}
		}
		/*
		 * Zeitabfrage;
		 */
		if (type->zeiten)
		{
			BOOLEAN ok=FALSE;
			acl_zeit_t *z=type->zeiten;
			while (z)
			{
				if (vglzeit>=z->from && vglzeit<=z->to)
				{
					ok=TRUE;
					break;
				}
				z=z->next;
			}
			if (!ok)
			{
				ERROR("zur Zeit ist der Zugriff nicht gestattet");
				return FALSE;
			}
		}
		/* 
		 * Usernamenbeschrnkung:
		 *    *    -> alle
		 *    !x   -> x nicht
		 *    x    -> x ja.
		 *    
		 */
		if (type->user)
		{
			BOOLEAN ok=FALSE;
			acl_user_t *u=type->user;
			while (u)
			{
				if (!strcmp(u->name,"*"))
					ok=TRUE;
				else if (u->name[0]=='!')
				{
					char *p=u->name+1;
					while (isspace(*p))
						p++;
					if (!strcmp(p,ubuf))
						ok=FALSE;
				}
				else
				{
					if (!strcmp(u->name,ubuf))
						ok=TRUE;
				}
				u=u->next;
			}
			if (!ok)
			{
				ERROR("fr Sie nicht gestattet");
				return FALSE;
			}
			return TRUE;
		}
	}

/* wenn wir hierhin kommen, ist kein ACL-Eintrag fr den Typ
 * vorhanden */
	ERROR("fr Sie nicht gestattet");	
	return FALSE;
}

static void
acl_free(acl_t *acl)
{
	int i;
	acl_class_t *type;
	for (i=0;i<ANZ_KLASSEN;i++)
	{
		acl_user_t *u;
		acl_zeit_t *z;
		type=acl->klasse[i];
		u=type->user;
		while (u)
		{
			acl_user_t *next_u=u->next;
			free(u->name);
			free(u);
			u=next_u;
		}
		z=type->zeiten;
		while (z)
		{
			acl_zeit_t *next_z=z->next;
			free(z);
			z=next_z;
		}

		free(type);
	}
	free(acl);
	acl_zahl--;
}

void
acl_invalidate_cachefile(const char *filename)
{
	acl_t *acl;
	for (acl=acl_anker;acl;acl=acl->next)
	{
		if (!strcmp(filename,acl->filename))
		{
			acl_free(acl);
			return;
		}
	}
	return;
}
	
