/* batch.c : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
*/
#include "tomsysdep.h"

#include <portab.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>

#include "prototyp.h"
#include "brett.h"
#include "modem.h"
#include "terminal.h"
#include "user.h"
#include "tompaths.h"
#include "box_conf.h"
#include "befehle.h" /* fr batch() */

#include "aktdata.h"
#include "log.h"

extern USER user;
extern BRETTINFO *br_ind;
extern BRETT akt_brett;
extern BRETTNR brettnrW;
extern BRETTART postfach_art;

extern BYTE
batchB,
offlineB;

extern WORD temp_levelW;
extern int logout_modus;



/*
	Batch-Verarbeitung

	16.11.92 Frank, erneuert, zeilenweise Abarbeitung
	17.02.93 Frank, eindeutiger Batch fr jeden User
*/
BOOLEAN batch(const char *para)
{
	char file[128];

	sprintf(file,"%s%ld.bat",box_conf.batch_dir,(long)user.nummer);

	if (!strcmp(para,"INST"))
	{
		batch_install(NULL);
		return TRUE;
	}

	if (*para)
	{
		parameter_falsch("BATCH",para);
		return FALSE;
	}

	if (!fileexist(file))
	{
		mbfehler("Kein Batch installiert (siehe 'BATch HILFE')");
		return FALSE;
	}

	set_status_line("Installierter Batch");
	zeige_datei(file);
	mboutput("\r\n >> installierten Batch wirklich starten? (J/N)");
	if (mbdecide("JN"," Ja, dann mal los| Nein, blo nicht\r\n") == 2)
		return FALSE;

	set_status_line("Batch-Beginn:");
	strcpy(akt_anruf.zuletzt,"BAT");

	batch_ausfuehrung(file);

	set_status_line("Batch-Ende:");
	if (!offlineB)
		strcpy(akt_anruf.zuletzt,"BOX");
	else
		logout_modus = 1;

	return TRUE;

} /* batch */



/*
	15.10.92 Uwe, Offline bei timeout-Abfrage beachtet
					  Level temporr auf 9
	08.12.92 Frank, konverterunabhngig
*/
void automatic_batch(const char *filename)
{
	char pfad[128];
	BYTE old_offlineB;
	WORD old_levelW;	/* autobatches laufen mit Userlevel 9 ab. */

	old_offlineB = offlineB;
	offlineB = TRUE;
	old_levelW = temp_levelW;
	temp_levelW = 9;
	sprintf(pfad,"%s%s.bat",box_conf.batch_dir,filename);
	if (!fileexist(pfad))
		sprintf(pfad,"%s%s",box_conf.batch_dir,filename);
	if (fileexist(pfad))
	{
		set_status_line(tprintf("Batch %s beginnt",filename));
		strcpy(akt_anruf.zuletzt,"BAT");

		batch_ausfuehrung(pfad);

		set_status_line(tprintf("Batch %s endet",filename));
		strcpy(akt_anruf.zuletzt,"BOX");
	}
	offlineB = old_offlineB;
	temp_levelW = old_levelW;

} /* automatic_batch */



/*
	16.11.92 Frank, ausgelagert
	01.12.92 Frank, C-like, Lib-unabhngig
	17.02.93 Frank, eindeutiger Batch fr jeden User
	29.12.03 Uwe, Dummyparameter eingebaut.
*/
BOOLEAN batch_install(const char *dummy TOM_ATTRIB_UNUSED)
{
	char file[PATH_MAX];

	set_status_line("Batch-Befehlsliste installieren");

	sprintf(file,"%s%ld.bat",box_conf.batch_dir,(long)user.nummer);

	call_editor(file, FALSE);

	/* leere Batches lschen */
	if (!len_datei(file))
		unlink(file);

	mboutput("\r\n\n");

	return TRUE;

} /* batch_install */



/*
	01.12.92 Frank, fast komplett neu, zeilenweise Abarbeitung
*/
void batch_ausfuehrung(const char *file)
{
	FILE *batch_f;
	char zeile[110];

	batchB = TRUE;

	if ((batch_f = fopen(file,BINREAD)) == NULL)
	{
		mboutput("kein Batch unter diesem Namen vorhanden\r\n");
		akt_anruf.batch_status=TRISTATE_FAILURE;
		batchB = FALSE;
		return;
	}
	akt_anruf.batch_status=TRISTATE_OK; /* wird bei Abbruch zurckgesetzt */

	batchB = TRUE;

	while (tom_fgetline(zeile,batch_f))
	{
		strtrim(zeile);
		if (!offlineB)
			check_timeout();
		if (logout_modus==1)
			break;

		mboutput(prepare_prompt());
		mboutput(zeile);
		mboutput("\r\n");

		write_info_protokoll(INFOPROT_TEXT,"<BATCH>",zeile);

		befehl_ausfuehrung(strupr(zeile));

		if (userbreak())			/* || fehler aus befehl_ausfuehrung() */
		{
			akt_anruf.batch_status=TRISTATE_FAILURE;
			break;
		}
	}
	fclose(batch_f);

	batchB = FALSE;

} /* batch_ausfuehrung */
