/* bb_create_b.c: create a new brett.
    Copyright (C) 1996 Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
*/

#include <portab.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <syslog.h>
#include <assert.h>
#include <errno.h>

#include "config.h"
#include "libquark.h"

#include "brett_db.h"
#include "brett_int.h"
#include "filelock.h"

static BRETTNR
bb_get_free_number /* and check name */ (const char *name)
{
	BRETT eines;
	char ecanon[sizeof(eines.name)*3];
	char ncanon[sizeof(eines.name)*3];
	BRETTNR max;
	char *flag;
	size_t gelesen;
	BRETTNR i;

	bb_canonical(name,ncanon);
	
	/* Wir mssen alle Bretter in der DB durchsuchen, um:
	 * - eine freie Brettnummer zu finden, und
	 * - festzustellen, ob der Brettname nicht mittlerweile vergeben ist.
	 * Es reicht nicht, die Daten des Caches zu lesen.
	 *
	 * Der Cache wird dabei nicht aktualisiert. Die boxinternen Zugriffsflags
	 * (B_NEU, B_READ ...) knnen nicht mit aktualisiert werden - wir haben
	 * hier keinen Zugriff auf die Userdaten. Irgendwann knnte man ein
	 * neues Flag einfhren (B_FLAGS_DIRTY) um der Box zu signalisieren, da
	 * sie die Zugriffsflags im betreffenden Brett aktualisieren mu.
	 */
	max=max_brettnr;
	if (-1==lseek(bb_fh,POS_INDEX(0),SEEK_SET))
	{
		syslog(LOG_ERR,"cannot lseek in brett.dat: %s",strerror(errno));
		return 0;
	}

	flag=XMALLOC(1+max);
	memset(flag,0,(1+max));

	while ((gelesen=read(bb_fh,&eines,sizeof(BRETT)))==sizeof(BRETT))
	{
		if (eines.geloescht || eines.nr<=0)
			continue;
		if (eines.nr > max)
		{
			/* Hmm ... jetzt mte man eigentlich die internen Daten des 
			 * Prozesses aktualisieren (br_ind und max_brettnr). Das ist
			 * nicht so frchterlich viel Aufwand, scheitert aber an
			 * den obigen Erwgungen bezglich der Cacheaktualisierung.
			 */
			BRETTNR old=max;
			max=eines.nr+100;
			flag=XREALLOC(flag,1+max);
			while (old+1<max+1)
				flag[++old]=0;
		}
		flag[eines.nr]=1;
		bb_canonical(eines.name,ecanon);
		if (0==strcmp(ecanon,ncanon))
		{
			/* Namenskonflikt: Der Name existiert in irgendeiner Schreibweise
			 * schon */
			syslog(LOG_WARNING,"trying to create existing BRETT");
			fprintf(stderr,"trying to create existing BRETT");
			free(flag);
			return 0;
		}
	}
	if (gelesen!=0 && gelesen!=sizeof(BRETT))
	{
		/* this must not happen */
		syslog(LOG_WARNING,"Cannot read BRETT in bb_create_brett: %m");
		fprintf(stderr,"bb_create_brett: read gescheitert\r\n");
		return 0;
	}

	/*
	 * Eine freie Brettnummer suchen.
	 */
 	srand(time(NULL));
 	/* zuerst in der Liste der Brettnummern, fr die in br_ind ohnehin
 	 * schon Platz vergeben ist. Das spart RAM ...
 	 * Der Zufall soll vermeiden da die Box die Nummern gelschter
 	 * Bretter sofort wieder neu vergibt. 100%ig ist das nicht, aber
 	 * es lindert das Problem (die Tauschinstallationen haben die alten
 	 * Nummern noch eine ganze Zeit lang drin) deutlich.
 	 */
 	i=rand();
 	i*=max_brettnr;
 	i/=RAND_MAX;
 	if (i==0)
 		i++;
 	while (i<max_brettnr)
 	{
		if (!flag[i])
		{
			free(flag);
			return i;
		}
		i++;
 	}
 	free(flag);
 	/* Pech. In dem Fall nehmen wir die hchste Brettnummer (aus der Datei!)
 	 * plus 1. Allerdings mu auch br_ind realloziert werden. */
 	{
 		BRETTNR old=max_brettnr;
		max_brettnr=max+10; /* 10 Pltze frei lassen ... */
		br_ind=(BRETTINFO *) XREALLOC(br_ind,sizeof(BRETTINFO)*(max_brettnr+1));
		while (old<max_brettnr)
		{
			memset(&br_ind[old],0,sizeof(BRETTINFO));
			old++; 
		}
 	}
 	return max+1;
}

/*
BB_FUNCTION(Funktion,BOOLEAN,bb_create_brett, [[(BRETT *@var{b})]],[[
    Legt ein neues Brett mit den Daten in @var{b} an und gibt dem Brett 
    seine Nummer (die vor Rckkehr in @var{b} eingetragen wird).

    Es werden maximal 10 Versuche gemacht, den Verwaltungsblock der
    Brettdatenbank zu locken. Smtliche Bretter werden noch einmal
    gelesen, um erkennen zu knnen, ob mittlerweile neu angelegte
    Bretter mit den neuen Brett kollidieren. Auf das Locking der
    einzelnen Bretter wird jedoch verzichtet (damit bleibt eine
    Racecondition brig - aber die sichere Variante frit zu viel
    Zeit).

    Liefert TRUE zurck, falls das Brett angelegt werden konnte, sonst
    FALSE.
]])
*/
BOOLEAN bb_create_brett(BRETT *b)
{
#if 0
	syslog(LOG_WARNING,"brett %s soll angelegt werden",b->name;
	fprintf(stderr,"Bretter knnen im NNTP-Client-Betrieb nicht angelegt werden\r\n");
	return FALSE;
#endif
	BOOLEAN ok=FALSE;
	int tries=10;
	char buf[sizeof(b->name)*2+1];
	size_t ende;
	BRETTNR high;

#ifdef CAN_LOCK_FILES
	/* Zuerst mal einen Lock auf die Verwaltungsdaten der Brettdatenbank
	 * setzen. */
	while (tries)
	{
		if (0!=fd_write_lock(bb_fh,0,SEEK_SET,BB_OFFSET_FIRST_BRETT))
		{
			if (errno == EDEADLK)
			{
				syslog(LOG_WARNING,"Deadlock beim in bb_create_brett");
				return FALSE;
			}
			if (errno == EINTR)
			{
				syslog(LOG_WARNING,"Interrupt beim in bb_create_brett");
				return FALSE;
			}
			fprintf(stderr,"bb_create_brett: lock detected. Trying again\r\n");
			sleep(1);
		}
		else
		{
			ok=TRUE;
			break;
		}
		tries--;
	}

	if (!ok)
	{
		syslog(LOG_WARNING,"Cannot lock in bb_create_brett");
		fprintf(stderr,"bb_create_brett: Kann neues Brett nicht anlegen\r\n");
		return FALSE;
	}
#endif

	b->nr=bb_get_free_number(b->name);
	if (!b->nr)
		return FALSE;

	/* Nun an das Ende der brett.dat springen */
	ende=lseek(bb_fh,0,SEEK_END);
	if (((size_t)-1)==ende)
	{
		syslog(LOG_WARNING,"Cannot seek in bb_create_brett: %m");
		fprintf(stderr,"bb_create_brett: seek gescheitert\r\n");
		ok=FALSE;
		goto unlock;
	}
	high=(ende-BB_OFFSET_FIRST_BRETT)/sizeof(BRETT);
#define bnr b->nr

	br_ind[bnr].brett = (BRETT *) XMALLOC(sizeof(BRETT));
	memcpy(br_ind[bnr].brett,b,sizeof(BRETT));
	br_ind[bnr].brett->index=high;
	br_ind[bnr].brett->nr=bnr;
	br_ind[bnr].brett->created = time(NULL);

	if (sizeof(BRETT)!=write(bb_fh,br_ind[bnr].brett,sizeof(BRETT)))
	{
		syslog(LOG_WARNING,"write to brett.dat failed: %m");
		fprintf(stderr,"bb_create_brett: kann neues Brett nicht schreiben!\r\n");
		ok=FALSE;
		free(br_ind[bnr].brett);
		br_ind[bnr].brett=NULL;
		goto unlock;
	}
	bb_canonical(b->name,buf);
	br_ind[bnr].canon=XSTRDUP(buf);

	if (!b->netzstatus)
		br_ind[bnr].brett->flags |= B_LOKAL;

	/* Lesen und Schreiben kann der, der das Brett anlegt, in jedem Fall. */
	br_ind[bnr].brett->flags |= B_LOCKED|B_READ|B_WRITE;

	bb_lock_brett(bnr,BB_WRITELOCK,FALSE);
	memcpy(b,br_ind[bnr].brett,sizeof(BRETT));

unlock:
#ifdef CAN_LOCK_FILES
	fd_un_lock(bb_fh,0,SEEK_SET,BB_OFFSET_FIRST_BRETT);
#endif
	if (ok)
		brettanzahl++;

	return ok;

} /* bb_create_brett */

