/* bb_init.c: load brett database.
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
*/
#include <portab.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <syslog.h>
#include <assert.h>
#include <errno.h>
#include <signal.h>

#include "config.h"
#include "libquark.h"

#include "brett_db.h"
#include "brett_int.h"
#include "filelock.h"

#ifndef O_BINARY
#define O_BINARY 0
#endif

int bb_fh=-1;

long brettanzahl = 0;

BRETTINFO   *br_ind = NULL;
long max_brettnr = 0;
char *bb_binaer_db_dir=NULL;
char *bb_text_db_dir=NULL;
char *bb_sys_dir=NULL;

bb_rename_t *bb_rename_anker=NULL;

/* Ein Pseudobrett fr das Postfach des Benutzers. Besser als das berall
 * gesondert abzufangen ... */
static BRETT postfach_pseudo;
static BOOLEAN bb_read_xbrett_dat(const char *path);
static BOOLEAN read_rename_dat(const char *path);




/*
BB_FUNCTION(Funktion,BOOLEAN,bb_init,
[[(const char *@var{dir}, const char *@var{text_db_dir}, 
    const char *@var{binaer_db_dir})]],[[
    Initialisiert die Brettdatenbank in @var{dir}. Dabei werden alle
    Brettdaten gelesen und in den internen Cache geschrieben.
	@var{binaer_db_dir} ist die Basis der Verzeichnishierarchie, in
	der die Zugriffsdaten und die Inhalte der Binrbretter untergebracht
	werden, @var{text_db_dir} enthlt die Zugriffslisten der Textbretter.

    Liefert FALSE bei Mierfolg, TRUE bei Erfolg, und kehrt bei
    Speichermangel nicht immer zurck.
]])
*/
#define BBUF_SIZE (16*1024)
BOOLEAN bb_init(const char *path, const char *db_text_dir, 
				const char *db_binaer_dir)
{
	char datei[PATH_MAX];
	long groesse;
	struct bb_verwalt v;
	long anzahl;
	BRETT *b;
	int i;
	BRETT *brett= NULL;

	if (path)
		strcpy(datei,path);
	else
		*datei = 0;
	strcat(datei,"brett.dat");
	bb_binaer_db_dir=XSTRDUP(db_binaer_dir);
	bb_text_db_dir=XSTRDUP(db_text_dir);
	bb_sys_dir=XSTRDUP(path);

	bb_fh=open(datei,O_RDWR | O_BINARY);
	if (bb_fh==-1)
	{
		syslog(LOG_ERR,"Fehler beim Oeffnen von brett.dat: %m");
		return FALSE;
	}

	groesse = lseek(bb_fh,0,SEEK_END);
	if (groesse!=-1)
		if (-1==lseek(bb_fh,0,SEEK_SET))
			groesse=-1;
	if (groesse < BB_OFFSET_FIRST_BRETT)
	{
		syslog(LOG_ERR,"korrupte Brettdatenbank");
		CLOSE_IT;
		return FALSE;
	}

#ifdef CAN_LOCK_FILES
	if (0!=fd_readw_lock(bb_fh,0,SEEK_SET,sizeof(v)))
	{
		syslog(LOG_ERR,"Kann brettdatenbank nicht locken.");
		CLOSE_IT;
		return FALSE;
	}
#endif
	if (sizeof(v)!=read(bb_fh,&v,sizeof(v)))
	{
		syslog(LOG_ERR,"Kann Brettdatensatzlaenge nicht lesen");
		CLOSE_IT;
		return FALSE;
	}
#ifdef CAN_LOCK_FILES
	fd_un_lock(bb_fh,0,SEEK_SET,sizeof(v));
#endif
	/* Gre der Datenstze prfen. Der Aufwand ist wahrscheinlich 
	 * berflssig, weil hoffentlich alle Compiler fr ein Zielsystem
	 * die gleichen Alignmentregeln benutzen.
	 * Aber in Anbetracht der Tatsache, da der Crosscompiler fr den
	 * Atari die Brettstruktur 254 statt 256 Bytes gro macht, schadet
	 * ein wenig Paranioa nicht.
	 */
	if (v.dat_size==0)
	{
		/* Das dient dem Update alter Versionen */
		v.dat_size=sizeof(BRETT);
		lseek(bb_fh,0,SEEK_SET);
#ifdef CAN_LOCK_FILES
		if (0!=fd_writew_lock(bb_fh,0,SEEK_SET,sizeof(v)))
		{
			syslog(LOG_ERR,"Kann brett.dat nicht locken: %m");
			CLOSE_IT;
			return FALSE;
		}
#endif
		if (sizeof(v)!=write(bb_fh,&v,sizeof(v)))
		{
			syslog(LOG_ERR,"Fehler beim Schreiben auf brett.dat: %m");
			CLOSE_IT;
			return FALSE;
		}
#ifdef CAN_LOCK_FILES
		fd_un_lock(bb_fh,0,SEEK_SET,sizeof(v));
#endif
	}
	if (v.dat_size!=sizeof(BRETT))
	{
		syslog(LOG_ERR,"Brettdatensatzlaenge %ld != vorheriger %ld",
			(long) sizeof(BRETT), v.dat_size);
		CLOSE_IT;
		return FALSE;
	}

	brettanzahl = (groesse-BB_OFFSET_FIRST_BRETT) / sizeof(BRETT);
	max_brettnr=0;
	if (brettanzahl)
	{
		brett = (BRETT *) XCALLOC(brettanzahl,sizeof(BRETT));

		lseek(bb_fh,POS_INDEX(0),SEEK_SET);
		/* jetzt mit groem Puffer am Stck lesen */
		{
			size_t to_read;
			size_t got;
			to_read=sizeof(BRETT) * brettanzahl;
#ifdef CAN_LOCK_FILES
			if (0!=fd_readw_lock(bb_fh,0,SEEK_SET,to_read))
			{
				syslog(LOG_ERR,"Kann brett.dat nicht locken: %m");
				CLOSE_IT;
				return FALSE;
			}
#endif
			got=xread(bb_fh,(void *)brett,to_read);
			if (got==(size_t)-1)
			{
				syslog(LOG_ERR,"Fehler beim Lesen von brett.dat: %m");
				CLOSE_IT;
				return FALSE;
			}
			if (got!=to_read) {
				syslog(LOG_ERR,"Fehler beim Lesen von brett.dat: Datei zu kurz");
				syslog(LOG_ERR,"wollte %lu, bekam %lu Bytes",
					   (ULONG) to_read, (ULONG) got);
				CLOSE_IT;
				return FALSE;
			}
#ifdef CAN_LOCK_FILES
			fd_un_lock(bb_fh,0,SEEK_SET,to_read);
#endif
		}

		for (b=brett,i=0;i<brettanzahl;b++,i++)
		{
			if (b->nr>=max_brettnr && !b->geloescht)
				max_brettnr=b->nr;
		}
	}
	/* 100 mehr */
	max_brettnr += 100;

	br_ind = (BRETTINFO *) calloc(max_brettnr,sizeof(BRETTINFO));
	if (!br_ind)
	{
		free(brett);
		CLOSE_IT;
		return FALSE;
	}

	/* brettinterne Eintrge initialisieren */
	anzahl = brettanzahl;
	b = brett;
	i = 0;
#define ONE \
		if (!anzahl--) break;  \
		if (b->nr >0 && !b->geloescht) { \
			char buf[sizeof(b->name)*2+1]; \
			b->index=i; \
			b->flags &= B_MASK_DATABASE; \
			b->aliasname = NULL; \
			b->sysopinfo = NULL; \
			b->lokalinfo = NULL; \
			b->userinfo = NULL; \
			if (!b->netzstatus) b->flags |= B_LOKAL; \
			if (br_ind[b->nr].brett) \
				{ fprintf(stderr,"bb_init.c: duplicate bnr %ld (%s,%s)\n",\
				  (long) b->nr,br_ind[b->nr].brett->name,b->name); \
				  free(br_ind[b->nr].canon); free(br_ind[b->nr].brett); } \
			br_ind[b->nr].brett = XMALLOC(sizeof(BRETT)); \
			memcpy(br_ind[b->nr].brett,b,sizeof(BRETT)); \
			bb_canonical(b->name,buf); \
			br_ind[b->nr].canon=XSTRDUP(buf); \
		} \
		b++; \
		i++; 

#ifdef TOM_DEMO
	do 
	{
        ONE; /* 1 */
        ONE; /* 2 */
        ONE; /* 3 */
        ONE; /* 4 */
        ONE; /* 5 */
        ONE; /* 6 */
        ONE; /* 7 */
        ONE; /* 8 */
        ONE; /* 9 */
        ONE; /* 10 */
        ONE; /* 11 */
        ONE; /* 12 */
        ONE; /* 13 */
        ONE; /* 14 */
        ONE; /* 15 */
        ONE; /* 16 */
        ONE; /* 17 */
        ONE; /* 18 */
        ONE; /* 19 */
        ONE; /* 20 */
	} while(0);
#else
	while (TRUE)
	{
		ONE;
	}
#endif
	memset(&postfach_pseudo,0,sizeof(postfach_pseudo));
	strcpy(postfach_pseudo.name,"Post/Programmfach");
	postfach_pseudo.flags=B_READ|B_WRITE;
	br_ind[0].canon = XSTRDUP("post/programmfach");  /* just for safety .. */
	br_ind[0].brett = &postfach_pseudo;

	if (brett)
		free(brett);

	/* Und nun noch die xbrett.dat lesen */
	if (!bb_read_xbrett_dat(path))
		return FALSE;
	if (!read_rename_dat(path))
		return FALSE;
	return TRUE;
} /* bb_init */

static BOOLEAN
read_rename_dat(const char *path)
{
	char datei[PATH_MAX];
	char buf[2100];
	int tries=5;
	FILE *f;
	if (path)
		strcpy(datei,path);
	else
		*datei = 0;
	strcat(datei,"rename.dat");

	f=fopen(datei,"rb");
	if (!f)	{
		if (errno==ENOENT)
			return TRUE;
		return FALSE;
	}
	/* Nicht warten! */
#ifdef CAN_LOCK_FILES
	while (0!=fd_read_lock(fileno(f),0,SEEK_SET,1))
	{
		if (--tries==0)
		{
			fclose(f);
			fprintf(stderr,"xbrett.dat is locked, giving up\r\n");
			return FALSE;
		}
		fprintf(stderr,"xbrett.dat is locked, sleeping 5 seconds.\r\n");
		sleep(5);
	}
#endif
	while (fgets(buf,sizeof(buf),f)) {
		bb_rename_t *akt,*n;
		char *at1,*at2;
		char cfrom[BDAT_GRP_NAME_LEN];
		char cto[BDAT_GRP_NAME_LEN];
		at1=strchr(buf,'@');
		if (!at1)
			continue;
		*at1++=0;
		at2=strchr(at1,'@');
		if (!at2)
			continue;
		*at2++=0;
		bb_canonical(at1,cfrom);
		bb_canonical(at2,cto);
		for (akt=bb_rename_anker;akt;akt=akt->next) {
			if (!strcmp(akt->old_name,cfrom)) 
				break;
		}
		if (akt) {
			akt->when=strtol(buf,NULL,10);
			free(akt->new_name);
			akt->new_name=XSTRDUP(cto);
			continue;
		}
		n=XMALLOC(sizeof(bb_rename_t));
		n->old_name=XSTRDUP(cfrom);
		n->new_name=XSTRDUP(cto);
		n->when=strtol(buf,NULL,10);
		n->next=bb_rename_anker;
		bb_rename_anker=n;
	}
	return TRUE;
}


static BOOLEAN
bb_read_xbrett_dat(const char *path)
{
	char rbuf[1024];
	char datei[PATH_MAX];
	FILE *f;
	BRETTNR bnr;
	int tries=3;
	BRETT *b;

	if (path)
		strcpy(datei,path);
	else
		*datei = 0;
	strcat(datei,"xbrett.dat");

	f=fopen(datei,"rb");

	if (!f)
	{
		struct stat st;
		if (0==stat(datei,&st))
		{
			syslog(LOG_ERR,"Fehler beim Oeffnen von xbrett.dat: %m");
			return FALSE;
		}
		f=fopen(datei,"w+b");
		if (!f)
		{
			syslog(LOG_ERR,"Fehler beim Anlegen von xbrett.dat: %m");
			return FALSE;
		}
		fclose(f);
		return TRUE; /* Wir *haben* soviel gelesen wie da war ... */
	}

	/* Nicht warten! */
#ifdef CAN_LOCK_FILES
	while (0!=fd_read_lock(fileno(f),0,SEEK_SET,1))
	{
		if (--tries==0)
		{
			fclose(f);
			fprintf(stderr,"xbrett.dat is locked, giving up\r\n");
			return FALSE;
		}
		fprintf(stderr,"xbrett.dat is locked, sleeping 5 seconds.\r\n");
		sleep(5);
	}
#endif

	b=NULL;	
	while (fgets(rbuf,sizeof(rbuf),f))
	{
		char *p=strchr(rbuf,'\n');
		if (p)
		{
			*p=0;
			if (p>rbuf)
			{
				p--;
				if (*p=='\r')
					*p=0;
			}
		}
		if (!b && *rbuf!='G')
			continue;
		switch(*rbuf)
		{
		case 'G':
			bnr=strtol(rbuf+1,NULL,10);
			if (bnr<0 || bnr>=max_brettnr)
			{
				bnr=0;
				b=NULL;
			}
			else
			{
				b=br_ind[bnr].brett;
				if (b)
				{
					/* Neuere Versionen der Texte hngen immer hinten */
					free(b->userinfo);
					free(b->sysopinfo);
					free(b->aliasname);
					free(b->lokalinfo);
					b->userinfo=NULL;
					b->lokalinfo=NULL;
					b->sysopinfo=NULL;
					b->aliasname=NULL;
				}
				else
					bnr=0; /* gelscht? */
			}
			break;
		case 'U': 
			if (b->userinfo)
				free(b->userinfo);
			b->userinfo=XSTRDUP(rbuf+1);
			break;
		case 'A':
			if (b->aliasname)
				free(b->aliasname);
			b->aliasname=XSTRDUP(rbuf+1);
			break;
		case 'S': 
			if (b->sysopinfo)
				free(b->sysopinfo);
			b->sysopinfo=XSTRDUP(rbuf+1);
			break;
		case ':':
			if (!b->lokalinfo)
			{
				b->lokalinfo=concat(rbuf+1,"\r\n",NULL);
				XNULLCHECK(b->lokalinfo);
			}
			else
			{
				char *old=b->lokalinfo;
				b->lokalinfo=concat(
					b->lokalinfo,rbuf+1,"\r\n",NULL);
				XNULLCHECK(b->lokalinfo);
				if (old)
					free(old);
			}
			break;
		/* Kompatibilitt zu alten Versionen */
		case 'E':
			if (!b->netzaenderung)
				mtzeit_to_time_t(rbuf+1,&b->netzaenderung);
			break;
		case 'F': 
			if (!b->followup_to_default)
			{
				BRETTNR fup=brettnr(rbuf+1);
				if (fup)
					b->followup_to_default=fup;
			}
			break;
		}
	} /* while fgets */
	fclose(f);
	return TRUE;
}
