/* bb_write.c: write a brett.
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
*/

#include <portab.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <syslog.h>
#include <assert.h>
#include <errno.h>


#include "config.h"
#include "libquark.h"


#include "brett_db.h"
#include "brett_int.h"
#include "filelock.h"

/*
BB_FUNCTION(Funktion,BOOLEAN,bb_write_brett,
[[(BRETTNR bnr,BOOLEAN @var{unlock})]],[[
    Schreibt die Daten des Brettes @var{bnr} aus dem Cache in die 
	Datenbank. Ein Lock wird freigegeben, wenn @var{unlock} TRUE
    ist.
	Ist @var{changed_xinfo} TRUE werden auch die Infotexte in die
	xbrett.dat geschrieben (aus Effizienzgrnden wird sonst darauf
	verzichten).

	Es wird vorrausgesetzt, da nur ein Proze, der die Schreibberechtigung 
	angefordert und erhalten hat, diese Funktion aufruft.

    Liefert TRUE bei Erfolg, sonst FALSE.
]])
*/
BOOLEAN 
bb_write_brett(BRETTNR bnr,BOOLEAN unlock, BOOLEAN changed_xinfo)
{
	ULONG old_flags;
	BRETT *b;
	if (!bnr)
		return TRUE;
	CHECK_BRETTNR(bnr,return FALSE);

	if (!(br_ind[bnr].brett->flags & B_LOCKED))
		return FALSE;

	if (-1==lseek(bb_fh,POS_NR(bnr),SEEK_SET))
	{
		syslog(LOG_ERR,"cannot seek pos of group %ld: %m", 
			(ULONG) bnr);
		CLOSE_IT;
		return FALSE;
	}
	b=br_ind[bnr].brett;

	old_flags=b->flags;
	b->flags &= (B_MASK_DATABASE);

	if (xwrite(bb_fh,b,sizeof(BRETT))==-1)
	{
		b->flags=old_flags;
		syslog(LOG_ERR,"Fehler beim Schreiben in brett.dat: %m");
		CLOSE_IT;
		return FALSE;
	}
	b->flags=old_flags;

	if (unlock)
		bb_unlock_brett(bnr);

	/* Wenn sich die Textinformationen unterscheiden */
	if (changed_xinfo)
	{
		char xname[PATH_MAX];
		FILE *f;
		sprintf(xname,"%sxbrett.dat",bb_sys_dir);
		f=fopen(xname,"ab");
		if (!f)
		{
			syslog(LOG_ERR,"kann %s nicht ffnen: %m",xname);
			CLOSE_IT;
			return FALSE;
		}
#ifdef CAN_LOCK_FILES
		fd_writew_lock(fileno(f),0,SEEK_SET,1);
#endif

		fprintf(f,"G%lu\n",(ULONG) b->nr);
		if (b->sysopinfo)
			fprintf(f,"S%s\n",b->sysopinfo);
		if (b->aliasname)
			fprintf(f,"A%s\n",b->aliasname);
		if (b->userinfo)
			fprintf(f,"U%s\n",b->userinfo);
		if (b->lokalinfo)
		{
			char *s;
			s=b->lokalinfo;
			while (s && *s)
			{
				char *cr_p;
				char *next_p=NULL;
				char old_char=0; /* only to keep gcc quiet */
				cr_p=strpbrk(s,"\r\n");
				if (cr_p)
				{
					old_char=*cr_p;
					next_p=cr_p;
					*next_p++='\0';
					if (*next_p=='\n')
						next_p++;
				}
				fprintf(f,":%s\n",s);
				if (cr_p)
					*cr_p=old_char;
				s=next_p;
			}
		}
		fclose(f);
	}

	return TRUE;

} /* bb_write_brett */

