/*
 * befehle/inst_vt.c: Farbeinstellung, Bildschirmzeilen.
 * Sollte eventuell ganz auf Farbeinstellungen reduziert werden, fr
 * den Rest gibt es den Usereditor.
 */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
       (might be the wrong person except for offline)

    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/


#include "tomsysdep.h"

#include <portab.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "prototyp.h"
#include "modem.h"
#include "io.h"
#include "terminal.h"
#include "user.h"
#include "msg.h"
#include "befehle.h"

extern USER user;

extern int col_ix[];

extern WORD
anz_lineW,
vtmodeW;

extern int user_in_fd;
extern int user_out_fd;

static void 
inst_vt_change_colors(void)
{
	int i;
	char dummyS[80];
	char zeile[80];
	char vorgabe[10];
	set_status_line("Farben ndern...");
	mbprintf("%s%s ",vt_pos_S(4,1),vt_S(DEL_SCR_REST));
	for (i = 0; i < NUMBER_COLORS; i++)
		mbprintf("%s  %d  %s",vt_S(col_ix[i]),i+1,vt_S(ATTR_OFF));
	mboutput("\r\n\n");
	for (i = 0; i < NUMBER_COLORS; i++)
	{
		int answer;

		sprintf(zeile," Schriftfarbe     #%d %s(momentan)%s: ",
				i+1,vt_S(col_ix[i]),vt_S(ATTR_OFF));
		if (col_ix[i]<FIRST_COLOR || col_ix[i]>LAST_COLOR)
			sprintf(vorgabe,"%d",i);
		else
			sprintf(vorgabe,"%d",col_ix[i]-FIRST_COLOR+1);
		do 
		{
			mbquestion(1,2,TRUE,zeile,vorgabe,dummyS);
			answer = atoi(dummyS);
			if (answer>0 && answer<NUMBER_COLORS+1)
				col_ix[i] = answer+FIRST_COLOR-1;
		} while (answer<0 || answer>8);
	}
}

static void
inst_vt_resize(int *zeilen, int *spalten)
{
	int ret;
	int mode=RESIZE_MODE_NORMAL;

	ret=io_resize_terminal(user_in_fd,user_out_fd,RESIZE_MODE_INQUIRE);
	if (ret==RESIZE_ANS_NOT_TERM)
	{
		/* das heit, es geht nicht. */
		mboutput("Automatische Terminalgrenerkennung unmglich: kein Terminal\r\n");
		return;
	}
	else if (ret==RESIZE_ANS_NOT_CAP)
	{
		mboutput(
"Laut Terminalbeschreibung kann die Terminalgre nicht erfragt werden.\r\n"
"VT100/Ansi/Linux-Console-Werte versuchen? (J/N)\r\n");
		ret = mbdecide("JN"," Ja\r\n\n| Nein\r\n\n");
		if (ret==2)
			return;
		mode=RESIZE_MODE_FORCE_VT;
	}
	ret=io_resize_terminal(user_in_fd,user_out_fd,mode);
	if (ret!=RESIZE_ANS_OK)
	{
		/* das heit, es geht nicht. */
		mboutput("Terminalgrenerkennung fehlgeschlagen\r\n");
		return;
	}
	io_get_term_size(user_in_fd,spalten,zeilen);

	mbprintf("Terminalgre: %d Zeilen zu %d Spalten\r\n",*zeilen,*spalten);
}

static void
inst_vt_change_size(int *zeilen, int *spalten)
{
	char vorgabe[10];
	char dummyS[80];
	int answer;
	set_status_line("Terminalgre einstellen ...");
	mbprintf("%s%s",vt_pos_S(4,1),vt_S(DEL_SCR_REST));

	sprintf(vorgabe,"%d",*zeilen);
	mbquestion(2,2,TRUE," Anzahl der Bildschirmzeilen: ",
			   vorgabe,dummyS);
	answer = atoi(dummyS);
	*zeilen = Max(answer,10);

	sprintf(vorgabe,"%d",*spalten);
	mbquestion(2,2,TRUE," Anzahl der Bildschirmspalten: ",
			   vorgabe,dummyS);
	answer = atoi(dummyS);
	*spalten = Max(answer,10);
	
	set_scroll_area(1);
}

static void
inst_vt_defaults(int *zeilen, int *spalten)
{
	*zeilen=24;
	*spalten=80;
	mbprintf("%s%s",vt_pos_S(4,1),vt_S(DEL_SCR_REST));
	set_scroll_area(1);
	mboutput(" Standardeinstellung gesetzt\r\n\n");
	sleep(1);
}

BOOLEAN install_vt(const char *para)
{
	int i;
	int zeilen=anz_lineW;
	int spalten=user.bs_spalten;

	if (0==stricmp(para,"farben")
		|| 0==stricmp(para,"farbe")
		|| 0==stricmp(para,"color")
		|| 0==stricmp(para,"colors"))
		inst_vt_change_colors();
	else if (0==stricmp(para,"size"))
		inst_vt_change_size(&zeilen,&spalten);
	else if (0==stricmp(para,"resize"))
		inst_vt_resize(&zeilen,&spalten);
	else if (0==stricmp(para,"defaults"))
		inst_vt_defaults(&zeilen,&spalten);
	else
	{
		if (vtmodeW != T_VT100)
		{
			mboutput(" Terminaleinstellungen nur fr VT-100 Emulation mglich\r\n");
			return FALSE;
		}

		do
		{
			set_status_line("VT-100 Einstellungen ndern");
			mboutput("\r\n");
			mbprintf(" %s0%s - Zurck ins Hauptprogramm\r\n",vt_S(ATTR_INVERS),vt_S(ATTR_OFF));
			mboutput("\r\n");
			mbprintf(" %s1%s - Farben ndern\r\n",vt_S(ATTR_INVERS),vt_S(ATTR_OFF));
			mboutput("\r\n");
			mbprintf(" %s2%s - Terminalgre einstellen\r\n",vt_S(ATTR_INVERS),vt_S(ATTR_OFF));
			mboutput("\r\n");
			mbprintf(" %s3%s - Terminalgre automatisch feststellen\r\n",vt_S(ATTR_INVERS),vt_S(ATTR_OFF));
			mboutput("\r\n");
			mbprintf(" %s4%s - Standard-Werte zurckholen\r\n",vt_S(ATTR_INVERS),vt_S(ATTR_OFF));
			mboutput("\r\n");
			mboutput(vt_S(STORE_CUR));
			mbprintf("%s -> Ihre Wahl : ",vt_pos_S(anz_lineW,1));
			i = mbdecide("01234",
						 " Abbruch| Terminalgre/manuell| "
						 "Terminalgre/automatisch| Standardwerte");
			mboutput(vt_S(RESTORE_CUR));
			mboutput("\r\n");
			mboutput("\r\n");
			i--;
			switch(i)
			{
			case 1:
				inst_vt_change_colors();
				break;
				
			case 2:
				inst_vt_change_size(&zeilen,&spalten);
				break;
				
			case 3:
				inst_vt_resize(&zeilen,&spalten);
				break;

			case 4:
				inst_vt_defaults(&zeilen,&spalten);
				break;
			}
			mbprintf("%s%s",vt_pos_S(4,1),vt_S(DEL_SCR_REST));
		}
		while (i);
	}

	mboutput(vt_S(STORE_CUR));
	mboutput(vt_pos_S(anz_lineW,1));
	mboutput(" nderungen dauerhaft speichern? (J/N) -> ");
	i = mbdecide("JN"," Ja, speichern| Nein, Abbruch");
	mboutput(vt_S(RESTORE_CUR));
	mboutput("\r\n");
	mboutput("\r\n");
	i--;
	if (!i)
	{
		int n;
		for (n = 0; n < NUMBER_COLORS; n++)
			user.colatt[n].fg_col = col_ix[n];

		user.bs_zeilen = zeilen;
		user.bs_spalten = spalten;
		anz_lineW = zeilen;
		ub_write_user(&user,FALSE);

		mboutput(" VT-100 nderungen aktiv und gespeichert\r\n\n");
	}
	else
	{
		anz_lineW = zeilen;
		mboutput(" VT-100 nderungen aktiv, aber nicht gespeichert\r\n\n");
	}

	sync_terminal();

	return TRUE;
}
