/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/

#include "tomsysdep.h"

#include <portab.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <limits.h>
#include <ctype.h>
#include <unistd.h>

#ifdef OFFLINE2
#include "offline/inc/version.h"
#include "offline/inc/mdb_global.h"
#include "offline/inc/message.h"
#else
#include "offline/version.h"
#include "offline/mdb_global.h"
#include "offline/message.h"
#endif
#include "prototyp.h"
#include "brett.h"
#include "user.h"
#include "modem.h"
#include "bin_msg.h"
#include "libquark.h"
#include "msg.h"
#include "befehle.h"
#include "tompaths.h"
#include "box_conf.h"
#include "libquark.h"
#include "routing.h"

extern USER user;
extern BRETT akt_brett;
extern BRETTNR brettnrW;
extern BRETTINFO *br_ind;

static BOOLEAN kopieren_binaer (const char *para);
static BOOLEAN kopieren_text (const char *para);


BOOLEAN
bef_kopieren (const char *para)
{
	if (user.typ != UTYP_USER && brettnrW == 0)
	{
		mbfehler ("Das Kopieren von Nachrichten ist nur Benutzern gestattet");
		return FALSE;
	}
	if (brettart (brettnrW) == BINAER_BRETT)
		return kopieren_binaer (para);
	else
		return kopieren_text (para);
}



static BOOLEAN
kopieren_binaer (const char *para)
{
	BOOLEAN paraB = FALSE;
	USERNR adressat;
	BRETTNR bnr;
	WORD iW;
	char empfaenger[80];
	long all;
	BIN_MSG bin_info;
	FILE *bin_f;
	BOOLEAN ist_pp;

	set_status_line ("Datei kopieren");

	iW = (strcmp (para, "*") && !strpbrk (para, "-,. ")) ? atoi (para) : 0;

	if (!binaer_dat_anzahl (brettnrW, 0, &all, NULL, NULL, NULL))
		goto abbruch;
	if (iW <= 0 || iW > all)
		goto abbruch;

	if ((bin_f = binaer_dat_oeffnen (brettnrW, FALSE)) == NULL)
		return FALSE;
	fseek (bin_f, (iW - 1) * sizeof (BIN_MSG), SEEK_SET);
	fread (&bin_info, sizeof (BIN_MSG), 1, bin_f);
	binaer_dat_schliessen (bin_f);

	if (!bin_read_access (&bin_info))
	{
		mbfehler ("Zugriff nicht erlaubt");
		return FALSE;
	}

	if (!fileexist (tprintf ("%s%ld.bin", brettpfad (brettnrW), bin_info.nr)))
	{
		mbfehler ("Datei auf Speichermedium nicht vorhanden");
	   /* ### Datei in index.dat als gelscht eintragen */
		return FALSE;
	}

	*empfaenger=0;
	do
	{
		bnr = 0;
		if (!mbfrage(60,MBF_LINE|MBF_TRIM," >> Empfnger: ", empfaenger)) {
			return FALSE;
		}
		if (strchr (empfaenger, '@')) {
			mbfehler ("Binrdateien knnen nicht bers Netz kopiert werden");
			return FALSE;
		} else if (!*empfaenger) {
			return TRUE;
		} else if (ub_find_username (&adressat, empfaenger)) {
			bnr = 0;
			ist_pp = TRUE;
		} else if ((bnr = find_brett (empfaenger)) == 0) {
			mbfehler ("User oder Brett unbekannt");
			return FALSE;
		}

		if (bnr && brettart (bnr) != BINAER_BRETT)
		{
			mbfehler ("Binrdateien knnen nicht in Textbretter kopiert werden");
			return FALSE;
		}
		if (bnr && !(br_ind[bnr].brett->flags & B_WRITE))
		{
			mbfehler ("Zielbrett ist schreibgeschtzt");
			return FALSE;
		}
		if (bnr && brettnrW == bnr)
		{
			mbfehler ("Kopie gleiches Brett nicht mglich");
			return FALSE;
		}
		if (adressat)
		{
			bin_info.empfaenger = adressat;
		}
		bin_info.datime = time (NULL);

		if (binaer_dat_kopieren (brettnrW, bnr, bin_info.lokal))
		{
			mboutput ("\r\n");
			mboutput (" -> Datei kopiert");
			bb_update_brett (bnr);	/* setzt Brettdatum */
		}


		return TRUE;
	}
	while (1);

  abbruch:
	if (!paraB)
		parameter_falsch ("KOPIEREN", para);

	return TRUE;

}								/* kopieren_binaer */


/*
	Mails kopieren
	bedeutet:
	- Mails exportieren
	- Mails importieren, mit Ursprungsdatum

	Das Kopieren knnte ab MDB 1.99 schneller geschehen, wenn nicht erst
	ber eine temporre Datei gearbeitet wrde, da die Nachricht nun 
	im Speicher vorliegt. FB 17.11.96
 */
static BOOLEAN
kopieren_text (const char *para)
{
	char empfaenger[75];
	char exportfname[PATH_MAX];
	char ofname[PATH_MAX];
	const char *name;
	BOOLEAN pm;
	int gh;
	long max;
	long nr;
	BOOLEAN weiterleiten = FALSE;
#ifdef OFFLINE_VERSION
	char *m;
	FILE *tmp_f;
#endif
	*ofname='\0';

	set_status_line ("Nachricht kopieren");
	if (brettnrW != 0 && !(br_ind[brettnrW].brett->flags & B_READ))
	{
		mbfehler ("Kopieren hier nicht erlaubt");
		return FALSE;
	}

	while (*para == '-')
	{
		char c = para[1];

		para+=2;
		switch (c)
		{
		case 'w':
		case 'W':
			weiterleiten = TRUE;
			break;
		default:
			mbfehler (tprintf ("Illegaler Parameter: %c", c));
			return FALSE;
			break;
		}
	   /* skip over whitespace */
		while (*para && isspace (*para))
			para++;
	}

	if (weiterleiten && brettnrW)
	{
		mboutput (
		  "ffentliche Nachrichten knnen nicht weitergeleitet werden\r\n");
		mboutput (
					 "Wenn Sie fortfahren wird die Nachricht kopiert\r\n");
		weiterleiten = FALSE;
	}

	if (!*para)
	{
		mbfehler ("Nachrichtennummer fehlt");
		return FALSE;
	}

	nr = strtol (para, NULL, 10);
	if (nr < 1)
	{
		mbfehler ("Ungltige Nachrichtennummer");
		return FALSE;
	}

	pm = (brettnrW == 0);
	name = (pm) ? user.name : akt_brett.name;

#ifdef OFFLINE_VERSION
	if ((m = group_get_article(nr)) == NULL)
	{
		mbfehler ("Nachricht nicht vorhanden");
		return TRUE;
	}
	sprintf (exportfname, "%s%s", tom_tmp_dir, "tmpmsg.txt");
	unlink (exportfname);
	if ((tmp_f = fopen(exportfname,"w")) == NULL)
	{
		free(m);
		mbfehler("Kann temporre Datei nicht anlegen");
		return FALSE;
	}
	fwrite(m,strlen(m),1,tmp_f);
	fclose(tmp_f);
	free(m);
#else
	if (!group_info_msgs (name, pm, &max, NULL) || !max)
	{
		mbfehler ("Keine Eintrge vorhanden");
		return TRUE;
	}
	if (nr > max)
	{
		mbfehler ("Nachricht nicht vorhanden");
		return TRUE;
	}

	if ((gh = group_open (getpid (), name, pm, FALSE)) < 0)
	{
		mbfehler ("Keine Eintrge vorhanden");
		return TRUE;
	}

#ifdef OFFLINE_VERSION
	if (!index_goto (gh, nr - 1) || !msg_read (gh,FALSE))
#else
	if (!index_goto (gh, nr - 1) || !msg_read (gh))
#endif
	{
		group_close (gh, getpid ());
		mbfehler ("Nachricht nicht vorhanden");
		return TRUE;
	}
	sprintf (exportfname, "%s%s", tom_tmp_dir, "tmpmsg.txt");
#ifdef OFFLINE_VERSION
	if (!msg_export_raw (gh, exportfname, NULL, FALSE) || !fileexist (exportfname))
#else
	if (!msg_export_raw (gh, exportfname, NULL) || !fileexist (exportfname))
#endif
	{
		group_close (gh, getpid ());
		unlink (exportfname);
		mbfehler ("Nachricht nicht lesbar");
		return TRUE;
	}
	group_close (gh, getpid ());
#endif

	*empfaenger=0;
	while (TRUE)
	{
		USERNR unr;
		BRETTNR bnr;

		unr = 0;
		bnr = 0;
		mboutput ("Empfnger:\r\n");
		if (!mbfrage(70,MBF_LINE|MBF_TRIM,">",empfaenger))
			break;
		if (!*empfaenger)
			break;
		if (!strchr (empfaenger, '@'))
		{
		   /* knnte eine ffentliche Nachricht sein */
			USERNR *uar = ub_find_username_multi (empfaenger, FALSE);

			if (uar && uar[1])
			{
				free (uar);
				mboutput ("Nicht eindeutig!\r\n");
				continue;
			}
			if (uar)
			{
				unr = uar[0];
				free (uar);
			}
			bnr = brettnr (empfaenger);
			if (unr && bnr)
			{
				mboutput ("Nicht eindeutig, pat auf Benutzer und Brett!\r\n");
				continue;
			}
			else if (unr)
			{
			   /* da mssen wir durch den Router um Forwarder zu beachten! */
				goto try_router;
			}
			else if (!bnr)
			{
				mboutput ("Unbekannter Empfnger!\r\n");
				continue;
			}
		   /* ffentliche Nachricht */
			pm = FALSE;
			if (brettart (bnr) == BINAER_BRETT)
			{
				mboutput ("Kopien von Texten in Binrbretter nicht mglich\r\n");
				continue;
			}
			if (brettnrW == bnr)
			{
				mboutput ("Kopie in gleiches Brett nicht mglich\r\n");
				continue;
			}
			if (!(br_ind[bnr].brett->flags & B_WRITE))
			{
				mboutput ("Zielbrett ist schreibgeschtzt");
				continue;
			}
		   /* 
		    * PM -> Brett
		    *   Hm - kann leicht etwas peinliches passieren.
		    */
			if (weiterleiten)
			{
			   /* Da geht bei MAUS *alles* schief. */
				mboutput (
					"Nachrichten knnen nicht in Bretter weitergeleitet werden.\r\n");
				mboutput (
					 "Wenn Sie fortfahren wird die Nachricht kopiert.\r\n");
				mboutput ("Wollen Sie fortfahren? (J/N) ");
				if (!JAnein (NULL))
					goto cleanup;
				weiterleiten = FALSE;
			}
			if (!brettnrW)
			{
				mbprintf (
					"Wollen Sie wirklich eine persnliche Nachricht in ein Brett %s? (J/N) ",
					weiterleiten ? "weiterleiten" : "kopieren");
				if (!JAnein (NULL))
					goto cleanup;
			}
		}
		else
		  try_router:
		{
			char *comment = NULL;

			pm = TRUE;
			if (!r_vrfy (empfaenger, &comment, level (), user.bezahlt))
			{
				mbprintf ("Nicht zustellbar: %s\r\n",
						  comment ? comment : "unbekannter Fehler");
				free (comment);
				continue;
			}
			free (comment);
		}

	   /* Beim Weiterleiten nochmal nachfragen */
		if (weiterleiten)
		{
			struct mbselect_t auswahl[] =
			{
				{'W', "Weiterleiten\r\n"},
				{'K', "Kopieren\r\n"},
				{'A', "Abbrechen\r\n"},
				{'\033', "Abbrechen\r\n"},
				{0, NULL}
			};
			char c;

			mboutput (
				"Wollen Sie wirklich eine persnliche Nachricht weiterleiten?\r\n"
				"Dabei wird das Orginal gelscht, Antworten gegen an den Originalabsender.\r\n"
				"Sie knnen die Nachricht auch kopieren, dabei bleibt da Original erhalten\r\n"
				"und Sie erhalten eventuelle Antworten.\r\n"
				"(W)eiterleiten, (K)opieren oder (A)bbrechen? ");
			c = mbselect (auswahl);
			if (c == 'K')
				weiterleiten = FALSE;
			else if (c != 'W')
				goto cleanup;
		}

	   /* Wir wissen wohin die Nachricht soll. Nun Text fertig machen. */
		{
			char buf[16384];
			char *betreff = NULL;
			char *orig_abs = NULL;
			char *orig_id = NULL;
			struct veinfachpara vpara[10];
			int vcount = 0;
			BOOLEAN neue_id = FALSE;
			BOOLEAN ok;
			FILE *in,
			 *out;

			in = xfopen (exportfname, "r");
			sprintf (ofname, "%s%s", tom_tmp_dir, "kopie.txt");
			out = xfopen (ofname, "w");
			while (fgets (buf, sizeof (buf), in))
			{
				char *p = strpbrk (buf, "\r\n");

				if (p)
					*p = 0;
				if (!*buf || *buf == ':')
					break;
				switch (*buf)
				{
				case '#':
					orig_id = XSTRDUP (buf + 1);
					break;
				case 'V':
					orig_abs = XSTRDUP (buf + 1);
					/* fprintf (out, "Original-Absender: %s\n", buf + 1); */
					break;
				case 'W':
					if (weiterleiten)
						betreff = concat (buf + 1, " (fwd)", NULL);
					else
						betreff = concat (buf + 1, " (Kopie)", NULL);
					XNULLCHECK (betreff);
					break;
				}
			}
		   /* So, ID, Betreff, Orig-Abs haben wir */
			if (weiterleiten)
			{
			   /* da kann es zu erheblichen Problemen kommen, wenn die
			    * Nachricht mit MAUS kollidiert. Die macht einen 
			    * Dupecheck bei PMs und kommt mit PM-IDs bei 
			    * ffentlichen Nachrichten ins Schleudern. 
			    */
#define UEBERTRIEBENE_RUECKSICHT_AUF_MAUS
#ifdef UEBERTRIEBENE_RUECKSICHT_AUF_MAUS
				if (pm)
				{
					mboutput (
						"MAUS erkennt weitergeleitete Nachrichten, die ihre Original-ID\r\n"
						"beibehalten, flschlicherweise als Dupes. Soll die ID verndert\r\n"
						"werden? Das ist nur ntig wenn die Nachricht durch eine MAUS geht,\r\n"
						"und kostet die Kommentarverkettung. ID verndern? (J/N)");
					if (JAnein (NULL))
						neue_id = TRUE;
				}
#endif
			}
			else
				neue_id = TRUE;
			if (weiterleiten)
			{
				fprintf (out, "Weitergeleitet durch: %s @ %s\n",
						 user.name, MYDOMAINPART);
			}
			else
			{
				fprintf (out, "Originalabsender: %s\n", orig_abs);
				fprintf (out, "Kopiert aus: %s\r\n",
						 brettnrW ? akt_brett.name : "Postfach");
			}
			if (neue_id)
			{
				fprintf (out, "Original-Id: %s\n",
						 orig_id ? orig_id : "unbekannt");
			}
			fputc('\n',out); /* Leerzeile als Trenner erzeugen */
			if (*buf==':') /* paranoia */
			{
				fputs (buf + 1, out);
				fputc ('\n', out);
			}
			while (fgets (buf, sizeof (buf), in))
			{
				char *q = strpbrk (buf, "\r\n");

				if (q)
					*q = 0;
				if (!*buf || *buf == '#')
					break;
				if (*buf == ':')
				{
					fputs (buf + 1, out);
					fputc ('\n', out);
				}
			}
			fclose (out);
			fclose (in);

			if (!neue_id)
			{
				vpara[vcount].kenn = '#';
				vpara[vcount].str = XSTRDUP (orig_id);
				vcount++;
			}

			vpara[vcount].kenn = 'V';
			if (!weiterleiten)
			{
				vpara[vcount].str = concat (user.name, "@", MYDOMAINPART, NULL);
				XNULLCHECK (vpara[vcount].str);
			}
			else
				vpara[vcount].str = XSTRDUP (orig_abs);
			vcount++;

			vpara[vcount].kenn = 'W';
			vpara[vcount].str = XSTRDUP (betreff);
			vcount++;

			if (pm)
			{
				vpara[vcount].kenn = 'A';
				vpara[vcount].str = XSTRDUP (empfaenger);
				vcount++;
			}
			else
			{
				vpara[vcount].kenn = 'G';
				vpara[vcount].str = XSTRDUP (empfaenger);
				vcount++;
			}

			vpara[vcount].kenn = '\0';
			vpara[vcount].str = NULL;
			vcount++;

			ok = veinfachnachricht (pm, level (), user.bezahlt, ofname, vpara);
			while (vcount)
			{
				vcount--;
				free (vpara[vcount].str);
			}
			free (orig_id);
			free (betreff);
			free (orig_abs);

			if (!ok)
				mboutput ("Die Nachricht konnte nicht versandt werden!\r\n");
			else
			{
				unlink (ofname);
				mboutput ("Nachricht versandt!\r\n");
				break;
			}
		}
	}
  cleanup:
	if (*ofname)
		unlink (ofname);
	unlink (exportfname);
	return TRUE;
}								/* kopieren_text */
