/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
#include "tomsysdep.h"

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <errno.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <signal.h>
#include <dirent.h>

#include <portab.h>

#include "prototyp.h"
#include "io.h"
#include "modem.h"
#include "terminal.h"
#include "user.h"
#include "brett.h"
#ifdef OFFLINE2
#include "offline/inc/version.h"
#include "offline/inc/mdb_global.h"
#else
#include "offline/version.h"
#include "offline/mdb_global.h"
#endif
#include "bin_msg.h"
#include "routing.h"
#include "business.h"
#include "box_conf.h"
#include "befehle.h"
#include "protokoll.h"
#include "aktdata.h"
#include "tompaths.h"

extern USER user;

WORD tagesanrufeW;

static void midnight_zahler_update(void);
static void markiere_inaktive_user(void);

/*
	07.01.93 Frank, mehrere Funktionen berarbeitet, lauffhig

	MultiPort-Betrieb beachten, Userhandling in cron/server verlagern
*/

BOOLEAN bef_midnight(const char *para)
{
	const char *such;
	BOOLEAN ok=TRUE;

	BOOLEAN counter_f=FALSE;
	BOOLEAN bin_f=FALSE;
	BOOLEAN mail_f=FALSE;
	BOOLEAN news_f=FALSE;
	BOOLEAN user_f=FALSE;
	BOOLEAN status_f=FALSE;
	BOOLEAN zahler_f=FALSE;

	such=para;
	do
	{
		const char *p;
		BOOLEAN nicht=FALSE;

		if (*such=='!')
		{
			nicht=TRUE;
			such++;
		}

		p=strpbrk(such,"\t ");

		if (!strnicmp(such,"*",1))
		{
			counter_f=TRUE ^ nicht;
			bin_f=TRUE ^ nicht;
			mail_f=TRUE ^ nicht;
			news_f=TRUE ^ nicht;
			user_f=TRUE ^ nicht;
			status_f=TRUE ^ nicht;
			zahler_f=TRUE ^ nicht;
		}
		else if (!strnicmp(such,"COUNTER",7))
		{
			counter_f=TRUE ^ nicht;
		}
		else if (!strnicmp(such,"BIN",3))
		{
			bin_f=TRUE ^ nicht;
		}
		else if (!strnicmp(such,"MAIL",4))
		{
			mail_f=TRUE ^ nicht;
		}
		else if (!strnicmp(such,"NEWS",4))
		{
			news_f=TRUE ^ nicht;
		}
		else if (!strnicmp(such,"USER",4))
		{
			user_f=TRUE ^ nicht;
		}
		else if (!strnicmp(such,"STATUS",4))
		{
			status_f=TRUE ^ nicht;
		}
		else if (!strnicmp(such,"ZAHLER",6) || !strnicmp(such,"$",1))
		{
			zahler_f=TRUE ^ nicht;
		}
		else
		{
			ok=FALSE;
		}

		if (p)
		{
			such=p;
			while (isspace(*such))
				such++;
		}
		else
			such=NULL;
	} while (such && *such);

	if (!ok)
	{
		parameter_falsch("MIDNIGHT",para);
		return FALSE;
	}

	set_status_line("Mitternachts-Update");
	if (counter_f)
	{
		FILE *f;
		struct status_dat st;
		update_statistik();
		f=open_read_status_dat(FALSE,&st);
		st.postertrag=0;
		st.tagesanrufe=0;
		write_close_status_dat(f,&st);
		ok=TRUE;
	}
	if (bin_f)
	{
		set_status_line("Programmfcher putzen");
		binaer_dat_putze_alle(TRUE);
		set_status_line("Binrbretter putzen");
		binaer_dat_putze_alle(FALSE);
		bin_global_putzen();
		set_status_line("globale Binrdatenbank erzeugen");
		bin_global_create();
		ok=TRUE;
	}

	if (mail_f)
	{
		set_status_line("Postfcher putzen");
#ifdef OFFLINE_VERSION
		mdb_putz("*",TRUE);
#else
		mdb_putze_alles(TRUE,FALSE);
#endif
		ok=TRUE;
	}

	if (news_f)
	{
		set_status_line("Gruppen putzen");
		bef_putz("GRUPPEN");
		ok=TRUE;
	}

	if (user_f)
	{
		/* usercrunch */
		set_status_line("inaktive User markieren");
		markiere_inaktive_user();
#if 0
		set_status_line("User crunchen");
		crunch_user();
#endif
		ok=TRUE;
	}

	if (status_f)
	{
		set_status_line("Nachrichtenstatus putzen");
		r_st_expire();
		ok=TRUE;
	}

	if (zahler_f)
	{
		set_status_line("Warnungen vor ablaufendem Zahlerstatus versenden");
		midnight_zahler_update();
		ok=TRUE;
	}
	return ok;
}

static void
midnight_zahler_update(void)
{
	time_t yet;
	struct tm *tm;
	char gname[PATH_MAX];
	FILE *g;
	BOOLEAN warned=FALSE;
	USER thisuser;
	char empfaenger[256];
	char absender[256];
	char fnamein[PATH_MAX];
	FILE *fin;

	thisuser.nummer=0;
	yet=time(NULL);

	sprintf(fnamein,"%szahlerwarn.msg",box_conf.texte_dir);
	fin=fopen(fnamein,"r");
	if (!fin)
		return;

	sprintf(absender,"Quark @ %s",MYDOMAINPART);

	sprintf(gname,"%szahlsum.msg",tom_tmp_dir);
#define FRIST_1 (30*86400L)
#define FRIST_2 (10*86400L)
	g=xfopen(gname,"w");
	fprintf(g,
"Folgende Benutzer wurden ber den bevorstehenden Verlust des Zahlerstatus\n"
"informiert:\n"
		);
	while (ub_walk_user(&thisuser))
	{
		BOOLEAN doit=FALSE;
		char fnameout[PATH_MAX];
		FILE *fout;
		USER u;

		if (!thisuser.bezahlt)
			continue;
		/* wenn noch mehr als 30 Tage Zahler */
		if (thisuser.bezahlt_bis>yet+FRIST_1)
			continue;
		/* Wenn 30 Tage Frist gerade berschritten */
		if (thisuser.bezahlt_bis-FRIST_1 <= yet
			&& thisuser.bezahlt_bis-FRIST_1+86400L >= yet)
			doit=TRUE;
		/* Wenn 10 Tage Frist gerade berschritten */
		if (thisuser.bezahlt_bis-FRIST_2 <= yet
			&& thisuser.bezahlt_bis-FRIST_2+86400L >= yet)
			doit=TRUE;
		if (!doit)
			continue;

		sprintf(empfaenger,"%s @ %s",thisuser.name,MYDOMAINPART);

		tm=localtime(&thisuser.bezahlt_bis);


		sprintf(fnameout,"%szahler.msg",tom_tmp_dir);

		fout=xfopen(fnameout,"w");
		fseek(fin,0,SEEK_SET);
		u=user;
		user=thisuser;
		interpret_file_to_file(fin,fout);
		user=u;
		fclose(fout);

		einfachnachricht(TRUE,9,TRUE,fnameout,
			'A', empfaenger,
			'V', absender,
			'W', "Zahlerstatus luft ab",
			0);
		unlink(fnameout);
		fprintf(g,"%02d.%02d.%d %s\n",
			tm->tm_mday,
			tm->tm_mon+1,
			tm->tm_year,
			thisuser.name);
		warned=TRUE;
	}

	fclose(g);
	fclose(fin);
	if (warned)
	{
		sprintf(empfaenger,"Sysop @ %s",MYDOMAINPART);
		einfachnachricht(TRUE,9,TRUE,gname,
			'A', empfaenger,
			'V', absender,
			'W', "Zahlerstatuswarnungen",
			0);
	}
	unlink(gname);
	return;
} /* midnight_zahler_update */

static void markiere_inaktive_user(void)
{
	USER u;
	time_t jetzt;
	long anzahl = 0;
	FILE *f;
	char fname[PATH_MAX];
	sprintf(fname,"%sloeschmail.txt",tom_tmp_dir);


	if (!box_conf.kill_inactive_days)
		return;

	f=xfopen(fname,"w");
	fprintf(f,"Heute wurden folgende Benutzer automatisch gelscht:\n\n");

	time(&jetzt);
	u.nummer = 0;
	while (ub_walk_user(&u))
	{
		if (!ub_read_user(u.nummer,&u,FALSE))
			continue;
		if (u.level>5 || u.geschlecht==NEUTRUM)
			continue;
		if ((jetzt - u.letzter_login) / 86400L < box_conf.kill_inactive_days)
			continue;
		if (u.anrufe > (ULONG) box_conf.kill_inactive_calls)
			continue;
		if (u.bezahlt)
			continue;
		if (u.flags & UB_FLAG_PROMI)
			continue;
		/* Lesen und sperren */
		if (!ub_read_user(u.nummer,&u,TRUE))
			continue;
		if (anzahl)
			fprintf(f,"\n");
		fprintf(f,"Name:    %s\n",u.name);
		fprintf(f,"Level:   %d\n",(int) u.level);
		fprintf(f,"Telefon: %s\n",u.telefon);
		fprintf(f,"Anrufe:  %ld\n",(long) u.anrufe);
		fprintf(f,"Erster:  %s",ctime(&u.erster_login));
		fprintf(f,"Letzter: %s\n",ctime(&u.letzter_login));
		anzahl++;
		delete_user(u.nummer);
	}
	fclose(f);
	if (anzahl)
	{
		char empfaenger[256];
		char absender[256];

		sprintf(absender,"Quark @ %s",MYDOMAINPART);
		sprintf(empfaenger,"Sysop @ %s",MYDOMAINPART);

		einfachnachricht(TRUE,9,TRUE,fname,
			'A', empfaenger,
			'V', absender,
			'W', "automatisch gelschte Benutzer",
			0);
	}
	unlink(fname);

} /* markiere_inaktive_user */
