/* route_befehl.c : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
*/
#include "config.h"

#include <portab.h>
#include <string.h>
#include <stdlib.h>

#include "tomsysdep.h"
#include "prototyp.h"
#include "befehle.h"
#include "user.h"
#include "msg.h"
#include "modem.h"
#include "routing.h"
#include "box_conf.h"

static BOOLEAN work_route(message_t *msg, const char *adresse);

extern USER user;
extern int logout_modus;

BOOLEAN 
bef_route(const char *para)
{
	char adr[1024];
	message_t *msg=msg_alloc();
	int old_grade;
	if (!msg)
	{
		mbfehler("out of memory");
		return FALSE;
	}
	old_grade=r_set_verbose_grade(0);
	r_set_verbose_grade(old_grade);

	if (!*para)
	{
		while (1)
		{
			if (logout_modus)
				break;
			mbquestion(60,1,TRUE," Ziel: ","",adr);
			if (!*adr)
			{
				msg_free(msg);
				return TRUE;
			}
			mboutput("\r\n");
			work_route(msg,adr);
		}
	}
	else if (!stricmp(para,"test"))
	{
		FILE *f;
		f=fopen(tprintf("%sroute.tst",box_conf.config_dir),"r");
		if (!f)
		{
			mbfehler("route.tst in tom/system nicht vorhanden");
				msg_free(msg);
			return FALSE;
		}
		while (fgets(adr,sizeof(adr),f))
		{
			char *p;
			p=strpbrk(adr,"\r\n");
			if (p)
				*p='\0';
			mbprintf(" %s:\r\n",adr);
			work_route(msg,adr);
		}
		fclose(f);
	}
	else if (para && *para)
	{
		work_route(msg,para);
	}
	else
		parameter_falsch("ROUTE",para);

	msg_free(msg);
	r_set_verbose_grade(old_grade);
	return TRUE;
}

static BOOLEAN 
work_route(message_t *msg, const char *adresse)
{
	r_adr_t *adr=NULL;
	r_adr_t *akt;

	if (!strnicmp(adresse,"grade ",6))
	{
		int g=strtol(adresse+6,NULL,10);
		int n;
		n=r_set_verbose_grade(g);
		mbprintf("Alt: %d, neu: %d\r\n",n,g);
		return TRUE;
	}

	adr=r_alloc();
	if (!adr)
	{
		mboutput("out of memory\r\n");
		return FALSE;
	}

	adr->erroradr=NULL;
	adr->inputadr=strdup(adresse);
	if (!adr->inputadr)
	{
		r_free(adr);
		mboutput("out of memory\r\n");
		return FALSE;
	}
	adr->level=level();
	adr->sec_level=level();
	adr->bezahlt=user.bezahlt;
	adr->sec_bezahlt=user.bezahlt;
	adr->owner=user.nummer;

	if (!r_router(msg,adr))
		mboutput(" Keine Route bekannt\r\n");

	for (akt=adr; akt; akt=akt->next)
	{
		mbprintf(" %s:\r\n",akt->inputadr);
		if (akt->resultadr)
			mbprintf(" Env: %s\r\n",akt->resultadr);
		if (akt->error)
			mbprintf("   Fehler aufgetreten\r\n");
		if (akt->logstr)
		{
			const char *s="???";
			switch(akt->loggrade)
			{
			case R_LOG_INFO: s="Info"; break;
			case R_LOG_NOTICE: s="Notiz"; break;
			case R_LOG_ERROR: s="Fehler"; break;
			case R_LOG_FATAL: s="Fatal"; break;
			}
			mbprintf("   %s (%s)\r\n",akt->logstr, s);
		}
		if (akt->dsn) 
			mbprintf("   DSN: %s\r\n",akt->dsn);
		if (akt->errorflags)
		{
			mboutput("  Fehler: ");
			if (akt->errorflags & R_E_TEMP)
				mboutput("temp. ");
			if (akt->errorflags & R_E_ADR)
				mboutput("Adresse ");
			if (akt->errorflags & R_E_ACCESS)
				mboutput("verboten ");
			if (akt->errorflags & R_E_INTERN)
				mboutput("intern");
			mboutput("\r\n");
		}
		mbprintf("  Loopcount: %lu\r\n",(ULONG) akt->loopcount);
		if (akt->flags)
		{
			mboutput("  Flags: ");
			if (akt->flags & R_F_NEUE_ID)
				mboutput("ID ");
			if (akt->flags & R_F_TAUSCHLOG)
				mboutput("#LOG ");
			if (akt->flags & R_F_SPAETER_AUFL)
				mboutput("spter ");
			if (akt->flags & R_F_LOKAL)
				mboutput("lokal ");
			if (akt->flags & R_F_NO_WILDCARD)
				mboutput("!* ");
			if (akt->flags & R_F_NEWENVELOPE)
				mboutput("+env ");
			if (akt->flags & R_F_PSEUDO)
				mboutput("pseudo ");
			if (akt->flags & R_F_AUFGELOEST)
				mboutput("aufg. ");
			if (akt->flags & R_F_GEROUTET)
				mboutput("rout. ");
			if (akt->flags & R_F_ZUGESTELLT)
				mboutput("zug. ");
			if (akt->flags & R_F_E_TRIED)
				mboutput("tried "); /* this should not happen */
			if (akt->flags & R_F_TO_POSTMASTER)
				mboutput("bounce-postmaster ");
			if (akt->flags & R_F_TO_OWNER)
				mboutput("bounce-owner ");
			if (akt->flags & R_F_TO_SENDER)
				mboutput("bounce-sender ");
			mboutput("\r\n");
		}
		if (akt->transport)
			mbprintf("  Transport: %s\r\n",akt->transport->name);
		else	
			mboutput("  Transport: keiner\r\n");
		mbprintf("  User/Gate: %lu\r\n",(ULONG) akt->user_gate_nr);
		mboutput("\r\n");
	}
	r_free_chain(adr);

	return TRUE;
}

