/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
*/
#include "tomsysdep.h"

#include <stdio.h>
#include <string.h>

#include "prototyp.h"
#include "befehle.h"
#include "brett.h"
#include "user.h"
#include "modem.h"
#include "terminal.h"
#include "bin_msg.h"
#include "bin_out.h"

extern USER user;
extern BRETTNR brettnrW;
extern BRETTINFO *br_ind;

extern BYTE batchB;
extern WORD count_crW;


static BOOLEAN suchen_text(const char *para);
static BOOLEAN suchen_binaer(const char *para);



/*
	16.12.92 Frank, vorlufig nur in Binrbrettern erlauben
*/
BOOLEAN bef_suchen(const char *para)
{
	if (!(br_ind[brettnrW].brett->flags & B_READ))
	{
		mbfehler("Suchfunktion hier nicht erlaubt");
		return FALSE;
	}
	if (!*para)
	{
		mbfehler("Suchbegriff fehlt");
		return FALSE;
	}

	if (brettart(brettnrW) == BINAER_BRETT)
		return suchen_binaer(para);
	else
		return suchen_text(para);

} /* bef_suchen */



static BOOLEAN suchen_text(const char *para TOM_ATTRIB_UNUSED)
{
	mbfehler("Suchen in Textbrettern nicht mglich");
	return FALSE;

} /* suchen_text */



static BOOLEAN
binaer_suchen_ausgabe(const BIN_MSG *bin_info,BOOLEAN *kopf,BOOLEAN joker,
                      const char *para);

static BOOLEAN suchen_binaer(const char *para)
{
	BOOLEAN gefunden = FALSE;
	BOOLEAN para_ok = FALSE;
	BOOLEAN kopf = FALSE;
	long all;
	FILE *bin_f;
	BIN_MSG bin_info;
	BOOLEAN joker = FALSE;

	joker = (strpbrk(para,"*?") != 0);

	if (!binaer_dat_anzahl(brettnrW,0,&all,NULL,NULL,NULL))
		return FALSE;
	if (!all)
	{
		mboutput(" Keine Eintrge in diesem Brett vorhanden\r\n\n");
		return FALSE;
	}

	if ((bin_f = binaer_dat_oeffnen(brettnrW,FALSE)) == NULL)
		return FALSE;
	while (fread(&bin_info,sizeof(BIN_MSG),1,bin_f))
	{
		if (userbreak())
			break;

		if (binaer_suchen_ausgabe(&bin_info,&kopf,joker,para))
			para_ok = TRUE;
	}
	binaer_dat_schliessen(bin_f);

	if (!gefunden)
	{
		char tmp[80];
		long nr;

		if (all)
			mboutput(" Keine Eintrge mit diesem Suchtext gefunden\r\n\n");

		para_ok = TRUE;

		if (!batchB)
		{
			mboutput(" Global suchen? (J/N) ");
			if (mbdecide("JN","Ja\r\n|Nein\r\n") == 2)
				goto abbruch;
		}

		strcpy(tmp,para);
		strlwr(tmp);

		nr = bin_global_suchen((joker) ? 'W' : 'N',tmp,0,&bin_info);
		while (nr >= 0)
		{
			char bef_para[10];

			para_ok = TRUE;

			if (br_ind[bin_info.brettnr].brett->flags & B_READ)
			{
				sprintf(bef_para,"%ld",bin_info.brettnr);
				bef_brett(bef_para);
				sprintf(bef_para,"%ld",bin_info.lokal);
				if (!download_intern(bef_para))
					break;
			}

			nr = bin_global_suchen((joker) ? 'W' : 'N',tmp,nr+1,&bin_info);
			if (nr < 0)
				mboutput("\r\n Keine weiteren Eintrge\r\n");
		}
	}

abbruch:
	set_scroll_area(1);

	if (!para_ok)  parameter_falsch("SUCHEN"," ");

	return TRUE;

} /* suchen_binaer */



static BOOLEAN binaer_suchen_ausgabe
(const BIN_MSG *bin_info,BOOLEAN *kopf,BOOLEAN joker,const char *para)
{
	BIN_MSG tmp;

	/* diese Abfrage funktioniert nur bei lokaler Suche */
	if (!bin_read_access(bin_info))
		return FALSE;
#if 0
	if (user.level < 9)
		return FALSE;
#endif
	/* da para in Groschrift vorliegt, Binrdatensatz ebenfalls wandeln */
	tmp = *bin_info;
	strupr(tmp.betreff);
	strupr(tmp.absender);
	strupr(tmp.name);

	if (!strstr(tmp.betreff,para)
	&&  !strstr(tmp.absender,para)
	&& (!joker || !wildcard(tmp.name,para)))
		return FALSE;

	if (!*kopf)
	{
		set_status_line("Brettinhalt:");
		bin_heading();
		*kopf = TRUE;

		set_scroll_area(2);
		count_crW = 0;
	}

	if (bin_info->brettnr)
	{
		if (br_ind[bin_info->brettnr].brett)
			mbprintf("Brett: %s\r\n",br_ind[bin_info->brettnr].brett->name);
		else
			mbprintf("Brett: ??? [interner Fehler]\r\n");
	}

	bin_msg_inhalt(bin_info);

	return TRUE;

} /* binaer_suchen_ausgabe */
