/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
#include "config.h"

#include <portab.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <syslog.h>
#include <errno.h>

#include "tomsysdep.h"
#include "prototyp.h"
#include "befehle.h"
#include "modem.h"
#include "user.h"
#include "brett.h"
#include "tausch.h"
#include "aktdata.h"
#include "tompaths.h"
#include "box_conf.h"

extern USER user;

extern BRETTNR brettnrW;
extern BRETT akt_brett;

/*
	Tausch
	- BRETT
	- DATUM
	- INST
	- ADD [Gruppenname]
	- DEL [Gruppenname]

	19.07.93 Frank, ADD und DEL
	25.10.93 Frank, Bugfix fr ADD/DEL
	30.11.93 Uwe, an nderungen im Tausch angepat.
*/
BOOLEAN bef_tausch(const char *para)
{
	if (!*para)
	{
		/* fr den aktuellen User tauschen */
		return ta_run_tausch(ta_modus_normal,NULL,NULL,NULL,NULL);
	}
	else if (!strnicmp(para,"BRETT",5))
	{
		strcpy(akt_anruf.zuletzt,"TAU");
		tausch_brett(&para[6]);
		strcpy(akt_anruf.zuletzt,"BOX");
		return TRUE;
	}
	else if (!stricmp(para,"INST"))
	{
		ta_do_inst();
		return TRUE;
	}
	else if (!strnicmp(para,"CMD",3))
	{
		char duplicate[256];
		FILE *f;
		char fname[PATH_MAX];
		USERNR unr;

		strcpy(duplicate,para+3);
		strtrim(duplicate);

		if (!*duplicate)
		{
			mboutput(" An wen soll das Kommando gehen?\r\n");
			mbquestion(sizeof(user.name),3,TRUE," > ","",duplicate);
			strtrim(duplicate);
			if (!*duplicate)
				return TRUE;
		}

		if (!ub_find_username(&unr,duplicate+1))
		{
			mbfehler("Unbekannter Benutzer");
			return FALSE;
		}

		if (unr!=user.nummer && level()<8)
		{
			mbfehler("Nicht erlaubt");
			return FALSE;
		}

		mboutput("\r\n Welches Kommando senden?\r\n");
		mbquestion(75,3,TRUE," > ","",duplicate);
		strtrim(duplicate);

		if (!*duplicate)
			return TRUE;

		sprintf(fname,"%s%lu.once",box_conf.tausch_dir,(ULONG) unr);
		f=fopen(fname,BINAPPEND);
		if (!f)
		{
			syslog(LOG_ERR,"cannot write to %s: %s",fname,strerror(errno));
			mbfehler("kann Kommando nicht speichern");
			return FALSE;
		}
		fprintf(f,"%s\r\n", duplicate);
		fclose(f);
		mboutput("OK\r\n");
		return TRUE;
	}
	else if (!strnicmp(para,"ADD",3) || !strncmp(para,"DEL",3))
	{
		char duplicate[256];
		BOOLEAN add=!strnicmp(para,"ADD",3);

		strcpy(duplicate,para+3);
		strtrim(duplicate);

		if (brettnrW==0)
		{
			mbfehler("Postfach ist/bleibt bestellt");
			return FALSE;
		}

		if (*duplicate=='@')
		{
			FILE *f;
			char fname[PATH_MAX];
			USERNR unr;
			const char *p;
			p=duplicate+1;
			while (*p=='@')
				p++;
			if (!ub_find_username(&unr,p))
			{
				mbfehler("Unbekannter Benutzer");
				return FALSE;
			}
			if (unr!=user.nummer && level()<8)
			{
				mbfehler("Nicht erlaubt");
				return FALSE;
			}
			sprintf(fname,"%s%lu.once",box_conf.tausch_dir,(ULONG) unr);
			f=fopen(fname,BINAPPEND);
			if (!f)
			{
				syslog(LOG_ERR,"cannot write to %s: %s",fname,strerror(errno));
				mbfehler("kann Bestellung nicht speichern");
				return FALSE;
			}
			fprintf(f,"G%s%s\r\n", add ? "+" : "-" , akt_brett.name);
			fclose(f);
			mboutput("OK\r\n");
			if (duplicate[1]!='@')
				return TRUE;
			/* else fall through nach @@ */
		}
		if (*duplicate && (*duplicate !='@' || duplicate[1]=='@'))
		{
			/* Fr einen User bestellen */
			FILE *f;
			char fname[PATH_MAX];
			USERNR unr;
			char *p;
			if (*duplicate=='@')
				p=duplicate+2;
			else
				p=duplicate;
			if (!ub_find_username(&unr,p))
			{
				mbfehler("Unbekannter Benutzer");
				return FALSE;
			}
			if (unr!=user.nummer && level()<8)
			{
				mbfehler("Nicht erlaubt");
				return FALSE;
			}
			sprintf(fname,"%s%lu.tau",box_conf.tausch_dir,(ULONG) unr);
			if (!fileexist(fname))
			{
				mbfehler("Zuerst mu TAUSCH INST aufgerufen werden");
				return FALSE;
			}
			f=fopen(fname,BINAPPEND);
			if (!f)
			{
				syslog(LOG_ERR,"cannot write to %s: %s",fname,strerror(errno));
				mbfehler("kann Bestellung nicht speichern");
				return FALSE;
			}
			fprintf(f,"%s %lu %s\r\n", 
				add ? "+" : "-" , 
				(ULONG) brettnrW, akt_brett.name);
			fclose(f);
			mboutput("OK\r\n");
			return TRUE;
		}
		else if (!*duplicate)
		{
			BOOLEAN ok;
			char fehl[128];
			if (add)
				ok=ta_brett_bestellen(brettnrW, fehl);
			else
				ok=ta_brett_abbestellen(brettnrW, fehl);
			if (ok)
				mbprintf(" Gruppe wurde %sbestellt\r\n",
					(add)? "" : "ab");
			else
				mbprintf(" %s\r\n",fehl);
			if (ok)
				/* Installation sichern */
				ta_sichere_installation(FALSE);
			return ok;
		}
		return TRUE;
	}
	else
	{
		parameter_falsch("TAUSCH",para);
		return FALSE;
	}
}
