/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact1: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/

#include "tomsysdep.h"

#include <portab.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/stat.h>

#include "prototyp.h"
#include "userbase.h"
#ifdef OFFLINE2
#include "offline/inc/version.h"
#include "offline/inc/mdb_global.h"
#else
#include "offline/version.h"
#include "offline/mdb_global.h"
#endif
#include "user.h"
#include "box_conf.h"
#include "brett.h"
#include "modem.h"

extern BOOLEAN bef_brett_sortierung_maybe_wrong;
#define UMASK                     (S_IRUSR | S_IWUSR | S_IXUSR)

extern USER user;

BRETT akt_brett;
BRETTNR brettnrW = 0;
BRETTART postfach_art;

#ifdef HAVE_DOSDRIVES
extern char known_drivesS[];
#endif

/*
	smtliche Zugriffsrechte und Flags eines Brettes fr den angegebenen
	User einstellen
*/
void b_set_flags_brett(USER *u, BRETTNR nr)
{
	BRETT *b;

	/*
		Lese- und Schreibberechtigung berprfen und ggf. eintragen.
		wg. der Brettlevel mu dazu die komplette BRETT-Struktur 
		angefordert werden
	*/
	b=br_ind[nr].brett;
	if (!b)
		return;
	b->flags &= ~(B_NEU);
	b->flags |= B_READ | B_WRITE;

	/* Level ok? */
	if (level()<b->lesen_level)
		b->flags &= ~B_READ;
	if (level()<b->schreib_level)
		b->flags &= ~B_WRITE;

	/* Zahlerstatus bercksichtigen */
	if (!u->bezahlt)
	{
		if (b->flags & B_ZAHLER)
			b->flags &= ~B_READ;
		if (b->flags & B_PAYWRITE)
			b->flags &= ~B_WRITE;
	}

	/* Mitgliedschaft */
	if (B_MITLISTE(b->flags))
	{	
		BOOLEAN ist_mitglied=b_tst_access(b->nr,u->nummer);
		if (!ist_mitglied)
		{
			/* Geheime Gruppe: Weder Lesen noch Schreiben ohne
			 * Mitgliedschaft erlaubt */
			if (b->flags & B_GEHEIM)
			{
				b->flags &= ~B_READ;
				b->flags &= ~B_WRITE;
			}
			if (b->flags & B_MITGLIED)
			{
				/* B_MITGLIED: Schreiben nur fr Mitglieder */
				b->flags &= ~B_WRITE;
			}
			/* B_LESMITGLIED: darf B_READ nicht lschen, sonst 
			 * wird das Brett in Infofiles und B* nicht angezeigt
			 * Schreibzugriff wird allerdings immer verwehrt.
			 */
			if (b->flags & B_LESMITGLIED)
				b->flags &= ~B_WRITE;
		}
	}

	/* Brettbearbeiter und Level 9 haben generell Lese- und Schreibrecht */
	if (u->nummer == b->grp_chef || level()==9)
	{
		b->flags |= B_READ;
		b->flags |= B_WRITE;
	}

	/*
	 * Gste bekommen alle Binrbretter als neu angezeigt.
	 */
	if (u->letzter_login < b->datum
		|| (u->level==0 && b->art==BINAER_BRETT && b->datum))
		b->flags |= B_NEU;

} /* b_set_flags_brett */

/*
	smtliche Zugriffsrechte und Flags aller Bretter fr den angegebenen
	User einstellen

	->			u	:	User, fr den die Einstellung gelten soll

	Durch die einmalige Berechnung aller relevanten Flags beim Login des 
	Users kann im nachfolgendem Betrieb auf die relativ zeitaufwendigen 
	berprfungen (insb. bei "BRETT *") verzichtet und stattdessen auf die 
	Flags der internen BRETTINFO-Struktur zugegriffen werden.

	Nachteil:
	ndert der Sysop an der Konsole den Userlevel hat das keine Auswir-
	kungen auf den aktuell eingeloggten Benutzer. Dazu mu diese Funktion 
	erneut aufgerufen werden.
*/
void b_set_flags(USER *u)
{
	BRETTNR nr;

	for (nr = 1; nr < max_brettnr; nr++)
	{
		if (!br_ind[nr].brett)
			continue;
		b_set_flags_brett(u,nr);
	}
} /* b_set_flags */



/*
	Standardbrettstruktur erzeugen
	19.10.93 Frank

	initialisiert eine Brettstruktur auf sinnvolle Standardwerte:
	- Chef lt. Config, sonst Sysop
	- Leselevel 2
	- Schreiblevel 4
	- Textbrett
	- max. 300 Mails
	- 14 Tage Aufbewahrzeit

	todo:
	- Standardbrett ber spezielles Brettflag oder -namen ("STANDARD")
	  vorgeben und bernehmen
	- Trennung zwischen Text- und Binrbrett
*/
void b_default_brett(BRETT *b)
{
	BRETTNR bnr;
	memset(b,0,sizeof(BRETT));

	(void) ub_find_username(&b->grp_chef,box_conf.default_chef);

	if (box_conf.default_brett
		&& (bnr=brettnr(box_conf.default_brett))!=0 
		&& bb_read_brett(bnr,FALSE))
	{
		BRETT *d;
		d=br_ind[bnr].brett;
		b->art=d->art;
		b->partition=d->partition;
		b->lesen_level=d->lesen_level;
		b->schreib_level=d->schreib_level;
		b->flags=d->flags;
		b->art=d->art;
		b->netzstatus=d->netzstatus;
		b->sprache=d->sprache;
		b->ursprungsnetz=d->ursprungsnetz;
		return;
	}
	b->art = TEXT_BRETT;
	b->lesen_level = 2;
	b->schreib_level = 4;
	/* b->aufbewahrzeit = 14; */

} /* b_default_brett */



/*
	feststellen, ob es mind. ein Brett mit neuem Inhalt gibt

	04.03.93 Frank, neu
	27.12.93 Uwe, nur solche Bretter, die man auch lesen kann.
*/
BOOLEAN b_check_new(void)
{
	long i = 1;

	while (i < max_brettnr)
	{
		if (br_ind[i].brett
		&& (br_ind[i].brett->flags & B_READ)
		&& (br_ind[i].brett->flags & B_NEU))
			return TRUE;
		i++;
	}

	return FALSE;

} /* b_check_new */



const char *b_verwalter_S(BRETTNR nr)
{
	static char verwalter[UDAT_NAME_LEN];
	USERNR unr;
	USER u;

	CHECK_BRETTNR(nr,return "");
	if (br_ind[nr].brett->grp_chef 
		&& ub_read_user(br_ind[nr].brett->grp_chef,&u,FALSE))
	{
		strcpy(verwalter,u.name);
		return verwalter;
	}
	else if (!box_conf.default_chef)
		return "";
	if (!ub_find_username(&unr,box_conf.default_chef))
		return "";
	strcpy(verwalter,box_conf.default_chef);
	return verwalter;
} /* b_verwalter_S */


BOOLEAN 
b_remove_brett(BRETTNR bnr)
{
	char folderS[PATH_MAX];
	BRETT *b;
	CHECK_BRETTNR(bnr,return FALSE);
	b=br_ind[bnr].brett;
	mboutput(" -> lsche Inhalt...\r\n");
	/* zwischen Text- und Binrbrettern unterscheiden */
	if (b->art == TEXT_BRETT)
		mdb_remove_group(b->name,FALSE);
	else
	{
		/* ### mssen via Binrverwaltung gelscht werden */
		system(tprintf("rm -f %s*.bin",brettpfad(bnr)));
		system(tprintf("rm -f %s*.txt",brettpfad(bnr)));
	}

	mboutput(" -> lsche Zugriffsliste\r\n");
	(void) unlink(tprintf("%s%s",brettpfad(bnr),"zugriff.dat"));
	(void) unlink(tprintf("%s%s",brettpfad(bnr),"zugriff.lst"));

	mboutput(" -> lsche Verzeichnis\r\n");
	strcpy(folderS,brettpfad(bnr));
	folderS[strlen(folderS)-1] = 0;
	(void) rmdir(folderS);

	mboutput(" -> lsche Eintrag aus brett.dat\r\n");
	bb_remove_brett(bnr); /* bernimmt auch br_ind[b->nr] */
	/* Sortierte Liste der Brettnamen lschen, da sie nun#
	 * nicht mehr stimmt.
	 */
	b_free_sortierung();
	return TRUE;
}

BOOLEAN 
b_create_brett(BRETT *b)
{
	BRETTNR bnr;
	mboutput(" -> installiere neues Brett\r\n");
	if (!bb_create_brett(b))
	{
		mbfehler("kann Brett in der Datenbank nicht anlegen");
		return FALSE;
	}
	bnr=b->nr;

	/*
	 * brettpfad kann noch nicht benutzt werden, da das Brett noch nicht
	 * geschrieben ist und brettart() auf das Brett nicht funktioniert.
	 */
	if (br_ind[bnr].brett->art != BINAER_BRETT)
#ifdef OFFLINE_VERSION
		mdb_create_group(br_ind[bnr].brett->name,FALSE);
#else
		mdb_create_group(br_ind[bnr].brett->name,FALSE,NULL);
#endif
	else
	{
#ifdef HAVE_DOSDRIVES
		char part[] = "\0\0";
		part[0] = 'A' + br_ind[bnr].brett->partition;
		if (!strchr(known_drivesS, part[0]))
			strcat(known_drivesS,part);
#endif
		{
			char pa[PATH_MAX];
			char *p;
			strcpy(pa,brettpfad(bnr));
			p=strrchr(pa,'/');
			if (p && !p[1])
				*p=0;
			mkdir(pa,UMASK);
		}
	}

	(void) unlink(tprintf("%s%s",brettpfad(bnr),"zugriff.dat"));
	if (B_MITLISTE(br_ind[bnr].brett->flags))
		b_add_access(bnr,br_ind[bnr].brett->grp_chef);

	br_ind[bnr].brett->configaenderung=time(NULL);
	bb_write_brett(bnr,TRUE,TRUE);

	/* Sortierte Liste der Brettnamen lschen, da sie nun
	 * nicht mehr stimmt.
	 */
	b_free_sortierung();
	bef_brett_sortierung_maybe_wrong=TRUE;
	return TRUE;
}

