/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
       (might be the wrong person except for offline)

    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
#include "tomsysdep.h"

#include <stdio.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <syslog.h>

#include "prototyp.h"
#include "modem.h"
#include "io.h"
#include "aktdata.h"
#include "user.h"

#ifdef WITH_JOBCONTROL
#include "jobs.h"
#endif

extern int user_out_fd;
extern int user_in_fd;

extern BOOLEAN modem_vorhanden;

static int child_exitstatus;
static pid_t child_pid;
extern int logout_modus;

static RETSIGTYPE 
alarm_handler(int sig TOM_ATTRIB_UNUSED)
{
	/* dummy */
}

static void 
child_handler(int sig TOM_ATTRIB_UNUSED)
{
	int status;

	if (waitpid(child_pid,&status,WUNTRACED|WNOHANG)<=0)
	{
		/* kein child mehr da bzw. Fehler */
		child_exitstatus=1; /* *irgendetwas* ist da falsch! */
		return;
	}

#ifdef SIGCONT
#ifdef WITH_JOBCONTROL
	if (!jc_shell_is_interactive) {
#endif
		/* gestoppte Prozesse wieder starten */
		if (WIFSTOPPED(status))
		{
			switch(WSTOPSIG(status))
			{
				case SIGSTOP:
				case SIGTSTP:
					kill(child_pid,SIGCONT);
				break;
			}
			return;
		}
#ifdef WITH_JOBCONTROL
	}
#endif
#endif

	if (WIFSIGNALED(status))
	{
		child_exitstatus=WTERMSIG(status);
		/*
		 * das passiert unter Linux genau dann, wenn ein Prozess mit 
		 * exit(128) beendet. Das tut Z-Modem gelegentlich.
		 */
		if (!child_exitstatus)
			child_exitstatus=1;
	}
	else
		child_exitstatus=WEXITSTATUS(status);
}

#ifndef HAVE_FORK
/* fork brauchen wir nunmal */
#error call_unix.c ohne fork?
#endif

extern char **environ;

/*
 * externes Programm aufrufen. Wenn irgendeine Routine scheulich aussieht,
 * dann diese.
 */
int 
call_extern(const char *command,
	const char *workdir, time_t killzeit, 
	BOOLEAN ignore_exitcode, BOOLEAN as_user)
{
	char old_zuletzt[10];
	BOOLEAN carrier_lost;
	RETSIGTYPE (*old_alarm)(int);
	int i;
#ifdef WITH_JOBCONTROL
	job_t *job;
#endif

	io_drain(user_out_fd);
	io_save_restore_attribs(user_out_fd,TRUE);
	io_mode_sane(user_out_fd);
	io_set_ONLCR(1,TRUE); /* NL -> CR/NL */
	if (akt_connection.sends_cr && akt_connection.sends_lf)
		io_set_IGNCR(0,TRUE);
	else if (akt_connection.sends_cr)
		io_set_ICRNL(0,TRUE);
	else
		io_set_IGNCR(0,TRUE); /* just for safety */

	child_exitstatus=-1;

#ifdef WITH_JOBCONTROL
	if (jc_shell_is_interactive) {
		job=XMALLOC(sizeof(job_t));
		memset(job,0,sizeof(job));
		job->next=jc_first_job;
		jc_first_job=job;
	}
#endif

	/* Der setuid mu fr beide Prozesse geschehen, da das Kind sonst nicht
	 * mit kill() beendet werden kann.
	 */
	if (as_user)
		set_user_ids();

	signal(SIGCHLD,child_handler);
	/*
	 * Wir benutzen vfork statt fork. Das bedingt, da wir keine 
	 * Operationen der Standardlibrary aufrufen (die ist unter Mint
	 * nicht threadsafe (und anderswo auch nicht unbedingt)), sondern
	 * uns auf Systemcalls beschrnken. Aber mehr ist auch nicht
	 * ntig.
	 */
	child_pid=vfork();

	if (child_pid==0)
	{
		/* child */
		const char *new_argv[4];
		const char * meldung;
		if (workdir && *workdir)
		{
			if (chdir(workdir))
			{
				meldung="kann Verzeichnis nicht wechseln\r\n";
				write(2,meldung,strlen(meldung));
				exit(1);
			}
		}
#ifdef WITH_JOBCONTROL
		if (jc_shell_is_interactive) {
			/* Put the process into the process group and give the process group
			 * the terminal, if appropriate.
			 * This has to be done both by the shell and in the individual
			 * child processes because of potential race conditions.  */
			child_pid = getpid ();
			setpgid (child_pid, child_pid);
			/* into foregrouund */
			tcsetpgrp (jc_shell_terminal, child_pid);

			/* Set the handling for job control signals back to the default.  */
			signal (SIGINT, SIG_DFL);
			signal (SIGQUIT, SIG_DFL);
			signal (SIGTSTP, SIG_DFL);
			signal (SIGTTIN, SIG_DFL);
			signal (SIGTTOU, SIG_DFL);
			signal (SIGCHLD, SIG_DFL);
		}
#endif
		/* alle Filedescriptoren schlieen, die sind mglicherweise
		 * wertvoll.
		 */
#ifndef OPEN_MAX
		for (i=3;i<256;i++)
#else
		for (i=3;i<OPEN_MAX;i++)
#endif
			close(i);

#if 0
		if (akt_anruf.cr_mapping)
			io_set_cr_mapping(0,TRUE);
#endif
#define SHELL_PATH  "/bin/sh"   /* Path of the shell.  */
#define SHELL_NAME  "sh"        /* Name to give it.  */
		new_argv[0] = SHELL_NAME;
		new_argv[1] = "-c";
		new_argv[2] = command;
		new_argv[3] = NULL;

		if (as_user && akt_connection.user_environ)
			(void) execve (SHELL_PATH, (char *const *) new_argv,
				(char *const *) akt_connection.user_environ);
		else
			(void) execv (SHELL_PATH, (char *const *) new_argv);
		/* _exit gibt im Gegensatz zu exit keine Stdlib-Strukturen frei,
		 * wie z.B. Streamdescriptors (die Dateien *sind* schon geschlossen)
		 * Dafr ist exit() ANSI und _exit Posix.
		 */
		meldung = "exec /bin/sh failed\n";
		write(2,meldung,strlen(meldung));
		_exit (127);
	}
	else if (child_pid<0)
	{
		io_save_restore_attribs(user_out_fd,FALSE);
		mboutput ("fork() gescheitert, kann externes Programm nicht starten\r\n");
		syslog (LOG_NOTICE, "fork() failed (%s)",
			strerror (errno));
		return 1; /* Fehler */
	}
#ifdef WITH_JOBCONTROL
	if (jc_shell_is_interactive) {
		/* Put the process into the process group and give the process group
		 * the terminal, if appropriate.
		 * This has to be done both by the shell and in the individual
		 * child processes because of potential race conditions.  */
		setpgid (child_pid, child_pid);
	}
#endif
	/* parent */
	carrier_lost=FALSE;

	if (child_exitstatus==-1) {
		/* das kann != -1 sein wenn sich das child beendet bevor
		 * der Mutterprozess die Prozess-ID erhlt */

		strncpy(old_zuletzt,akt_anruf.zuletzt,sizeof(akt_anruf.zuletzt));
		if (!strcmp(old_zuletzt,"BOX"))
			strcpy(akt_anruf.zuletzt,"EXT");
		
#ifdef EXPLIZITE_CARRIERABFRAGE
		old_alarm=signal(SIGALRM,alarm_handler);
		while (TRUE)
		{
			/* der *muss* hier stehen! */
			signal(SIGALRM,alarm_handler);
			alarm(10);
			pause();
			
			if (child_exitstatus!=-1)
				break;
			
			if (!carrier_still_up())
			{
				carrier_lost=TRUE;
				modem_vorhanden=FALSE;
				break;
			}
			
			if (killzeit!=(time_t) 0 && time(NULL)>killzeit)
				break;
		}
		
		alarm(0);
		signal(SIGALRM,old_alarm);
#else
		old_alarm=signal(SIGALRM,alarm_handler);
		if (killzeit != (time_t) 0)
			alarm(killzeit - time(NULL));
		pause();
		carrier_checkpoint(FALSE);
		if (logout_modus)
		{
			carrier_lost=TRUE;
			modem_vorhanden=FALSE;
		}
		alarm(0);
		signal(SIGALRM,old_alarm);
#endif
	} else {
		/* don't let a zombie survive */
		int status;
		waitpid(child_pid,&status,WUNTRACED|WNOHANG);
	}
	signal(SIGCHLD,SIG_DFL);

	io_save_restore_attribs(user_out_fd,FALSE);

	if (carrier_lost)
	{
		int status;
		RETSIGTYPE (*old_hup)(int);
		if (0==kill(child_pid,SIGKILL))
		{
			(void) waitpid(child_pid,&status,WUNTRACED);
		}
		/* Hangupsignal an die ganze Prozessgruppe */
		old_hup=signal(SIGHUP,SIG_IGN);
		kill(-getpgrp(),SIGHUP);
#ifdef SIGCONT
		kill(-getpgrp(),SIGCONT);
#endif
		signal(SIGHUP,old_hup);
		/* kill(getpid(),SIGHUP); */
		child_exitstatus=1;
		user_in_fd=-1;
		user_out_fd=-1;
	}

	if (child_exitstatus==-1)
	{
		int status;
		RETSIGTYPE (*old_hup)(int);

		/* Das war eine Onlinezeitberschreitung. 
		 * Und die Box ist keine freundliche nette Mutter aus dem 
		 * Bilderbuch, und auch keine Rabenmutter, sondern eine
		 * passionierte Kindermrderin.
		 * Wir schicken nun dem geforkten Kind ein SIGKILL. Das Kind
		 * beendigt sich und das OS sendet SIGHUP/SIGCONT an alle
		 * Mitglieder der Prozessgruppe.
		 */
		strcpy(akt_anruf.zuletzt,"TIM");
		child_exitstatus=1;
		if (0==kill(child_pid,SIGHUP))
			(void) waitpid(child_pid,&status,WUNTRACED);
		/* Hangupsignal an die ganze Prozessgruppe */
		old_hup=signal(SIGHUP,SIG_IGN);
		kill(-getpgrp(),SIGHUP);
#ifdef SIGCONT
		kill(-getpgrp(),SIGCONT);
#endif
		signal(SIGHUP,old_hup);
		user_in_fd=-1;
		user_out_fd=-1;
	}
	else
	{
		strncpy(akt_anruf.zuletzt,old_zuletzt,sizeof(akt_anruf.zuletzt));
	}

	io_drain(user_out_fd);

	if (child_exitstatus && !ignore_exitcode)
	{
		fprintf(stderr,"Externes Programm beendete sich mit Code %d\n",
			child_exitstatus );
		mbprintf("Externes Programm beendete sich mit Code %d\r\n",
			child_exitstatus );
	}

	if (as_user)
		set_box_ids();

	return ignore_exitcode ? 0 : child_exitstatus;

} /* call_extern */


