/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/

#include <portab.h>
#include "tomsysdep.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <syslog.h>
#include <fnmatch.h>

#include "prototyp.h"
#include "io.h"			/* wg. io_set_baud */
#include "modem.h"
#include "box_conf.h"
#include "aktdata.h"
#include "tompaths.h"
#include "libquark.h"
#include "log.h"

static int test_connect(char *test, BOOLEAN *errorp);
static int check_clid(char *from, char *to);

/*
 * Wie der Name schon sagt. Returned TRUE bei Connect, sonst False.
 */
int 
warte_auf_carrier(int fd)
{
	time_t startzeit;
	char modemdaten[80];
	int clid_command=0;

	akt_connection.carrierstring=NULL;
	akt_connection.carrierstringraw=NULL;
	akt_connection.carriercmd=NULL;
	akt_connection.clid_from=NULL;
	akt_connection.clid_to=NULL;

	if (!akt_connection.is_modem)
	{
		/* das sollte nicht passieren */
		syslog(LOG_INFO,"trying to make connect, but no modem");
		akt_connection.carrierstring=XSTRDUP("????");
		return CARRIER_LOGIN;
	}

	/* in-band-signaling :-/
	 * Das Zyxel liefert nach dem *ersten* Ring die Caller-ID. Wenn wir den
	 * ersten verpassen (zu spaet gestartet?) erhalten wir also keine CLID.
	 */
	if (box_conf.modem_get_clid &&
		0==stricmp(box_conf.modem_get_clid,"zyxel")) {
		int st;
		*modemdaten = 0;
		while (1) {
			char zeichen;
			size_t len;
			struct timeval zeit;
			fd_set set;
			zeit.tv_sec = 0;
			zeit.tv_usec = 100000;
			FD_ZERO(&set);
			FD_SET(fd,&set);

			st = select(fd+1,&set,NULL,NULL,&zeit);
			if (st==-1 || st == 0)
				break;
			if (read(fd,&zeichen,1) <= 0)
			{
				syslog(LOG_INFO,"trying to make connect, read: %m");
				akt_connection.carrierstring=XSTRDUP("RERR"); /* Read ERRor */
				return CARRIER_ERROR;
			}

			len = strlen(modemdaten);
			if (!len && isspace(zeichen)) /* Spaces am Anfang berspringen */
				continue;
			modemdaten[len++] = zeichen;
			modemdaten[len]   = 0;
			if (len>70)
			{
				memmove(modemdaten,modemdaten+10,70);
				len-=10;
			}
		};

		if (*modemdaten)  {
			char *p=modemdaten;
			while (isspace(*p))
				p++;
			p=strstr(modemdaten,"FM:");
			if (p) {
				char *q;
				char *r;
				q=strstr(p,"TO:");
				r=strpbrk(p,"\r\n");
				if (r)
					*r=0;
				write_info_protokoll(INFOPROT_FREE,NULL,p);
				if (q)
					q[-1]=0;
				strtrim(p);
				strtrim(q);
				akt_connection.clid_from=XSTRDUP(p+3);
				akt_connection.clid_to=XSTRDUP(q+3);
			} else if ((p=strstr(modemdaten,"TO:"))!=NULL) {
				char *r;
				r=strpbrk(p,"\r\n");
				if (r)
					*r=0;
				write_info_protokoll(INFOPROT_FREE,NULL,p);
				strtrim(p);
				akt_connection.clid_to=XSTRDUP(p+3);
			}
			if (akt_connection.clid_to || akt_connection.clid_from) {
				clid_command=check_clid(akt_connection.clid_from,akt_connection.clid_to); 
				if (clid_command==3) { /* reject */
					akt_connection.carrierstring=XSTRDUP("REJE");
					akt_connection.carrierstringraw=XSTRDUP("REJECT");
					io_flush(fd);
					if (box_conf.modem_takeoff) {
						send_modem(fd,box_conf.modem_takeoff);
						send_modem(fd,"\r");
					}
					io_drain(fd);
					syslog(LOG_INFO,"CLID %s/%s -> reject",
						akt_connection.clid_from ? akt_connection.clid_from : "NONE",
						akt_connection.clid_to ? akt_connection.clid_to : "NONE");
					return CARRIER_ERROR;
				} else if (clid_command==2) {
					/* exec now */
					syslog(LOG_INFO,"CLID %s/%s -> exec-now %s",
						akt_connection.clid_from ? akt_connection.clid_from : "NONE",
						akt_connection.clid_to ? akt_connection.clid_to : "NONE",
						akt_connection.carriercmd ? akt_connection.carriercmd : "NONE");
					if (akt_connection.carriercmd)
						return CARRIER_SPEC;
				}
			}
		}
	}

	/* Mll (CR/LF) nach "RING" verwerfen */
	io_flush(fd);
	answer_modem(fd);
	io_drain(fd);

	*modemdaten = 0;
	startzeit = time(NULL);

	while (time(NULL)-startzeit < box_conf.modem_connect_wartezeit)
	{
		char zeichen;
		size_t len;

		if (!io_charavail(fd))
		{
			msleep(100);
			continue;
		}
		if (read(fd,&zeichen,1) <= 0)
		{
			syslog(LOG_INFO,"trying to make connect, read: %m");
			akt_connection.carrierstring=XSTRDUP("RERR"); /* Read ERRor */
			return CARRIER_ERROR;
		}

		len = strlen(modemdaten);
		if (!len && isspace(zeichen)) /* Spaces am Anfang berspringen */
			continue;

		modemdaten[len++] = zeichen;
		modemdaten[len]   = 0;
		if (len>70)
		{
			memmove(modemdaten,modemdaten+10,70);
			len-=10;
		}

		if (strstr(modemdaten,box_conf.modem_nocarrier))
		{
			akt_connection.carrierstring=XSTRDUP("NCAR");	/* No CARrier */
			akt_connection.carrierstringraw=XSTRDUP(modemdaten);
			return CARRIER_ERROR;
		}

		if (strstr(modemdaten,box_conf.modem_ring))	/* ? RING */
		{
			/* Das kann nur passieren, wenn das Modem ein ATA berhrt
			 * hat, und das ist *eigentlich* "this can not happen".
			 */
			io_drain(fd);
			answer_modem(fd);
			strcpy(modemdaten," ");
			len = 1;								/* ### 07.06.94 Frank, neu */
		}
		/* smtliche CONNECT-Meldungen durchsuchen */
		if (zeichen == '\r' || zeichen == '\n')
		{
			BOOLEAN errorflag;
			int erg;
			erg=test_connect(modemdaten,&errorflag);
			if (errorflag==0 && clid_command==1) {
				syslog(LOG_INFO,"CLID %s/%s -> exec %s",
					akt_connection.clid_from ? akt_connection.clid_from : "NONE",
					akt_connection.clid_to ? akt_connection.clid_to : "NONE",
					akt_connection.carriercmd ? akt_connection.carriercmd : "NONE");
				if (box_conf.modem_set_baud && akt_connection.carrier_baud)
				{
					io_set_baud(fd,akt_connection.carrier_baud);
					io_drain(fd);		/* notwendig? */
				}
				if (akt_connection.carriercmd)
					return CARRIER_SPEC;
				return CARRIER_LOGIN;
			}
			if (1==erg)
			{
				if (box_conf.modem_set_baud && akt_connection.carrier_baud)
				{
					io_set_baud(fd,akt_connection.carrier_baud);
					io_drain(fd);		/* notwendig? */
				}
				if (akt_connection.carriercmd)
					return CARRIER_SPEC;
				return CARRIER_LOGIN;
			}
			else if (2==erg)
			{
				/* ignore */
				len=0;
				*modemdaten=0;
				continue;
			}
			if (errorflag)
			{
				syslog(LOG_ERR,"out of memory");
				fprintf(stderr,"out of memory\r\n");
				if (!akt_connection.carrierstring)
					akt_connection.carrierstring=XSTRDUP("EMEM");
				return CARRIER_ERROR;
			}
			/* in jedem Fall nach CR oder LF den Connectversuch beenden. */

			/*
			 * an dieser Stelle haben wir mal den Carrier getestet - bis
			 * wir auf eine Maschine trafen, die das Carrier-erkannt-Signal
			 * etwas spter setzte. Brauchen tun wir es ohnehin nicht,
			 * sptestens bei "sind sie eingetragenes Looser" stellen
			 * wir das sowieso fest.
			 */

			{
				/* unbekannter Connect */
				char *pos;
				FILE *err_f;

				/* Mgliche Racecondition im Multiportbetrieb. Aber Locking
				 * wre wohl doch bertrieben 
				 */
				err_f = xfopen("err_con.dat","a+");
				while ((pos = strpbrk(modemdaten,"\r\n")) != NULL)
					memcpy(pos,pos+1,strlen(pos));
				fprintf(err_f," ? %s %s >%s\n",dd_mm_yy(),hh_mm_ss(),modemdaten);
				fclose(err_f);

				akt_connection.carrierstring=XSTRDUP("NKNC");	/* No KNown Connect */
				return CARRIER_LOGIN;
			}
		}
	}

	/* das ist ein Timeout */
	akt_connection.carrierstring=XSTRDUP("TIME");	/* Timed out */
	return CARRIER_ERROR;
}


/*
 * liefert:
 * 0 -> keine Besonderheit
 * 1 -> exec Programm nach Connect
 * 2 -> exec Programm, ohne Versuch, den Connect herzustellen
 * 3 -> reject (na, so etwa)
 */
static int
check_clid(char *from, char *to)
{
	FILE *con_f = NULL;
	char eintrag[1024];					/* Eintrag aus connect.cnf */

	if (tom_port_config_dir)
		con_f = fopen(tprintf("%s%s",tom_port_config_dir,"clid.cnf"),
					  BIN_RW_OPEN);
	else
		con_f = NULL;

	if (!con_f)
		con_f = fopen(tprintf("%s%s",box_conf.config_dir,"clid.cnf"),
					  BIN_RW_OPEN);
	if (!con_f)
		return 0;

syslog(LOG_INFO,"check_clid: check %s/%s",from?from: "NONE", to ? to : "TO");

	while (fgets(eintrag,sizeof(eintrag),con_f))
	{
		char *p;
		char *p_from,*p_to,*p_command,*p_what;

		p=strpbrk(eintrag,"\r\n");
		if (p)
			*p=0;

		strtrim(eintrag);

		if (!*eintrag || *eintrag == ';' || *eintrag ==';')
			continue;
		p_command=NULL;

		/*
			nacheinander CONNECT-String, Baudrate, Protokollstring, erwartete 
			CPS-Rate und optional Fehlerkorrekturflag auslesen
		*/
		if ((p_from  = strtok(eintrag,"#")) == NULL)
			continue;
		if ((p_to = strtok(NULL,"#")) == NULL)
			continue;
		if ((p_what  = strtok(NULL,"\t ")) == NULL)
			continue;
		p_command  = strtok(NULL,"#");
		strtrim(p_from);
		strtrim(p_to);
		strtrim(p_what);
		if (p_command) {
			strtrim(p_command);
		}
		if (!*p_what || !*p_from || !*p_to)
			continue;

		if (p_from && *p_from) {
			if (0!=fnmatch(p_from,from ? from : "NONE" ,0))
				continue;
		}
		if (p_to && *p_to) {
			if (0!=fnmatch(p_to,to ? to : "NONE" ,0))
				continue;
		}
		if (!stricmp(p_what,"reject")) {
			fclose(con_f);
syslog(LOG_INFO,"check_clid: return 3");
			return 3;
		}
		if (!p_command)
			continue;
		if (!stricmp(p_what,"exec-now")) {
			akt_connection.carriercmd=XSTRDUP(p_command);
			fclose(con_f);
syslog(LOG_INFO,"check_clid: return 2");
			return 2;
		}
		if (!stricmp(p_what,"exec")) {
			akt_connection.carriercmd=XSTRDUP(p_command);
			fclose(con_f);
syslog(LOG_INFO,"check_clid: return 1");
			return 1;
		}
	}

syslog(LOG_INFO,"check_clid: return 0");
	fclose(con_f);
	return 0;
}

/* Sucht String in der connect.cnf
 * Liefert:
 * 0 -> unbekannt
 * 1 -> CONNECT
 * 2 -> ignore this string.
 */
static int 
test_connect(char *test, BOOLEAN *errorp)
{
	static FILE *con_f = NULL;
	size_t len;
	char *p_err;
	char eintrag[1024];					/* Eintrag aus connect.cnf */
	char *p_msg,*p_baud,*p_str,*p_cps,*p_cmd;


	if (tom_port_config_dir)
		con_f = fopen(tprintf("%s%s",tom_port_config_dir,"connect.cnf"),
					  BIN_RW_OPEN);
	else
		con_f = NULL;

	if (!con_f)
		con_f = fopen(tprintf("%s%s",box_conf.config_dir,"connect.cnf"),
					  BIN_RW_OPEN);
	if (!con_f)
		return 0;

	*errorp=FALSE;

	/* Leerzeichen und -zeilen an Anfang und Ende berlesen */
	while (isspace(*test))
		test++;
	while ((len = strlen(test)) != 0)
		if (isspace(test[len-1]))
			test[len-1] = 0;
		else
			break;

	rewind(con_f);
	while (fgets(eintrag,sizeof(eintrag),con_f))
	{
		char *p;
		p=strpbrk(eintrag,"\r\n");
		if (p)
			*p=0;

		if (!strlen(eintrag) || *eintrag == ';')
			continue;
		p_cmd=NULL;

		/*
			nacheinander CONNECT-String, Baudrate, Protokollstring, erwartete 
			CPS-Rate und optional Fehlerkorrekturflag auslesen
		*/
		if ((p_msg  = strtok(eintrag,"#")) == NULL)
			continue;
		if ((p_baud = strtok(NULL,"#")) == NULL)
			continue;
		if ((p_str  = strtok(NULL,"#")) == NULL)
			continue;
		if ((p_cps  = strtok(NULL,"#")) == NULL)
			continue;
		if ((p_err  = strtok(NULL,"#")) != NULL)
			p_cmd  = strtok(NULL,"#");

		strtrim(p_msg);
		if (!strcmp(test,p_msg) || !strnicmp(test,"@CLID",5)) {
			char *last;
			strtrim(p_baud);
			strtrim(p_str);
			strtrim(p_cps);
			if (p_cmd)
			{
				strtrim(p_cmd);
				if (0==strcasecmp(p_cmd,"ignore"))
					return 2;
				if (!akt_connection.carriercmd)
					akt_connection.carriercmd=strdup(p_cmd);
				if (!akt_connection.carriercmd)
				{
					*errorp=TRUE;
					return 0;
				}
			}
			akt_connection.error_free=FALSE;
			if (p_err)
			{
				strtrim(p_err);
				if (atoi(p_err))
					akt_connection.error_free=TRUE;
			}
			akt_connection.carrier_baud = atol(p_baud);
			akt_connection.carrier_cps = atol(p_cps);
			akt_connection.carrierstring=strdup(p_str);
			if (!akt_connection.carrierstring)
			{
				*errorp=TRUE;
				return 0;
			}

			/*
			 * den zum Connect fhrenden String isolieren und aufbewahren 
			 */
			p=test+strlen(test)-1;
			while (*p=='\r' || *p=='\n')
				*p--=0;
			last=test;
			while ((p=strpbrk(last,"\r\n"))!=NULL)
				last=p+1;
			akt_connection.carrierstringraw=strdup(last);
			if (!akt_connection.carrierstringraw)
			{
				*errorp=TRUE;
				return 0;
			}

			fclose(con_f);

			return 1;
		}
	}

	fclose(con_f);
	return 0;

} /* test_connect */
