/* cslib.c : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
       (might be the wrong person except for offline)

    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <stdio.h>
#include <ctype.h>

#include "cslib.h"

#ifdef __TOS__
#include "atarist.h"
#else
#include "iso-8859-1.h"
#endif

#ifdef __PUREC__
#define strcasecmp stricmp
#define strncasecmp strnicmp
#endif

	/* Null -> kein Zeichensatz eingestellt, Ausgaben entumlauten */
static cstl_t *tr_cs;

static unsigned char nativetable[256];

int (*ausprintf)(const char *, ... )=printf;

cstl_t *get_charsettable(void)
{
	return cstl_table;
}

void set_charset(cstl_t *cs)
{
	tr_cs=cs;
}

void unset_charset(void)
{
	tr_cs=get_charset_from_name("ascii");
}

void charset_init(int (*out_func)(const char *, ... ))
{
	int i;
	cstl_t *p;
	if (out_func != (int (*)(const char *, ... )) 0)
		ausprintf=out_func;

	/* bersetzungstabelle fr den native-charset */
	for (i=0;i<256;i++)
		nativetable[i]=i;

	/* Aliaspointer zuordnen */
	p=cstl_table;
	while (p && p->name)
	{
		if (p->flags & CSF_ALIAS)
		{
			/* Ist ein alias. Jetzt das Original suchen */
			cstl_t *p2;
			p2=cstl_table;
			while (p2->csid)
			{
				if (p2->csid==p->csid && !(p2->flags & CSF_ALIAS))
				{
					p->aliasp=p2;
					break;
				}
				p2++;
			}
		}
		else if (p->flags & CSF_NATIVE)
		{
			p->tr_to=nativetable;
			p->tr_from=nativetable;
		}
		p++;
	}
}

cstl_t *get_charset_from_name(const char *charsetname)
{
	cstl_t *p;

	p=cstl_table;
	while (p && p->name)
	{
		if (!strcasecmp(p->name, charsetname))
		{
			/* Gefunden. */
			cstl_t *p2;
			/* Wenn kein Alias */
			if (!(p->flags & CSF_ALIAS))
			{
				return p;
			}
			p2=cstl_table;
			while (p2->csid)
			{
				if (p2->csid==p->csid && !(p2->flags & CSF_ALIAS))
					return p2;
				p2++;
			}
			return NULL; /* kann eigentlich nicht passieren */
		}
		p++;
	}
	return NULL; /* nicht gefunden */
}

int get_charset_nr_from_name(const char *charsetname)
{
	cstl_t *p;

	p=cstl_table;
	while (p && p->name)
	{
		if (!strcasecmp(p->name, charsetname))
		{
			/* Gefunden. */
			return p->csid;
		}
		p++;
	}
	return 0; /* nicht gefunden */
}

cstl_t *get_charset_from_nr(int charsetid)
{
	cstl_t *p;

	p=cstl_table;
	while (p && p->name)
	{
		if (charsetid==p->csid)
			return p;
		p++;
	}
	return NULL; /* nicht gefunden */
}

/*
	Basisroutine zur Umwandlung von Blcken.
	to_foreign: Aus eigenem in fremden Zeichensatz wandeln.
*/
static size_t
int_block_translate(int to_foreign, unsigned char *to, 
	const unsigned char *from, size_t len)
{

	if (!len)
		return len;

	/* 
		Wenn kein Zeichensatz initialisiert ist oder entumlautet werden
		soll
	*/
	if (!tr_cs || tr_cs->flags & CSF_ENTUMLAUTEN)
	{
		/* 
			Wenn vom fremden in unseren eigenen Zeichensatz gewandelt
			werden soll, dann einfach kopieren
		*/
		if (!to_foreign)
		{
			memcpy(to,from,len);
			return len;
		}
		else
		{
			/* entumlauten */
			size_t i;
			size_t l;
			const unsigned char *pf;
			unsigned char *pt;
			pf=from;
			pt=to;
			l=0;
			for (i=0;i<len; i++)
			{
				switch(*pf)
				{
				case (unsigned char) '':
					*pt++='u';
					*pt++='e';
					l+=2;
					break;
				case (unsigned char) '':
					*pt++='a';
					*pt++='e';
					l+=2;
					break;
				case (unsigned char) '':
					*pt++='o';
					*pt++='e';
					l+=2;
					break;
				case (unsigned char) '':
					*pt++='U';
					*pt++='e';
					l+=2;
					break;
				case (unsigned char) '':
					*pt++='A';
					*pt++='e';
					l+=2;
					break;
				case (unsigned char) '':
					*pt++='O';
					*pt++='e';
					l+=2;
					break;
				case (unsigned char) '':
					*pt++='s';
					*pt++='s';
					l+=2;
					break;
				default:
					*pt++=*pf;
					l++;
					break;
				} /* switch */
				pf++;
			} /* for */
			return l;
		} /* entumlauten? */
	}
	else if (tr_cs->flags & CSF_NATIVE)
	{
		/* 
			wenn der Fremde Zeichensatz "native" ist, dann einfach
			kopieren.
		*/
		memcpy(to,from,len);
		return len;
	}
	else
	{
		register const unsigned char *pf;
		register unsigned char *pt;
		register unsigned char const *tab;
		register size_t i;
		if (to_foreign)
			tab=tr_cs->tr_to;
		else
			tab=tr_cs->tr_from;
		pt=to;
		pf=from;
#if 0
		for (i=0;i<len;i++)
		{
			*pt++=tab[*pf++];
		}
#else
#define WORK *pt++=tab[*pf++]
		i=(len+16-1) >> 4;
		switch(len & (16 - 1)) {
		case 0: do {
			WORK;  
		case 15: WORK; 
		case 14: WORK; 
		case 13: WORK; 
		case 12: WORK; 
		case 11: WORK; 
		case 10: WORK; 
		case 9: WORK; 
		case 8: WORK; 
		case 7: WORK; 
		case 6: WORK;  
		case 5: WORK;
		case 4: WORK;  
		case 3: WORK;
		case 2: WORK;  
		case 1: WORK;
		} while (--i);
		}
#undef WORK
#endif
		/* *pt='\0'; */
		return len;
	}
}

size_t block_to_native(unsigned char *to,
	const unsigned char *from,size_t maxlen, size_t len)
{
	if (len>maxlen)	
		return (size_t) -1;
	return int_block_translate(0,to,from,len);
}

size_t block_to_foreign(unsigned char *to,
	const unsigned char *from,size_t maxlen, size_t len)
{
	if (len>maxlen)	
		return (size_t) -1;
	return int_block_translate(1,to,from,len);
}

size_t string_to_native(unsigned char *to,
	const unsigned char *from,size_t maxlen)
{
	size_t len;
	len=strlen((const char *)from)+1;
	if (len>maxlen)	
		return (size_t) -1;
	return int_block_translate(0,to,from,len)-1; /* -1 wg. \0 */
}

size_t string_to_foreign(unsigned char *to,
	const unsigned char *from,size_t maxlen)
{
	size_t len;
	len=strlen((const char *)from)+1; /* +1 wg. \0 */
	if (len>maxlen)	
		return (size_t) -1;
	return int_block_translate(1,to,from,len)-1; /* -1 wg. \0 */
}

size_t char_to_native(unsigned char *to, unsigned char from)
{
	/* Kein Zeichensatz angegeben? */
	to[1]='\0';
	if (!tr_cs || tr_cs->flags & CSF_ENTUMLAUTEN)
	{
		/* dann kopieren */
#if 0
		to[0]=from;
#else
		to[0]=from & 0x7f;
#endif
	}
	else
		*to=tr_cs->tr_from[from];
	return 1;
}

size_t char_to_foreign(unsigned char *to, unsigned char from)
{
	to[1]='\0';
	/* Kein Zeichensatz angegeben? */
	if (!tr_cs || tr_cs->flags & CSF_ENTUMLAUTEN)
	{
		/* dann kopieren */
		switch(from)
		{
		case (unsigned char) '':
			*to='u';
			to[1]='e';
			to[2]='\0';
			break;
		case (unsigned char) '':
			*to='a';
			to[1]='e';
			to[2]='\0';
			break;
		case (unsigned char) '':
			*to='o';
			to[1]='e';
			to[2]='\0';
			break;
		case (unsigned char) '':
			*to='U';
			to[1]='e';
			to[2]='\0';
			break;
		case (unsigned char) '':
			*to='A';
			to[1]='e';
			to[2]='\0';
			break;
		case (unsigned char) '':
			*to='O';
			to[1]='e';
			to[2]='\0';
			break;
		case (unsigned char) '':
			*to='s';
			to[1]='s';
			to[2]='\0';
			break;
		default:
			*to=from;
			to[1]='\0';
			break;
		} /* switch */
		return 2;
	}
	else
		*to=tr_cs->tr_from[from];
	return 1;
}


void cst_list(int mit_alias)
{
	cstl_t *cs;
	cs=cstl_table;
	while (cs && cs->name)
	{
		if (cs->flags & CSF_ALIAS && !mit_alias)
		{
			cs++;
			continue;
		}
		ausprintf("%3d %-12.12s ",cs->csid, cs->name);
		if (cs->flags)
		{
			int cnt;
			ausprintf("[");
			cnt=0;
			if (cs->flags & CSF_NATIVE)
			{
				ausprintf("keine Wandlung ntig");
				cnt++;
			}
			if (cs->flags & CSF_ENTUMLAUTEN)
			{
				ausprintf("%s%s",cnt ? ", " : "", 
					"entumlauten bei Ausgabe");
				cnt++;
			}
			if (cs->flags & CSF_ALIAS)
			{
				ausprintf("%s%s%s",cnt ? ", " : "",
					"Alias fr ",cs->aliasp->name);
				cnt++;
			}
			ausprintf("]");
		} /* Wenn Flags angegeben */
		ausprintf("\r\n");
		cs++;
	} /* while */
}
