/* cslib.h : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
       (might be the wrong person except for offline)

    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
#ifndef CS_LIB_H
#define CS_LIB_H

/*
	Spezielle Zeichensatz-IDs:
	0: illegal.
	1: MausCode (z.Z. nicht implementiert, aber reserviert)
	1..99: reserviert.
*/
#define CSF_NATIVE 		1
#define CSF_ENTUMLAUTEN	2
#define CSF_ALIAS		4

typedef struct cstl_t_struct
{
	const char *name;
	int flags;
	int csid;
	unsigned char const *tr_from;
	unsigned char const *tr_to;
	struct cstl_t_struct *aliasp;
} cstl_t; /* CharSetTransLate_Type */

cstl_t *get_charset_from_name(const char *charsetname);
int get_charset_nr_from_name(const char *charsetname);
cstl_t *get_charset_from_nr(int charsetid);

/* 
	Mu unbedingt aufgerufen werden. Bei Verwendung aus der Quark
	mu mbprintf bergeben werden. Sonst kann auch 0 bergeben werden,
	dann wird printf verwendet.
 */
void charset_init(int (*out_func)(const char *, ... ));

void set_charset(cstl_t *cs);
	/* Umlautwandlung deaktivieren -> keine Umlaute bei Ausgabe */
void unset_charset(void);
cstl_t *get_charsettable(void);

void cst_list(int mit_alias);

/* cs == NULL -> "Keine Umlaute" */
/* 
	maxlen gibt die maximale Lnge des Resultats an (= Lnge des 
	Ausgabepuffers)
	Wenn das Resultat lnger ist als maxlen, scheitert die Umwandlung,
	die Funktionen liefern (size_t) -1 zurck.
	Die Ausgabe kann durchaus lnger werden als als die Eingabe
	(keine Umlaute, Mauscode ...)
	Auf der sicheren Seite ist man, wenn der Ausgabepuffer viermal so
	gro ist wie die Eingabe.
*/
size_t block_to_native(unsigned char *to, const unsigned char *from, 
	size_t max, size_t len);
size_t block_to_foreign(unsigned char *to, const unsigned char *from, 
	size_t max, size_t len);
size_t string_to_native(unsigned char *to, const unsigned char *from, 
	size_t max);
size_t string_to_foreign(unsigned char *to, const unsigned char *from, 
	size_t max);
/* auch hier: to==NULL -> statischer Puffer. 
	to braucht hchstens 4 Byte zu sein 
*/
size_t char_to_native(unsigned char *to, unsigned char from);
size_t char_to_foreign(unsigned char *to, unsigned char from);

#endif
