/* einfachnachricht.c : *sending a message, simple interface */
/*
    Copyright (C) 1995 1996 Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
*/

#include "tomsysdep.h"

#include "portab.h"
#include <stdio.h>
#include <limits.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <ctype.h>

#include "prototyp.h"
#include "modem.h"
#include "routing.h"
#include "route_int.h"

BOOLEAN
einfachnachricht(BOOLEAN pm, UWORD lv, BOOLEAN bezahlt, const char *fname, ...)
{
	va_list ap;
	struct veinfachpara *vpara=NULL;
	int zahl=0;
	BOOLEAN errorflag;

	va_start(ap,fname);

	while (TRUE)
	{
		char kenn;
		char *str;

		/* char/int wegen promotion rules: ein bergebenes "char" wird zu
		 * int erweitert.
		 */
		zahl++;
		kenn = (char ) va_arg(ap,int);
		if (kenn==0)
			break;
		str = va_arg(ap,char *);
	}
	va_end(ap);

	vpara=XMALLOC(zahl*sizeof(struct veinfachpara));

	va_start(ap,fname);

	zahl=0;
	while (TRUE)
	{
		char kenn;
		char *str;

		/* char/int wegen promotion rules: ein bergebenes "char" wird zu
		 * int erweitert.
		 */
		kenn = (char ) va_arg(ap,int);
		if (kenn==0)
			break;
		str = va_arg(ap,char *);
		vpara[zahl].kenn=kenn;
		vpara[zahl].str=str;
		zahl++;
	}
	va_end(ap);
	vpara[zahl].kenn=0;
	vpara[zahl].str=NULL;

	errorflag=veinfachnachricht(pm, lv, bezahlt, fname, vpara);
	free(vpara);
	return errorflag;
}

BOOLEAN
veinfachnachricht(BOOLEAN pm, UWORD lv, BOOLEAN bezahlt, const char *fname, 
	struct veinfachpara *para)
{
	message_t *msg;
	r_adr_t *adr=NULL;
	BOOLEAN errorflag=FALSE;

	if (!r_init_done)
		r_do_init();

	msg=msg_alloc();
	if (!msg)
	{
		mboutput("out of memory\r\n");
		return FALSE;
	}
	msg->fname=strdup(fname);
	if (!msg->fname)
	{
		msg_free(msg);
		mboutput("out of memory\r\n");
		return FALSE;
	}

	while (para->kenn)
	{
		char kenn=para->kenn;
		char *str=para->str;
		para++;

		switch(kenn)
		{
#define EINFACH(kennz,which) case kennz: strcpy(msg->which,str); break
#define DYN_EINFACH(kennz,which) case kennz: \
	if (msg->which) free(msg->which); msg->which=XSTRDUP(str); break;
		EINFACH('#',msg_id);
		EINFACH('-',ref_id);
		EINFACH('S',sender);
		EINFACH('V',absender);
		EINFACH('T',reply_to);
		EINFACH('W',betreff);
		EINFACH('E',eingabezeit);
		EINFACH('B',bearbeitungsstatus);
		EINFACH('O',organisation);
		EINFACH('N',realname);
		DYN_EINFACH('Y',gateway_zeile);
		DYN_EINFACH('p',path_zeile);
		DYN_EINFACH('X',flags_zeile);
		DYN_EINFACH('M',mimeinfo);
#undef EINFACH
#undef DYN_EINFACH	
#define SLIST(kennz,which) \
		case kennz: { simplelist_t *n=simple_alloc(str); \
			if (!n) errorflag=TRUE; \
			else { n->next=msg->which; msg->which=n; } } break;
		SLIST('A',to);
		SLIST('G',gruppen);
		SLIST('F',followup);
#undef SLIST
		case 'D': msg->distribution_byte=*str; break;
		}
	}

	if (errorflag)
	{
		msg_free(msg);
		mboutput("out of memory\r\n");
		return FALSE;
	}

#define STRCHECK(which) \
	if (!msg->which[0]) { errorflag=TRUE; mboutput(#which "fehlt\r\n"); }
#define LISTCHECK(which) \
	if (!msg->which) { errorflag=TRUE; mboutput(#which "fehlt\r\n"); }

	STRCHECK(betreff);
	STRCHECK(absender);

	if (pm)
	{
		LISTCHECK(to);
	}
	else
	{
		LISTCHECK(gruppen);
		if (!msg->distribution_byte)
			msg->distribution_byte='N';
	}
#undef STRCHECK
#undef LISTCHECK

	if (errorflag)
	{
		msg_free(msg);
		return FALSE;
	}

	time(&msg->importzeit);
	if (!*msg->eingabezeit)
	{
		time_t_to_mtzeit(&msg->importzeit,msg->eingabezeit);
		msg->eingabezeit_tt=msg->importzeit;
	}
	else
	{
		mtzeit_to_time_t(msg->eingabezeit,&msg->importzeit);
	}


	msglog("#einfachnachricht\n");
	if (pm)
	{
		simplelist_t *l;
		if (!msg->bearbeitungsstatus[0])
		{
			*msg->bearbeitungsstatus='N';
			time_t_to_mtzeit(&msg->importzeit,msg->bearbeitungsstatus+1);
		}
		adr=NULL;
		for (l=msg->to;l;l=l->next)
		{
			r_adr_t *neu;
			neu=r_alloc();
			if (neu)
				neu->inputadr=strdup(l->data);
			if (!neu || !neu->inputadr)
			{
				msg_free(msg);
				if (neu)
					r_free(neu);
				if (adr)
					r_free_chain(adr);
				mboutput("out of memory\r\n");
				return FALSE;
			}
			neu->level=lv;
			neu->sec_level=lv;
			neu->bezahlt=bezahlt;
			neu->sec_bezahlt=bezahlt;
			neu->owner=0;
			neu->next=adr;
			adr=neu;
		}
		if (!r_router(msg,adr))
		{
			r_adr_t *akt;

			for (akt=adr;akt;akt=akt->next)
			{
				if (akt->logstr)
					mbprintf("  %s\r\n",akt->logstr);
			}
			r_free_chain(adr);
			msg_free(msg);
			mboutput("Keine Route gefunden\r\n");
			return FALSE;
		}
	}

	if (!msg->msg_id[0])
		make_msg_id(msg->msg_id,pm,NULL,NULL);

	if (pm)
		msg->msgtyp=pers_msg;
	else
		msg->msgtyp=gruppen_msg;

	errorflag=!r_transport(msg,adr, FALSE);
	if (errorflag && adr)
	{
		r_adr_t *akt;

		for (akt=adr;akt;akt=akt->next)
		{
			if (akt->logstr)
				mbprintf("  %s\r\n",akt->logstr);
		}
	}
	r_flush();

	if (adr)
		r_free_chain(adr);
	msg_free(msg);
	return !errorflag;
}
