/* event.c : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
#include "tomsysdep.h"

#include <portab.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <time.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "prototyp.h"
#include "aktdata.h"
#include "tompaths.h"
#include "box_conf.h"
#include "event.h"
#include "modem.h"

static event_t *anker;
static time_t last_mtime_cnf;
static time_t last_mtime_dat;

static void free_list (void);
static BOOLEAN read_event_dat (void);
static void write_event_dat (void);
static BOOLEAN read_event_cnf (char *fname);
static long parse_ev_zeit (char *s);
static void calc_next (void);
static BOOLEAN do_event (event_t * ev);

extern BYTE offlineB;

event_t *
event_anker (void)
{
	if (!anker)
		event_load ();
	return anker;
}

static void
free_list (void)
{
	event_t *akt,
	*next;

	akt = anker;
	while (akt)
	{
		if (akt->id)
			free (akt->id);
		if (akt->cmd)
			free (akt->cmd);
		next = akt->next;
		free (akt);
		akt = next;
	}
	anker = NULL;
}

static BOOLEAN
do_event (event_t * ev)
{
	if (!strnicmp (ev->cmd, "BATCH ", 6))
	{
		automatic_batch (ev->cmd + 6);
		return FALSE;
	}
	else if (!stricmp (ev->cmd, "LEAVE"))
		return TRUE;
	syslog (LOG_NOTICE, "Event %s: unbekanntes Kommando %s", ev->id, ev->cmd);
	return FALSE;
}

static long
parse_ev_zeit (char *s)
{
	char *p;

	p = strpbrk (s, ": ");
	if (!p)
		return -1;
	*p++ = '\0';
	return strtol (s, NULL, 10) * 100 + strtol (p, NULL, 10);
}

static BOOLEAN
read_event_cnf (char *fname)
{
	FILE *f;
	int zeile=0;
	char buf[1024];

	f=fopen(fname,"r");
	if (!f)
	{
		syslog (LOG_ERR, "kann event.cnf nicht lesen");
		return FALSE;
	}

	while (fgets(buf,sizeof(buf),f))
	{
		char *id;
		char *soft,*wurf,*hard,*status,*cmd;
		char *p;
		int softz,wurfz,hardz;
		event_t *new;
		enum auto_enum
		{
			MANUELL_AUTO,
			NUR_MANUELL,
			NUR_AUTO
		} statusflag = MANUELL_AUTO;

		zeile++;

		p=strpbrk(buf,"\r\n");
		if (p)
			*p=0;
		id=buf;
		while (isspace(*id))
			id++;
		if (*id=='#' ||*id=='\0')
			continue;
		soft=strpbrk(id,"\t ");
		if (!soft)
		{
			syslog (LOG_NOTICE, "Eventkonfig Zeile %d: unparsbar", zeile);
			continue;
		}
		*soft++=0;
		while (isspace(*soft))
			soft++;

		wurf=strpbrk(soft," \t");
		if (!wurf)
		{
			syslog (LOG_NOTICE, "Eventkonfig Zeile %d: unparsbar", zeile);
			continue;
		}
		*wurf++=0;
		while (isspace(*wurf))
			wurf++;

		hard=strpbrk(wurf," \t");
		if (!hard)
		{
			syslog (LOG_NOTICE, "Eventkonfig Zeile %d: unparsbar", zeile);
			continue;
		}
		*hard++=0;
		while (isspace(*hard))
			hard++;

		status=strpbrk(hard," \t");
		if (!status)
		{
			syslog (LOG_NOTICE, "Eventkonfig Zeile %d: unparsbar", zeile);
			continue;
		}
		*status++=0;
		while (isspace(*status))
			status++;

		cmd=strpbrk(status," \t");
		if (!cmd || !*cmd)
		{
			syslog (LOG_NOTICE, "Eventkonfig Zeile %d: unparsbar", zeile);
			continue;
		}
		*cmd++=0;
		while (isspace(*cmd))
			cmd++;

		if ((softz = parse_ev_zeit (soft)) == -1)
		{
			syslog (LOG_NOTICE, 
				"Eventkonfig Zeile %d: Weiche Startzeit %s unparsbar", 
				zeile, soft);
			continue;
		}
		if ((wurfz = parse_ev_zeit (wurf)) == -1)
		{
			syslog (LOG_NOTICE, 
				"Eventkonfig Zeile %d: Wurfzeit %s unparsbar", 
				zeile,soft);
			continue;
		}
		if ((hardz = parse_ev_zeit (hard)) == -1)
		{
			syslog (LOG_NOTICE, 
				"Eventkonfig Zeile %d: Harte Startzeit %s unparsbar", 
				zeile,soft);
			continue;
		}
		if (!stricmp("both",status))
			statusflag=MANUELL_AUTO;
		else if (!stricmp("manuell",status))
			statusflag=NUR_MANUELL;
		else if (!stricmp("auto",status))
			statusflag=NUR_AUTO;
		else 
		{
			syslog (LOG_NOTICE, "Eventkonfig Zeile %d: Typ %s unparsbar",
				zeile);
			continue;
		}

		new = XMALLOC (sizeof (event_t));
		new->soft_begin = softz;
		new->wurf_begin = wurfz;
		new->hard_begin = hardz;
		new->id = XSTRDUP (id);
		new->cmd = XSTRDUP (cmd);
		new->manuell_ok = TRUE;
		new->auto_ok = TRUE;
		if (statusflag == NUR_AUTO)
			new->manuell_ok = FALSE;
		else if (statusflag == NUR_MANUELL)
			new->auto_ok = FALSE;

		if (!anker)
		{
			new->next = anker;
			anker = new;
		}
		else if (!new->auto_ok)
		{
			event_t *last;

			for (last = anker; last->next; last = last->next)
				;
			last->next = new;
			new->next = NULL;
		}
		else
		{
			event_t *akt;
			event_t *last = NULL;

			for (akt = anker; akt; akt = akt->next)
			{
				if (akt->soft_begin > new->soft_begin)
					break;
				last = akt;
			}
			if (!last)
			{
				new->next = anker;
				anker = new;
			}
			else
			{
				new->next = akt;
				last->next = new;
			}
		}
	}
	fclose(f);
	return TRUE;
}								/* read_event_cnf */

static void
write_event_dat (void)
{
	char fname[PATH_MAX];
	FILE *f;
	event_t *akt;

	sprintf (fname, "%s/event.dat", box_conf.var_dir);

	f = xfopen (fname, "w");
	akt = anker;
	while (akt)
	{
		fprintf (f, "%s %ld\n", akt->id, akt->last_started);
		akt = akt->next;
	}
	fclose (f);
}


static BOOLEAN
read_event_dat (void)
{
	char fname[PATH_MAX];
	FILE *f;
	char buf[1024];
	event_t *akt;

	sprintf (fname, "%s/event.dat", box_conf.var_dir);

	akt = anker;
	while (akt)
	{
		akt->last_started = 0;
		akt = akt->next;
	}

	f = fopen (fname, "r");
	if (!f)
	{
		syslog (LOG_NOTICE, "event.dat existiert nicht");
		/* Kein Grund zur Sorge */
		return TRUE;
	}
	while (fgets (buf, sizeof (buf), f))
	{
		char *p;
		char *n;

		p = strpbrk (buf, "\r\n");
		if (p)
			*p = '\0';
		n = buf;
		while (isspace (*n))
			n++;
		p = strpbrk (buf, " \t");
		if (!p)
		{
			syslog (LOG_NOTICE, "event.dat: unparsbare Zeile %s", buf);
			continue;
		}
		*p++ = '\0';
		akt = anker;
		while (akt)
		{
			if (!stricmp (akt->id, n))
				akt->last_started = strtol (p, NULL, 10);
			akt = akt->next;
		}
	}
	fclose (f);
	return TRUE;
}

	/* Eventliste laden */
BOOLEAN
event_load (void)
{
	char fname[PATH_MAX];
	struct stat st_dat;
	struct stat st_cnf;

	sprintf (fname, "%s/event.dat", box_conf.var_dir);
	if (stat (fname, &st_dat) != 0)
		st_dat.st_mtime=0;
	sprintf (fname, "%sevent.cnf", box_conf.config_dir);
	if (stat (fname, &st_cnf) != 0)
		st_cnf.st_mtime=0;
	if (st_cnf.st_mtime != last_mtime_cnf) {
		last_mtime_cnf=st_cnf.st_mtime;
		last_mtime_dat=st_dat.st_mtime;
		/* eventuell alte Liste verwerfen */
		free_list ();
		if (!read_event_cnf (fname))
			return FALSE;
		if (!read_event_dat ())
			return FALSE;
		calc_next ();
		return TRUE;
	}

	if (st_dat.st_mtime != last_mtime_dat)  {
		last_mtime_dat=st_dat.st_mtime;
		if (!read_event_dat ())
			return FALSE;
	}
	return TRUE;
}

/* "naechste" Zeitpunkte fr die events bestimmen */
static void
calc_next (void)
{
	time_t yet,
	 heute;
	struct tm yettm,
	 heutetm;
	event_t *ev;

	/* Das "jetzt" bestimmen */
	yet = time (NULL);
	yettm = *localtime (&yet);

	/* und heute, 0 Uhr */
	heutetm = yettm;
	heutetm.tm_hour = 0;
	heutetm.tm_sec = 0;
	heutetm.tm_min = 0;
	heute = mktime (&heutetm);

	for (ev = anker; ev != NULL; ev = ev->next)
	{
		struct tm worktm;
		time_t last;

		if (!ev->auto_ok)
			continue;

		while (yet > ev->next_hard) {
			ev->next_wurf+=86400L;
			ev->next_soft+=86400L;
			ev->next_hard+=86400L;
		}

		worktm = heutetm;

		/* Softstart des Events holen */
		worktm.tm_hour = ev->soft_begin / 100;
		worktm.tm_min = ev->soft_begin % 100;
		(void) mktime (&worktm);	/* setzt tm_isdst */
		worktm.tm_hour = ev->soft_begin / 100;
		worktm.tm_min = ev->soft_begin % 100;
		ev->next_soft = mktime (&worktm);

		if (ev->next_soft == -1)
		{
			/* z.B. Sommerzeitumstellung */
			ev->next_soft = time (NULL) + 86400L;
			ev->next_wurf = time (NULL) + 86400L;
			ev->next_hard = time (NULL) + 86400L;
			continue;
		}

		/* Wurfstart des Events holen */
		if (ev->wurf_begin != (ULONG) -1)
		{
			worktm.tm_hour = ev->wurf_begin / 100;
			worktm.tm_min = ev->wurf_begin % 100;
			(void) mktime (&worktm);	/* is_dst */
			worktm.tm_hour = ev->wurf_begin / 100;
			worktm.tm_min = ev->wurf_begin % 100;
			ev->next_wurf = mktime (&worktm);
			if (ev->next_soft > ev->next_wurf)
				ev->next_wurf = ev->next_wurf + 86400L;
		}
		else
			ev->next_wurf = ev->next_soft;

		/* Harten Startzeitpunkt holen */
		worktm.tm_hour = ev->hard_begin / 100;
		worktm.tm_min = ev->hard_begin % 100;
		(void) mktime (&worktm);	/* dst */
		worktm.tm_hour = ev->hard_begin / 100;
		worktm.tm_min = ev->hard_begin % 100;
		ev->next_hard = mktime (&worktm);
		if (ev->next_hard < ev->next_soft)
			ev->next_hard = ev->next_hard + 86400L;
		if (ev->next_hard < ev->next_wurf)
			ev->next_hard = ev->next_hard + 86400L;


		/* Letzte Ausfhrung des Events */
		last = ev->last_started;

		/* Schon alles zu spt oder schon ausgefhrt? */
		if (yet > ev->next_hard || last > ev->next_soft)
		{
			/* Alles auf den nchsten Tag legen */
			do
			{
				ev->next_wurf = ev->next_wurf + 86400L;
				ev->next_soft = ev->next_soft + 86400L;
				ev->next_hard = ev->next_hard + 86400L;
			}
			while (ev->next_hard < yet || last > ev->next_soft);
		}
	}
}

event_t *
event_next_soft (time_t * offset)
{
	event_t *best = NULL;
	time_t shortest = (time_t) 999999999;
	event_t *ev;
	time_t t;
	time_t dif;

	if (!anker)
		event_load ();

	t = time (NULL);
	for (ev = anker; ev != NULL; ev = ev->next)
	{
		if (!ev->auto_ok)
			continue;

		while (t > ev->next_hard) {
			ev->next_wurf+=86400L;
			ev->next_soft+=86400L;
			ev->next_hard+=86400L;
		}

		if (t > ev->next_soft && ev->next_soft != -1)
			dif = 0;
		else
			dif = ev->next_soft - t;
		if (dif && dif < shortest)
		{
			shortest = dif;
			best = ev;
		}
	}
	if (offset)
		*offset = shortest;
	return best;
}

event_t *
event_next_wurf (time_t * offset)
{
	event_t *best = NULL;
	time_t shortest = (time_t) 999999999;
	event_t *ev;
	time_t t;
	time_t dif;

	if (!anker)
		event_load ();

	t = time (NULL);
	for (ev = anker; ev != NULL; ev = ev->next)
	{
		if (!ev->auto_ok)
			continue;

		while (t > ev->next_hard) {
			ev->next_wurf+=86400L;
			ev->next_soft+=86400L;
			ev->next_hard+=86400L;
		}

		if (t > ev->next_wurf)
			dif = 0;
		else
			dif = ev->next_wurf - t;
		if (dif && dif < shortest)
		{
			shortest = dif;
			best = ev;
		}
	}
	if (offset)
		*offset = shortest;
	return best;
}

char *
event_execute (const char *para)
{
	event_t *ev;
	BOOLEAN recalc = FALSE;

	if (!anker)
		event_load ();

	ev = anker;
	while (ev)
	{
		time_t t;
		BOOLEAN soll = FALSE;

		t = time (NULL);

		while (t > ev->next_hard) {
			ev->next_wurf+=86400L;
			ev->next_soft+=86400L;
			ev->next_hard+=86400L;
		}

		if (para && !stricmp (ev->id, para) && ev->manuell_ok)
			soll = TRUE;
		else if (!para && ev->auto_ok && t >= ev->next_soft
				 && ev->next_soft != -1)
		{
			soll = TRUE;
			recalc = TRUE;
		}
		if (soll)
		{
			if (para || t <= ev->next_hard)
			{
				time_t started=time(NULL);
				/* Racecondition bei >= Abfrage vermeiden */
				ev->last_started=started+30;

				/* starten */
				if (do_event (ev))
				{
					if (recalc)
					{
						calc_next ();
						write_event_dat ();
					}
					return ev->id;
				}
				/* das war ein Batch */
				if (recalc)
				{
					calc_next ();
					write_event_dat ();
				}

				/* wenn kein User eingeloggt (also nicht manuell ausgelst) */
				if (!akt_anruf.startzeit)
				{
					anrufdaten_init(ANRUF_INTERN);
					strcpy(akt_anruf.name,ev->id);
					akt_anruf.startzeit=started;
					login_protokoll_write();
				}
				return NULL;
			}
		}
		ev = ev->next;
	}
	if (recalc)
	{
		calc_next ();
		write_event_dat ();
	}
	else if (para)
	{
		if (!offlineB)
			mbfehler ("Event nicht gefunden");
		syslog (LOG_NOTICE, "Event %s nicht gefunden", para);
	}
	return NULL;
}
