/* gebuehren.c */
/*
    Copyright (C) 1996 Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
*/
#include "tomsysdep.h"

#include <unistd.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <ctype.h>
#include <syslog.h>
#include <errno.h>
#include "libquark.h"
#include "itbcompile.h"
#include "long-options.h"

#define MAX_ZONEN 20

const char *lineend;
const char *program_name;

static BOOLEAN
zeitraum_tag(char *str, int *tag)
{
	if (isdigit(*str))
	{
		*tag=(int) strtol(str,NULL,10);
		return TRUE;
	}
	else
	{
		if (0==strcasecmp(str,"so")
			|| 0==strcasecmp(str,"sonntag")
			|| 0==strcasecmp(str,"sunday")
			|| 0==strcasecmp(str,"sun"))
		{
			*tag=0;
			return TRUE;
		}
		if (0==strcasecmp(str,"mo")
			|| 0==strcasecmp(str,"montag")
			|| 0==strcasecmp(str,"mon")
			|| 0==strcasecmp(str,"monday"))
		{
			*tag=1;
			return TRUE;
		}
		if (0==strcasecmp(str,"di")
			|| 0==strcasecmp(str,"dienstag")
			|| 0==strcasecmp(str,"tuesday")
			|| 0==strcasecmp(str,"tue"))
		{
			*tag=2;
			return TRUE;
		}
		if (0==strcasecmp(str,"mi")
			|| 0==strcasecmp(str,"mittwoch")
			|| 0==strcasecmp(str,"wednesday")
			|| 0==strcasecmp(str,"wed"))
		{
			*tag=3;
			return TRUE;
		}
		if (0==strcasecmp(str,"do")
			|| 0==strcasecmp(str,"donnerstag")
			|| 0==strcasecmp(str,"thursday")
			|| 0==strcasecmp(str,"thu"))
		{
			*tag=4;
			return TRUE;
		}
		if (0==strcasecmp(str,"fr")
			|| 0==strcasecmp(str,"freitag")
			|| 0==strcasecmp(str,"friday")
			|| 0==strcasecmp(str,"fri"))
		{
			*tag=5;
			return TRUE;
		}
		if (0==strcasecmp(str,"sa")
			|| 0==strcasecmp(str,"samstag")
			|| 0==strcasecmp(str,"saturday")
			|| 0==strcasecmp(str,"sat"))
		{
			*tag=6;
			return TRUE;
		}
	}
	return FALSE;

}

static BOOLEAN
zeitraum_hhmm(char *str,long *res)
{
	char *p=strchr(str,':');
	size_t l=strlen(str);
	*res=0;
	if (!p)
	{
		/* HH[MM[SS]] */
		if (l>6 || l==4 || (l&1))
			return FALSE;
		if (l==6)
			*res=strtol(str+4,NULL,10);
		if (l>2)
		{
			str[4]=0;
			*res+=strtol(str+2,NULL,10)*60;
		}
		str[2]=0;
		*res+=strtol(str,NULL,10)*60*60;
		return TRUE;
	}
	else
	{
		/* HH[:MM[:SS]] */
		/* 01 234 567 */
		if (l!=2 && l!=5 && l!=8)
			return FALSE;
		if (l==8)
			*res=strtol(str+6,NULL,10);
		if (l>2)
		{
			str[5]=0;
			*res+=strtol(str+3,NULL,10)*60;
		}
		str[2]=0;
		*res+=strtol(str,NULL,10)*60*60;
		return TRUE;
	}
}

static void
zeitraum_auswerten(int tag, long user_from, long user_to, char *buf,
	long *zonenkosten, int *max_zone)
{
	char *p;
	char *next;
	long bereich_from;
	long bereich_to;
	long sekunden;
	long kosten_pro_takt;
	int i;

	p=strpbrk(buf,"\r\n");
	if (p)
		*p=0;

	/*
	 * Tage holen 
	 */

	next=strpbrk(buf,"\t ");
	if (!next)
		return;
	*next++=0;
	while (isspace(*next))
		next++;
	if (!*next)
		return;

	p=strchr(buf,'-');
	if (p)
	{
		int tfrom; /* "tag von" */
		int tto;   /* "tag bis" */
		*p++=0;
		if (!zeitraum_tag(buf, &tfrom))
			return;
		if (!zeitraum_tag(p, &tto))
			return;
		if (tag<tfrom || tag >tto)
			return;
	}
	else
	{
		BOOLEAN found=FALSE;
		int tfrom; /* "tag, fr den die Zeile gilt" */
		while (buf && *buf)
		{
			p=strchr(buf,',');
			/* so */
			if (p)
				*p++=0;
			if (!zeitraum_tag(buf, &tfrom))
				return;
			buf=p;
			if (tfrom==tag)
				found=TRUE;
		}
		if (!found)
			return;
	}

	/* So, die erste Angabe traf zu. Jetzt die Uhrzeiten */
	p=next;
	next=strpbrk(p,"\t ");
	if (!next)
		return;
	*next++=0;
	while (isspace(*next))
		next++;
	if (!*next)
		return;

	if (!zeitraum_hhmm(p,&bereich_from))
		return;

	p=next;
	next=strpbrk(p,"\t ");
	if (!next)
		return;
	*next++=0;
	while (isspace(*next))
		next++;
	if (!*next)
		return;

	if (!zeitraum_hhmm(p,&bereich_to))
		return;

	/* Outside given range? */
	if (bereich_from > user_to || bereich_to <user_from)
		return;

	/*
	 * It's our time ...
	 */
	{
		long f,t;
		if (user_from > bereich_from)
			f=user_from;
		else
			f=bereich_from;
		if (user_to < bereich_to)
			t=user_to;
		else
			t=bereich_to;
		sekunden=t-f;
		if (sekunden < 0)
			/* da hat jemand etwas falsch eingegeben */
			return;
	}

	/* Kosten holen (12 Pf pro Tag -> 12) */
	p=next;
	next=strpbrk(p,"\t ");
	if (!next)
		return;
	*next++=0;
	while (isspace(*next))
		next++;
	if (!*next)
		return;
	kosten_pro_takt=strtol(p,NULL,10);

	/* So, jetzt kommt die Taktdauer pro Zone */
	p=next;
	i=0;
	while (p && *p)
	{
		double seks=strtod(p,&next);
		if (seks!=0)
		{
			int takte=sekunden/seks;
			if (sekunden - ((int)(takte * seks))!=0)
/*			if ((sekunden % seks) > 0.001) */
				takte++;
			zonenkosten[i]+=kosten_pro_takt * takte;
		}
		if (next && *next)
		{
			while (!isdigit(*next))
				next++;
		}
		p=next;
		i++;
		if (i>*max_zone)
			*max_zone=i;
		if (i==MAX_ZONEN)
			break;
	}
	return;
}

static void 
gebuehrenanzeige(const char *cnffile, time_t from, time_t to)
{
	char buf[1024];
	FILE *f;
	struct tm tf;
	struct tm tt;
	int i;
	int max_zone=0;
	long zonenkosten[MAX_ZONEN];
	char *zonenname[MAX_ZONEN];
	long this_from,this_to;
	size_t maxlen_zn=0;
	const char *einheit="DM";
	long faktor=100;
	int nachkomma;
	int vorkomma;
	long max_kosten;


	tf=*localtime(&from);
	tt=*localtime(&to);
	/* Taktinformationen lesen */
	f=fopen(cnffile,"r");
	if (!f)
		error(1,errno,"cannot open %s",cnffile);

	for (i=0;i<MAX_ZONEN;i++)
	{
		zonenkosten[i]=0;
		zonenname[i]=NULL;
	}

	i=0;
	while (fgets(buf,sizeof(buf),f))
	{
		char *p=strpbrk(buf,"\r\n");
		char *q;
		if (p)
			*p=0;
		if (*buf!='=')
			continue;
		p=buf+1;
		while (isspace(*p))
			p++;
		if (!*p)
			continue;
		q=strpbrk(p,"\t ");
		if (!q)
			continue;
		*q++=0;
		while (isspace(*q))
			q++;
		if (!*q)
			continue;
		if (!strcasecmp(p,"zone"))
		{
			strtrim(q);
			if (strlen(q)>maxlen_zn)
				maxlen_zn=strlen(q);
			zonenname[i++]=XSTRDUP(q);
		}
		if (!strcasecmp(p,"faktor"))
		{
			faktor=strtol(q,NULL,10);
			if (faktor<=0)
				faktor=1; /* division durch 0? */
		}
		if (!strcasecmp(p,"einheit"))
		{
			strtrim(q);
			einheit=XSTRDUP(q);
		}
	}
	if (maxlen_zn < 3)
		maxlen_zn=3; /* Platz fr ??? */

	rewind(f);
	if (tf.tm_mday!=tt.tm_mday)
	{
		/* Arg - ber die Tagesgrenze hinweg. */
		this_from=tf.tm_hour*3600+tf.tm_min*60+tf.tm_sec;
		while (fgets(buf,sizeof(buf),f))
		{
			char *p=strpbrk(buf,"\r\n");
			if (p)
				*p=0;
			if (*buf=='=')
				continue;
			zeitraum_auswerten(tf.tm_wday, this_from, 24 * 3600, buf, 
				zonenkosten,&max_zone);
		}
		tf.tm_hour=0;
		tf.tm_min=0;
		tf.tm_sec=0;
		rewind(f);
	}

	this_from=tf.tm_hour*3600+tf.tm_min*60+tf.tm_sec;
	this_to=tt.tm_hour*3600+tt.tm_min*60+tt.tm_sec;
	while (fgets(buf,sizeof(buf),f))
	{
		char *p=strpbrk(buf,"\r\n");
		if (p)
			*p=0;
		if (*buf=='=')
			continue;
		zeitraum_auswerten(tf.tm_wday, this_from, this_to, buf, zonenkosten,
			&max_zone);
	}

	tf=*localtime(&from); /* tf wird oben eventuell verndert */


	max_kosten=0;
	for (i=0;i<max_zone;i++)
	{
		if (zonenkosten[i]>max_kosten)
			max_kosten=zonenkosten[i];
	}
	max_kosten/=faktor;
	if (faktor <= 10)
		nachkomma=1;
	else if (faktor <= 100)
		nachkomma=2;
	else if (faktor <= 1000)
		nachkomma=3;
	else
		nachkomma=4;

	if (max_kosten <= 10)
		vorkomma=1;
	else if (max_kosten <= 100)
		vorkomma=2;
	else if (max_kosten <= 1000)
		vorkomma=3;
	else
		vorkomma=4;

	printf(
		"Die Gebhren fr %ld:%02d Minuten von %02d:%02d bis %02d:%02d betragen:\n",
		(to-from) / 60, (int) ((to-from) %60), 
		tf.tm_hour, tf.tm_min, tt.tm_hour, tt.tm_min);
	for (i=0;i<max_zone;i++)
	{
		const char *s= zonenname[i] ? zonenname[i] : "???";
		printf("    %s:%*.*s %*ld.%0*ld %s\n", s,
			(int) (maxlen_zn-strlen(s)), (int) (maxlen_zn-strlen(s)),
			"                                                               ",
			vorkomma, zonenkosten[i] / faktor, 
			nachkomma, zonenkosten[i] % faktor ,
			einheit);
	}

}

static void
usage(int ex)
{
	FILE *f;
	if (ex)
		f=stderr;
	else
		f=stdout;
	fprintf(f,
		"usage: %s [--cnffile configfile] [--from fromtime] [--to totime]\n",
		program_name);
	fprintf(f, 
		"defaults:\n"
		"  cnffile: config/gebuehren.cnf\n"
		"  to: now\n"
		"there is no default for from, you have to provide it\n"
		);

	exit(ex);
}

int 
main(int argc, char **argv)
{
	BOOLEAN got_version=FALSE;
	const char *cnffile="config/gebuehren.cnf";
	time_t from=0;
	time_t to=time(NULL);
	char *next;
	int option_index = 0;

	struct option opts[] =
	{
		{"cnffile", required_argument, NULL, 'c'},
		{"from", required_argument, NULL, 'f'},
		{"to", required_argument, NULL, 't'},
		{NULL, 0, NULL, 0}
	};

	program_name=strrchr(argv[0],'/');
	if (!program_name)
		program_name = argv[0];
	else
		program_name++;

	parse_long_options (argc, argv, program_name, PACKAGE_VERSION, usage);

	while (1)
	{
		int c;

		c = getopt_long (argc, argv, "c:f:t:", opts, &option_index);
		if (c == -1)
			break;
		switch (c)
		{
		case 'c':
			cnffile = optarg;
			break;
		case 'f':
			from=strtol(optarg,&next,10);
			if (!next || *next)
			{
				error(2,0,"can't understand %s",optarg);
				exit(1);
			}
			break;
		case 't':
			to=strtol(optarg,&next,10);
			if (!next || *next)
			{
				error(2,0,"can't understand %s",optarg);
				exit(1);
			}
			break;
		case 'h':
			usage(0);
			break;
		case 'v':
			printf ("%s - %s\n", program_name, PACKAGE_VERSION);
			printf("  %s %s\n", __DATE__,__TIME__);
			exit(0);
			break;
		default:
			usage (2);
			break;
		}
	}

	if (!cnffile || !from || !to)
	{
		if (got_version)
			exit(0);
		if (!cnffile)
			fprintf(stderr,"%s: cnffile not given%s",program_name,lineend);
		if (!to)
			fprintf(stderr,"%s: to not given%s",program_name,lineend);
		if (!from)
			fprintf(stderr,"%s: from not given%s",program_name,lineend);
		usage(2);
	}
	gebuehrenanzeige(cnffile,from,to);
	exit(0);
}

