/* getty.c: Der Teil der Quark, der den getty ersetzt */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact1: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/

#include "tomsysdep.h"

#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <errno.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#ifdef HAVE_SYS_TERMIO_H
#include <sys/termio.h> /* HPUX braucht das fr TIOCSCTTY */
#endif
#include <signal.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <netinet/in.h>

#include "tomsysdep.h"

#include "tom_termcap.h"
/* manche curses definieren TRUE/FALSE unbedingt */
#include <portab.h>

#include "prototyp.h"
#include "io.h"
#include "modem.h"
#include "terminal.h"
#include "event.h"
#include "box_conf.h"
#include "protokoll.h"
#include "aktdata.h"
#include "msg.h"
#include "log.h"
#include "tsignal.h"
#include "getty.h"
#include "argument.h"
#include "io.h"
#include "uosock.h"

#include "lockfile.h"

static int modem_fd;
extern int user_in_fd;
extern int user_out_fd;
extern int sysop_in_fd;
extern int sysop_out_fd;

extern BOOLEAN modem_vorhanden;

extern char *device_lockfile;
extern BOOLEAN device_locked;
extern volatile int got_sigusr2; /* wird TRUE bei Empfang von SIGUSR2 */
extern volatile int got_sigterm; /* wird TRUE bei Empfang von SIGTERM */
static char *device_name=NULL;
static BOOLEAN is_netmodem=FALSE;


static time_t myfiledate;
static char *myname;

extern char *global_event_hack;

extern int old_stdin_fileno;

extern volatile int got_sigusr1;

extern RETSIGTYPE SIGUSR1_handler(int siechtum);

/* 
 * Warteschleife des boxinternen gettys. Wartet auf Anruf, ESC l, ESC n,
 * q oder Q, oder einen Event.
 */
int getty_wait(char **eventid, int *ctyp)
{
	char modemdaten[50];
	int modem;
	int annehmenB;

	*eventid=NULL;
	*ctyp=CARRIER_ERROR;

	signal(SIGUSR1,SIGUSR1_handler);

	modem = 0;

	annehmenB = -1;		/* ungltiger Wert um Aktualisierung zu erzwingen */

	init_menue();						/* Bildschirm-Maske installieren */

	while (TRUE)
	{
		struct timeval zeit;
		fd_set set;
		BYTE neuB;
		int ret;
		int select_nfds;

		/* Event-Verarbeitung */
		if (argument_execute_events)
		{
			*eventid=event_execute(NULL);
			if (*eventid)
				return LEAVE_EVENT;
		}

		/* max. 30 s auf Schnittstelleneingaben warten */
		/* stdin + Port */
		FD_ZERO(&set);
		FD_SET(STDIN_FILENO,&set);
		select_nfds=STDIN_FILENO;
		if (modem_fd >= 0 && modem_vorhanden) {
			FD_SET(modem_fd,&set);
			if (modem_fd > STDIN_FILENO)
				select_nfds=modem_fd;
		}

		/* 60s timeout fr Statusauffrischung und Mitternachtsupdate */
		if (annehmenB >= 0)
			zeit.tv_sec = 60;
		else
			zeit.tv_sec = 0;
		zeit.tv_usec = 0;
		while (TRUE)
		{
			ret=select(select_nfds+1,&set,NULL,NULL,&zeit);
			if (ret>=0) /* timed out or input available */
				break;
			/* jemand anderes will hflich an die Box */
			if (got_sigusr1)
				return LEAVE_RESTART;
			/* und das ist die heftigere Variante */
			if (got_sigusr2 || got_sigterm)
				return LEAVE_RESTART;
		}

		if (ret && FD_ISSET(STDIN_FILENO,&set))
		{
			static int escape = FALSE;
			char pressed_key;

			read(STDIN_FILENO,&pressed_key,1);
			if (escape)
			{
				switch(pressed_key)
				{
		        case ' ':
					init_menue();
					break;
				case 'h':
					printf("%s",
						"k Console (un)besetzt\n"
						"l Consolenlogin\n"
						"Q Ende mit Abheben\n"
						"q Ende ohne Abheben\n"
						"SPACE - Bildschirm neu aufbauen\n"
					);
					break;
				case 'k':
					box_conf.console_besetzt = !box_conf.console_besetzt;
					break;
				case 'l': /* l Consolenlogin */
					if (!lock_tom(10))
					{
						/* das ist aber rgerlich - jemand ist in
						 * der Box.
						 */
						if (device_name)
							lock_dev(device_name,TRUE);
						return LEAVE_RESTART;
					}
					display_status("Consolen-Login aktiviert");
					anrufdaten_init(ANRUF_USER);
					if (modem_vorhanden) {
						/* Modem freigeben */
						close(modem_fd);
						modem_fd=-1;
						modem_vorhanden=FALSE;
					}
					akt_connection.is_console = TRUE;
					akt_connection.startzeit = time(NULL);
					akt_connection.carrierstring=XSTRDUP("CONS");
					tprint_fd(user_out_fd,"cl");

					*ctyp=CARRIER_LOGIN;
					return LEAVE_LOGIN;
					break;

				case 'Q':
					display_status("Modem abheben");
					takeoff_modem(modem_fd);
					/* FALLTHROUGH */
				case 'q':
					display_status(">>> Programmende <<<");
					*ctyp=CARRIER_LOGIN;
					return LEAVE_OK;
				}
				escape=FALSE;
				annehmenB = -1;
			}
			else if (pressed_key==27)
				escape=TRUE;
		}
		/* Zeichen an serieller Schnittstelle? */
		else if (modem_vorhanden && modem_fd >= 0 
			&& ret && FD_ISSET(modem_fd,&set))
		{
			if (!lock_dev(device_name,FALSE))
			{
				void (*old_handler)(int);
				if (!read(modem_fd,modemdaten+modem,1))
				{
					int nr = errno;
					fprintf(stderr,"Fehler bei read() in main:%s\r\n",strerror(nr));
					syslog(LOG_INFO,"Fehler bei read() in main:%s\r\n",strerror(nr));
					tom_leave(LEAVE_ERROR,NULL);
				}
				modemdaten[++modem] = 0;
				if (modem > 30)
				{
					memcpy(modemdaten,&modemdaten[10],30);
					modem -= 10;
				}
				if (strstr(modemdaten,box_conf.modem_ring) 
					&& anrufe_annehmen())
				{
					device_locked=TRUE;	/*  */
					anrufdaten_init(ANRUF_USER);
					if (box_conf.nologin_device_file 
						|| box_conf.nologin_global_file) {
						char *buf=NULL;
						char *p=strrchr(device_name,'/');
						ret=1;
						if (!p)
							p=device_name;
						else
							p++;
						if (box_conf.nologin_device_file) {
							asprintf(&buf,box_conf.nologin_device_file,p);
							XNULLCHECK(buf);
							ret=access(buf,F_OK);
							strcpy(akt_anruf.name,"nologin: device");
						}
						if (ret && box_conf.nologin_global_file) {
							free(buf);
							buf=XSTRDUP(box_conf.nologin_global_file);
							ret=access(buf,F_OK);
							strcpy(akt_anruf.name,"nologin: global");
						}
						if (ret == 0) {
							BOOLEAN t_o=FALSE;
							/* nologin-File exists */
							display_status(buf);
							free(buf);
							/* warten bis zum Auflegen */
							while ((p=modem_readline(modem_fd,10,&t_o))!=NULL) {
								if (*p!='\r' && *p 
									&& !strstr(p,box_conf.modem_ring))
									break;
								free(p);
							}
							free(p);
							modem_vorhanden=FALSE;
							strcpy(akt_anruf.zuletzt,"NLOG");
							login_protokoll_write();
							return LEAVE_RESTART;
						}
						free(buf);
					}
					/* CLOCAL setzen */
					io_set_CLOCAL(STDIN_FILENO,TRUE);
					old_handler=signal(SIGHUP,SIG_IGN);

					display_status("Modem: Ein Anruf wird registriert");
					modem = 0;


					/* "interne I/O-Umlenkung" */
					sysop_in_fd=dup(STDIN_FILENO);
					sysop_out_fd=dup(STDOUT_FILENO);
					old_stdin_fileno=sysop_in_fd;
#define OUT(meld) { int e=errno; \
	syslog(LOG_ERR,"%s: %m",meld); \
	errno=e; \
	}
					/* 
					 * nun versuchen wir, unser ctty loszuwerden und
					 * session leader zu werden.
					 */
#ifdef HAVE_SETSID
					if (setsid()<0)
					{
						/* Einzige Mglichkeit: Wir sind bereits SL 
						 * und mssen nur noch das ctty loswerden.
						 */
						/* OUT("setsid"); */
						if (!isatty(0))
							OUT("isatty: stdin is no tty");
#if defined(TIOCNOTTY)
						if (ioctl(STDIN_FILENO,TIOCNOTTY IOCTL_THIRD_PARA)<0)
						{
							OUT("stdin");
							exit(1);
						}
#endif
#if defined(TIOCSCTTY)
						if (ioctl(modem_fd,TIOCSCTTY IOCTL_THIRD_PARA))
						{
							OUT("ioctl(modem_fd,TIOCSCTTY)");
							if (close(open(device_name,O_RDONLY)))
								OUT("close(open(modem))");
						}
#else
						if (close(open(device_name,O_RDONLY)))
							OUT("close(open(modem))");
#endif
					}
#else
#  if defined(HAVE_SETPGID)
					setpgid(0,0);
#  elif defined(HAVE_SETPGRP)
					setpgrp(0,0);
#  else
#    error no way to make me session leader
#   endif
					if (ioctl(STDIN_FILENO,TIOCNOTTY IOCTL_THIRD_PARA))
					{
						OUT("close(open(modem))");
						exit(1);
					}
#endif
					/* OK, wir haben kein ctty mehr und sind SL */

					/* Bildschirm lschen */
					tprint_fd(sysop_out_fd,"cl");

					/* stdin/out/err auf das Modem */
					dup2(modem_fd,STDIN_FILENO);
					dup2(modem_fd,STDOUT_FILENO);
					dup2(modem_fd,STDERR_FILENO);
					close(modem_fd); /* not longer needed */
					user_in_fd = 0;
					user_out_fd = 1;
					signal(SIGHUP,old_handler);

					inquire_remote(STDIN_FILENO);

					/* Modemdevice wird nun controlling terminal */
#if defined(TIOCSCTTY) && !defined(__hpux)
					if (ioctl(STDIN_FILENO,TIOCSCTTY IOCTL_THIRD_PARA))
						OUT("ioctl(stdin,TIOCSCTTY)");
#else
					OUT("kein TIOCSCTTY");
					if (close(open(device_name,O_RDONLY)))
						OUT("close(open(modem))");
#endif
					if (!lock_tom(1))
					{
						/* das ist aber rgerlich - jemand ist in
						 * der Box.
						 */
						lock_dev(device_name,TRUE);
						return LEAVE_RESTART;
					}
					{
						const char *s="Abhebebefehl wird gesendet\n";
						write(sysop_out_fd,s,strlen(s));
					}
					if (is_netmodem)
						akt_connection.is_modem=TRUE;
					*ctyp=warte_auf_carrier(user_in_fd);
					if (*ctyp==CARRIER_ERROR)
					{
						strcpy(akt_anruf.name,"No Carrier");
						/* CONnect Wait */
						strcpy(akt_anruf.zuletzt,"CONW");
						login_protokoll_write();
						close(0);
						close(1);
						close(2);
						user_in_fd = -1;
						user_out_fd = -1;
						if (box_conf.log_quality_level)
						{
							time_t t;
							char *p,*q;
							time(&t);
							p=ctime(&t);
							q=strpbrk(p,"\r\n");
							if (q)
								*q=0;

							write_info_protokoll(INFOPROT_FREE,NULL,p);
							write_info_protokoll(INFOPROT_FREE,NULL,
"Verbindungsdaten nach NO CARRIER beim Verbindungsaufbau");
							sleep(2); /* das Modem knnte Zeit brauchen */
							get_connection_quality(device_name,
								box_conf.log_quality_command);
						}
						return LEAVE_ERROR;
					}
					/* CLOCAL weg,von jetzt an wollen wir SIGHUP */
					io_set_CLOCAL(STDIN_FILENO,FALSE);

					return LEAVE_LOGIN;
				}
				/* unlock device */
				lock_dev(device_name,TRUE);

				device_locked=FALSE;
			}
			else
			{
				/* jemand anderes ist am Modem ... */
				modem_vorhanden=FALSE;
				return LEAVE_RESTART;
			}
		}
		else
			update_menue();

		neuB = anrufe_annehmen();
		if (annehmenB != neuB)
		{
			annehmenB = neuB;
			if (annehmenB)
				display_status("< Box ist online >");
			else
				display_status("- Box ist OFFLINE -");
		}

		/* Falls mittlerweile eine neuere Programmversion vorhanden ist */
		{
			struct stat st;
			if (!stat(myname,&st))
			{
				if (st.st_mtime > myfiledate)
				{
					display_status(">>> Programmende wegen Update <<<");
					return LEAVE_RESTART;
				}
			}
		}
	} /* while (TRUE) */

} /* wait_for_event */


void
getty_deinit(void)
{
	if (device_locked)
	{
		if (!akt_connection.is_console) {
			hangup_modem(user_out_fd);
			close(user_out_fd);
			close(user_in_fd);
		}

		lock_dev(device_name,TRUE);
		device_locked=FALSE;
	}
}


void 
getty_init(const char *argv0)
{
	struct stat st;
	const char *p;
	if (argument_port)
		p=argument_port;
	else
		p=box_conf.modem_portname;
	if (!p)
		device_name=NULL;
	else if (*p=='/')
		device_name=XSTRDUP(p);
	else if (!strstr(p,"://")) {
		device_name=concat("/dev/",p,NULL);
		XNULLCHECK(device_name);
	} else
		device_name=XSTRDUP(p);
	if (!stat(argv0,&st))
		myfiledate=st.st_mtime;
	else
		myfiledate=time(NULL);
	if (*argv0=='/')
	{
		myname=strdup(argv0);
		if (!myname)
		{	
			syslog(LOG_ERR,"out of memory");
			tom_leave(LEAVE_RESTART,NULL);
		}
	}
	else
	{
		char tmp[PATH_MAX];
		size_t l;
		getcwd(tmp,sizeof(tmp));
		l=strlen(tmp)+2+strlen(argv0);
		myname=malloc(l);
		if (!myname)
		{	
			syslog(LOG_ERR,"out of memory");
			tom_leave(LEAVE_RESTART,NULL);
		}
		strcpy(myname,tmp);
		strcat(myname,"/");
		strcat(myname,argv0);
	}
	if (argument_tty)
	{
		/* es ist davon auszugehen, da stdin/out/err nicht definiert
		 * sind */
		uid_t e,r;
		e=geteuid();
		r=getuid();
		if (0!=strncmp(argument_tty,"/dev/",5))
			argument_tty=concat("/dev/",argument_tty,NULL);
		XNULLCHECK(argument_tty);
		/* Zugriffsrechte auf die Gerte in Ordnung bringen - 
         * mglicherweise verstellen externe Programme sie. sliplogin ist
         * ein Beispiel.
         */
		if (r!=e && r==0)
		{
			setuid(0);
			if (stat(argument_tty,&st)==0) {
				if ((st.st_mode & (S_IRUSR | S_IWUSR)) != (S_IRUSR | S_IWUSR))
					chmod(argument_tty,st.st_mode | (S_IRUSR | S_IWUSR));
				if (st.st_uid!=e)
					chown(argument_tty,e,getegid());
			}
			if (device_name) {
				if (stat(device_name,&st)==0) {
					if ((st.st_mode & (S_IRUSR | S_IWUSR)) != (S_IRUSR | S_IWUSR))
						chmod(device_name,st.st_mode | (S_IRUSR | S_IWUSR));
					if (st.st_uid!=e)
						chown(device_name,e,getegid());
				}
			}
			setuid(e); /* we aren't root anymore - this sets saved uid */
		}
		close(0);
		/* nonblock wegen seriellen Terminals - das kme bestimmt nicht
		 * immer gut ... */
		if (0!=open(argument_tty,O_RDWR|O_NONBLOCK))
		{
			syslog(LOG_ERR,"cannot open %s: %m",argument_tty,strerror(errno));
			exit(1);
		}
		/* Nonblockingmode verlassen, nun berflssig */
		fcntl(0,F_SETFL, O_RDWR);

		close(1);
		close(2);
		dup2(0,1);
		dup2(0,2);
		if (box_conf.console_init_prg)
			system(box_conf.console_init_prg);
		io_mode_work(0, TRUE);
		/* io_set_ICRNL(0,TRUE); */
		io_set_ONLCR(0,TRUE);
	}
}

void getty_devinit(void)
{
	if (!device_name)
		modem_vorhanden=FALSE;
	else if (0==memcmp(device_name,"modem://",8)) {
		is_netmodem=TRUE;
		modem_vorhanden=TRUE;
	} else 
		modem_vorhanden=TRUE; /* check_modem korrigiert das noch */

	/* 
	 * nun wissen wir, ob ein Modem konfiguriert ist. Falls ja, ist 
	 * ein Lockfile anzulegen.
	 */
	if (modem_vorhanden)
	{
		BOOLEAN war_gelockt=FALSE;
		char  lockfname[PATH_MAX];
		int cnt=0;
		if (!is_netmodem) {
			sprintf(lockfname,"%s%s",box_conf.lockdir,
				lock_dev_filename(device_name));
			signal(SIGUSR1,SIGUSR1_handler);
			while (lock_dev(device_name,FALSE))
			{
				war_gelockt=TRUE;
				sleep(5);
				cnt++;
				if (cnt>=3600/5) /* eine Stunde */
				{
					syslog(LOG_ERR,"kann Port %s nicht locken",
						device_name);
					tom_leave(LEAVE_ERROR,NULL);
				}
				if (got_sigusr2 || got_sigusr1 || got_sigterm)
					tom_leave(LEAVE_RESTART,NULL);
			}
			signal(SIGUSR1,SIG_IGN);

			device_lockfile=XSTRDUP(lockfname);
			device_locked=TRUE;

			if (war_gelockt)
			{
				/* etwas warten. Es ist nicht gut, die Schnittstelle
				 * sofort wieder anzusprechen. */
				sleep(1);
			}
			/* 
			 * "Nonblocking" oeffnen, sonst wartet open, bis der Carrier
			 * steht.
			 */
#ifdef O_NONBLOCK
			modem_fd=open(device_name,O_RDWR|O_NONBLOCK);
#elif defined(O_NDELAY)
			/* MiNT kennt nur O_NDELAY -> non posix */
			modem_fd=open(device_name,O_RDWR|O_NDELAY);
#else
#  error something is seriously broken here
#endif
			if (modem_fd<0) {
				syslog(LOG_ERR,"cannot open %s: %s", device_name,
					strerror(errno));
				tom_leave(LEAVE_ERROR,NULL);
			}
			/* Nonblockingmode verlassen, nun berflssig */
			fcntl(modem_fd,F_SETFL, O_RDWR);
		} else {
			int r;
			struct in_addr ina;
			unsigned short port;
			struct sockaddr_in s_in;

			port=htons(2500);
			ina.s_addr=INADDR_ANY;
			r=parse_ip_port(device_name+8, /* modem:// */
				&ina,&port,"tcp");
			if (r==-1) {
				syslog(LOG_ERR,"cannot understand %s", device_name);
				_exit(1);
			}
			s_in.sin_family=AF_INET;
			s_in.sin_addr=ina;
			s_in.sin_port=port;

			modem_fd=uosock_connect(&s_in, 10);
			if (modem_fd == -1) {
				syslog(LOG_ERR,"socket_connect to %s: %s", 
					device_name,strerror(errno));
				tom_leave(LEAVE_ERROR,NULL);
			}
		}

		io_set_baud(modem_fd,19200);
		io_set_CLOCAL(modem_fd,TRUE);
		io_mode_work(modem_fd, TRUE);

		display_status("berprfe Modem-Status");
		check_modem(modem_fd);
		display_status("Initialisiere Modem");
		init_modem(modem_fd);

		if (!is_netmodem) {
			lock_delete(device_lockfile);
			device_locked=FALSE;
		}
	}
}

