#ifndef TOM_JOBS_H
#define TOM_JOBS_H

#ifdef WITH_JOBCONTROL

#ifdef HAVE_TERMIOS_H
#  include <termios.h>
#elif defined(HAVE_SYS_TERMIOS_H)
#  include <sys/termios.h>
#else
#  error expected termios.h or sys/termios.h!
#endif

/* 
 * warning: to simplify things our jobs have just one process!
 */
typedef struct job_t {
	struct job_t *next;			/* next active job */
	char completed;				/* true if process has completed */
	char stopped;				/* true if process has stopped */
	int status;					/* reported status value */
	pid_t pid;					/* process ID */
	pid_t pgid;					/* process group ID - under TOM equal to pid!*/
	char notified;				/* true if user told about stopped job */
	struct termios tmodes;		/* saved terminal modes */
	int stdin, stdout, stderr;	/* standard i/o channels */
} job_t;

/* The active jobs are linked into a list.  This is its head.   */
extern job_t         *jc_first_job;
extern pid_t          jc_shell_pgid;
extern struct termios jc_shell_tmodes;
extern int            jc_shell_terminal;
extern int            jc_shell_is_interactive;

/* Find the active job with the indicated PGID.  */
job_t * jc_find_job (pid_t pgid);
/* Return true if all processes in the job have stopped or completed.  */
int jc_job_is_stopped (job_t * j);
/* Return true if all processes in the job have completed.  */
int jc_job_is_completed (job_t * j);
/* Make sure the shell is running interactively as the foreground job
 * before proceeding. */
void jc_init_shell (void);



#endif

#endif
