#ifndef LIBQUARK_H
#define LIBQUARK_H

#include "tomsysdep.h"
#include "stdlib.h" /* need this for xmalloc */
#include "error.h"
#include "mtzeit.h"
#include "entumlauten.h"

extern char * fast_strftime(char *buf,size_t buflen, 
	const char *format,struct tm *tm);
extern int strhalfcasecmp(const char *uplow, const char *low);
extern char * strtrim (char *s);
extern char *string_make(size_t anzahl,char zeichen);
char *libquark_strupr (char *s);
char *libquark_strlwr (char *s);
int strnhalfcasecmp(const char *uplow, const char *low, size_t n);
int strhalfcasecmp(const char *uplow, const char *low);

ssize_t xwrite(int fd, const void *data, size_t i);
ssize_t xread(int fd, void *data, size_t bytes);
BOOLEAN file_copy(const char *src,const char *dst);
BOOLEAN file_move(const char *src,const char *dst);
int libquark_litoa(long i, char *a);

#ifndef HAVE_STRUPR
/* could be defined in string.h */
#define strupr(x) libquark_strupr(x)
#endif
#ifndef HAVE_STRLWR
/* could be defined in string.h */
#define strlwr(x) libquark_strlwr(x)
#endif



extern char *tprintf(const char *format,...);

extern char *concat (const char *first,...);
extern char *xconcat (const char *first,...);

#if !defined(__GNUC__) || !defined(__OPTIMIZE__)
extern void *libquark_xmalloc (size_t size, const char *fname, ULONG line);
extern char *libquark_xstrdup (const char *s, const char *fname, ULONG line);
extern void *libquark_xrealloc (void *p, size_t size, const char *fname, ULONG line);
extern void *libquark_xcalloc (size_t nmemb, size_t size, const char *fname, ULONG line);
#else
static __inline__ void * 
libquark_xmalloc(size_t size, const char *fname, ULONG line)
{
	void *p;
	p=malloc(size);
	if (!p)
		error_at_line(1,0,fname,line,"out of memory");
	return p;
}
static __inline__ char * 
libquark_xstrdup (const char *s, const char *fname, ULONG line) {
	char *p;
	size_t size=strlen(s)+1;
	p=malloc(size);
	if (!p)
		error_at_line(1,0,fname,line,"out of memory");
	memcpy(p,s,size);
	return p;
}
static __inline__ void * 
libquark_xrealloc(void *block, size_t size, const char *fname, ULONG line)
{
	void *p;
	p=realloc(block,size);
	if (!p)
		error_at_line(1,0,fname,line,"out of memory");
	return p;
}
static __inline__ void *
libquark_xcalloc(size_t nmemb, size_t size,const char *fname, ULONG line)
{
	void *p;
	p=calloc(nmemb,size);
	if (!p)
		error_at_line(1,0,fname,line,"out of memory");
	return p;
}
#endif

#define XMALLOC(x)    libquark_xmalloc(x,__FILE__,__LINE__)
#define XSTRDUP(x)    libquark_xstrdup(x,__FILE__,__LINE__)
#define XREALLOC(x,y) libquark_xrealloc(x,y,__FILE__,__LINE__)
#define XCALLOC(x,y)  libquark_xcalloc(x,y,__FILE__,__LINE__)

#define XNULLCHECK(ptr) \
	do { if (!(ptr)) \
		error_at_line(1,0,__FILE__, __LINE__, "out of memory"); \
	} while(0)

#endif
