/* log.c : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
       (might be the wrong person except for offline)

    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
/*
 * log.c
 */

#include "tomsysdep.h"

#include <portab.h>
#include <stdio.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <unistd.h>
#include <syslog.h>
#include <stdarg.h>

#include "tompaths.h"
#include "modem.h"
#include "prototyp.h"
#include "box_conf.h"
#include "protokoll.h"
#include "aktdata.h"
#include "log.h"

extern USER user;

extern BRETT akt_brett;

extern int sysop_out_fd;


static FILE *mlog_f;
extern USER user;


void 
write_info_protokoll (int was, const char *frage, const char *text)
{
	char buf[1024];
	char protname[PATH_MAX];
	FILE *prot_f;

	sprintf (protname, "%s%s.inf", box_conf.log_dir, datum_pro_S ());

	prot_f = xfopen (protname, "a");
	if (ftell (prot_f) == 0)
	{
		fprintf (prot_f, "------------------------------------------------------------------------------\n");
		fprintf (prot_f, "-- Info-Protokoll vom: %s %s\n", dd_mm_yy (), hh_mm_ss ());
		fprintf (prot_f, "------------------------------------------------------------------------------\n");
	}
	#define OUTPUT(format) fprintf(prot_f,format)

	switch(was)
	{
	case INFOPROT_LOGIN:
		OUTPUT("--\n");
		OUTPUT(
"------------------------------------------------------------------------------\n"
			);
		sprintf(buf,"-- Login am: %s um: %s\n" , dd_mm_yy (), hh_mm_ss ());
		OUTPUT(buf);
		OUTPUT("--\n");
		break;
	case INFOPROT_LOGOUT:
		OUTPUT("--\n");
		sprintf(buf,"-- Logout am: %s um: %s\n", dd_mm_yy (), hh_mm_ss ());
		OUTPUT(buf);
		OUTPUT(
"------------------------------------------------------------------------------\n"
			);
		break;
	case INFOPROT_USER:
		sprintf(buf,"-- User: %s\n", user.name);
		OUTPUT(buf);
		break;
	case INFOPROT_FREE:
		sprintf(buf,"%s\n", text);
		OUTPUT(buf);
		break;
	case INFOPROT_TEXT:
		sprintf(buf,"%-4ld  %-24s%-38s%s\n",
				 (long) akt_brett.nr, 
				 frage ? frage : "", 
				 text, hh_mm_ss ());
		OUTPUT(buf);

		if (box_conf.view_user && sysop_out_fd>=0)
		{
			sprintf (buf,"%-4ld %-30.30s %-30.30s %s\n",
					(long) akt_brett.nr, 
					frage ? frage : "", 
					text, hh_mm_ss ());
			write(sysop_out_fd,buf,strlen(buf));
		}
		break;
	}
	fflush(prot_f);

	fclose (prot_f);

}	/* write_info_protokoll */


static char 
tristate_char (TRISTATE what, char c)
{
	switch (what)
	{
	case TRISTATE_NOT_TRIED:
		return ' ';
	case TRISTATE_FAILURE:
		return tolower (c);
	case TRISTATE_OK:
		return toupper (c);
	}
	return '?';
}


void 
login_protokoll_write (void)
{
	char s[200]; /* large enough to not overflow */
	char fname[PATH_MAX];
	size_t l = 0;
	FILE *f;
	time_t dauer;
	char port[3];
	char anruftypstring[20];
	char dauerstring[20];
	char flagstring[20];

	int sysloglevel=LOG_DEBUG;

	static BOOLEAN no_file_protokoll;	/* Von nun an keine Fileprotokolle mehr */

	if (akt_anruf.endzeit == 0)
		akt_anruf.endzeit = time (NULL);

	sprintf (port, "  ");

	switch (akt_anruf.anruftyp)
	{
	case ANRUF_SPEC:
		sprintf (anruftypstring, "Spec");
		break;
	case ANRUF_FAX:
		sprintf (anruftypstring, "Fax");
		break;
	case ANRUF_INTERN:
		sprintf (anruftypstring, "Intern");
		break;
	case ANRUF_CCOMMAND:
		sprintf (anruftypstring, "Command");
		break;
	case ANRUF_CBATCH:
		sprintf (anruftypstring, "CBatch");
		break;
	case ANRUF_CTAUSCH:
		sprintf (anruftypstring, "CTausch");
		break;
	case ANRUF_USER:			/* Userlogin */
		sprintf (anruftypstring, "%6ld %c",
				 (long) akt_anruf.anrufnummer,
				 (akt_anruf.ist_zahler) ? '$' : ' ');
		break;
	case ANRUF_EVENT:			/* Event */
		sprintf (anruftypstring, "Event");
		break;

	default:					/* nur fr den Fall */
		sprintf (anruftypstring, "??%02d??", akt_anruf.anruftyp);
		break;
	}
	/* Dauer berechnen und String daraus vorfertigen */
	dauer = akt_anruf.endzeit - akt_anruf.userstartzeit; /* -3600? */
#if 0
	/* das gab bei libc-5 eine Stunde offset. */
	strftime (dauerstring, sizeof (dauerstring), "%H:%M", localtime (&dauer));
#else
	sprintf(dauerstring,"%02ld:%02ld", (dauer / 60) % 60, dauer % 60);
#endif

	/* Nun die diversen Flags */
	flagstring[0] = tristate_char (akt_anruf.tausch_status, 't');
	flagstring[1] = tristate_char (akt_anruf.chat_status, 'C');
	flagstring[2] = tristate_char (akt_anruf.batch_status, 'B');
	flagstring[3] = (akt_anruf.neuuser) ? 'N' : ' ';
	flagstring[4] = '\0';

	/*
	 * Vorbereitung vorbei, nun schreiben der Protokolldatei
	 */
	l = sprintf (s, "%-8.8s", anruftypstring);

	l += sprintf (s + l, "%-27.27s ", akt_anruf.name);

	/* Startzeit eintragen */
	l += strftime (s + l, 7, "%H:%M ", localtime (&akt_anruf.userstartzeit));

	/* Endzeit eintragen */
	l += strftime (s + l, 7, "%H:%M ", localtime (&akt_anruf.endzeit));

	/* Dauer eingetragen */
	l += sprintf (s + l, "%-5.5s ", dauerstring);

	/* Portnummer schreiben. Port 0, wenn kein Eintrag */
	l += sprintf (s + l, "%s ", port);

	if (akt_anruf.anruftyp == ANRUF_USER || akt_anruf.anruftyp == ANRUF_SPEC) {
		l += sprintf (s + l, "%5.5s ", akt_connection.carrierstring);
		/* nun Flags und Carrier Lost */
		l += sprintf (s + l, "%-6.6s %3.3s",
					  flagstring,
					  akt_anruf.zuletzt);
	} else {
		l += sprintf (s + l, " %s", akt_anruf.free_string);
	}

	if (!no_file_protokoll)
	{
		sprintf (fname, "%s%s.log", box_conf.log_dir, datum_pro_S ());
		/* 
		   Kein Bin-Append, weil keine Binrdaten.
		 */
		f = fopen (fname, "a");
		if (f == NULL)
		{
			no_file_protokoll = TRUE;	/* Um Loops zu vermeiden */
			dateifehler (fname);
		}
		fprintf (f, "%s\n", s);
		fclose (f);
		sysloglevel=LOG_DEBUG;
	}
	else
		sysloglevel=LOG_INFO;
	/* 
	   Nun den Syslogeintrag schreiben. Da syslog einige Daten selbst
	   einsetzt, mu das Format zwangsweise ein anderes sein.
	   Syslog loggt: Datum und Uhrzeit, Host, "tom", Prozess-ID.
	   Es ist auch keineswegs wichtig, das Format besonders schn zu machen.
	   Mit einiger Wahrscheinlichkeit sind das nicht die einzigen Ausgaben
	   auf derselben virtuellen Console oder demselben Fenster, folglich
	   mu man erkennen knnen, was diese Meldung sagen will.
	   Verfgbare Zeilenlnge:
	   = 79 - sizeof("Feb  2 14:29:19 hostname tom[12345]: ")
	   1234567890123456789012345678901234567"
	   Man kann also von hchstens 39 verfgbaren Zeichen ausgehen
	   Format: 
	   1. Zeile: 123456  $irgendein sehr langer Benutzer (cut)
	   2. Zeile:  port HH:MM:SS CONN ???                 (useranruf)
	   2. Zeile:  port HH:MM:SS <irgendein wilder Text>  (sonstwas)
	   123456789012345678901234567890123456789
	 */
	/* Was los war, und wie derjenige heit (auf 30 Zeichen gekrzt) */
	syslog (sysloglevel, "%-8.8s %-30.30s", anruftypstring, akt_anruf.name);

	/* Dauer eingetragen */
	l += sprintf (s + l, "%-8.8s ", dauerstring);
	if (akt_anruf.anruftyp == ANRUF_USER)
	{
		syslog (sysloglevel, " %-4.4s %8.8s %-5.5s %3.3s %s",
				port, dauerstring, akt_connection.carrierstring, akt_anruf.zuletzt,
				flagstring);
	}
	else
	{
		syslog (sysloglevel, " %-4.4s %8.8s %-24.24s",
				port, dauerstring, akt_anruf.free_string);
	}
}

 

/*
 * einen Eintrag ins Msglog schreiben.
 */
void
msglog (const char *format,...)
{
	va_list param;
	char fname[PATH_MAX];

	if (!mlog_f)
	{
		time_t t;
		sprintf (fname, "%s%s.msg", box_conf.log_dir, datum_pro_S());
		mlog_f = xfopen (fname, "a");

		fprintf(mlog_f,"U %lu %s\n",(ULONG) user.nummer, user.name);
		time(&t);
		fputs("- ",mlog_f);
		fputs(ctime(&t),mlog_f);
	}
	if (format && strchr(format,'%'))
	{
		va_start (param, format);
		(void) vfprintf (mlog_f, format, param);
		va_end (param);
	}
	else if (!format || 0==strcmp(format,"%s"))
	{
		char *p;
		va_start (param, format);
		p=va_arg(param,char *);
		if (p)
			fputs(p,mlog_f);
		va_end (param);
	}
	else /* format ohne % */
	{
		fputs(format,mlog_f);
	}
}

void
msglog_end (void)
{
	if (mlog_f)
	{
		fclose (mlog_f);
		mlog_f = NULL;
	}
}

