/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact1: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/

#include "tomsysdep.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <signal.h>
#include <setjmp.h>
#include <errno.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <syslog.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>

#if defined(HAVE_TERMIOS_H)
#  include <termios.h>
#elif defined(HAVE_SYS_TERMIOS_H)
#  include <sys/termios.h>
#elif defined(HAVE_TERMIO_H)
#  include <termio.h>
#elif defined(HAVE_SYS_TERMIO_H)
#  include <sys/termio.h>
#endif

#ifdef HAVE_SYS_MODEM_H
#include <sys/modem.h>
#endif
#ifdef HAVE_SYS_TERMIOX_H
#include <sys/termiox.h>
#endif

#if defined(__TOS__)
#include <tos.h>
#endif

#include "prototyp.h"
#include "aktdata.h"
#include "box_conf.h"
#include "io.h"

extern BOOLEAN modem_vorhanden;

#ifdef USE_SOFT_CARRIER
static void io_soft_carrier(BOOLEAN soft);
#endif

static BOOLEAN no_tty=FALSE;


BOOLEAN io_test_modem_lines(int fd)
{
#if defined(TIOCM_DTR) && defined(TIOCMGET)
	int status;
	if (ioctl(fd,TIOCMGET,&status)==-1)
		return FALSE;
	return TRUE;

#elif defined(MCGETA) && defined(MDTR)

	long mflags;
	if (ioctl(fd,MCGETA,&mflags))
		return FALSE;
	return TRUE;
#endif
	/* just to be sure */
	return FALSE;
}


#ifdef USE_SOFT_CARRIER
/*
	Carrier-Modus auf Soft- oder Hardware setzen

	Klartext: CLOCAL setzen (TRUE) oder lschen (FALSE), jedenfalls
	laut Linuxkernel

	->			hard	:	TRUE = Software, FALSE = Hardware
*/
static void io_soft_carrier(BOOLEAN soft)
{
	int soft_carrier = soft;

#ifdef TIOCSSOFTCAR
	ioctl(remote_in_fd,TIOCSSOFTCAR,&soft_carrier);
#elif defined(__TOS__)
#else
#error "TIOCSSOFTCAR missing"
#endif

} /* io_soft_carrier */
#endif

#ifdef EXPLIZITE_CARRIERABFRAGE
/*
	Carrier abfragen. Sighup wird ohnehin abgefangen.
	- call_extern()
	- mbchar()
*/
BOOLEAN io_test_carrier(int fd)
{
	if (no_tty)
		return TRUE;
#if defined(TIOCMGET) && defined(TIOCM_CAR)
	{
		int status;
		ioctl(fd,TIOCMGET,&status);
		if (status & TIOCM_CAR)
			return TRUE;
		else
			return FALSE;
	}
#elif defined(MCGETA) && defined(MDCD)
	{
		long mflags;
		ioctl(fd,MCGETA,&mflags);
		if (mflags & MDCD)
			return TRUE;
		else
			return FALSE;
	}
#else
#error no way to detect carrier loss
#endif

} /* io_test_carrier */
#endif


/*
	Portstatus ermitteln (POLL)
	Ersatz fr Bconstat

	12.06.93 Frank
	21.10.93 Frank, auf portables select umgestellt

	Achtung: Diese Fn. pollt (Gefahr von busy-waiting)
	Aufrufer sind:
	- main.c (nur TOS)
	- chat.c
	- boxio.c (userbreak,mbchar)
	- modem.c (modem_ok)
	- connect.c
*/
BOOLEAN io_charavail(int device)
{
	fd_set set;
	struct timeval zeit;

	if (device<0)
		return FALSE;

	FD_ZERO(&set);
	FD_SET(device,&set);
	zeit.tv_sec  = 0;
	zeit.tv_usec = 0;

	return (select(device+1,&set,0,0,&zeit) > 0);

} /* io_charavail */



/*
	warten, bis Puffer an port_fd bertragen wurde
	UND ungelesene Eingaben verwerfen
*/
void io_flush(int fd)
{
	if (no_tty)
		return;
	tcflush(fd,TCIOFLUSH);
	tcflush(fd,TCIOFLUSH); /* sollte nicht ntig sein, aber schadet nicht */

} /* io_flush */



/*
	warten, bis Puffer an port_fd bertragen wurde
*/
void io_drain(int fd)
{
	if (no_tty)
		return;
	tcdrain(fd);

} /* io_drain */



/*
	Break senden
*/
void io_send_break(int fd)
{
	if (no_tty)
		return;
	tcsendbreak(fd,0);

} /* io_send_break */



BOOLEAN
io_mode_work(int fd, BOOLEAN getty_mode)
{
	struct termios work;

	if (tcgetattr(fd,&work))
		return FALSE;

	if (getty_mode && (fd == STDIN_FILENO) && isatty(STDIN_FILENO))
	{
		work.c_cflag &= ~CSIZE;
		work.c_cflag |= CS8;
		work.c_lflag &= ~ICANON;
		work.c_lflag &= ~ECHO;
		work.c_cc[VMIN]  = 1;
		work.c_cc[VTIME] = 0;
		if (tcsetattr(STDIN_FILENO,TCSADRAIN,&work))
			return FALSE;
		return TRUE;
	}

	/* cflags: clocal in Ruhe lassen! das ist weiter oben schon 
	 * gesetzt worden!  Speed in Ruhe lassen!
	 */
	if (getty_mode)
	{
		work.c_cflag &= ~(CSIZE);
		work.c_cflag |= CS8;     /* 8 Bit breite Zeichen */
		work.c_cflag |= CREAD;   /* von diesem Terminal kann input gelesen werden */
		work.c_cflag |= HUPCL;
		work.c_cflag &= ~CSTOPB; /* nur 1 Stopbit */
		work.c_cflag &= ~PARENB; /* no parity */
		work.c_cflag &= ~PARODD; /* even parity, berflssig hier */
#ifdef CRTSCTS
		/* undefined under HP/UX */
		if (box_conf.modem_xonxoff)
			work.c_cflag &= ~CRTSCTS;
		else
			work.c_cflag |= CRTSCTS;
#endif
	}

	/*
	 * c_iflag:
	 */

	if (getty_mode)
	{
		/*
		 * Parity Fehler ignorieren, mit dem erwnschten Nebeneffekt, da 
		 * ESC als EIN Zeichen statt als \033\033 geliefert wird.
		 * Parity Fehler liefern eine 0 beim auslesen.
		 * ntig bei local?
		 */
		work.c_iflag &= INPCK;  /* parity check disabled */
		work.c_iflag &= ~ISTRIP;/* 8. Bit nicht killen */
		work.c_iflag |= IGNBRK; /* Eine Reihe 0bits ist *kein* Break */
		if (box_conf.modem_xonxoff)
		{
			work.c_iflag |= IXON;  /* ^S beachten */
			work.c_iflag |= IXOFF; /* ^Q beachten */
		}
		else
		{
			work.c_iflag &= ~IXON;  /* ^S ^Q nicht besonders beachten */
			work.c_iflag &= ~IXOFF; /* dito */
		}
		work.c_iflag &= ~(ICRNL);
		work.c_iflag &= ~(IGNCR);
		work.c_iflag &= ~(INLCR);
	}

#if 0
	/*
	 * CR/LF -> LF Wandlung beim Input wre hilfreich, kann aber 
	 * problematisch sein, da nicht alle Stellen der Quark mit LF 
	 * statt CR klar kommen drften. Abhilfe: mbchar kann LD wieder
	 * nach CR konvertieren - das klingt zwar idiotisch, ist aber
	 * kaum Overhead.
	 */
	work.c_iflag |= IGNCR | ICRNL;
	work.c_iflag &= ~INLCR;
#endif

	if (getty_mode)
	{
		work.c_oflag = OPOST;
	}
	/*
	 * c_lflag:
	 */
	work.c_lflag &= ~(ICANON);
	work.c_lflag &= ~(ECHO);
#ifdef ECHOE
	work.c_lflag &= ~(ECHOE);
#endif
#ifdef ECHOPRT
	work.c_lflag &= ~(ECHOPRT);
#endif
#ifdef ECHOK
	work.c_lflag &= ~(ECHOK);
#endif
#ifdef ECHOKE
	work.c_lflag &= ~(ECHOKE);
#endif
#ifdef ECHONL
	work.c_lflag &= ~(ECHONL);
#endif
#ifdef ECHOCTL
	work.c_lflag &= ~(ECHOCTL);
#endif
	work.c_lflag &= ~(ISIG); /* ### change that! - userunfreundlich */
	work.c_lflag &= ~(IEXTEN);
	work.c_lflag &= ~(NOFLSH);
	work.c_lflag &= ~(TOSTOP);
#ifdef ALTWERASE
	work.c_lflag &= ~(ALTWERASE);
#endif
#ifdef FLUSHO
	work.c_lflag &= ~(FLUSHO);
#endif
#ifdef NOKERNINFO
	work.c_lflag &= ~(NOKERNINFO);
#endif
#ifdef PENDIN
	work.c_lflag &= ~(PENDIN);
#endif

	work.c_cc[VMIN]  = 1;
	work.c_cc[VTIME] = 0;
	work.c_cc[VSTART] = 17; /* ^Q */
	work.c_cc[VSTOP] = 19; /* ^S */
	if (tcsetattr(fd,TCSADRAIN,&work))
		return FALSE;

#if defined (TCGETX)
	/* HP-UX way of setting RTS/CTS. Braindead? At least strange. */

	/* wenn ich mich recht erinnere mssen RTSXOFF und CTSXON 
	 * auch gesetzt sein, wenn nur XON/XOFF verwendet wird. */
	if (!getty_mode)
	{
		struct termiox tx;
		if (ioctl(fd,TCGETX,&tx))
			return FALSE;
		tx.x_hflag|=RTSXOFF;
		tx.x_hflag|=CTSXON;
		if (ioctl(fd,TCSETX,&tx))
			return FALSE;
	}
#endif
	return TRUE;
}



	/* Diese Funktion wird aufgerufen, wenn io_mode_work schonmal
	 * aufgerufen wurde. Sie sollte tunlichst nur Flags setzen,
	 * die ein spterer io_mode_work-Aufruf zurcksetzen kann!
	 */
BOOLEAN
io_mode_sane(int fd)
{
	struct termios work;

	if (tcgetattr(fd,&work))
		return FALSE;

	/* Inputflags:  */
	work.c_iflag &= INPCK;  /* parity check disabled */
	work.c_iflag &= ~ISTRIP;/* 8. Bit nicht killen */
	work.c_iflag |= IGNBRK; /* Eine Reihe 0bits ist *kein* Break */
	work.c_iflag |= IXON;  /* ^S ^Q beachten */
#ifdef IXANY
	work.c_iflag |= IXANY; /* und jedes Zeichen starten die Ausgabe */
#endif

	/* Control flags: (CLOCAL in Ruhe lassen, damit schlagen wir uns
	 * weiter oben herum, *hier* wissen wir nichts darber
	 */
    work.c_cflag &= ~(CSIZE);
	work.c_cflag |= CS8;     /* 8 Bit breite Zeichen */
	work.c_cflag |= CREAD;   /* von diesem Terminal kann input gelesen werden */
	work.c_cflag |= HUPCL;
    work.c_cflag &= ~CSTOPB; /* nur 1 Stopbit */
    work.c_cflag &= ~PARENB; /* no parity */
    work.c_cflag &= ~PARODD; /* even parity, berflssig hier */

	work.c_oflag = OPOST;
#ifdef TAB3
	work.c_oflag |= TAB3;
#endif
#ifdef ONLCR
	/* GNU/BSD-Erweiterung, aber notwendig! */
	work.c_oflag |= ONLCR;
#endif

	work.c_lflag = ECHOK | ECHOE | ECHO | ISIG | ICANON;

	work.c_cc[VMIN]  = 1;
	work.c_cc[VTIME] = 0;
	if (tcsetattr(fd,TCSADRAIN,&work))
		return FALSE;
	return TRUE;
}

/*
	Baudrate setzen
	XON/XOFF + 8N1
*/
BOOLEAN 
io_set_baud(int fd, unsigned long baud)
{
	struct speed
	{
		unsigned long baudrate;
		int baudcode;
	}
	baudmap[] =
	{
		{	    0,	B0			},
		{	   50,	B50		},
		{	   75,	B75		},
		{	  110,	B110		},
#ifdef B134
		{	  134,	B134		},
#endif
#ifdef B135
		{	  135,	B135		},
#endif
		{	  150,	B150		},
		{	  200,	B200		},
		{	  300,	B300		},
		{	  600,	B600		},
		{	 1200,	B1200		},
		{	 1800,	B1800		},
		{	 2400,	B2400		},
		{	 4800,	B4800		},
		{	 9600,	B9600		},
		{	19200L,	B19200	},
		{	38400L,	B38400	},
#ifdef B57600
		{	57600L, B57600  },
#endif
#ifdef B115200
		{	115200L, B115200  },
#endif
#ifdef B230400
		{	230400L, B230400  },
#endif
#ifdef B460800
		{	460800L, B460800  },
#endif
	};
	int i = 0;

	if (no_tty)
		return TRUE;

	while (i < (int) (sizeof(baudmap)/sizeof(struct speed)))
	{
		if (baudmap[i].baudrate == baud)
		{
			struct termios ta;
			tcgetattr(fd,&ta);
			cfsetospeed(&ta,baudmap[i].baudcode);
			cfsetispeed(&ta,baudmap[i].baudcode);
			tcsetattr(fd,TCSADRAIN,&ta);
#ifdef __TOS__
			/* wg. Synchronisationsfehler zwischen Rsconf und MFP 40ms warten */
			msleep(40);
#endif
			return TRUE;
		}
		i++;
	}

	return FALSE;
} /* io_set_baud */

BOOLEAN 
io_save_restore_attribs(int fd, BOOLEAN save)
{
	static struct termios tio;
	if (no_tty)
		return TRUE;
	if (save)
	{
		if (tcgetattr(fd,&tio))
			return FALSE;
	}
	else
		if (tcsetattr(fd,TCSADRAIN,&tio))
			return FALSE;
	return TRUE;
}

static struct termios stdin_tio;
void 
io_save_stdin(void)
{
	tcgetattr(STDIN_FILENO,&stdin_tio);
}

int old_stdin_fileno=STDIN_FILENO;

void
io_restore_stdin(void)
{
	/* tricky thing - ab einem gewissen Punkt ist stdin != altes Stdin */
	tcsetattr(old_stdin_fileno,TCSADRAIN,&stdin_tio);
}

void
io_tty_test(int fd)
{
	if (!isatty(fd))
		no_tty=TRUE;
}

/*
 * 1. nicht auf der Console (die Werte, die da eingestellt sind, 
 *    stimmen mit viel hherer Wahrscheinlichkeit als die in den
 *    Userdaten.
 * 2. Mindestens 80x24, weil sonst zu viele Programme versagen.
 */
void
io_change_term_size (int fd, int x, int y)
{
#ifdef TIOCGSIZE
	struct ttysize win;

#elif defined(TIOCGWINSZ)
	struct winsize win;

#endif

#if 0
	/* 
	 * Da kommt das *berhaupt* nicht gut.
	 */
	if (akt_connection.is_console)
		return;
#endif

#ifdef TIOCGSIZE
	if (ioctl (fd, TIOCGSIZE, &win))
		return;
	if (y && y>24)
		win.ts_lines = y;
	else
		win.ts_lines = 24;
	if (x && x>80)
		win.ts_cols = x;
	else
		win.ts_cols = 80;
	ioctl (fd, TIOCSSIZE, &win);
#elif defined TIOCGWINSZ
	if (ioctl (fd, TIOCGWINSZ, &win))
		return;
	if (y && y >24)
		win.ws_row = y;
	else
		win.ws_row = 24;
	if (x && x>80)
		win.ws_col = x;
	else
		win.ws_col = 80;
	ioctl (fd, TIOCSWINSZ, &win);
#endif
}

BOOLEAN
io_get_term_size (int fd, int *x, int *y)
{
#ifdef TIOCGSIZE
	struct ttysize win;

#elif defined(TIOCGWINSZ)
	struct winsize win;

#endif

#ifdef TIOCGSIZE
	if (ioctl (fd, TIOCGSIZE, &win))
		return FALSE;
	if (y)
		*y=win.ts_lines;
	if (x)
		*x=win.ts_cols;
#elif defined TIOCGWINSZ
	if (ioctl (fd, TIOCGWINSZ, &win))
		return FALSE;
	if (y)
		*y=win.ws_row;
	if (x)
		*x=win.ws_col;
#else
	{
		const char *s;
		s=getenv("LINES");
		if (s)
			*y=strtol(s,NULL,10);
		else
			*y=25;
		s=getenv("COLUMNS");
		if (s)
			*x=strtol(s,NULL,10);
		else
			*x=80;	
	}
#endif
	return TRUE;
}

#define MKFLAG(which) \
static BOOLEAN io_tio_set_flag_##which(int fd, int val, BOOLEAN on) \
{ struct termios tio; \
	if (tcgetattr(fd,&tio)) return FALSE; \
	if (on) tio.which |= (val); \
	else tio.which &= ~(val); \
	if (tcsetattr(fd,TCSADRAIN,&tio)) return FALSE; \
	return TRUE; \
}

MKFLAG(c_lflag)
MKFLAG(c_iflag)
MKFLAG(c_oflag)
MKFLAG(c_cflag)

BOOLEAN 
io_set_CLOCAL(int fd, BOOLEAN on)
{
	return io_tio_set_flag_c_cflag(fd,CLOCAL,on);
}

BOOLEAN
io_set_ICRNL(int fd, BOOLEAN on)
{
	return io_tio_set_flag_c_iflag(fd,ICRNL,on);
}

BOOLEAN
io_set_INLCR(int fd, BOOLEAN on)
{
	return io_tio_set_flag_c_iflag(fd,INLCR,on);
}

BOOLEAN
io_set_IGNCR(int fd, BOOLEAN on)
{
	return io_tio_set_flag_c_iflag(fd,IGNCR,on);
}

BOOLEAN
io_set_ONLCR(int fd, BOOLEAN on)
{
#ifdef ONLCR
	return io_tio_set_flag_c_oflag(fd,ONLCR,on);
#else
	return TRUE;
#endif
}

BOOLEAN
io_set_ECHO(int fd, BOOLEAN on)
{
	return io_tio_set_flag_c_lflag(fd,ECHO,on);
}

BOOLEAN
io_set_ISIG(int fd, BOOLEAN on)
{
	return io_tio_set_flag_c_lflag(fd,ISIG,on);
}

/*
	DTR (de)aktivieren
*/
BOOLEAN
io_set_DTR(int fd, BOOLEAN on)
{
#if defined(TIOCM_DTR) && defined(TIOCMGET) && defined(TIOCMSET)

	int status;
	if (ioctl(fd,TIOCMGET,&status))
		return FALSE;
	if (on)
		status |= TIOCM_DTR;
	else
		status &= ~TIOCM_DTR;
	if (ioctl(fd,TIOCMSET,&status))
		return FALSE;

#elif defined(TIOCSDTR)
	struct termios term;

	if (tcgetattr(fd,&term))
		return FALSE;
	if (on)
		ioctl(fd,TIOCSDTR,&term);
	else
		ioctl(fd,TIOCCDTR,&term);
	if (tcsetattr(fd,TCSADRAIN,&term))
		return FALSE;

#elif defined(MCGETA) && defined(MDTR)

	long mflags;
	if (ioctl(fd,MCGETA,&mflags))
		return FALSE;
	if (on)
		mflags |= MDTR;
	else
		mflags &= ~MDTR;
	if (ioctl(fd,MCSETA,&mflags))
		return FALSE;

#elif defined(WINNT)
#pragma warn "no way to set/clear dtr under WinNT"
#else
#error "no way to set/clear dtr"
#endif

	/* ### sicherheitshalber etwas warten */
	msleep(on ? 200 : 500);
	return TRUE;

} /* io_set_dtr */


