/* mtzeit2tt.c : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
       (might be the wrong person except for offline)

    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
/* Funktionen rund um das MausTauschzeitformat */

#include "config.h"

#include "portab.h"
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <ctype.h>

#include "mtzeit.h"

/* LIBQUARK_FUNCTION(Funktion,BOOLEAN,mtzeit_to_time_t,
[[(char *@var{buf}, time_t *@var{tim})]],[[
Die Funktion konvertiert die MausTauschzeit in @var{buf} (mit oder
ohne Sekunden) in @code{Sekunden seit 1970} und legt das Ergebnis
in @var{tim} ab.

Diese Funktion ist deutlich schneller als einfach, auf mktime basierende
Varianten (und ignoriert smtliche mgliche nderungen von Zeitzonen
und Environmentvariablen). 

Liefert TRUE bei Erfolg, sonst FALSE.
]] */


struct dst
{
	int start_mon;
	int start_day;
	int end_mon;
	int end_day;
};

static struct dst dst_ar[]=
{
	/* 1990 */ { 3,25, 9,30},
	/* 1991 */ { 3,31, 9,29},
	/* 1992 */ { 3,29, 9,27},
	/* 1993 */ { 3,28, 9,26},
	/* 1994 */ { 3,27, 9,25},
	/* 1995 */ { 3,26, 9,24},
	/* 1996 */ { 3,31, 10,27},
	/* 1997 */ { 3,30, 10,26},
	/* 1998 */ { 3,29, 10,25},
	/* 1999 */ { 3,28, 10,31},
	/* 2000 */ { 3,26, 10,29},
	/* 2001 */ { 3,25, 10,28},
	/* 2002 */ { 3,31, 10,27},
	/* 2003 */ { 3,30, 10,26},  
	/* 2004 */ { 3,28, 10,31}
};
/*
	Zeit der Form YYYYMMDDHHMM[ss] in time_t umwandeln

	Sinn: Verzicht auf mktime und der damit verbundenen zeitaufwendigen 
	Auswertung von timezone, Environment etc.
	Vor 1990 wird weiterhin mktime() verwendet, danach allerdings die 
	Anzahl der seit 1.1.1970 vergangenen Sekunden auf Basis von mktime(1990) 
	selber berechnet.

	Autor: fb. Fixes: uo.
*/
/*
 * Zerlegt MausTauschzeit nach time_t.
 */
BOOLEAN
mtzeit_to_time_t(const char *s, time_t *tt)
{
	static time_t ref = 0;
	struct tm zeit;
	long days_to_month[] = { 0,31,59,90,120,151,181,212,243,273,304,334,365 };
	long days;
	struct dst *dst;
	int downset=0;
	size_t l;
	if (!tt)
		return FALSE;

	l = strlen (s);

	if (l != 12 && l != 14)
		return FALSE;

	/* Referenz-time_t vom 1.1.1990 berechnen */
	if (!ref)
	{
		zeit.tm_year = 1990 - 1900;
		zeit.tm_mon  = 0;
		zeit.tm_mday = 1;
		zeit.tm_hour = 0;
		zeit.tm_min  = 0;
		zeit.tm_sec  = 0;
		zeit.tm_isdst= 0; /* ja, 0: keine DST */
		zeit.tm_yday = 0;
		zeit.tm_wday = 0;
		ref = mktime(&zeit);
	}

	zeit.tm_year = (*s - '0')*1000 + (*++s - '0')*100 + (*++s - '0')*10 + (*++s - '0');
	if (zeit.tm_year < 1990)
	{
		/* DST etc. ermitteln */
		if (!mtzeit_to_tm(s-4,&zeit))
			return FALSE;

		*tt=mktime(&zeit);
		if (*tt==(time_t) 0)
			return FALSE;
		return TRUE;
	}

	zeit.tm_mon  = (*++s - '0')*10 + (*++s - '0') -1;
	zeit.tm_mday = (*++s - '0')*10 + (*++s - '0');
	zeit.tm_hour = (*++s - '0')*10 + (*++s - '0');
	zeit.tm_min  = (*++s - '0')*10 + (*++s - '0');
	if (isdigit(*++s))
		zeit.tm_sec = (*s - '0')*10 + (*++s - '0');
	else
		zeit.tm_sec = 0;

	/*
		eigene Zeitberechnung inkl. Schaltjahre auf Basis des time_t Wertes 
		vom 1.1.1990
	*/
	/* Anzahl Tage ohne Beruecksichtigung der Schaltjahre */
	days = (zeit.tm_year - 1990) * 365L + days_to_month[zeit.tm_mon] + zeit.tm_mday -1;
	/* Anzahl Schalttage berechnen, letztes (vor 1990) war 1988 
	 * Berechnung ist gltig bis 2400 - das drfte reichen :-) */
	days += (zeit.tm_year - 1988) /4;
	if ((zeit.tm_year - 1988) %4 == 0)
	{
		/* wir sind in einem Schaltjahr */

		/* In Januar und Februar um einen Tag nach unten korrigieren */
		if (zeit.tm_mon<2)
			days--;
	}

	if (zeit.tm_year-1990 <(int)(sizeof(dst_ar)/sizeof(dst)))
	{
		dst=dst_ar+(zeit.tm_year-1990);
		downset=-3600;
		if (zeit.tm_mon+1<dst->start_mon)
			downset=0;
		else if (zeit.tm_mon+1==dst->start_mon 
			&& zeit.tm_mday<dst->start_day)
			downset=0;
		else if (zeit.tm_mon+1==dst->start_mon 
			&& zeit.tm_mday==dst->start_day
			&& zeit.tm_hour<3)
			downset=0;
		else if (zeit.tm_mon+1>dst->end_mon)
			downset=0;
		else if (zeit.tm_mon+1==dst->end_mon 
			&& zeit.tm_mday>dst->end_day)
			downset=0;
		else if (zeit.tm_mon+1==dst->end_mon 
			&& zeit.tm_mday==dst->end_day
			&& zeit.tm_hour>2)
			downset=0;
	}

	*tt=ref + days * 86400L + zeit.tm_hour * 3600L + zeit.tm_min * 60 + zeit.tm_sec + downset;
	return TRUE;

} /* mtzeit_to_time_t */
