/* infile.c : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB3 (mausnet)
       (might be the wrong person except for offline)

    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
/*
	Offline - MausTausch-FrontEnd

	Infile-Erstellung/Auswertung
	(c) 1992-94 bei Frank Baumgart

	begonnen am 01.05.1992
	Version vom 01.11.1994
*/


#include "portab.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>
#include <sys/types.h>
#include <unistd.h>

#include "system.h"

#ifdef HAVE_DIRENT_H
# include <dirent.h>
# define NAMLEN(dirent) strlen((dirent)->d_name)
#else
# define dirent direct
# define NAMLEN(dirent) (dirent)->d_namlen
# ifdef HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# ifdef HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# ifdef HAVE_NDIR_H
#  include <ndir.h>
# endif
# ifdef HAVE_DIRECT_H
#  include <direct.h>
# endif
#endif

#include "frontend.h"
#include "struktur.h"
#include "mdb/msgid.h"
#include "message.h"
#include "post_fe.h"
#include "mdb/mt_cmd.h"
#include "mdb_conf.h"
#include "screen/io.h"
#include "tool.h"
#include "fehler.h"
#include "missing/missing.h"

extern CONFIG config;


static BOOLEAN copy_message(FILE *inf_f,long nr,const char *net_id);

static BOOLEAN export_status2box(FILE *infile,const char *username);

#if !defined(SERVER)
static BOOLEAN infile_packen(void);
#endif
#if !defined(CLIENT)
static void infile_status_reset(const char *name,const char *status_log);
#endif



#if !defined(SERVER)
/*
	Kommando fr nchstes Infile anhngen

	->			cmd	:	Kommando, ohne fhrendes ':' und \r\n, z.B. "G+"
				par	:	Parameter
*/
void infile_cmd(const char *cmd,const char *par)
{
	char pfad[PATH_MAX];
	FILE *cmd_f;

	strcpy(pfad,config.messages);
	strcat(pfad,"tausch.cmd");
	if ((cmd_f = fopen(pfad,BINAPPEND)) == NULL)
		return;
	fprintf(cmd_f,":%s%s\r\n",cmd,par);
	fclose(cmd_f);

} /* infile_cmd */



void bestelle_gruppe(const char *gruppe,BOOLEAN an)
{
	infile_cmd((an) ? "G+" : "G-",gruppe);

} /* bestelle_gruppe */



void mitglied(const char *gruppe,const char *user,BOOLEAN an)
{
	char tmp[MAX_HDR];

	sprintf(tmp,"%s%c%s",user,(an) ? '>' : '<',gruppe);
	infile_cmd("GU",tmp);

} /* mitglied */



/*
	Infile einpacken

	wird nur verwendet, wenn CRUNCH definiert ist

	mgliche Fehler:
	- Packer nicht gefunden
	- Packen fehlerhaft
*/
static BOOLEAN infile_packen(void)
{
	char name[128];
	char *cmd = NULL;
	char temp[PATH_MAX];

	if (!*config.crunch)
		return TRUE;

	strcpy(name,config.crunch);
	/* Kommandozeile ? */
	if ((cmd = strchr(name,' ')) != NULL)
		*cmd++ = 0;
	puts("packe Infile");
	printf("%s %s\n",name,cmd);
	sprintf(temp,"%s %s",name,(cmd) ? cmd : "");
	(void) shell(temp);

	return TRUE;

} /* infile_packen */



/*
	Infile zusammenstellen

	- Infofile-CRCs
	- Gruppen(ab)bestellungenn
	- Messagestatus
	- Kommentare, Antworten und neue Messages
	- Kopien

	- ein evtl. bereits existierendes Infile wird dabei gelscht
	- sollten ITI und/oder ITG noch nicht vorhanden sein, werden sie bestellt
*/
BOOLEAN infile_erstellen(void)
{
	char inf_name[PATH_MAX];
#ifdef __WATCOMC__
	char ts[PATH_MAX];
#endif
	FILE *inf_f;
	FILE *grp_f;
	DIR *dir;
	struct dirent *entry;

	strcpy(inf_name,config.export);
	strcat(inf_name,"infile.txt");
	if ((inf_f = fopen(inf_name,BINWRITE)) == NULL)
	{
		fehler(INF_CREATE);
		return FALSE;
	}
	/* infile puffern */
	pufferung(inf_f,NULL,_IOFBF,16384);

	/* Kommandoblock schreiben */
	fputs("#CMD\r\n",inf_f);
	fputs(":PSA\r\n",inf_f);				/* neue Statusmeldungen */
	if (config.req_pm)
		fputs(":PN\r\n",inf_f);				/* neue PMs */
	if (config.req_om)
		fputs(":ON\r\n",inf_f);				/* neue OMs */
	if (config.req_rename)
		fputs(":REN\r\n",inf_f);			/* Gruppenumbenennungen */

	/* Infofile-CRCs eintragen */
#if defined(__WATCOMC__)
	strcpy(ts,config.infofiles);
	strcat(ts,"*.*");
	if ((dir = opendir(ts)) != NULL)
#else
	if ((dir = opendir(config.infofiles)) != NULL)
#endif
	{
		while ((entry = readdir(dir)) != NULL)
		{
			size_t len;

			/* Header-Dateien auslesen */
			len = strlen(entry->d_name);

#if defined(EMX)
			/* Bug: DJGPP liefert nur kleingeschriebene Dateinamen */
			strupr(entry->d_name);
#endif

			if ((entry->d_name[0] == 'I' || entry->d_name[0] == 'J')
			&& (len > 6)
			&& (!stricmp(entry->d_name+len-3,"crc")))
			{
				char eintrag[80];
				FILE *crc_f;
				char *p;

				strcpy(inf_name,config.infofiles);
				strcat(inf_name,entry->d_name);

				if ((crc_f = fopen(inf_name,"r")) != NULL)
				{
					entry->d_name[len-4] = 0;		/* 4 = strlen(".crc") */
					if (!fgets(eintrag,sizeof(eintrag),crc_f))
						/* leere CRC-Datei -> Anforderung erzwingen */
						strcpy(eintrag,"-1");

					if ((p = strchr(eintrag,' ')) != NULL)
						*p = 0;
					fprintf(inf_f,":%s %s\r\n",entry->d_name,eintrag);
					fclose(crc_f);
				}
			}
		}
		closedir(dir);
	}

	/* falls noch keine ITI/ITG vorhanden sind, diese bestellen */
	strcpy(inf_name,config.infofiles);
	strcat(inf_name,"ITI.txt");
	if (!fileexist(inf_name))
		fputs(":ITI -1\r\n",inf_f);
	strcpy(inf_name,config.infofiles);
	strcat(inf_name,"ITG.txt");
	if (!fileexist(inf_name))
		fputs(":ITG -1\r\n",inf_f);

	/* Tausch-Kommandos, z.B. Gruppen(ab)bestellungen, Mitgliedschaften */
	strcpy(inf_name,config.messages);
	strcat(inf_name,"tausch.cmd");
	if ((grp_f = fopen(inf_name,BINREAD)) != NULL)
	{
		char eintrag[MAX_CPG];

		while (fgets(eintrag,MAX_CPG,grp_f))
			fwrite(eintrag,strlen(eintrag),1,inf_f);
		fclose(grp_f);
	}

	/* Messagestatus schreiben */
	(void) export_status2box(inf_f,akt_username());

#if defined(__WATCOMC__)
	strcpy(ts,config.messages);
	strcat(ts,"*.*");
	if ((dir = opendir(ts)) != NULL)
#else
	if ((dir = opendir(config.messages)) != NULL)
#endif
	{
		while ((entry = readdir(dir)) != NULL)
		{
			/* Header-Dateien auslesen */
#if defined(__WATCOMC__)
			strlwr(entry->d_name);
#endif
			if ((entry->d_name[0] == 'm') && (atoi(entry->d_name+1))
			&& (strlen(entry->d_name) == strlen("m0000001.hdr"))
			&& (!strcmp(entry->d_name+9,"hdr")))
			{
				copy_message(inf_f,atol(entry->d_name+1),NULL);
			}
		}
		closedir(dir);
	}

	fputs("#\r\n",inf_f);
	fclose(inf_f);

	return infile_packen();

} /* infile_erstellen */

#endif /* !SERVER */



/*
	Nachricht ins Infile kopieren

	->			inf_f		:	stream der Zieldatei
				nr			:	Nummer der Nachrichtendatei
				net_id	:	zu bernehmende Netz-ID fr Infile-PM-Import
								inkl. Zeilenende; sonst NULL

	WICHTIG:
	jede Zeile wird mit <cr><lf> abgeschlossen bzw. umgewandelt

	todo:
	beim Tausch an Maus-Systemen TABs in Leerzeichen wandeln
*/
static BOOLEAN copy_message(FILE *inf_f,long nr,const char *net_id)
{
	char *buf;
	char hdr_name[PATH_MAX];
	char msg_name[PATH_MAX];
	char cc_name[PATH_MAX];
	FILE *hdr_f,*msg_f;
	FILE *cc_f = NULL;

	sprintf(hdr_name,"%sm%07ld.hdr",config.messages,nr);
	sprintf(msg_name,"%sm%07ld.txt",config.messages,nr);
	sprintf(cc_name, "%sm%07ld.cc", config.messages,nr);

	if ((hdr_f = fopen(hdr_name,BINREAD)) == NULL)
		return FALSE;

	if ((msg_f = fopen(msg_name,BINREAD)) == NULL)
	{
		fclose(hdr_f);
		return FALSE;
	}

	if (!net_id)
		cc_f = fopen(cc_name,BINREAD);

	if ((buf = (char *) malloc(MSG_SIZE)) == NULL)
	{
		fclose(msg_f);
		fclose(hdr_f);
		return FALSE;
	}

	/* bei Reimport eigener Infile-PMs die FE-ID durch die Net-ID ersetzen */
	if (net_id)
	{
		fgets(buf,MAX_CPG,hdr_f);
		fputs(net_id,inf_f);
	}
	/* (restlichen) Header komplett bernehmen */
	/* (liegt bereits mit <cr><lf> vor */
	while (fgets(buf,MAX_CPG,hdr_f))
		fputs(buf,inf_f);
	fclose(hdr_f);

	if (cc_f)
	{
		while (fgets(buf,MAX_CPG,cc_f))
		{
			char *p;
			if ((p = strpbrk(buf,"\r\n")) != NULL)
				*p = 0;
			if (strlen(buf) < 2)
				continue;
			fprintf(inf_f,"K%s\r\n",buf);
			fprintf(inf_f,">Kopie\r\n");
		}
		fclose(cc_f);
	}

	/* Nachricht bernehmen, optional Zeilen zu Endloszeilen zusammenfassen */
	if (!config.join_lines)
	{
		while (fgets(buf,(int)MSG_SIZE,msg_f))
		{
			char *p = buf;

			p += strlen(buf)-1;
			while (p > buf && (*(p-1) == '\r' || *(p-1) == '\n'))
				*--p = 0;
			*p++ = '\r';
			*p++ = '\n';
			*p++ = 0;

			fputc(':',inf_f);
			fputs(buf,inf_f);
		}
	}
	else
	{
		BOOLEAN endlos = FALSE;

		while (fgets(buf,(int)MSG_SIZE,msg_f))
		{
			char *p;
			size_t len;

			if ((p = strpbrk(buf,"\r\n")) != NULL)
				*p = 0;

			if (endlos && (*buf == ' '))
			{
				fputs("\r\n",inf_f);
				endlos = FALSE;
			}

			if (!endlos)
				fputc(':',inf_f);

			len = strlen(buf);
			if ((len > 1) && (buf[len-1] == ' ') && (buf[len-2] != ' '))
				endlos = TRUE;
			else
				endlos = FALSE;

			fputs(buf,inf_f);

			if (!endlos)
				fputs("\r\n",inf_f);
		}
	}
	fclose(msg_f);
	free(buf);

	return TRUE;

} /* copy_message */



#if !defined(CLIENT)

/*
	GEAENDERT-Flag exportierter Statusmeldungen zurcksetzen

	->			name	:	Username
				status:	Pfad+Name der Status-LOGdatei
*/
static void infile_status_reset(const char *name,const char *status_log)
{
	FILE *stat_f;
	GLOBAL_REF gref;
	char id[MAX_HDR];
	int gh;

	if ((stat_f = fopen(status_log,"r")) == NULL)
		return;

	if ((gh = group_open(getpid(),name,TRUE,TRUE)) < 0)
	{
		fclose(stat_f);
		return;
	}
	while (fgets(id,sizeof(id),stat_f))
	{
#ifndef CLIENT		/* ### z.Z. ist msgid_find() zu kompliziert */
		if (!msgid_find(id,&gref))
#endif
			continue;
		index_goto(gh,gref.msg_nr);
		index_read(gh);
		/* dieser Aufruf ist korrekt, da das GEAENDERT-Flag zurckgesetzt wird */
		msg_set_status(gh,GEAENDERT);
	}
	group_close(gh,getpid());

	unlink(status_log);

} /* infile_status_reset */



/*
	eigene Infile-PMs importieren

	->			messages	:	Pfad zum "messages" Verzeichnis, mit abschlieendem 
								(Back-)slash

   Beim Import wird der Log-Block in einer Datei gespeichert und am Ende des
   normalen Outfile-Imports ausgewertet.
   Zunchst mal mu mglichst einfach festgestellt werden, ob es sich um
   eine ffentliche Nachricht handelt, am besten dadurch, da nach einer
   Gruppenzeile innerhalb des Headers gesucht wird. Oder besser: ID...
   Von der Box akzeptierte OMs werden gelscht.
   Bleiben die PMs: diese werden in eine Datei namens "outfile.txt" kopiert,   
   die sich aber nicht im import- sondern im messages-Verzeichnis befindet.
   Die einzelnen Nachrichten werden gelscht (besser wre: erst spter).
   Der Import wird mit verndertem Importpfad (nmlich messages-Verzeichnis)
   auf das eigene Pseudo-Outfile angesetzt.
*/
BOOLEAN infile_import(const char *messages)
{
	char log_name[PATH_MAX];
	char out_name[PATH_MAX];
	char hdr_name[PATH_MAX];
	char msg_name[PATH_MAX];
	char cc_name[PATH_MAX];
	char status_log[PATH_MAX];
	char zeile[256];
	FILE *log_f;
	FILE *out_f;
	long nr = 0;
	BOOLEAN ok;

	mputs("Infile-Import:");

	/* alte Tausch-Kommandos lschen */
	strcpy(log_name,messages);
	strcat(log_name,"tausch.cmd");
	unlink(log_name);

	/* LOG-Block ffnen */
	strcpy(log_name,messages);
	strcat(log_name,"log");
	if ((log_f = fopen(log_name,"r")) == NULL)
		return TRUE;

	/* neues Outfile im messages-Verzeichnis anlegen */
	strcpy(out_name,messages);
	strcat(out_name,"outfile.txt");
	if ((out_f = fopen(out_name,"w")) == NULL)
	{
		fclose(log_f);
		return FALSE;
	}

	while (1)
	{
		char net_id[256];

		/* eigene, temporre FrontEnd-ID */
		if (!fgets(zeile,sizeof(zeile),log_f))
			break;
		if (*zeile != '#')
			continue;
		nr = atol(zeile+3);

		/* von der Box vergebene Netz-ID */
		if (!fgets(net_id+1,sizeof(net_id)-1,log_f))
			break;
		net_id[0] = '#';

		sprintf(hdr_name,"%sm%07ld.hdr",messages,nr);
		sprintf(msg_name,"%sm%07ld.txt",messages,nr);
		sprintf(cc_name, "%sm%07ld.cc", messages,nr);

		/* persnliche Nachrichten in pseudo-Outfile bernehmen */
		if (zeile[1] == 'P')
			copy_message(out_f,nr,net_id);

		/*
			ffentliche Nachrichten kommen per Echo zurck und knnen
			ohne weiteres gelscht werden.
			Da im Log sowieso nur IDs akzeptierter Nachrichten stehen, 
			bleiben beanstandete im messages-Verzeichnis erhalten.
		*/
		unlink(hdr_name);
		unlink(msg_name);
		unlink(cc_name);
	}
	fputs("#\r\n",out_f);
	fclose(out_f);
	fclose(log_f);

	ok = infile_einfuegen(out_name);

	/* es wird davon ausgegangen, da Infile-Importe erfolgreich sind */
	unlink(out_name);
	unlink(log_name);

	/* Zurcksetzen eigener Statusmeldungen */
	strcpy(status_log,messages);
	strcat(status_log,"status");
	infile_status_reset(akt_username(),status_log);

	return ok;

} /* infile_import */
#endif /* !CLIENT */



/*
	Mailstatus von User -> Box exportieren

	->			infile	:	stream fr export-Datei
				username	:	Username

	vgl. export_status2user() in mdb/export.c
*/
static BOOLEAN export_status2box(FILE *infile,const char *username)
{
	extern WORK *msg[];					/* ### durch msg_* FN.'s ersetzen */

	char status[] = "?GZBWMAYK";
	char status_log[PATH_MAX];
	FILE *stat_f;
	char id[MAX_HDR];
	char *pos;
	WORK *mp;
	int gh;

	if ((gh = group_open(getpid(),username,TRUE,FALSE)) < 0)
		return FALSE;

	mp = msg[gh];
	if (!mp->kennung.status_neu)
	{
		group_close(gh,getpid());
		return TRUE;
	}

	/*
		exportierte Statusmeldungen merken, um beim Infile-Import das 
		GEAENDERT-Flag entfernen zu knnen
	*/
	strcpy(status_log,config.messages);
	strcat(status_log,"status");
	stat_f = fopen(status_log,"w");

	while (index_read(gh))
	{
		/* Status nur bei nderung fremder PMs exportieren */
		if ((mp->ix.status & GEAENDERT)
		&& !(mp->ix.flags & YOURMAIL)
		&& !(mp->ix.flags & BEANTWORTET)
		&&  (mp->ix.status & ~GEAENDERT))
		{
			msg_read_header(gh);
			strncpy(id,mp->puffer,MAX_HDR-1);
			id[MAX_HDR-1] = 0;
			if ((pos = strpbrk(id,"\r\n")) != NULL)
				*pos = 0;
			fprintf(infile,"%s\r\n",id);
			fprintf(infile,"B%c\r\n",status[mp->ix.status & ~GEAENDERT]);

			if (stat_f)
				fprintf(stat_f,"%s\r\n",id);
		}

		if (!index_next(gh))
			break;
	}

	mp->kennung.status_neu = 0;
	group_close(gh,getpid());

	if (stat_f)
		fclose(stat_f);

	return TRUE;

} /* export_status2box */
