/* support.c : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB3 (mausnet)
       (might be the wrong person except for offline)

    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
/*
	Offline - MausTausch-FrontEnd

	allgemeines Datenbankhandling
	(c) 1992-95 bei Frank Baumgart

	begonnen am 05.06.1992
	Version vom 10.01.1995
*/


#include "portab.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <limits.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "system.h"
#include "mdb_defs.h"
#include "kennung.h"
#include "mdb_glob.h"
#include "mdb/mt_cmd.h"				/* wg. head_getinfo */
#include "tool.h"
#include "mdb_conf.h"
#include "fehler.h"

extern BOOLEAN interaktiv;

extern CONFIG config;



/*
	Username ermitteln

	Reihenfolge:
	- HEAD-Eintrag
	- CNF-Var "Username"
	- $FULLNAME
	- "Frank Baumgart"
*/
char *akt_username(void)
{
	static char name[100] = "Frank Baumgart";

#if !defined(BOX)
	if (head_getinfo('U',name,FALSE))
		return name;
#endif

	if (*config.username)
		return config.username;

	if (getenv("FULLNAME"))
		return getenv("FULLNAME");

	return name;

} /* akt_username */



/*
	stream-Pufferung vornehmen bzw. weglassen

	Je nach Einstellung wird versucht, den angegebenen stream zu puffern 
	oder wegen Speichersparens eben nicht
*/
int pufferung(FILE *fp,void *buf,int type,size_t size)
{
#if !defined(CLIENT) && !defined(SERVER)

	/* wir sind sparsam und nehmen wenig RAM in Anspruch ? */
	if (!config.pufferung)
		return 0;

#ifdef linux	/* ltere Linux C-Libraries haben einen Bug in setvbuf */
	if (config.linux_bug)
	{
		char *p;

		/*
			im nicht-interaktivem Betrieb selber Puffer anfordern, da dieser 
			bei Programmende automatisch wieder freigegeben wird.
		*/
		if (interaktiv)
			return 0;
		p = (char *) malloc(size);
		if (p)
			return setvbuf(fp,p,type,size);
		else
			return 0;
	}
#else
	return setvbuf(fp,buf,type,size);
#endif
#endif

	return 0;

} /* pufferung */



/*
	Kennung ein- bzw. berlesen

	->			ind_f		:	Indexstream
			<-	kennung	:	Speicher fr Kennung bzw. NULL

	Diese Funktion kmmert sich um Indexkennungen unterschiedlicher 
	Gre, der Puffer mu jedoch gengend Platz fr die grtmgliche 
	Kennung bieten.
	Die Streamposition wird nicht zurckgesetzt, also kein rewind
	Nach Beendigung steht die Streamposition hinter der Kennung

	Die Existenz dieser Funktion bedeutet nicht, da Datenbanken
	unterschiedlicher Versionen gemischt verwendet werden knnen.
*/
void lies_kennung(FILE *ind_f,KENNUNG *kennung)
{
	UWORD groesse = 0;				/* Gre der Kennung in Bytes */

	rewind(ind_f);
	fread(&groesse,sizeof(groesse),1,ind_f);
	rewind(ind_f);

	if (sizeof(KENNUNG) != groesse)
	{
		fprintf(stderr,"Gruppe: %s\n",kennung->name);
		fputs("Sie verwenden eine alte Programmversion zusammen mit einer neueren Datenbank.\n",stderr);
		fputs("Bitte besorgen sie sich eine aktuelle Version von Offline.\n",stderr);
		fputs("",stderr);
		fprintf(stderr,"erwartete Gre: %d, gefundene Gre: %d\n",sizeof(KENNUNG),groesse);
	}

	if (kennung)
		fread(kennung,groesse,1,ind_f);

	/* auch r/w Wechsel */
	fseek(ind_f,(long) groesse,SEEK_SET);

} /* lies_kennung */



/*
	String mit aktuellem Datum zurckliefern
	Format: yymmdd

	Fr Logdateien u..
*/
char *yymmdd(void)
{
	static char buf[8];
	time_t zeit = time(NULL);

	strftime(buf,8,"%y%m%d",localtime(&zeit));

	return buf;

} /* yymmdd */



/*
	Zeit der Form YYYYMMDDHHMM[ss] in time_t umwandeln

	Sinn: Verzicht auf mktime und der damit verbundenen zeitaufwendigen 
	Auswertung von timezone, Environment etc.
	Vor 1990 wird weiterhin mktime() verwendet, danach allerdings die 
	Anzahl der seit 1.1.1970 vergangenen Sekunden auf Basis von mktime(1990) 
	selber berechnet.
*/
time_t calctime(const char *s)
{
	static time_t ref = 0;
	struct tm zeit;
	long days_to_month[] = { 0,31,59,90,120,151,181,212,243,273,304,334,365 };
	long days;

	/* Referenz-time_t vom 1.1.1990 berechnen */
	if (!ref)
	{
		zeit.tm_year = 1990 - 1900;
		zeit.tm_mon  = 0;
		zeit.tm_mday = 1;
		zeit.tm_hour = 0;
		zeit.tm_min  = 0;
		zeit.tm_sec  = 0;
		zeit.tm_isdst= 0 /* Frank meint -1 */;
		zeit.tm_yday = 0;
		zeit.tm_wday = 0;
		ref = mktime(&zeit);
	}

	zeit.tm_year = (*s - '0')*1000 + (*++s - '0')*100 + (*++s - '0')*10 + (*++s - '0');
	zeit.tm_mon  = (*++s - '0')*10 + (*++s - '0') -1;
	zeit.tm_mday = (*++s - '0')*10 + (*++s - '0');
	zeit.tm_hour = (*++s - '0')*10 + (*++s - '0');
	zeit.tm_min  = (*++s - '0')*10 + (*++s - '0');
	if (isdigit(*++s))
		zeit.tm_sec = (*s - '0')*10 + (*++s - '0');
	else
		zeit.tm_sec = 0;

	if (zeit.tm_year < 1990)
	{
		zeit.tm_year -= 1900;
		return mktime(&zeit);
	}

	/*
		eigene Zeitberechnung inkl. Schaltjahre auf Basis des time_t Wertes 
		vom 1.1.1990
	*/
	/* Anzahl Tage ohne Beruecksichtigung der Schaltjahre */
	days = (zeit.tm_year - 1990) * 365L + days_to_month[zeit.tm_mon] + zeit.tm_mday -1;
	/* ### Anzahl Schalttage berechnen, letztes (vor 1990) war 1988 */
	days += (zeit.tm_year - 1988) /4;

#if 0		/* nur zum Debugging */
	{
		time_t secs = ref + days * 86400L + zeit.tm_hour * 3600L + zeit.tm_min * 60L + zeit.tm_sec;
		time_t orig;

		zeit.tm_year -= 1900;
		orig = mktime(&zeit);
		printf("mytime:%ld , mktime:%ld , diff:%ld\r\n",
				(long)secs,(long)orig,(long)(orig-secs));

		printf(ctime(&secs));
		printf(ctime(&orig));
		
		return orig;
		return secs;
	}		
#endif

	return ref + days * 86400L + zeit.tm_hour * 3600L + zeit.tm_min * 60 + zeit.tm_sec;

} /* calctime */



#if !defined(BOX) || defined(SERVER)
/*
	Gre einer Datei ermitteln

	->			stream	:	FILE-Pointer auf geffneten stream

	<-			Dateigre in Bytes
*/
long filesize(FILE *stream)
{
	long save_pos;
	long laenge;

	save_pos = ftell(stream);

	fseek(stream,0,SEEK_END);
	laenge = ftell(stream);

	fseek(stream,save_pos,SEEK_SET);

	return laenge;

} /* filesize */
#endif



#ifndef BOX		/* Doppel-Implementation vermeiden */
/*
	Existenz einer Datei berprfen

	->			name	:	Name der Datei

	<-			TRUE	:	existiert
				FALSE	:	existiert nicht
*/
BOOLEAN fileexist(const char *name)
{
	struct stat statbuf;

	if (stat(name,&statbuf) < 0)
		return FALSE;
	else
		return TRUE;

} /* fileexist */
#endif



/*
	Zeile aus einem stream lesen
	das Zeilenende wird abgeschnitten
	maximal 100 Zeichen werden gelesen

	27.11.92 Frank
*/
char *freadline(char *buf,size_t len,FILE *stream)
{
	char *p;

	if (!fgets(buf,(int) len,stream))
		return NULL;
	if ((p = strpbrk(buf,"\r\n")) != NULL)
		*p = 0;

	return buf;

} /* freadline */



/*
	Abbruch wichtiger Funktionsblcke verhindern (TRUE) bzw. zulassen (FALSE)
*/
void protect(BOOLEAN on)
{
	static BOOLEAN aktiv = FALSE;
	static BOOLEAN ignore_int = FALSE;
	static BOOLEAN ignore_hup = FALSE;

	if (on && aktiv)
	{
		fputs("Abbruch-Schutz bereits aktiv\r\n",stderr);
		return;
	}
	else if (!on && !aktiv)
	{
		fputs("Abbruch-Schutz war nicht aktiv\r\n",stderr);
		return;
	}

	if (on)
	{
		aktiv = TRUE;
		/*
			falls die Signale bereits von einer Shell ohne (korrekte) Job-Control 
			gendert sein sollten, drfen sie durch Offline nicht wieder bei 
			"on = FALSE" verstellt werden.
			Das merken wir, wenn sich beim Versuch, selbst SIG_IGN einzustellen, 
			herausstellt, da das Signal bereits auf SIG_IGN steht.
		*/
		if (signal(SIGINT,SIG_IGN) == SIG_IGN)
			ignore_int = TRUE;
#ifdef SIGHUP
		if (signal(SIGHUP,SIG_IGN) == SIG_IGN)
			ignore_hup = TRUE;
#endif
	}
	else
	{
		aktiv = FALSE;
		if (!ignore_int)
			signal(SIGINT,SIG_DFL);
#ifdef SIGHUP
		if (!ignore_hup)
			signal(SIGHUP,SIG_DFL);
#endif
	}

} /* protect */



/*
	Kommando via Shell ausfhren
*/
int shell(const char *cmd)
{
	char tmp[PATH_MAX];

	if (!cmd)
#if defined(DJGPP)
		/* DOS kennt kein system(NULL) */
		return 1;
#else
		return system(NULL);
#endif

	strcpy(tmp,cmd);

#if !defined(TEXT_FRONTEND)
	strcat(tmp," >/dev/null");
#endif

	return system(tmp);

} /* shell */



/*
	Zeitpunkt von ASCII in normalisiertes Format umrechnen

	->			zeit	:	ASCII-Format: dd.mm.yyyy[,hh:mm:ss]
				jetzt	:	Referenzzeitpunkt

	<-			Zeitpunkt im time_t Format

	Unvollstndige Angaben, z.B. "23.7." oder "23.7" werden durch die 
	time_t Angabe in "jetzt" ergnzt
*/
time_t berechne_zeit(const char *zeit,time_t jetzt)
{
	struct tm *neu;
	struct tm kopie;
	time_t ret;
	char *p;

	if (!zeit || !*zeit || !isdigit(*zeit))
		return jetzt;

	/* negativ ? dann Tage zurckrechnen */
	if (atoi(zeit) < 0)
		return jetzt - atoi(zeit) *3600L *24;

	neu = localtime(&jetzt);

	/* Tag */
	neu->tm_mday = atoi(zeit);
	p = strchr(zeit,'.');
	if (p && *(p+1))
		zeit = p+1;
	else
		goto end_calc;

	/* Monat */
	neu->tm_mon = atoi(zeit)-1;
	p = strchr(zeit,'.');
	if (p && *(p+1))
		zeit = p+1;
	else
		goto end_calc;

	/* Jahr */
	neu->tm_year = atoi(zeit);
	if (neu->tm_year > 1900)
		neu->tm_year -= 1900;
	p = strchr(zeit,',');
	if (p && *(p+1))
		zeit = p+1;
	else
		goto end_calc;

	/* Stunden */
	neu->tm_hour = atoi(zeit);
	if (strchr(zeit,':'))
		zeit = strchr(zeit,':')+1;
	else
		goto end_calc;

	/* Minuten */
	neu->tm_min = atoi(zeit);
	if (strchr(zeit,':'))
		zeit = strchr(zeit,':')+1;
	else
		goto end_calc;

	/* Sekunden */
	neu->tm_sec = atoi(zeit);

end_calc:
	kopie=*neu;
	ret=mktime(neu);
	if (neu->tm_isdst!=kopie.tm_isdst)
	{
		kopie.tm_isdst=neu->tm_isdst;
		*neu=kopie;
		ret=mktime(neu);
	}
	return ret;

} /* berechne_zeit */
