/* decode.c : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB3 (mausnet)
       (might be the wrong person except for offline)

    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
/*
	Offline - MausTausch-FrontEnd

	Kodier- und Dekodier-Routinen
	(c) 1992,93 bei Frank Baumgart

	begonnen am  5.11.1992
	Version vom  6.11.1992
*/


#include "portab.h"
#include <stdlib.h>


/*
	dafr sorgen, da in der Mailbox eine Kodierung eingesetzt wird, die 
	nicht mit den ffentlichen Offline-Versionen dekodiert werden knnen.
*/
#if defined(BOX) || (defined(TEXT_FRONTEND) && defined(DEVELOPER))
#define PRIM		97
#else
#define PRIM		139
#endif




#if 0
void encode(UBYTE *puffer,size_t len);
void decode(UBYTE *puffer,size_t len);

/*
	Kodieren eines Speicherbereiches
	simpel aber schnell, dient nur zum Schutz vor unbeabsichtigtem Lesen

	->			puffer	:	Zeiger auf zu kodierenden Speicherbereich
				len		:	Gre des Bereiches in Bytes

	Prinzip:
	Die Kodierung erfolgt von links nach rechts (aufsteigend)
	Fr jedes Zeichen wird das vorangehende, bereits kodierte, als auch 
	das nachfolgende unkodierte Zeichen verwendet.
	1.) Addition von 2*(aktuelle_Lnge & $EF)
	2.) Zurckschneiden auf Byte-Bereich
	3.) XOR-Verknpfung mit dem nchsten Byte (falls nicht letztes)
	    dieses ist noch unkodiert !
*/
void encode(UBYTE *puffer, size_t len)
{
	UWORD wert,next,last = 0;
	UWORD i;

	if (!len)  return;

	while (--len)
	{
		wert = *puffer;
		next = *(puffer+1);
		i = (UWORD) len;

		wert += last + i + (i * PRIM & next)*2;
		wert ^= last ^ next;

		*puffer++ = (UBYTE) wert;
		last = wert;
	}

	/* das letzte Zeichen wird separat verschlsselt */
	wert = *puffer;
	wert ^= last;
	*puffer = wert;

} /* encode */

/*
	Dekodieren eines Speicherbereiches

	->			puffer	:	Zeiger auf zu kodierenden Speicherbereich
				len		:	Gre des Bereiches in Bytes (min. 1)

	Prinzip:
	Die Kodierung erfolgt von rechts nach links (absteigend)
	1.) XOR-Verknpfung mit dem nchsten Byte (falls nicht letztes)
	    dieses ist bereits kodiert !
	2.) Subtraktion von 2*(aktuelle_Lnge & $EF)
	3.) Zurckschneiden auf Byte-Bereich
*/
void decode(UBYTE *puffer,size_t len)
{
	UWORD wert,last,next;
	UWORD i = 1;

	if (!len)  return;

	puffer += len-1;

	/* rechts stehendes Zeichen dekodieren */
	wert = *puffer;
	last = *(puffer-1);
	wert ^= last;
	*puffer-- = (UBYTE) wert;
	len--;

	if (!len)  return;

	while (--len)
	{
		wert = *puffer;
		last = *(puffer-1);
		next = *(puffer+1);
		wert ^= last ^ next;
		wert -= last + i + (i*PRIM & next)*2;
		wert &= 255;

		*puffer-- = (UBYTE) wert;
		i++;
	}

	/* erstes Zeichen */
	wert = *puffer;
	next = *(puffer+1);
	wert ^= next;
	wert -= i + (i*PRIM & next)*2;
	wert &= 255;
	*puffer = (UBYTE) wert;

} /* decode */
#endif

UWORD encode (UBYTE * puffer, size_t offset, UWORD last, size_t len);
UWORD decode (UBYTE * puffer, size_t offset, UWORD last, size_t len);
/* wesentlich schwchere Kodierung */
UWORD
encode (UBYTE * puffer, size_t offset, UWORD last, size_t len)
{
	UWORD i;

	if (!len)
		return 0;

	while (len--) {
		i = (UWORD) offset++;
		last = (*puffer) ^ (((i * PRIM) & last) + 31);
		*puffer++ = last;
	}
	return last;
}								/* encode */

UWORD
decode (UBYTE * puffer, size_t offset, UWORD last, size_t len)
{
	if (!len)
		return 0;

	while (len--) {
		UWORD yet = *puffer;
		*puffer++ = yet ^ (((offset * PRIM) & last) + 31);
		last = yet;
		offset++;
	}
	return last;
}

