/* export.c : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB3 (mausnet)
       (might be the wrong person except for offline)

    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
/*
	Offline - MausTausch-FrontEnd

	Export-Routinen
	(c) 1992-95 bei Frank Baumgart

	begonnen am 02.04.1992
	Version vom 25.01.1995
*/

#include "portab.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <limits.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "system.h"
#include "mdb_conf.h"
#include "offline.h"
#include "struktur.h"
#include "mdb_glob.h"
#include "mdb_intr.h"
#include "message.h"
#include "screen/io.h"
#include "tool.h"
#include "filter.h"
#include "fehler.h"
#include "dynhash.h"

#define	verbose		0

extern CONFIG config;				/* globale Einstellungen (Verkettung...) */


static BOOLEAN
exportiere_gruppe(FILE *exp_f,const char *name,BOOLEAN pm,BOOLEAN neuer,
                  ULONG von,ULONG bis,void *skip_ids,char *buf,
                  ULONG *bytes,ULONG *mails,EXPORT *opt);

static void
export_status2user(FILE *exp_f,FILE *ind_f,FILE *dat_f,BOOLEAN alle);



/*
	Gruppen im Outfile-Format exportieren

	->			liste		:	Name+Pfad der Exportlistendatei
				export	:	Name+Pfad der Exportdatei ("Outfile")
				opt		:	Parameter + Optionen

	<-			TRUE		:	Export erfolgreich
				FALSE		:	Fehler

	Die Distribution gibt die gewnschten Nachrichten an, ist dist also 
	V_MAUSNET, werden nur MAUSNET und NET Mails, aber keine lokalen exportiert.

	Im Boxmodus wird an ein evtl. vorhandenes Outfile angehngt.

	Zum einmaligen crossposting-Export werden via dynhash die IDs exportierter 
	Nachrichten in einer Hashtabelle gemerkt.
*/
BOOLEAN mdb_export(const char *liste,const char *export,EXPORT *opt)
{
	FILE *lst_f = NULL;
	FILE *exp_f;
	FILE *noid_f;
	char name[MAX_CPG];
	time_t ref_zeit = opt->zeit;
	void *no_ids = NULL;
	BOOLEAN ok = TRUE;			/* keep gcc happy */
	int anz_gruppen = 0;			/* Anzahl exportierter Gruppen */
	char *buf;
	ULONG bytes = 0;				/* Outfile-Gre in Bytes */
	ULONG mails = 0;				/* Anzahl exportierter Messages */
	EXPORT l_opt = *opt;			/* Optionen fr aktuelle Gruppe */

	if (!liste || !*liste)
		return FALSE;

	/* Datei mit Liste zu exportierender Gruppen ffnen */
	if ((lst_f = fopen(liste,BINREAD)) == NULL)
		return FALSE;

	/*
		Speicher fr Messagepuffer anfordern. Der eigentliche Export liest
		immer in Blcken mit max. dieser (MSG_SIZE) Lnge.
		Fr PMs ist MSG_SIZE die maximal mgliche Gre, da sie kpl. in den 
		Buffer passen mu, um dekodiert zu werden.
	*/
	if ((buf = (char *) malloc(MSG_SIZE)) == NULL)
	{
		fclose(lst_f);
		fehler(NO_MEMORY);
		return FALSE;
	}

	/* bei Box-Betrieb anhngen, denn vorher ex. evtl. andere Blcke */
	if ((exp_f = fopen(export,(opt->box) ? BINAPPEND : BINWRITE)) == NULL)
	{
		fclose(lst_f);
		fehler(NO_EXPFILE);
		return FALSE;
	}
	/* am wichtigsten ist die Pufferung des Exports */
	if ((size_t)65536L==0 || pufferung(exp_f,NULL,_IOFBF,(size_t)65536L))
		if ((size_t)32768L==0 || pufferung(exp_f,NULL,_IOFBF,(size_t)32768L))
			pufferung(exp_f,NULL,_IOFBF,16384);

	/* Filter einlesen */
	if (!opt->box && config.filter_export)
		lade_filter();

	/*
		Tabelle schon exportierter Nachrichten fhren. Im Boxmodus wird es 
		bereits beim "ohne Echo"-Feature fr User-Outfiles gebraucht, anson-
		sten generell bei Crosspostings, um diese nur ein Mal zu exportieren
	*/
	no_ids = hhcreate(5000);

	/*
		ID-Liste nicht zu exportierender Nachrichten ? ("ohne Echo")
		Dann wird aus der ASCII-Liste eine Hashtabelle mit den IDs der 
		Nachrichten aufgebaut und beim Export bercksichtigt
	*/
	if (opt->noidfname && (noid_f = fopen(opt->noidfname,BINREAD)) != NULL)
	{
		char id[256];
		DYN_ENTRY item;

		while (fgets(id,sizeof(id),noid_f))
		{
			char *p;

			if ((p = strpbrk(id,"\r\n")) != NULL)
				*p = 0;

			item.crc = crc32ansi(id,strlen(id));
			item.u.nr  = 0;

			hhsearch(no_ids,&item,DYN_ENTER);
		}
		fclose(noid_f);
	}

	while (ok && fgets(name,MAX_CPG,lst_f))
	{
		const char *tmp_name;
		BOOLEAN pm;
		ULONG von,bis;
		char *pos;

		von = bis = 0;

		if ((pos = strpbrk(name,"\r\n")) != NULL)
			*pos = 0;
		/* versehentlich leerer Eintrag ? */
		if (!*name)
			continue;

		ref_zeit = 0;

		if ((pos = strchr(name,',')) != NULL)
		{
			*pos = 0;
			if (strchr(pos+1,'-'))
			{
				von = atol(pos+1);
				bis = atol(strchr(pos+1,'-')+1);
				if (!bis)
					bis = (ULONG) -1;
				if (verbose)
					mprintf("exportiere: von %lu bis %lu\r\n",von,bis);
			}
			else
			{
				ref_zeit = berechne_zeit(pos+1,ref_zeit);
				if (verbose)
/*					mprintf("exportiere: %s ab %s\r\n",name,ctime(&zeit));*/
					mprintf("exportiere: %s ab %s\r\n",name,ctime(&opt->zeit));
			}
		}
		else
		{
			ref_zeit = opt->zeit;
			if (verbose)
				mprintf("exportiere: %s",name);
		}

		pm = (*name == '*');
		tmp_name = (pm) ? name+1 : name;

		l_opt.zeit = ref_zeit;
		ok = exportiere_gruppe(exp_f,tmp_name,pm,TRUE,von,bis,no_ids,buf,
		                       &bytes,&mails,&l_opt);

		anz_gruppen++;
	}

	/* mehr als eine Gruppe exportiert ? */
	if (anz_gruppen > 1)
	{
		/* dann auch Gesamtanzahl ausgeben */
		mputs("===================================");
		mprintf("%-27s%8lu\r\n","",mails);
	}


	if (no_ids)
		hhdestroy(no_ids);

	if (!opt->box)
	{
		time_t t = time(NULL);
		char tmp[20];

		/* Autoren- und Themenfilter freigeben */
		verwerfe_filter();

		/* #LOG-Kennung schreiben */
		fputs("#LOG\r\n",exp_f);
		fprintf(exp_f,":!V%s\r\n",MAJOR_VERSION "." MINOR_VERSION);
/*		fprintf(exp_f,":!D%02s.%02s.%04s\r\n",__DATE__);*/
		strftime(tmp,12,"%d.%m.%Y",localtime(&t));
		fprintf(exp_f,":!D%s\r\n",tmp);
		fprintf(exp_f,":!T%s\r\n",__TIME__);
		fputs("#\r\n",exp_f);
	}

	fclose(exp_f);
	free(buf);
	fclose(lst_f);

	return ok;

} /* mdb_export */



/*
	Nachrichten als Outfile exportieren

	->			exp_f		:	stream der Export-Datei ("outfile.txt")
				name		:	Gruppen/Username
				pm			:	TRUE = User, FALSE = Gruppe
				zeit		:	Referenzzeitpunkt fr ltere/neuere Nachrichten
				neuer		:	Bearbeitung neuerer oder lterer Datenstze als <zeit>
				von		:	Nummer der ersten zu exp. Msg. oder 0
				bis		:	Nummer der letzten zu exp. Msg. oder 0
				skip_ids	:	Hashtabelle mit nicht zu exportierenden IDs
				buf		:	Puffer fr Msg.Text etc., Gre = MSG_SIZE
				opt		:	Parameter und Optionen, s. EXPORT

	<-			TRUE		:	kein Fehler
				FALSE		:	Fehler/Gruppe ex. nicht/max. Outfilegre erreicht

	Achtung:
	es knnten nur alle (alten) Mailstatus gefordert sein, dann darf die 
	Funktion nicht abbrechen, wenn keine neuen Nachrichten vorhanden sind.
*/
static BOOLEAN
exportiere_gruppe(FILE *exp_f,const char *name,BOOLEAN pm,BOOLEAN neuer,
                  ULONG von,ULONG bis,void *skip_ids,char *buf,
                  ULONG *bytes,ULONG *mails,EXPORT *opt)
{
	FILE *ind_f,*dat_f;
	KENNUNG kennung;
	INDEX ix;
	BOOLEAN ok = TRUE;
	long anzahl = 0;				/* Anzahl exportierter Messages */
	char dat_pfad[PATH_MAX];
	char ind_pfad[PATH_MAX];
	time_t jetzt = 0;				/* fr PM-Tauschdatum; keep gcc happy */
	size_t sitelen;				/* fr PATH-Zeile */

	sitelen = (opt->site) ? strlen(opt->site) : 0;



	if (group_path(name,dat_pfad,pm) < 0)
		return TRUE;
	strcpy(ind_pfad,dat_pfad);
	strcat(ind_pfad,"ind");
	strcat(dat_pfad,"dat");

	if (opt->zeit && neuer)
	{
		struct stat status;
		if (!stat(dat_pfad,&status) && (status.st_mtime < opt->zeit)
			&& !opt->box && !pm && !opt->status_mode)
		{
			mprintf("%-30.30s    -\r\n",name);
			return TRUE;
		}
	}

	/* Indexdatei ffnen (Box: Schreibzugriff wg. PM-Status) */
	if ((ind_f = fopen(ind_pfad,(opt->box && pm) ? BIN_RW_OPEN : BINREAD)) == NULL)
	{
		/* keine Fehlermeldung, evtl. ex. die Gruppe nur nicht mehr */
		return TRUE;
	}

	pufferung(ind_f,NULL,_IOFBF,(size_t)((opt->box && pm) ? 2048 : 16384));

	lies_kennung(ind_f,&kennung);

	/*
		Zeitpunkt des letzten Nachrichtenimports ermitteln, um festzustellen, 
		ob berhaupt neue Nachrichten vorliegen (knnen)
	*/
	if (opt->zeit && neuer)
	{
#if 0
		int i = 0;
		time_t last_import = 0;
		time_t *ref = kennung.imported;

		while ((i < 10) && *ref)
		{
			last_import = *ref++;
			i++;
		}

		if ((last_import < opt->zeit) || (kennung.changed < opt->zeit))
#else	/* Fix von Uwe folgt: */
		if (kennung.changed < opt->zeit 
			&& !opt->box && !pm && !opt->status_mode)
#endif
		{
			mprintf("%-30.30s    -\r\n",name);
			fclose(ind_f);
			return TRUE;
		}
	}

	/* Datendatei zum Lesen ffnen */
	if ((dat_f = fopen(dat_pfad,BINREAD)) == NULL)
	{
		fclose(ind_f);
		/* kein Grund, aufzuhren, s.o. */
		return TRUE;
	}
	/* Datendatei puffern */
	pufferung(dat_f,NULL,_IOFBF,MSG_SIZE);

	if (opt->box && pm && opt->status_mode)
	{
		jetzt = time(NULL);
		export_status2user(exp_f,ind_f,dat_f,
			opt->status_mode==2 ? TRUE : FALSE );
	}


	/*
		ggf. vorhandene Startposition in Indexdatei anfahren
	*/
	if (neuer && opt->zeit)
	{
		int i = 0;
		MESSAGE start = 0;	/* Position der erstmglichen neuen Nachricht */
		time_t *ref = kennung.imported;

		while ((i < 10) && (*ref && *ref++ < opt->zeit))
			start = kennung.imp_anf[i++];

		/*
			Position der frhestmglichen neuen Nachricht anfahren.
			Das bedeutet, da VOR dieser Nachricht mit Sicherheit keine liegen, 
			die neuer sind als das Export-Referenzdatum. Es mu sich aber nicht 
			zwangslufig um die erste zu exportierende Nachricht handeln.
		*/
		if (start)  fgoto(start);
	}


	/*
		jetzt Msgs. mit richtigem Datum/Uhrzeit suchen
	*/
	while (get_index(&ix))
	{
		BOOLEAN exp_status,exp_flags;
		BOOLEAN aut_filter,thm_filter;

		aut_filter = FALSE;
		thm_filter = FALSE;

		/* Beschrnkung nach Msg.nummern ? */
		if (von || bis)
		{
			MESSAGE akt = akt_msg() -1;
			if ((von > akt) || (bis < akt))
				continue;
			if (bis && (akt > bis))
				break;
		}

		/* Distribution berprfen */
		if (!pm && (ix.flags < opt->dist))
			continue;

		/* Datum/Uhrzeit vergleichen */
		if ((neuer && (ix.zeit < opt->zeit))
		|| (!neuer && (ix.zeit > opt->zeit)))
			continue;

		/* Nachricht als gelscht gekennzeichnet ? */
		if (ix.flags & (TEIL | TOTAL))
			continue;

		/* Messagefilter */
		if (!opt->box && config.filter_export)
		{
			/* CRC im Filter ? */
			aut_filter |= msg_crc(AUTOR,ix.author_crc);
			thm_filter |= msg_crc(TITEL,ix.topic_crc);
		}

		/*
			Nachrichten mit IDs, die in "skipid.lst" stehen, verwerfen.
			Dabei handelt es sich i.A. um PM-Echos oder Crosspostings, von 
			denen schon eine Ausgabe gemacht wurde.
		*/
		if (opt->box && skip_ids && !pm)
		{
			DYN_ENTRY item;

			item.crc = ix.id_crc;
			item.u.nr  = 0;
			if (hhsearch(skip_ids,&item,DYN_FIND))
				continue;
		}

		/* fr einmaligen crossposting-Export ID-CRCs merken */
		/* ebenso alle anderen exportierten IDs merken - fuer T BRETT */
		if (!pm && skip_ids)
		{
			DYN_ENTRY item;

			item.crc = ix.id_crc;
			item.u.nr  = 0;
			hhsearch(skip_ids,&item,DYN_ENTER);
		}

		/* Outfilegrenbeschrnkung erreicht ? */
		if (opt->maximum && (*bytes + ix.len > opt->maximum))
		{
			mputs("vorgebene max. Outfilegre erreicht");
			ok = FALSE;
			break;
		}

		/* Entsprechende Dateiposition anfahren */
		if (ftell(dat_f) != ix.pos)
			fseek(dat_f,ix.pos,SEEK_SET);

		/*
			falls Msg. > MSG_SIZE, dann erst nur den Header einlesen, da dann 
			der fseek schneller als das berlesen der unntigen Daten ist
		*/
		if (ix.len >= MSG_SIZE)
		{
			fread(buf,1,ix.t_text,dat_f);
			buf[ix.t_text] = 0;
		}
		else
			fread(buf,1,ix.len,dat_f);

		if (!opt->box)
		{
			if (aut_filter)
			{
				/* dann auch Datensatz untersuchen, um 100%ig sicher zu sein */
				aut_filter = msg_filter(AUTOR,ix.author_crc,buf + ix.t_author);
				if (aut_filter)
					mputs("Autor gefiltert");
			}
			/* dito Thema (falls nicht schon Autor gefiltert) */
			if (thm_filter && !aut_filter)
			{
				/* dann auch Datensatz untersuchen, um 100%ig sicher zu sein */
				thm_filter = msg_filter(TITEL,ix.topic_crc,buf + ix.t_topic);
				if (thm_filter)
					mputs("Thema gefiltert");
			}
			/* trifft mind. einer der beiden Filter zu ? */
			if (thm_filter || aut_filter)
				continue;
		}

		/* Echovermeidung fr Boxen und Gates */
		if (opt->box && opt->site)
		{
			char *p;
			/* Path ist die Zeile, die mit einem kleinen 'p' beginnt. */
			p = strstr(buf,"\r\np");
			if (p)
			{
				char *q;
				char *crlf;
				BOOLEAN found=FALSE;

				crlf = strpbrk(p+3,"\r\n");
				q = strstr(p+3,opt->site);
				while (q && (!crlf || q<crlf))
				{
					/* psite1!site2 !site3, isspace wg. \r\n und Space */
					if ((q-3==p || q[-1]=='!' || q[-1]==' ')
						&& (q[sitelen]=='!' || isspace(q[sitelen]) || !q[sitelen]))
						found = TRUE;
					q = strstr(q+1,opt->site);
				}
				if (found)
					continue;
			}
		}

		if (!opt->box)
		{
			/*
				im FrontEnd-Modus Flags exportieren

				's'-Zeile, Flags:
				G gelesen
				F fixiert (lschgeschtzt)
				X gelscht (bzw. markiert)

				bei PMs wird auerdem eine neue 'B'-Zeile mit dem aktuellen 
				Status sowie dem Zeitpunkt der letzten nderung generiert.
			*/
			exp_status = (pm && config.exp_status && (ix.status & 0x7fff));
			exp_flags  = (config.exp_flags && (ix.flags & (GELESEN | TOTAL | GESCHUETZT)));
			if (exp_status || exp_flags)
			{
				char status[] = " GZBWMAYK";
				char bzeile[20];
				struct tm *tp;

				/*
					die letzten beiden Bedingungen zielen auf alte Datenbanken, in
					denen statt des t_status ein anderer Wert gespeichert wurde
				*/
				if (exp_status && !(ix.flags & YOURMAIL)
				&& ix.t_status && (ix.t_status < ix.t_text))
				{
					tp = localtime(&ix.bearbeit);
					/* alte B-Zeile mit neuer berschreiben */
					bzeile[0] = status[ix.status & 0x7fff];
					strftime(bzeile+1,13,"%Y%m%d%H%M",tp);
					memcpy(buf + ix.t_status,bzeile,13);
				}

				/* Header ausgeben */
				fwrite(buf,ix.t_text,1,exp_f);

				/*
					bei selbst geschriebenen PMs B-Zeile erzeugen, sofern nicht 
					vorhanden, und mit aktuellem Bearbeitungsstatus versehen
				*/
				if (exp_status && (ix.flags & YOURMAIL) && !ix.t_status
				&& (ix.status & 0x7fff))
				{
					tp = localtime(&ix.bearbeit);
					strftime(bzeile,19,"B %Y%m%d%H%M\r\n",tp);
					bzeile[1] = status[ix.status & 0x7fff];
					fwrite(bzeile,strlen(bzeile),1,exp_f);
				}

				if (exp_flags)
				{
					/* Flags ausgeben */
					fputc('s',exp_f);
					if (ix.flags & GELESEN)
						fputc('G',exp_f);
					if (ix.flags & TOTAL)
						fputc('X',exp_f);
					else
					if (ix.flags & GESCHUETZT)
						fputc('F',exp_f);
					fputs("\r\n",exp_f);
				}
			}
			else if (!fwrite(buf,ix.t_text,1,exp_f))
			{
				fehler(NO_SPACE);
				ok = FALSE;
				break;
			}
		}
		else /* opt->box */
		{
			/*
				Im Boxmodus Kleinbuchstabenzeilen NICHT ausgeben. uo,30.6.94
				auerdem ggf. bei Crosspostings nur einen Gruppennamen liefern.
				Basiert darauf, da Zeilen immer mit \r\n abgeschlossen sind,
				fwrite eines greren Blocks schneller ist als viele kleine
				fputs, weniger Kleinbuchstabenzeilen als sonstige Zeilen
				kommen, und das Schreiben auf den Messagepuffer wegen 
				mglichen mmap(MAP_SHARED) tdlich wre bzw. MAP_PRIVATE
				dann Speicher verschlingen wrde.
				Der Zeitbedarf hlt sich in Grenzen.
			 */
			char *s = buf;
			char *last=NULL;

			while (s && *s && s-buf < ix.t_text)
			{
				/* Kleinbuchstabenzeilen sowie eventuell G-Zeilen filtern */
				if (islower(*s) || (!pm && opt->one_group && *s=='G'))
				{
					/* Diese Zeile beginnt mit einem Kleinbuchstaben */
					if (last)
					{
						/* Noch nicht gesicherte Zeilen weg */
						fwrite(last,(size_t)(s-last),1,exp_f);
						last=NULL;
					}
				}
				else if (!last)
					last=s;
				s=strchr(s,'\r');
				if (s)
					s+=2; /* \r\n bergehen */
			}
			/* Irgendetwas noch nicht weggeschrieben? */
			if (last)
				fwrite(last,(size_t)(ix.t_text-(last-buf)),1,exp_f);
			if (opt->one_group && !pm)
				fprintf(exp_f,"G%s\r\n",name);
		}

		/*
			Header ist jetzt geschrieben, nun body schreiben
			Der steht, falls die Msg. < MSG_SIZE war, schon kpl. in buf, 
			ansonsten wird die Nachricht (mit beliebiger Lnge) in Blcken 
			von MSG_SIZE Bytes ausgelesen.
		*/
		if (ix.len < MSG_SIZE)
		{
			/* kodierte Nachrichten dekodieren */
			if (ix.flags & CODIERT)
				decode((UBYTE *) &buf[ix.t_text],0,0,ix.len - ix.t_text);

			if (!fwrite(buf + ix.t_text,ix.len - ix.t_text,1,exp_f))
			{
				fehler(NO_SPACE);
				ok = FALSE;
				break;
			}
		}
		else
		{
			size_t body_bytes;
			size_t block;
			size_t dec_offset=0;
			UWORD dec_last=0;

			body_bytes = ix.len - ix.t_text;
			while (body_bytes)
			{
				block = (body_bytes > MSG_SIZE) ? MSG_SIZE : body_bytes;
				fread(buf,block,1,dat_f);
				if (ix.flags & CODIERT) {
					dec_last=decode((unsigned char *)buf,dec_offset,dec_last,block);
					dec_offset+=block;
				}
				if (!fwrite(buf,block,1,exp_f))
				{
					fehler(NO_SPACE);
					ok = FALSE;
					break;
				}
				body_bytes -= block;
			}
			if (!ok)
				break;
		}

		/* bei PMs im Boxmodus das Tauschdatum sichern */
		if (opt->box && pm)
		{
			fback();
			ix.getauscht = jetzt;
			put_index(&ix);
			fseek(ind_f,0,SEEK_CUR);

			/* Mailstatus ("im Tausch") erzeugen */
#if 0
			make_mailstatus(buf);
#endif
		}

		*bytes += ix.len;
		anzahl++;

	} /* while */

	mprintf("%-30s%5ld\r\n",name,anzahl);

	*mails += anzahl;

	fclose(ind_f);
	fclose(dat_f);

	return ok;

} /* exportiere_gruppe */



/*
	Mailstatus von Box -> User exportieren

	->			exp_f		:	export-stream
				ind_f		:	Indexdatei-stream
				dat_f		:	Datendatei-stream
				alle		:	TRUE = alle, FALSE = nur neue

	vgl. export_status2box() in frontend/infile.c
*/
static void export_status2user(FILE *exp_f,FILE *ind_f,FILE *dat_f,BOOLEAN alle)
{
	char status[] = "?GZBWMAYKT";
	char id[MAX_HDR];
	KENNUNG kennung;
	INDEX ix;
	char datum[20];
	char *p;

	rewind(ind_f);
	rewind(dat_f);
	lies_kennung(ind_f,&kennung);

	if ((alle && !kennung.status_anz)
	|| (!alle && !kennung.status_neu))
		return;

	/* nur NEUE Statusinfos exportieren? dann zur ersten NEUEN Position */
	if (!alle)  fgoto(kennung.erst_stat);

	while (get_index(&ix))
	{
		if ((alle && ix.status)  ||  ix.status & GEAENDERT)
		{
			if (ix.status & GEAENDERT)
			{
				ix.status &= ~GEAENDERT;
				fback();
				fflush(ind_f);
				put_index(&ix);
				fflush(ind_f);
			}
			fseek(dat_f,ix.pos,SEEK_SET);
			fgets(id,sizeof(id),dat_f);
			if ((p = strpbrk(id,"\r\n")) != NULL)
				*p = 0;
			if (ix.status > 0 && ix.status < sizeof(status)-1)
			{
				fprintf(exp_f,"%s\r\n",id);
				strftime(datum,19,"%Y%m%d%H%M",localtime(&ix.bearbeit));
				fprintf(exp_f,"B%c%s\r\n",status[ix.status],datum);
				strftime(datum,19,"%Y%m%d%H%M",localtime(&ix.eingabe));
				fprintf(exp_f,"E%s\r\n",datum);
			}
		}
	}

	rewind(ind_f);
	if (kennung.status_neu)
	{
		kennung.status_neu = 0;
		fwrite(&kennung,sizeof(KENNUNG),1,ind_f);
		rewind(ind_f);
	}
	rewind(dat_f);
	lies_kennung(ind_f,NULL);

} /* export_status2user */



#include "msgid.h"

/*
	->			req_list	:	Datei mit ID-Zeilen
				outfile	:	Name der Ausgabedatei

	<-			TRUE		:	Kommando ok
				FALSE		:	Kommando fehlerhaft

	Daten werden an <outfile> angehngt, falls es bereits existiert.
	Das Funktionsergebnis ist unabh. davon, ob berhaupt eine ID gefunden 
	wurde. Wenn man daran interessiert ist, muss man sich das Outfile ansehen.

	todo:
	- abwgen, ob diese Funktion eher in Export, message- oder  msgid-Verwaltung 
	  gehrt (Favorit: message)
*/
BOOLEAN mdb_get_article(const char *req_list,const char *outfile, BOOLEAN dummy)
{
	char id[MAX_HDR];
	char name[MAX_CPG];
	GLOBAL_REF ref;
	FILE *req_f;
	BOOLEAN pm;
	BOOLEAN found=FALSE;
	int gh;

	if (!req_list || !outfile)
		return FALSE;

	if ((req_f = fopen(req_list,"r")) == NULL)
		return FALSE;

	while (fgets(id,sizeof(id),req_f))
	{
		if (!msgid_find(id,&ref))
			continue;

		pm = (ref.grp_nr & 0x8000) != 0;
		if (!gruppen_nr2name(ref.grp_nr & 0x7fff,pm,name))
			continue;

		if ((gh = group_open(getpid(),name,pm,FALSE)) < 0)
			continue;

		if (index_goto(gh,ref.msg_nr) && msg_read(gh))
		{
			if (msg_export_raw(gh,outfile,NULL))
				found=TRUE;
		}

		group_close(gh,getpid());
	}

	fclose(req_f);

	return found;

} /* mdb_get_article */
