/* import.c : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB3 (mausnet)
       (might be the wrong person except for offline)

    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
/*
	Offline - MausTausch-FrontEnd

	Import-Routinen
	(c) 1992-95 bei Frank Baumgart

	begonnen am 28.03.1992
	Version vom 11.01.1995
*/

#define UWE_WILL_GROESSER_ZEILEN


#include "portab.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if defined(__UNIX__)
#include <memory.h>
#endif
#include <unistd.h>
#include <time.h>
#include <limits.h>
#include <errno.h>

#ifdef HAVE_MMAP
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif

#include "system.h"
#include "missing/missing.h"
#include "mdb_conf.h"
#include "struktur.h"
#include "import.h"
#include "mdb_glob.h"
#include "mdb_intr.h"
#include "msgid.h"
#include "message.h"
#include "screen/io.h"
#include "tool.h"
#include "filter.h"
#include "fehler.h"
#include "mt_cmd.h"


extern CONFIG config;				/* globale Einstellungen (Verkettung...) */
extern KENNUNG leere_kennung;		/* initialisierte Datenbank-Kennung */


char username[80] = { 0 };			/* akt. Username aus HEAD-Block */


/*
	Struktur fr Zuordnungstabelle zur Sortierung der Reihenfolge der
	Abarbeitung bei Import des Outfiles
*/
typedef struct
{
	CRC crc;								/* CRC ber Gruppennamen */
	long offset;						/* Offset des Messageanfangs */

} SORT;


static BOOLEAN sortierung(FILE *imp_f,SORT **zuord,long maximal,long *anzahl);
static int     vergleich(const void *eins,const void *zwei);
static void		init_import_var(void);
static BOOLEAN process_block(FILE *imp_f,BOOLEAN box);
static void    kommentarverkettung(const char *id,INDEX *ix_neu,GRUPPE gnr);
static GRUPPE  gruppenwechsel(const char *name,BOOLEAN pm);
static GRUPPE  oeffne_gruppe(const char *name,BOOLEAN pm);
static BOOLEAN schliesse_gruppe(void);
static void    status_set(const char *id,UWORD status,time_t datum);

static KENNUNG kennung;				/* Kennung der akt. Gruppe */

static char *buf;						/* Messagepuffer von MSG_SIZE Bytes */
static INDEX leerindex;				/* leere Indexstruktur */
static char akt_grp[MAX_CPG];		/* Name der zuletzt bearbeiteten Gruppe */

static FILE *dat_f,*ind_f;			/* Daten- und Indexdatei */
static FILE *log_f = NULL;			/* Protokolldatei */

static ULONG imp_msgs;				/* Anzahl importierter Msgs. */
static ULONG cross_msgs;			/* Anzahl Crosspostings im gesamten Import */
static MESSAGE neue_msgs;			/* Anzahl neuer Msgs. in einer Gruppe */
static ULONG neue_bytes;			/* Anzahl neuer Bytes in einer Gruppe */
static ULONG dupe_msgs;				/* Anzahl Dupe-Msgs. je Importdatei */
static MESSAGE imp_filt;			/* Anzahl gelscht importierter Msgs. (Filter) */
static ULONG gefiltert;				/* Anzahl gefilterter Msgs. je Importdatei */
static ULONG anz_status;			/* Anzahl Mailstatusmeldungen je Importdatei */

static time_t import_zeit;			/* Importzeitpunkt */

#ifdef HAVE_MMAP
static caddr_t mmaddr;
static size_t mmsize;
static int is_mmapped;
#endif



#if 0
/*
	einfache CRC-Routine zur Gruppennamen-Unterscheidung
*/
static CRC simple_crc(const char *text,size_t len)
{
	CRC crc = 0;

	while (len--)
		crc = (crc << 1) ^ *text++;

	return crc;

} /* simple_crc */
#endif



/*
	Sortierung des Outfiles ermitteln und ggf. Offsets der einzelnen
	Messages merken

	->			imp_f		:	stream der Importdatei (Outfile.txt)
			<-	zuord		:	Speicher fr Zuordnungstabelle
				maximal	:	Gre der Tabelle (Anzahl mglicher Eintrge)
			<-	anzahl	:	Anzahl vorhandener Eintrge

	<-			TRUE		:	Outfile sortiert (Quark)
				FALSE		:	Outfile unsortiert (MAUS)

	Die Feststellung der Sortierung wird anhand der ersten 256 Gruppen-
	nummern vorgenommen.
	Handicap des gesamten Verfahrens ist sicherlich, da smtliche 
	neuen Gruppen die Nummer 0 bekommen, voneinander also nicht zu 
	unterscheiden sind. Dies wre zu umgehen, indem die bislang nicht 
	existierenden Gruppen bereits HIER angelegt wrden.

	Ansatz zur Minderung dieses Nachteiles:
	CRC ber neuen Gruppennamen modulo speicherbare Gruppennummern.
	Damit werden Msgs. neuer Gruppen zumindest zusammen eingeordnet, 
	wenngleich auch Kollisionen entstehen. Der Vorteil sollte jedoch 
	berwiegen.
	Die Sortierung selbst findet aber anhand der 32 Bit CRC anstelle
	der lediglich 10 Bit Nummer statt, da dies eindeutiger ist.

	Es wird verfolgt, ob das Outfile wirklich unsortiert war.
	Dazu wird beim Gruppenwechsel anhand der CRCs in der Hashtabelle 
	schonmal[] verglichen, ob die Gruppe bereits im Outfile vorkam.
	Eine Kollisionsbehandlung erscheint unntig.

	Falls die anfngliche Tabellengre nicht ausreichend sein sollte 
	wird versucht, diese zu verdoppeln. Sollte dies fehlschlagen, wird 
	das Outfile so behandelt, als ob es sortiert wre, so da durch die 
	dann erfolgende sequentielle Bearbeitung keine Nachrichten bersehen 
	werden.

	Verbesserungsmglichkeiten:
	- 16 Bit CRC statt 32 Bit berechnen. Reicht aus und ist schneller.
	- PMs ebenfalls nach Empfnger unterscheiden
*/
static BOOLEAN sortierung(FILE *imp_f,SORT **zuord,long maximal,long *anzahl)
{
	char grp_name[MAX_CPG];		/* Gruppenname */
	long save_pos;
	long offset = 0;				/* Messageoffset */
	CRC crc;							/* CRC ber Gruppenname */
	CRC last_crc = 0;				/* CRC der letzten Gruppe */
	size_t msg_nr = 0;			/* Nr. der aktuellen Message */
	UWORD nr;						/* Nr. der aktuellen Gruppe */
	char schonmal[1024];			/* Gruppe schon mal bearbeitet ? */
	BOOLEAN sortiert = TRUE;	/* Flag, ob Outfile sortiert */
	BOOLEAN weitere = TRUE;
	char c;
	long wechsel = 0;				/* Anzahl Positionswechsel */
	long zeile = 0;

	*anzahl = 0;
	memset(schonmal,FALSE,sizeof(schonmal));

	save_pos = ftell(imp_f);
	rewind(imp_f);
	if ((size_t)65536L==0 || pufferung(imp_f,NULL,_IOFBF,(size_t)65536L))
		if ((size_t)32768L==0 || pufferung(imp_f,NULL,_IOFBF,(size_t)32768L))
			pufferung(imp_f,NULL,_IOFBF,16384);

	/* Outfiles, die nicht mit '#' beginnen, sind keine */
	if ((*buf = fgetc(imp_f)) != '#')
	{
		*anzahl = -1;
		return 0;
	}
	ungetc(*buf,imp_f);

	/*
		Sortierungsinformation im HEAD-Block, sofern angegeben, auswerten
	*/
	if (fgets(buf,MAX_HDR,imp_f) && !strncmp(buf,"#HEAD",5))
	{
		while (fgets(buf,MAX_HDR,imp_f) && *buf == ':')
		{
			if (buf[1] == 'S')
			{
				if (strchr(buf+2,'U') && strchr(buf+2,'G'))
					/* sortiert lt. HEAD-Block */
					return TRUE;
				else
					/* unsortiert lt. HEAD-Block */
					break;
			}
		}
	}
	rewind(imp_f);

	/* Outfile nach '#', 'B', 'G' durchsuchen */
	while (weitere  &&  (c = fgetc(imp_f)) != EOF)
	{
		zeile++;

		switch (c)
		{
			case '#' :
				offset = ftell(imp_f) - 1;
				if (!fgets(buf,MAX_HDR,imp_f))
					weitere = FALSE;
				else
				if (*buf=='\r' || *buf=='\n')
					weitere = FALSE;
				else
				{
					if (msg_nr == maximal)
					{
						/* Zuordnungspuffer verdoppeln */
						*zuord = (SORT *) realloc(*zuord,sizeof(SORT) * maximal * 2);
						/* fail safe, wie sortiertes Outfile behandeln */
						if (!*zuord)
							return TRUE;

						maximal *= 2;
					}

					/* 
						Infofiles, Statusmeldungen etc. werden durch die 
						hchstmgliche ID-CRC am Schlu ausgewertet, es sei 
						denn, der ID folgt noch eine PM- oder Gruppenangabe
					*/
					(*zuord)[msg_nr].crc    = (CRC) -1;
					(*zuord)[msg_nr].offset = offset;
					msg_nr++;
				}
				break;


			case 'B' :
				/* persnliche Mitteilung bzw. Mailstatus */
				/* erhalten smtlich die "Gruppen"-Kennung 0, da sie 
					stets am Anfang des Outfiles stehen. Um Verwechslungen 
					mit spteren neuen Gruppen, deren Nummer ebenfalls 0 
					ist, zu vermeiden, wird schonmal[0] nicht verndert */
				fgets(grp_name,MAX_CPG,imp_f);

				/* bei # ID vergebene CRC verwerfen */
				if (msg_nr)
				{
					(*zuord)[msg_nr - 1].crc    = 0;
					(*zuord)[msg_nr - 1].offset = offset;
				}
				break;


			case 'G' :
				/* ffentliche Nachricht */
				fgets(grp_name,MAX_CPG,imp_f);
				{
					char *pos;
					if ((pos = strpbrk(grp_name,"\r\n")) != NULL)
						*pos = 0;
				}
				/* gar kein korrekter Name ? evtl. gar kein Outfile ? */
				if (!*grp_name)
					break;
				/* CRC ber den Gruppennamen bilden */
				crc = crc32ansi(grp_name,strlen(grp_name));
/*				crc = simple_crc(grp_name,strlen(grp_name));*/
				/* nur PMs haben CRC 0 (damit sie zuerst kommen) */
				if (!crc)  crc++;

				/* gab es diese Gruppe bislang noch nicht ? */
/*				if (!nr)*/  nr = (UWORD) (crc & 1023L);

				/* bei # ID vergebene CRC verwerfen */
				if (msg_nr)
				{
					(*zuord)[msg_nr - 1].crc    = crc;
					(*zuord)[msg_nr - 1].offset = offset;
				}

				/* aktuelle Gruppe ungleich letzter ? */
				if (crc != last_crc)
				{
					/* Anzahl Gruppenwechsel zhlen */
					wechsel++;

					/* berprfung auf Sortierung */
					if (sortiert)
					{
						/* wurde diese Gruppe schon mal bearbeitet ? */
						if (schonmal[nr])
							sortiert = FALSE;
						schonmal[nr] = TRUE;
						last_crc = crc;
					}
				}
				break;


			default :
				fgets(buf,(int)MSG_SIZE,imp_f);
				{
					size_t len = strlen(buf);
					/* Zeilen ohne CR und/oder LF weisen auf Outfile-Defekt hin */
					if (!len || (buf[len-1]!='\r' && buf[len-1]!='\n'))
					{
						fprintf(stderr,"Fehler (kein CR/LF in Zeile %ld\n",zeile);
						fprintf(stderr,"(%ld): %s\n",zeile,buf);
						*anzahl = -1;
						return 0;
					}
				}
		}
	}

	*anzahl = msg_nr;

	fseek(imp_f,save_pos,SEEK_SET);

	if (!sortiert)
		qsort(*zuord,msg_nr,sizeof(SORT),vergleich);

	return sortiert;

} /* sortierung */



/*
	Vergleichsroutine
	Sortierung nach aufsteigenden CRCs (Gruppennummern) und Dateipositionen
*/
static int vergleich(const void *eins,const void *zwei)
{
	if (((const SORT *)eins)->crc < ((const SORT *)zwei)->crc)
		return -1;
	if (((const SORT *)eins)->crc > ((const SORT *)zwei)->crc)
		return  1;

	/* Gruppennummern sind also gleich. Bleiben die offsets */
	if (((const SORT *)eins)->offset < ((const SORT *)zwei)->offset)
		return -1;
	if (((const SORT *)eins)->offset > ((const SORT *)zwei)->offset)
		return  1;

	return 0;

} /* vergleich */



/*
	einige Import-spezifische Variablen initialisieren
	wird sowohl vom Outfile- als auch vom Infile-Import bentigt
*/
static void init_import_var(void)
{
	/* leeren Index generieren */
	memset(&leerindex,0,sizeof(INDEX));
	/*
		nicht bestehende Verkettungen werden durch 0xFFFF(FFFF) markiert 
		so da 0x0 als Verweis auf die erste Message dienen kann
	*/
	leerindex.oben			= LEER;
	leerindex.unten		= LEER;
	leerindex.letzter		= LEER;
	leerindex.links		= LEER;
	leerindex.rechts		= LEER;

	imp_msgs = cross_msgs = neue_msgs = neue_bytes = dupe_msgs = 0;
	imp_filt = 0;
	gefiltert = 0;
	anz_status = 0;

	/* Bezeichnung der letzten Gruppe lschen */
	akt_grp[0] = 0;

	/* Bezeichnung des aktiven Users lschen */
	*username = 0;

} /* init_import_var */



/*
	outfile.txt importieren

	->			outfile	:	Dateiname inkl. Pfad des Outfiles
				box		:	TRUE = Boxmodus, FALSE = FrontEnd-Modus

	<-			TRUE		:	import verlief erfolgreich
				FALSE		:	import schlug fehl
*/
BOOLEAN mdb_import(const char *outfile,BOOLEAN box)
{
	char head[PATH_MAX];
	char tmp[PATH_MAX];
	FILE *imp_f;
	FILE *tmp_f;
	BOOLEAN ok = FALSE;
	size_t i = 0;								/* Anzahl aller Nachrichten */

	long wechsel = 0;							/* Dateipositionswechsel zhlen */
	BOOLEAN sortiert;							/* Outfile sortiert oder nicht */
	SORT *zuord = NULL;
	long anzahl = 1,gesamt = 0;

	if ((buf = (char *) malloc(MSG_SIZE)) == NULL)
	{
		fehler(NO_MEMORY);
		return FALSE;
	}

	import_zeit = time(NULL);

	/* Filter einlesen */
	if (!box)  lade_filter();

	/*
		Outfile ffnen und spter evtl. puffern
		statt komplett in den Speicher einzulesen
	*/

	/* Outfile ffnen (falls nicht stdin) */
	if (!strcmp(outfile,"-"))
		imp_f = stdin;
	else
	{
		if ((imp_f = fopen(outfile,BINREAD)) == NULL)
		{
			if (!box)
				verwerfe_filter();
			free(buf);
			fehler(OUT_OPEN);
			return FALSE;
		}
	}

	/* keinesfalls alten HEAD-Block verwenden */
	strcpy(head,config.messages);
	strcat(head,"head");
	unlink(head);

	/* soll der Import mitprotokolliert werden ? */
	if (config.imp_log)
	{
		char traffic[PATH_MAX];

		sprintf(traffic,"%str%s.log",config.log,yymmdd());
		if ((log_f = fopen(traffic,"a")) != NULL)
		{
			char zeile[80];

			strftime(zeile,70,"Import: %A, %d. %B %Y, %H:%M\n",localtime(&import_zeit));
			fputs(zeile,log_f);
		}
	}

	/*
		Datei "touchen", die das letzte Importdatum darstellt.
		z.B. zum spteren Vergleich, ob ein Infofile "neu" ist.
	*/
#ifndef BOX
	strcpy(tmp,config.infofiles);
	strcat(tmp,"import");
	if ((tmp_f = fopen(tmp,"w")) != NULL)
		fclose(tmp_f);
#endif

	init_import_var();

	/*
		alle Nachrichten einfgen
	*/
	/* Sortierung berprfen und ggf. Tabelle aufbauen */
	sortiert = config.sortiert;
	if (imp_f == stdin)
		sortiert = TRUE;
	else
	if (!sortiert)
	{
		zuord = (SORT *) malloc((size_t)(sizeof(SORT) * 2048));
		if (!zuord)
			sortiert = TRUE;	/* fail safe */
		else
			sortiert = sortierung(imp_f,&zuord,2048,&anzahl);
		if (anzahl == -1)
		{
			if (zuord)
				free(zuord);
			fehler(ILL_OUTFILE);
			goto abbruch;
		}
		gesamt = anzahl;
	}

	/* import-stream erst hier abhngig von der Sortierung puffern */
	/* stdin ??? */
	rewind(imp_f);
	if (!sortiert)
		pufferung(imp_f,NULL,_IOFBF,4096);
	else
		if ((size_t)65536L==0
		|| pufferung(imp_f,NULL,_IOFBF,(size_t)65536L))
			if ((size_t)32768L==0
			|| pufferung(imp_f,NULL,_IOFBF,(size_t)32768L))
				pufferung(imp_f,NULL,_IOFBF,16384);

	do
	{
		/* Dateiposition der nchsten Nachricht anspringen */
		if (!sortiert && i < gesamt && ftell(imp_f)!=zuord[i].offset)
		{
			fseek(imp_f,zuord[i].offset,SEEK_SET);
			wechsel++;
		}

		/* aktuelle Nachricht einfgen */
		ok = process_block(imp_f,box);
		i++;
		if (!sortiert)  anzahl--;
	}
	while (ok && (sortiert || anzahl) && !feof(imp_f));

	schliesse_gruppe();				/* Gruppendatei schlieen */
	grp_update(FALSE);				/* Gruppenliste updaten */
	grp_update(TRUE);					/* dito Userliste */
	if (zuord)							/* Speicher fr Zuordnungstabelle */
		free(zuord);

	/*
		diverse Import-Statistiken ausgeben
	*/
#ifdef TEXT_FRONTEND
	mputs("");
	mprintf("Blcke        : %ld\r\n",(long)i);
	if (imp_msgs)
		mprintf("Mails         : %lu\r\n",imp_msgs);
	if (cross_msgs)
		mprintf("Crosspostings : %lu\r\n",cross_msgs);
	if (dupe_msgs)
		mprintf("Dupes         : %lu\r\n",dupe_msgs);
	if (gefiltert)
		mprintf("gefiltert     : %lu\r\n",gefiltert);
	if (anz_status)
		mprintf("Mailstatus    : %lu\r\n",anz_status);
	if (wechsel)
		printf("fseeks: %ld\n",wechsel);
	taste();
#endif


abbruch:

	/* Outfile schlieen */
	if (imp_f != stdin)
		fclose(imp_f);

	/* Autoren- und Themenfilter freigeben */
	if (!box)  verwerfe_filter();

	free(buf);

	/* Logdatei, falls vorhanden, schlieen */
	if (log_f)
	{
		fputs("==============================================\n",log_f);
		fprintf(log_f,"%-38s%8ld\n","Blcke:",(long)i);
		if (imp_msgs)
			fprintf(log_f,"%-38s%8lu\n","Mails:",imp_msgs);
		if (cross_msgs)
			fprintf(log_f,"%-38s%8lu\n","Crosspostings:",cross_msgs);
		if (dupe_msgs)
			fprintf(log_f,"%-38s%8lu\n","Dupes:",dupe_msgs);
		if (gefiltert)
			fprintf(log_f,"%-38s%8lu\n","gefiltert:",gefiltert);
		if (anz_status)
			fprintf(log_f,"%-38s%8lu\n","Mailstatus:",anz_status);
		fputs("\n\n",log_f);
		fclose(log_f);
		log_f = NULL;
	}

	if (ok && config.rm_outfile)
	{
		/* outfile lschen */
		unlink(outfile);
	}

	return TRUE;

} /* mdb_import */



/*
	Nachricht in die Datenbasis einfgen

	->			imp_f		:	outfile-stream (einzufgende Datei bzw. stdin)
				box		:	TRUE = Boxmodus, FALSE = FrontEnd-Modus

	<-			TRUE		:	es gibt evtl. noch weitere Nachrichten
				FALSE		:	das Ende der Nachrichten wurde erreicht bzw. Fehler
*/
static BOOLEAN process_block(FILE *imp_f,BOOLEAN box)
{
	static GRUPPE nr;						/* Nr. von akt. Gruppe/User */
												/* static wg. globaler ID-Verwaltung */
	GLOBAL_REF gref;
	INDEX ix_neu = leerindex;
	char *pos = buf;						/* aktuelle Position */
	int zeichen;
	char id[MAX_HDR];						/* ID-Text */
	char referenz[MAX_HDR];				/* Puffer fr Msg.Referenz */
	char grp_name[MAX_XP][MAX_CPG];	/* Gruppenname(n) */
	GRUPPE xp = 0;							/* Gruppenanzahl (crosspostings) */
	char adressat[MAX_HDR];				/* Empfngername bei PM */
	char adr_lokal[MAX_CPG];			/* Box: lokaler Empfnger/Pseudouser */
	char grp_lokal[MAX_CPG];			/* Box: Gruppenname (junk) */
	FILE *xp_f;								/* tmp. Datei fr Crosspostingtext */
	UWORD t_absender = 0;				/* Offset der A-Zeile, fr PM-Re-Import */
	size_t len;								/* aktuelle Zeilenlnge */
	BOOLEAN pm = FALSE;					/* ffentlich oder PM */
	BOOLEAN filtern = FALSE;			/* Flag, ob Msg. filtern */
	BOOLEAN mailstatus = TRUE;
	BOOLEAN first_dk = FALSE;			/* Flag, ob bereits ein # gelesen */
	char *p;

	UWORD enc_last=0;
	size_t enc_offset=0;

	*referenz = 0;
	*adressat = 0;
	*adr_lokal = 0;
	*grp_lokal = 0;

	while ((zeichen = fgetc(imp_f)) != ':')
	{
		*pos++ = zeichen;
		*pos   = 0;			/* fr korrektes Stringende sorgen */

		if (zeichen == EOF)
			return FALSE;
		if ((zeichen == '#') && first_dk && mailstatus)		/* Mailstatus */
			break;
		if (!fgets(pos,(int)(MSG_SIZE - (pos - buf)),imp_f))
			return FALSE;
		len = strlen(pos);
		pos[len] = 0;		/* fr korrektes Stringende sorgen */

		switch (zeichen)
		{
			/* ID */
			case '#' :
				first_dk = TRUE;
				if (strchr(pos+1,'@'))
				{
					/* CRC ber ID bilden (ohne CR/LF) */
					size_t id_len = strcspn(pos+1,"\r\n") +1;
					ix_neu.id_crc = crc32ansi(pos,id_len);
					strncpy(id,pos,id_len);
					id[id_len] = 0;
				}
				else
				{
					size_t l;

					if ((l = strcspn(pos,"\r\n")) == 0)
						l = len;

					/* LOG-Block ? */
					if (!strncmp(pos,"LOG",l))
					{
						log_auswertung(imp_f);
						return TRUE;
					}

					/* Kommando an MT-Prozessor ? */
					if (!strncmp(pos,"CMD",l))
					{
						skip_block(imp_f,buf);
						return TRUE;
					}

					/* HEAD-Block ? */
					if (!strncmp(pos,"HEAD",l))
					{
						head_auswertung(imp_f);
						return TRUE;
					}

					/* REN-Block (Gruppenumbenennung) ? */
					if (!strncmp(pos,"REN",l))
					{
						ren_auswertung(imp_f);
						return TRUE;
					}

					/* vielleicht ein Infofile ? */
					if (*pos == 'I' || *pos == 'J')
					{
						if (infofile(imp_f,pos,buf))
							return TRUE;
					}

					/* Ende der Nachrichten erreicht ? */
					if (*pos == '\r' || *pos == '\n' || !*pos)
						return FALSE;
					else
					{
						/* unbekannter Block */
						fprintf(stderr,"unbekannter Block: %s\r\n",pos);
						skip_block(imp_f,buf);
						return TRUE;
					}
				}
				pos += len;
				break;

			/* Adressat */
			case 'A' :
				ix_neu.header |= 1L << ('A'-'A');
				t_absender = (UWORD) (pos - buf);
				strcpy(adressat,pos);
				if ((p = strpbrk(adressat,"\r\n")) != NULL)
					*p = 0;
				pos += len;
				break;

			/* Organisation */
			case 'O' :
				pos += len;
				break;

			/* Name des Absenders, sofern ungleich Adresse */
			case 'N' :
				ix_neu.header |= 1L << ('N'-'A');
				pos += len;
				break;

			/* Von (Absender) */
			case 'V' :
				ix_neu.header |= 1L << ('V'-'A');
				ix_neu.t_author = (UWORD) (pos - buf);
				ix_neu.author_crc = crc32ansi(pos,strcspn(pos,"\r\n"));
				if (!box && config.filter_author)
					filtern |= msg_filter(AUTOR,ix_neu.author_crc,pos);
				pos += len;
				break;

			/* Betreff */
			case 'W' :
				ix_neu.header |= 1L << ('W'-'A');
				ix_neu.t_topic = (UWORD) (pos - buf);
				ix_neu.topic_crc = crc32ansi(pos,strcspn(pos,"\r\n"));
				if (!box && config.filter_topic)
					filtern |= msg_filter(TITEL,ix_neu.topic_crc,pos);
				pos += len;
				break;

			/* MessageID fr Mitteilung via Gateway */
			case 'I' :
				ix_neu.header |= 1L << ('I'-'A');
				pos += len;
				break;

			/* MessageID fr Kommentar via Gateway */
			case 'R' :
				ix_neu.header |= 1L << ('R'-'A');
				pos += len;
				break;

			/* Kopie */
			case 'K' :
				mputs("Kopien werden derzeit nicht untersttzt");
				pos += len;
				break;


			/* Verteilungsinfo ? ("distribution") */
			case 'D' :
				ix_neu.header |= 1L << ('D'-'A');
				switch (*pos)
				{
					/* lokal */
					case 'L' :
						ix_neu.flags |= V_LOKAL;
						break;

					/* Mausnet */
					case 'M' :
						ix_neu.flags |= V_MAUSNET;
						break;

					/* Net */
					case 'N' :
						ix_neu.flags |= V_NET;
						break;

					default	:
						fprintf(stderr,"unbekannte Distribution:%c\r\n",*pos);
				}
				pos += len;
				break;

			/* Kommentar */
			case '-' :
				if (config.verkettung)
					strcpy(referenz,pos);
				pos += len;
				break;

			/* Eingabezeit */
			case 'E' :
				ix_neu.header |= 1L << ('E'-'A');
				mtzeit_to_time_t(&ix_neu.eingabe,pos);
				pos += len;
				break;
#if 0
			/* expire-Zeit */
			case 'e' :
				/* Datum extrahieren */
				mtzeit_to_time_t(&ix_neu.expire,pos);
				pos += len;
				break;
#endif

			/* Bearbeitungsstatus, ebenso Kennung fr PM bzw. Mailstatus */
			case 'B' :
				ix_neu.header |= 1L << ('B'-'A');
				pm = TRUE;
				/*
					mgliche Bearbeitungsstatus:

					'N' = nicht gelesen
					'Z' = zurckgestellt
					'B' = beantwortet
					'G' = gelesen
					'W' = weitergegeben
					'M' = MausNet
					'A' = angekommen in Zielbox
					'Y' = angekommen ber Gateway
					'K' = kopiert
					'T' = getauscht
				*/
				switch (*pos)
				{
					case 'N' :	ix_neu.status &= ~GELESEN;
									break;
					case 'Z' :	ix_neu.status = ZURUECKGESTELLT;
									break;
					case 'B' :	ix_neu.status = BEANTWORTET;
									break;
					case 'G' :	ix_neu.status = GELESEN;
									break;
					case 'W' :	ix_neu.status = WEITERGEGEBEN;
									break;
					case 'M' :	ix_neu.status = MAUSNET;
									break;
					case 'A' :	ix_neu.status = ANGEKOMMEN;
									break;
					case 'Y' :	ix_neu.status = GATEWAY;
									break;
					case 'K' :	ix_neu.status = KOPIERT;
									break;
					case 'T' :	ix_neu.status = GETAUSCHT;
									break;
					default	:
						fprintf(stderr,"unbekannter Bearbeitungsstatus:%c\r\n",*pos);
				}

				/* Datum ermitteln und eintragen */
				if (pos[1])
					mtzeit_to_time_t(&ix_neu.eingabe,pos+1);
				ix_neu.t_status = (UWORD) (pos - buf);
				pos += len;
				break;

			/* Gruppe */
			case 'G' :
				ix_neu.header |= 1L << ('G'-'A');
				pm = FALSE;
				mailstatus = FALSE;
				/* ber MAX_XP hinausgehende werden nur durchgeschleift */
				if (xp < MAX_XP)
				{
					if (xp)
						ix_neu.flags |= XP;

					strcpy(grp_name[xp],pos);
					if ((p = strpbrk(grp_name[xp],"\r\n")) != NULL)
						*p = 0;
					/* 
					 * Nicht vorhandene Gruppen zwar durchschleifen, aber
					 * hier nicht eintragen, um den Platz fr vorhandene
					 * freizulassen. Problem: Wenn die erste Gruppe nicht 
					 * vorhanden ist, belegt sie trotzdem einen Platz. 
					 * Knnte man durch Tricks abfangen, aber wichtig ist
					 * das nicht. -- Uwe
					 */
					if (!xp || group_path(grp_name[xp],NULL,FALSE) >= 0)
						xp++;
					else
						*grp_name[xp]='\0'; /* wohl nicht ntig. */

					/* Gruppennamen ggf. umstellen, um Gruppenwechsel zu minimieren */
					if ((xp > 1) && !stricmp(akt_grp,grp_name[xp-1]))
					{
						char hilf[MAX_CPG];

						/* Gruppe der aktuellen 'G'-Zeile mit Gruppe #0 vertauschen */
						strcpy(hilf,grp_name[xp-1]);
						strcpy(grp_name[xp-1],grp_name[0]);
						strcpy(grp_name[0],hilf);
					}
				}
				pos += len;
				break;

			/* FrontEnd-Statuszeile */
			case 's' :
				if (strchr(pos,'G'))
					ix_neu.flags |= GELESEN;
				if (strchr(pos,'X'))
					ix_neu.flags |= TOTAL;
				else
				if (strchr(pos,'F'))
					ix_neu.flags |= GESCHUETZT;
				pos--;
				break;

			/*
				lokaler Empfnger, falls abweichend von Empfngeradresse.
				dadurch kann der Adressat weiterhin nach z.B. "xy@pb" und 
				"xy@pb.sub.org" unterschieden werden, falls das fr ein 
				lokales Gate bentigt wird.
				Auerdem werden Pseudouser fr Netz-DBs hierdurch angegeben.
				(war: n-Zeile)
			*/
			case 'a' :
			case 'n' :
				if (box)
				{
					strcpy(adr_lokal,pos);
					if ((p = strpbrk(adr_lokal,"\r\n")) != NULL)
						*p = 0;
				}
				pos--;
				break;

			/*
				junk-Gruppe, in die alle Nachrichten importiert werden, die 
				in keine bestehende bzw. zu erzeugende Gruppe gehren, aber 
				mit allen weiteren Angaben erhalten werden sollen
				(z.B. Fehladressierungen)
			*/
			case 'g' :
				if (box)
				{
					strcpy(grp_lokal,pos);
					if ((p = strpbrk(grp_lokal,"\r\n")) != NULL)
						*p = 0;
				}
				pos--;
				break;

			/* Path-Zeile */
			case 'p' :
				if (box)
					pos += len;	/* mu beim Export umstndlich gefiltert werden */
				else
					pos--;
				break;

			/* Xtension-Zeile */
			case 'x' :
				pos--;			/* wird derzeit einfach verworfen */
				break;

			case 'X' :			/* Flags-Zeile */
			case 'Y' :			/* Gateway-Zeile */
			case 'S' :			/* Sender-Zeile */
			case 'T' :			/* reply-To-Zeile */
			case '!' :			/* Uwe Ohse Kommentarzeile */
			case 'M' :          /* MIME-Zeile */
				pos += len;
				break;

			/* Followup-to-Zeile */
			case 'F' :
				/* ix_neu.header |= 1L << ('F'-'A'); ??? */
				pos += len;
				break;

			/* '>' Zeilen kommentarlos verwerfen */
			case '>' :
#define UWE_WILL_GROESSER_ZEILEN
#if defined(UWE_WILL_GROESSER_ZEILEN)
				pos += len;
#else
				pos--;
#endif
				break;

			default :
				mprintf("unbekannte %c-Zeile\r\n",*(pos-1));
				mprintf("ID:%.*s\r\n",(int) strcspn(buf,"\r\n"),buf);
				pos += len;
		}
	}

	ungetc(zeichen,imp_f);

	/*
		Damit bei PMs nicht flschlich ein Mailstatus erkannt wird.
		Grund: ':' in PMs ist zwingend, aber 'A' im Mailstatus optional
		Somit keine Mailstatuserkennung aufgrund der 'A'-Zeile mglich.
	*/
	if (mailstatus && zeichen != ':')
	{
printf("setze Status zu Msg.: %s\r\n",id);
		status_set(id,ix_neu.status,ix_neu.eingabe);
		anz_status++;

		return TRUE;
	}

	/* fehlende W-Zeile ergnzen */
	if (!(ix_neu.header & (1L << ('W'-'A'))))
	{
		const char betreff[] = "Wmessage created by a buggy newsreader\r\n";
		ix_neu.header |= 1L << ('W'-'A');
		ix_neu.t_topic = (UWORD) (pos - buf);
		ix_neu.topic_crc = crc32ansi(betreff+1,strlen(betreff)-3);
		strcat(pos,betreff);
		pos += strlen(betreff);
	}

	/* fehlende E-Zeile fr den Export erzeugen */
	if (!(ix_neu.header & (1L << ('E'-'A'))))
	{
		ix_neu.header |= 1L << ('E'-'A');
		ix_neu.zeit = import_zeit;
		strftime(pos,13,"E%Y%m%d%H%M\r\n",localtime(&ix_neu.zeit));
		pos += 13;
	}

	ix_neu.t_text = (UWORD) (pos - buf);

	/* auf gltige und ausreichende Headerzeilen berprfen */
	/* Mailstatusmeldungen sind bereits abgearbeitet */
	if ((pm
	&& (((ix_neu.header & NOTWENDIG_PM) != NOTWENDIG_PM)
	||  ((ix_neu.header | ERLAUBT_PM)   != ERLAUBT_PM)))
	|| (!pm
	&& (((ix_neu.header & NOTWENDIG_OM) != NOTWENDIG_OM)
	||  ((ix_neu.header | ERLAUBT_OM)   != ERLAUBT_OM))))
	{
		/*
			handelt es sich um eine selbst verfate Nachricht eines PM-Outfile-
			Imports ? (bei bernahme der PM-Daten aus einem anderen FrontEnd)
		*/
		char *user = username;
		if (!*user)
			user = config.username;
		if (((ix_neu.header & EIGENE_PM) == EIGENE_PM)
		&&  ((ix_neu.header | ERLAUBT_EIGENE) == ERLAUBT_EIGENE)
		&& (user && *user))
		{
			ix_neu.flags |= YOURMAIL;
			ix_neu.t_author = t_absender;
			strcpy(adressat,user);
			pm = TRUE;
		}
		else
		{
			fehler(ILL_HDR);
			fprintf(stderr,"(%s: %s)\r\n",
					(pm) ? "PM" : "Gruppe",
					(pm) ? adressat : grp_name[0]);
			fprintf(stderr,"%*.*s\r\n",ix_neu.t_text,ix_neu.t_text,buf);
			skip_block(imp_f,buf);
			return TRUE;
		}
	}

	/* soll die Msg. gefiltert werden ? */
	if (!box && filtern)
	{
		gefiltert++;
		if (config.import_filtered)
		{
			/*
				Unterscheidung:
				bei Kommentaren sollte die Msg. nur als teilzulschen importiert 
				werden, so da die Verkettung erhalten bleibt.
				Problem:
				wie findet man heraus, da es sich um einen Kommentar handelt ?
				die Verkettungsinfos sind hier noch nicht initialisiert.
				Evtl. anhand vorhandener Ref-ID entscheiden.
			*/
/*			if (ix_neu.oben)
				ix_neu.flags |= TEIL;*/
				ix_neu.flags |= TOTAL;
			imp_filt++;
		}
		else
		{
			skip_block(imp_f,buf);
			return TRUE;
		}
	}

	/* Mail- statt Importdatum eintragen? zur Datenbernahme im Boxmodus */
	if (config.maildate)
		ix_neu.zeit = ix_neu.eingabe;
	else
		ix_neu.zeit = import_zeit;

	/*
		Auswahl der Datenbank
		Vergleich, ob Wechsel gg. letzter Nachricht

		- im Boxbetrieb lokalen Adressaten als Usernamen verwenden
		- im Boxbetrieb ggf. junk-Gruppe verwenden
		- ansonsten entsprechenden "realen" User bzw. Gruppe verwenden
	*/
	if (!pm)
	{
		if (box && *grp_lokal)
		{
			if (stricmp(akt_grp,grp_lokal))
				nr = gruppenwechsel(grp_lokal,pm);
		}
		else
		{
			/* bei Crosspostings nicht, wird unten erledigt */
			if (stricmp(akt_grp,grp_name[0]) && (xp < 2))
				nr = gruppenwechsel(grp_name[0],pm);
		}
	}
	else
	{
		if (box)
		{
			if (stricmp(akt_grp,(*adr_lokal) ? adr_lokal : adressat))
				nr = gruppenwechsel((*adr_lokal) ? adr_lokal : adressat,pm);
		}
		else
		{
			/* falls in HEAD ein Username stand, wird dieser verwendet */
			if (*username)
				strcpy(adressat,username);
			if ((p = strchr(adressat,'@')) != NULL)
			{
				while (p > adressat && *(p-1) == ' ')
					p--;
				if (strnicmp(akt_grp,adressat,(size_t)(p - adressat)))
					nr = gruppenwechsel(adressat,pm);
			}
			else
				if (stricmp(akt_grp,adressat))
					nr = gruppenwechsel(adressat,pm);
		}
	}

	if (nr < 0)
	{
		/* ### Fehlermeldung schreiben */
		skip_block(imp_f,buf);
		mprintf("Gruppen/Postfachwechsel fehlgeschlagen\r\n");
		return TRUE;
	}

	/*
		Distribution ggf. ergnzen
		Im Unterschied zu Maus 7 wird nicht die Distribution der evtl. 
		kommentierten Nachricht verwendet, da diese evtl. nicht mehr vorliegt
	*/
	if (!pm && !(ix_neu.flags & (V_LOKAL | V_MAUSNET | V_NET)))
		ix_neu.flags |= V_NET;

	/*
		Dupecheck durchfhren
		- im Boxmodus nicht bei PMs
		- im Boxmodus sind OMs mit derselben ID wie die einer PM legal
		(in newsgroup gegatete Mails, z.B. aus Mailingliste)
	*/
	if (config.dupe_chk)
	{
		GLOBAL_REF duperef;
		BOOLEAN weg_damit=FALSE;
		if (!box)
		{
			/* wenn die Nachricht schon da ist, dann weg damit */
			if (msgid_find(id,&duperef))
				weg_damit=TRUE;
		}
		else
		{
#if 0
			if (pm)
			{
				/* Bei PMs wissen wir nur dann, da die Nachricht ein
				 * Dupe ist, wenn die ID in der DB auf eine PM beim
				 * selben Empfnger verweist. (Mailinglistenproblematik).
				 */
				if (msgid_find(id,&duperef) 
					&& (duperef.grp_nr == (nr | 0x8000)))
					weg_damit=TRUE;
			}
			else
#else
			if (!pm)
#endif
			{
				/* wenn die Nachricht ffentlich ist, und die ID in 
				 * in der Datenbank auch zu einer ffentlichen Nachricht
				 * gehrt, dann ist die Nachricht ein Dupe. Mail2News-
				 * Gateways (und umgekehrt) lassen gren.
				 */
				if (msgid_find(id,&duperef) && !(duperef.grp_nr & 0x8000))
					weg_damit=TRUE;
			}
		}

		if (weg_damit)
		{
			dupe_msgs++;
			skip_block(imp_f,buf);
			return TRUE;
		}
	}

	/*
		die Nachricht wurde akzeptiert, darum wird jetzt der Header 
		geschrieben und smtliche Textzeilen ausgelesen und direkt in die 
		Datendatei geschrieben.
		Falls es sich um ein crossposting handelt, wird eine temporre 
		Datei geschrieben, um spter in jede xp-Gruppe eingefgt zu werden.
		todo:
		JETZT nur den Header und einen Verweis in jede XP-Datei schreiben
	*/

	/* Lnge des Textes eintragen (bislang erst des Headers) */
	ix_neu.len = (size_t) ix_neu.t_text;

	if (!pm && xp > 1 && !*grp_lokal)
	{
		if ((xp_f = tmpfile()) == NULL)
			fehler(NO_TMP);
	}
	else
		xp_f = NULL;

	/* Offset des Textanfangs innerhalb des Datenfiles eintragen */
	if (!xp_f)
		ix_neu.pos = ftell(dat_f);

	/* Header sichern */
	fwrite(buf,ix_neu.t_text,1,(xp_f) ? xp_f : dat_f);

	/*
		body auslesen und schreiben
	*/
	while ((zeichen = fgetc(imp_f)) == ':')
	{
		size_t bytes;

		ix_neu.zeilen++;
		*buf = zeichen;
		/* max. 32 KB/Zeile bei 16 Bit int Systemen */
		fgets(buf+1,MSG_SIZE,imp_f);
		bytes = strlen(buf);

		/* Lnge der lngsten Zeile fr Umbruch merken */
		if (bytes > ix_neu.max_col)
			ix_neu.max_col = bytes;

		/* Nachrichten-body ggf. kodieren */
		if (pm && config.coded_pm)
		{
			ix_neu.flags |= CODIERT;
			enc_last=encode((unsigned char *)buf,enc_offset,enc_last,bytes);
		}
		ix_neu.len += bytes;
		enc_offset+=bytes;

		/* Bei Crosspostings nur in die Temporrdatei schreiben */
		if (!fwrite(buf,bytes,1,(xp_f) ? xp_f : dat_f))
		{
			fehler(WRT_DAT);
			return FALSE;
		}
	}

	/* das zuviel gelesene Zeichen an den Puffer zurckgeben */
	ungetc(zeichen,imp_f);

	/* nicht-crosspostings werden hier abgehandelt, da <nr> bekannt ist */
	if (!xp_f)
	{
		/* globale Referenz eintragen */
		gref.grp_nr = (pm) ? nr | 0x8000 : nr;
		gref.msg_nr = akt_msg();
		msgid_add(id,&gref);

		/* Kommentarverkettung aufbauen */
		if (config.verkettung && referenz[0])
			kommentarverkettung(referenz,&ix_neu,(pm) ? nr | 0x8000 : nr);	/* ??? */

		/* Index an Indexdatei anhngen */
		if (!put_index(&ix_neu))
		{
			if (ferror(ind_f))
				clearerr(ind_f);
			fehler(WRT_IND);
			return FALSE;
		}

		/* wird bei Crosspostings pro Gruppe erledigt */
		imp_msgs++;
		if (box)
			neue_msgs++;
		/* FE-Importe knnten via 's'-Zeile den GELESEN-Status gesetzt haben */
		else if (!(ix_neu.flags & GELESEN))
		{
			neue_bytes += ix_neu.len;
			neue_msgs++;
		}
	}
	else /* xp_f */
	{
		/*
			Crosspostings in alle Gruppen schreiben, die es schon gibt.
			Crosspostings legen NIE neue Gruppen an.
		*/
		BOOLEAN eingetragen=FALSE;

		while (xp-- && xp_f)
		{
			size_t bytes;

			if (*grp_lokal)
				break;

			if (group_path(grp_name[xp],NULL,FALSE) < 0)
				continue;

			/* eine EXIST. Gruppe kann (und mu nat.) gelockt werden */
			(void) group_lock(grp_name[xp],FALSE,TRUE);

			nr = gruppenwechsel(grp_name[xp],FALSE);
			if (nr < 0)
				continue;

			/* Text aus tmp-File lesen und an neues Datenfile anhngen */
			fseek(xp_f,0,SEEK_SET);

			ix_neu.pos = ftell(dat_f);
			/* Kommentarverkettung aufbauen */
			if (config.verkettung && referenz[0])
			{
				/* KV der Vorgruppe nicht bernehmen. */
				ix_neu.oben = (MESSAGE) LEER;
				ix_neu.unten = (MESSAGE) LEER;
				ix_neu.rechts = (MESSAGE) LEER;
				ix_neu.links = (MESSAGE) LEER;
				/* und die anderen Flags? Was ist, wenn dies die 
				 * kommentierte Nachricht ist, und die Kommentare
				 * schon vorher gekommen sind?
				 */
				kommentarverkettung(referenz,&ix_neu,nr);
			}
			while ((bytes = fread(buf,1,MSG_SIZE,xp_f)) != 0)
				if (!fwrite(buf,1,bytes,dat_f))
				{
					fehler(WRT_DAT);
					return FALSE;
				}

			if (!eingetragen)
			{
				gref.grp_nr = (pm) ? nr | 0x8000 : nr;
				gref.msg_nr = akt_msg();
				msgid_add(id,&gref);
				eingetragen=TRUE;
			}

			/* Index an Indexdatei anhngen */
			if (!put_index(&ix_neu))
			{
				if (ferror(ind_f))
					clearerr(ind_f);
				fehler(WRT_IND);
				return FALSE;
			}

			group_unlock(grp_name[xp],FALSE);

			imp_msgs++;
			cross_msgs++;
			if (box)
				neue_msgs++;
			/* FE-Importe knnten via 's'-Zeile den GELESEN-Status gesetzt haben */
			else if (!(ix_neu.flags & GELESEN))
			{
				neue_bytes += ix_neu.len;
				neue_msgs++;
			}
		}

		fclose(xp_f); /* lscht sich selbst, da tempfile() */
		/* kein neuer Gruppenwechsel! */
	}

	return TRUE;

} /* process_block */



#if !defined(BOX) || defined(SERVER)
/*
	Infile-Nachrichten in die PM-Datenbasis einfgen

	->			infile		:	Infile Dateiname inkl. Pfad

	<-			TRUE			:	ok
				FALSE			:	Infile-Import fehlerhaft, abgebrochen

	Dies ist eine abgespeckte Version des obigen Outfile-Imports und unbe-
	dingt mit diesem synchron zu halten.
	Da eigene _ffentliche_ Infile-Msgs. durch das Box-Echo via Outfile 
	importiert werden, bruchten hier im Prinzip nur PMs verarbeitet werden, 
	z.T. (ref-Eintrag) knnen auch nur PMs verarbeitet werden.
*/
BOOLEAN infile_einfuegen(const char *infile)
{
	FILE *imp_f;
	INDEX ix_neu;
	GLOBAL_REF gref;
	char id[MAX_HDR];						/* ID-Text */
	size_t id_len;
	char *pos;								/* aktuelle Position */
	int zeichen;
	char referenz[MAX_HDR];				/* Puffer fr Msg.Referenz */
	GRUPPE nr;								/* Nr. akt. User */
	size_t len;								/* aktuelle Zeilenlnge */
	BOOLEAN ok;

	/*
		falls kein Name aus dem gerade erfolgten Import eines HEAD-Blockes 
		zur Verfgung steht, werden weitere, schlechtere Mglichkeiten 
		verwendet, s. akt_username(), notfalls "Frank Baumgart".
	*/
	if (!*username)
		strcpy(username,akt_username());
		
	nr = oeffne_gruppe(username,TRUE);

	/*
		fr den Fall, da fr jemanden importiert werden soll, der noch gar 
		keine PM-Daten besitzt. Dann wird praktisch nur der User angelegt
	*/
	grp_update(TRUE);

	if ((buf = (char *) malloc(MSG_SIZE)) == NULL)
	{
		fehler(NO_MEMORY);
		return FALSE;
	}

	if ((imp_f = fopen(infile,BINREAD)) == NULL)
	{
		free(buf);
		fehler(NO_INFILE);
		return FALSE;
	}

	/* wichtige Import-Variblen initialisieren */
	init_import_var();

	while (1)
	{
		ix_neu = leerindex;
		pos = buf;
		*referenz = 0;

		while ((zeichen = fgetc(imp_f)) != ':')
		{
			*pos++ = zeichen;
			*pos   = 0;			/* fr korrektes Stringende sorgen */

			if ((zeichen == EOF)
			|| (!fgets(pos,(int)(MSG_SIZE - (pos - buf)),imp_f)))
			{
				ok = FALSE;
				goto quit_import;
			}
			len = strlen(pos);
			pos[len] = 0;		/* fr korrektes Stringende sorgen */

			switch (zeichen)
			{
				/* ID */
				case '#' :
					/* Ende der Nachrichten erreicht ? */
					if (!*pos || *pos == '\r' || *pos == '\n')
					{
						ok = TRUE;
						goto quit_import;
					}
					/* CRC ber ID bilden (ohne CR/LF) */
					id_len = strcspn(pos,"\r\n");
					ix_neu.id_crc = crc32ansi(pos,id_len);
					strncpy(id,pos,id_len);
					id[id_len] = 0;
					pos += len;
					break;

				/* Adressat (Empfnger) wird statt Absender gespeichert */
				/* denn einen expliziten Absender gibt es nicht */
				case 'A' :
					ix_neu.header |= 1L << ('A'-'A');
					ix_neu.t_author = (UWORD) (pos - buf);
					ix_neu.author_crc = crc32ansi(pos,strcspn(pos,"\r\n"));
					pos += len;
					break;

#if 0
				/* Von (Absender) */
				case 'V' :
					ix_neu.header |= 1L << ('V'-'A');
					ix_neu.t_author = (UWORD) (pos - buf);
					ix_neu.author_crc = crc32ansi(pos,strcspn(pos,"\r\n"));
					pos += len;
					break;
#endif

				/* Organisation */
				case 'O' :
					pos += len;
					break;

				/* Name des Absenders, sofern ungleich Adresse */
				case 'N' :
					ix_neu.header |= 1L << ('N'-'A');
					pos += len;
					break;

				/* Betreff */
				case 'W' :
					ix_neu.header |= 1L << ('W'-'A');
					ix_neu.t_topic = (UWORD) (pos - buf);
					ix_neu.topic_crc = crc32ansi(pos,strcspn(pos,"\r\n"));
					pos += len;
					break;

				/* MessageID fr Mitteilung via Gateway */
				case 'I' :
					ix_neu.header |= 1L << ('I'-'A');
					pos += len;
					break;

				/* MessageID fr Kommentar via Gateway */
				case 'R' :
					ix_neu.header |= 1L << ('R'-'A');
					pos += len;
					break;

				/* Kopie-Zeilen verwerfen */
				case 'K' :
					pos--;
					break;

				/* Kommentar ? */
				case '-' :
					if (config.verkettung)
						strcpy(referenz,pos);
					pos += len;
					break;


				/* Eingangszeit */
				case 'E' :
					ix_neu.header |= 1L << ('E'-'A');
					/* Datum eintragen */
					/* Maildatum statt Importdatum ? */
					if (config.maildate)
						mtzeit_to_time_t(&ix_neu.zeit,pos);
					else
						ix_neu.zeit = import_zeit;
					pos += len;
					break;

#if 0
				/* expire-Zeit */
				case 'e' :
					/* Datum extrahieren */
					ix_neu.expire = mktime(pos);
					pos += len;
					break;
#endif

				/* Bearbeitungsstatus, ebenso Kennung fr PM bzw. Mailstatus */
				case 'B' :
					ix_neu.header |= 1L << ('B'-'A');
					/*
						mgliche Bearbeitungsstatus:

						'N' = nicht gelesen
						'Z' = zurckgestellt
						'B' = beantwortet
						'G' = gelesen
						'W' = weitergegeben
						'M' = MausNet
						'A' = angekommen in Zielbox
						'Y' = angekommen ber Gateway
						'K' = kopiert
					*/
					switch (pos[1])
					{
						case 'N' :	ix_neu.status &= ~GELESEN;
										break;
						case 'Z' :	ix_neu.status = ZURUECKGESTELLT;
										break;
						case 'B' :	ix_neu.status = BEANTWORTET;
										break;
						case 'G' :	ix_neu.status = GELESEN;
										break;
						case 'W' :	ix_neu.status = WEITERGEGEBEN;
										break;
						case 'M' :	ix_neu.status = MAUSNET;
										break;
						case 'A' :	ix_neu.status = ANGEKOMMEN;
										break;
						case 'Y' :	ix_neu.status = GATEWAY;
										break;
						case 'K' :	ix_neu.status = KOPIERT;
										break;
						case 'T' :	ix_neu.status = GETAUSCHT;
										break;
					}

					/* Datum ermitteln und eintragen */
					mtzeit_to_time_t(&ix_neu.eingabe,pos+1);
					ix_neu.t_status = (UWORD) (pos - buf);
					pos += len;
					break;

				/* '>' Zeilen werden kommentarlos verworfen */
				case '>' :
					pos--;
					break;

				default	:
					pos += len;
			}
		}

		ungetc(zeichen,imp_f);

		/* fehlende W-Zeile ergnzen */
		/* dieser Eintrag wird bei den Headerflags nicht vermerkt */
		if (!(ix_neu.header & (1L << ('W'-'A'))))
		{
			const char betreff[] = "Wmessage created by a buggy newsreader\r\n";
			ix_neu.header |= 1L << ('W'-'A');
			ix_neu.t_topic = (UWORD) (pos - buf);
			strcat(pos,betreff);
			pos += strlen(betreff);

			/* Kennzeichen fr "ohne Betreff" */
			ix_neu.topic_crc = (CRC)0;
		}

		/* fehlende E-Zeile fr den Export erzeugen */
		if (!(ix_neu.header & (1L << ('E'-'A'))))
		{
			ix_neu.header |= 1L << ('E'-'A');
			ix_neu.zeit = import_zeit;
			strftime(pos,13,"E%Y%m%d%H%M\r\n",localtime(&ix_neu.zeit));
			pos += 13;
		}

		ix_neu.t_text = (UWORD) (pos - buf);

		/* fngt die Zeile mit einem Doppelpunkt an ? */
		while ((zeichen = fgetc(imp_f)) == ':')
		{
			size_t bytes;

			ix_neu.zeilen++;					/* Anzahl Zeilen des Msg.TEXTes */
			*pos++ = zeichen;
			/* der folgende cast hat eine Grenbeschrnkung auf 
				32 KB/Zeile bei 16 Bit int Systemen zur Folge */
			fgets(pos,(int)MSG_SIZE,imp_f);
			bytes = strlen(pos);
#if 0
			/* Nachricht ggf. kodieren, NUR den MessageTEXT, nicht den Header */
			if (config.coded_pm)
			{
				ix_neu.flags |= CODIERT;
				encode((UBYTE *) &buf[ix_neu.t_text],ix_neu.len - ix_neu.t_text);
			}
#endif
			pos += bytes;
		}

		/* das zuviel gelesene Zeichen an den Puffer zurckgeben */
		ungetc(zeichen,imp_f);

		/* Lnge des Textes eintragen */
		ix_neu.len = (size_t) (pos - buf);

		/* Offset des Textanfangs innerhalb des Datenfiles eintragen */
		ix_neu.pos = ftell(dat_f);

		/* globale Referenz eintragen */
		gref.grp_nr = nr | 0x8000;
		gref.msg_nr = akt_msg();
		msgid_add(id,&gref);

		/* Kommentarverkettung aufbauen */
		if (config.verkettung && referenz[0])
			kommentarverkettung(referenz,&ix_neu,nr | 0x8000);

		/* vermerken, da es sich um eine eigene PM handelt */
		ix_neu.flags |= YOURMAIL;

		/* gesamte Msg. an das Datenfile anhngen */
		if (!fwrite(buf,ix_neu.len,1,dat_f))
		{
			fehler(WRT_DAT);
			ok = FALSE;
			goto quit_import;
		}
		/* Index an Indexdatei anhngen */
		if (!put_index(&ix_neu))
		{
			if (ferror(ind_f))
				clearerr(ind_f);
			fehler(WRT_IND);
			ok = FALSE;
			goto quit_import;
		}

		imp_msgs++;
		neue_msgs++;

	} /* while */

quit_import:
	free(buf);
	schliesse_gruppe();
	fclose(imp_f);

	return ok;

} /* infile_einfuegen */
#endif



/*
	Kommentarverkettung herstellen

	->			id			:	Zeiger auf ID-String (opt. mit CR/LF)
				ind_neu	:	anzuhngender Indexdatensatz
				gnr		:	Gruppennummer
*/
static void kommentarverkettung(const char *id,INDEX *ind_neu,GRUPPE gnr)
{
	GLOBAL_REF gref;
	INDEX ix;								/* aktuell durchsuchter Indexdatensatz */
	MESSAGE akt_nr = LEER;				/* Nummer des Kommentars */
	MESSAGE v_nr = 0;						/* Nummer des Vorgngers, keep gcc happy */

	if (!msgid_find(id,&gref))
		return;
#if 0
	printf("ref-DB: Kom. zu %s Nr.%ld, Mail Nr.%ld\r\n",
			(gref.grp_nr & 0x8000) ? "User" : "Gruppe",
			(long) (gref.grp_nr & 0x7fff),
			(long) gref.msg_nr);
#endif
	/* liegt die referenzierte Nachricht in einer anderen Gruppe ? */
	if (gref.grp_nr != gnr)
	{
		/* nur ein Flag in der akt. Mail setzen, da die Ref. existiert */
		ind_neu->flags |= GUEK;
		return;
	}

	/* aktuelle Messagenummer */
	akt_nr = (MESSAGE) akt_msg();

	/* Indexposition der kommentierten Nachricht anfahren */
	if ((gref.msg_nr > kennung.anzahl+imp_msgs) || fgoto(gref.msg_nr))
	{
		/* FEHLER! sofort abbrechen, blo keine sparsed files */
		fgoto(akt_nr);
		return;
	}

	/* dessen Index-Struktur einlesen */
	get_index(&ix);

	/* Index der kommentierten Msg. holen und merken */
	ind_neu->oben = (MESSAGE) gref.msg_nr;

	/*
		jetzt wird die kommentierte Msg. aktualisiert
	*/
	if (++ix.kommentare == 1)
	{
		/* Index des Kommentars eintragen */
		ix.unten = akt_nr;
	}
	else
	{
		/*
			da schon mind. ein Kommentar ex., mu der Vorgnger angepat 
			werden, dazu letzten Kommentar vor dem eigenen ermitteln.
			Dafr lohnt sich auch der Eintrag in der Indexstruktur !
		*/
		/* Nummer des Vorgngers holen */
		v_nr = ix.letzter;

		if (v_nr==LEER || v_nr == 0 || v_nr > kennung.anzahl+imp_msgs)
		{
			/*
				ALARM! ### hier sollte index.letzter korrigiert werden, wenn 
				man schon merkt, da der Wert falsch ist
			*/
			fgoto(akt_nr);
			return;
		}

		/* und als solchen eintragen */
		ind_neu->links = v_nr;
	}

	/* Index dieses Kommentars als den des letzten Kommentars speichern */
	ix.letzter = akt_nr;

	/* auf korrekte Position zurckfahren */
	fback();
	/* speichern des aktualisierten Indexdatensatzes */
	put_index(&ix);

	if (ix.kommentare > 1)
	{
		/* Indexdatensatz des Vorgngers einlesen */
		fgoto(v_nr);
		get_index(&ix);
		ix.rechts = akt_nr;
		/* auf korrekte Position zurckfahren */
		fback();
		/* speichern des aktualisierten Vorgngerdatensatzes */
		put_index(&ix);
	}

	fgoto(akt_nr);

} /* kommentarverkettung */



/*
	Gruppen/Userwechsel beim Einfgen neuer Nachrichten

	->				name	:	Name der neuen Gruppe
					pm		:	Gruppentyp (PM = TRUE , ffentlich = FALSE)

	<-				-1		:	Fehler (fatal, vmtl. kein Plattenplatz mehr)
					>= 0	:	Nummer der neuen Gruppe/User
*/
static GRUPPE gruppenwechsel(const char *name,BOOLEAN pm)
{
	/* war berhaupt schon eine Gruppe/User geffnet oder ist das
		die/der erste ? dann ist natrlich kein Schlieen ntig */
	if (ind_f && !schliesse_gruppe())
		return -1;

	return oeffne_gruppe(name,pm);

} /* gruppenwechsel */



/*
	Gruppe/User (er)ffnen
	ind und dat werden geffnet/angelegt sowie Kennung eingerichtet

	->				name	:	Name der neuen Gruppe
					pm		:	Gruppentyp (PM = TRUE , ffentlich = FALSE)

	<-				>= 0	:	Nummer der neuen Gruppe/User
					-1		:	Fehler
*/
static GRUPPE oeffne_gruppe(const char *name,BOOLEAN pm)
{
	char dat_pfad[PATH_MAX];
	char ind_pfad[PATH_MAX];
	BOOLEAN neu = FALSE;
	GRUPPE nr;

	mputs(name);

	/*
		im Boxmodus:
		Liste mit den Gruppen schreiben, in die Mails importiert wurden.
		Genau genommen mte der Dupe-Check nach oben in den Tausch, nur 
		fllt der dann fr den Server und somit im FE-Modus flach(?), was 
		auch nicht geht. Also mu die ID-Verwaltung parallel zu Offline 
		bzw. darber laufen.
	*/
	/* ... */

	/* Bezeichnung der aktuellen Gruppe/User merken */
	strcpy(akt_grp,name);

	/* bei PMs at-sign (@) und diesem vorangehende Leerzeichen abschneiden */
	if (pm)
	{
		char *pos = strchr(name,'@');
		if (pos)
		{
			*pos = '\0';
			while (*--pos == ' ');
				*(pos +1) = '\0';
		}
	}

	/*
		ggf. neue Gruppe (Index und Daten) (er-)ffnen
	*/
	if ((nr = group_path(name,dat_pfad,pm)) < 0)
	{
		if (!mdb_create_group(name,pm,NULL)
		|| ((nr = group_path(name,dat_pfad,pm)) < 0))
		{
			fputs("fataler Fehler (kann Gruppe nicht anlegen)",stderr);
			return -1;
		}
		neu = TRUE;
	}
	strcpy(ind_pfad,dat_pfad);
	strcat(ind_pfad,"ind");
	strcat(dat_pfad,"dat");

	/* Gruppe bereits vorhanden: Soll automatisch geputzt werden ? */
	if (!neu && config.auto_putz)
	{
		time_t bis_zeit;

		/* Kennung einlesen */
		if ((ind_f = fopen(ind_pfad,BIN_RW_OPEN)) == NULL)
		{
			fehler(IND_OPEN);
			return -1;
		}
		lies_kennung(ind_f,&kennung);
		fclose(ind_f);

		bis_zeit = (kennung.max_time) ? kennung.max_time : config.max_time;
		bis_zeit*= 86400L;

		/* Gruppe putzen */
		/* bei PMs natrlich keine generellen Lschmarkierungen */
		if (!pm || config.tausch_pm)
		{
			/* zu lschende Messages werden markiert */
			mdb_markiere(name,pm,TRUE,TRUE,(time_t) 0,
							(time_t) (time(NULL) - bis_zeit));
		}

		/*
			gengend zu putzende Msgs. vorhanden ?
			- entweder gruppenbezogen, falls definiert, oder
			- durch allgemeine Konfiguration
		*/
		if (group_info(name,pm,&kennung)
		&& ((kennung.min_putz && kennung.putzen >= kennung.min_putz)
		|| (!kennung.min_putz && config.min_putz && kennung.putzen >= config.min_putz)))
		{
			mdb_putze_gruppe(name,pm,FALSE,NULL);
		}
	}

	/* Indexdatei (wieder) ffnen */
	if ((ind_f = fopen(ind_pfad,BIN_RW_OPEN)) == NULL)
	{
		fehler(IND_OPEN);
		return -1;
	}

	/* Kennung erneut einlesen */
	lies_kennung(ind_f,&kennung);

	/* erster Import des heutigen Tages ? */
	{
		/* time_t des letzten Imports umwandeln und
			den Tag mit dem aktuellen vergleichen */
		time_t last,zeit;
		char tmp[30];			/* ctime liefert _statischen_ string */
		int pos = 10;

		while (--pos)
			if (kennung.imported[pos])
				break;
		last = kennung.imported[pos];		/* letztes Importdatum bzw. 0 */
		time(&zeit);
		if (last)
			strcpy(tmp,ctime(&last));
		/* Monatsnamen und Tag vergleichen */
		if (!last || strncmp(tmp+4,ctime(&zeit)+4,6))
		{
			/* unterschiedlich */
			/* Tabelle schon voll ? */
			if (pos == 9)
			{
				memcpy(kennung.imported,&kennung.imported[1],9 * sizeof(kennung.imported));
				memcpy(kennung.imp_anf ,&kennung.imp_anf [1],9 * sizeof(kennung.imp_anf));
			}
			kennung.imported[pos] = zeit;
			kennung.imp_anf [pos] = kennung.anzahl;

			/* Kennung zurckschreiben */
			rewind(ind_f);
			fwrite(&kennung,sizeof(KENNUNG),1,ind_f);
		}
	}

	/* Datei (lesen und anhngen) anlegen */
	if ((dat_f = fopen(dat_pfad,BIN_R_APPEND)) == NULL)
	{
		fclose(ind_f);
		fehler(DAT_CREATE);
		return -1;
	}

	/* Datendatei-Ausgabe puffern */
	pufferung(dat_f,NULL,_IOFBF,MSG_SIZE);

	/* Dateiende anfahren */
	fseek(ind_f,0,SEEK_END);
	fseek(dat_f,0,SEEK_END);

	/* protokollieren */
	if (config.trf_log && log_f)
	{
		if (pm)
			fprintf(log_f,"%-40.40s ","PM");
		else
			fprintf(log_f,"%-40.40s ",akt_grp);
	}

	return nr;

} /* oeffne_gruppe */



/*
	Gruppe/User schlieen
	- Kennung aktualisieren (Aufruf daher nur beim Import gestattet)
	- ind und dat schlieen

	<-			TRUE	:	alles ok
				FALSE	:	mind. eine der Dateien (ind oder dat) konnte nicht 
							korrekt geschlossen werden, vmtl. out of disc space
*/
static BOOLEAN schliesse_gruppe(void)
{
	BOOLEAN ok = TRUE;

	if (imp_msgs)
	{
		time_t zeit;
		struct tm *ts;

		/*
			Index-Kennung aktualisieren:
			- Zeitpunkt der letzten Verarbeitung (also jetzt)
			- Anzahl Gesamtmessages
			- Anzahl neuer Messages
			- monatliche Trafficdaten
		*/
		time(&zeit);
#if 0
		rewind(ind_f);
		lies_kennung(ind_f,&kennung);
#endif
		/* aktuelles Datum und Uhrzeit eintragen */
		kennung.changed  = zeit;
		kennung.anzahl  += (MESSAGE) imp_msgs;
		kennung.overall += imp_msgs;
		kennung.neue    += (MESSAGE) neue_msgs;
		kennung.putzen  += imp_filt;		/* gefiltert, aber gelscht importiert */

		/* Trafficdaten, monatlich */
		ts = localtime(&zeit);
		kennung.imp_mon_anz[ts->tm_mon] += neue_msgs;
		kennung.imp_mon_anz[(ts->tm_mon+1) % 12] = 0;
		kennung.imp_mon_byt[ts->tm_mon] += neue_bytes;
		kennung.imp_mon_byt[(ts->tm_mon+1) % 12] = 0;

		/*
			### rewind ist keine gute Idee, da es im Ggs. zu fseek keine Erfolgs-
			meldung gibt, gerade dieses rewind aber fehlschlagen kann, wenn die 
			Platte voll ist, da durch rewind der Buffer geflusht wird.
		*/
		rewind(ind_f);
		fwrite(&kennung,sizeof(KENNUNG),1,ind_f);
	}

	/* Bezeichnung der letzten Gruppe lschen */
	akt_grp[0] = 0;

	/* protokollieren */
	if (config.trf_log && log_f)
		fprintf(log_f,"%5lu\n",imp_msgs);

	imp_msgs = neue_msgs = neue_bytes = imp_filt = 0;

	/*
		berprfung auf Gltigkeit der offenen streams, da das Outfile keine 
		Nachrichten, sondern nur z.B. den LOG-Block enthalten haben kann
	*/
	if (dat_f)
	{
		if (fclose(dat_f) == EOF)
		{
			perror("dat-fclose failed");
			ok = FALSE;
		}
		dat_f = NULL;
	}
	if (ind_f)
	{
		if (fclose(ind_f) == EOF)
		{
			perror("ind-fclose failed");
			ok = FALSE;
		}
		ind_f = NULL;
	}

	return ok;

} /* schliesse_gruppe */



/*
	Mailstatus setzen

	->			id			:	ID-Text
				status	:	neuer Mailstatus
				datum		:	Zeitpunkt des neuen Status

	<-			TRUE		:	eingetragen
				FALSE		:	Eintrag nicht mglich/nicht vorhanden
*/
static void status_set(const char *id,UWORD status,time_t datum)
{
	GLOBAL_REF gref;
	char ind_pfad[PATH_MAX];
	KENNUNG k;
	INDEX ix;
	FILE *ind_f;

	if (!msgid_find(id,&gref))
		return;

	printf("ref-DB: Status zu %s Nr.%ld, Mail Nr.%ld\r\n",
			(gref.grp_nr & 0x8000) ? "User" : "Gruppe",
			(long) (gref.grp_nr & 0x7fff),
			(long) gref.msg_nr);
	if (!(gref.grp_nr & 0x8000))
	{
		mputs("Status in Gruppe !? Kann ja wohl nicht sein");
		return;
	}

	if (!gruppen_nr2pfad(gref.grp_nr & 0x7fff,TRUE,ind_pfad))
		return;
	strcat(ind_pfad,"ind");

	if ((ind_f = fopen(ind_pfad,BIN_RW_OPEN)) == NULL)
	{
		fehler(IND_OPEN);
		return;
	}

	lies_kennung(ind_f,&k);
	fgoto(gref.msg_nr);
	get_index(&ix);

	if (ix.status != status)
	{
		ix.status = status | GEAENDERT;
		ix.bearbeit = datum;
		fback();
		put_index(&ix);
		fflush(ind_f);
	
		/* Kennung updaten */
		if (gref.msg_nr < k.erst_stat)
			k.erst_stat = gref.msg_nr;
		k.status_anz++;
		k.status_neu++;

		rewind(ind_f);
		fwrite(&k,sizeof(KENNUNG),1,ind_f);
	}
	fclose(ind_f);

} /* status_set */

