/* infofile.c : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB3 (mausnet)
       (might be the wrong person except for offline)

    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
/*
	Offline - MausTausch-FrontEnd

	Infofile-Verwaltung
	(c) 1992-95 bei Frank Baumgart

	begonnen am 12.06.1992
	Version vom 14.01.1995
*/


#include "portab.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <unistd.h>

#include "system.h"
#include "mdb_intr.h"			/* nur wg. eigenem Prototypen */
#include "screen/io.h"
#include "tool.h"
#include "mdb_conf.h"

extern CONFIG config;

static void infofile_diff(const char *new_name,const char *old_name);
static void parse_iti(const char *pfad);
static void parse_itg(const char *pfad);



/*
	feststellen, ob es sich um ein Infofile handelt und dieses dann
	als separate Datei speichern

	->				imp_f		:	Importstream
					zeile		:	Vergleichsstring aus der Importdatei
					buf		:	Puffer von MSG_SIZE Bytes

	<-				TRUE		:	war ein Infofile
					FALSE		:	kein Infofile

	Mit mind. drei Grobuchstaben anfangende ID-Zeilen werden als 
	(unbekanntes) Infofile betrachtet und als solches gespeichert.

	Beschrnkungen:
	- Infofilename max. 70 Zeichen
*/
BOOLEAN infofile(FILE *imp_f,const char *zeile,char *buf)
{
	char pfad[PATH_MAX];
	char old[PATH_MAX] = { 0 };
	char name[80];						/* Name fr unbekanntes Infofile */
	size_t len;
	FILE *info_f;
	char c;

	len = strcspn(zeile,"\r\n");

	/* _knnte_ es ein Infofile sein ? */
	if ((len < 3) || (len > 70)
	|| !isupper(zeile[0]) || !isupper(zeile[1]) || !isupper(zeile[2]))
		return FALSE;

	strncpy(name,zeile,len);
	name[len] = 0;

	strcpy(pfad,config.infofiles);
	strcpy(old,config.infofiles);
	mprintf("Infofile: %s\n",name);
	strcat(pfad,name);
	strcat(old,name);
	strcat(pfad,".txt");
	strcat(old,".bak");

	unlink(old);

	/* Backup der alten Infodatei fhren ? */
	if (config.infobackup)
		rename(pfad,old);

	/*
		Informationen auslesen und in entsprechende Datei schreiben
	*/
	if ((info_f = fopen(pfad,BIN_RW_CREATE)) == NULL)
		return FALSE;
	while ((c = fgetc(imp_f)) == ':')
	{
		fgets(buf,MSG_SIZE,imp_f);
		fwrite(buf,strlen(buf),1,info_f);
	}
	ungetc(c,imp_f);
	fclose(info_f);

	/* Unterschiede ermitteln ? */
	if (*config.diff && config.autodiff /* && fileexist(old)*/)
		infofile_diff(pfad,old);

	/* technische Infofile-Liste und Gruppenliste parsen */
	if (!strcmp(name,"ITI"))
		parse_iti(pfad);
	else
	if (!strcmp(name,"ITG"))
		parse_itg(pfad);

	return TRUE;

} /* infofile */



/*
	Unterschiede zwischen altem und neuem Infofile herausfiltern
*/
static void infofile_diff(const char *new_name,const char *old_name)
{
	char diff[PATH_MAX];
	char cmd[255];
	char *p;

	if (!*config.diff)
		return;

	strcpy(diff,new_name);
	if ((p = strrchr(diff,'.')) == NULL)
		return;
	*(p+1) = 0;
	strcat(diff,"dif");
	unlink(diff);

	printf("diff: %s %s\r\n",new_name,old_name);

	sprintf(cmd,"%s %s %s > %s",config.diff,new_name,old_name,diff);
	shell(cmd);

} /* infofile_diff */



/*
	verfgbare Infofiles ermitteln
*/
static void parse_iti(const char *pfad)
{
	char liste[PATH_MAX];
	char zeile[256];
	FILE *inf_f,*lst_f;
	char *p;

	if ((inf_f = fopen(pfad,"r")) == NULL)
		return;
	sprintf(liste,"%s%s",config.infofiles,"infofile.lst");
	if ((lst_f = fopen(liste,"w")) == NULL)
	{
		fclose(inf_f);
		return;
	}

	while (fgets(zeile,sizeof(zeile),inf_f))
	{
		if ((p = strpbrk(zeile,"\r\n")) != NULL)
			*p = 0;

		if (zeile[0] == '#')
			fprintf(lst_f,"%s",zeile+1);
		else
		if (zeile[0] == ':')
			fprintf(lst_f,",%s\n",zeile+1);
	}

	fclose(lst_f);
	fclose(inf_f);

} /* parse_iti */



/*
	verfgbare Gruppen ermitteln
*/
static void parse_itg(const char *pfad)
{
	char liste[PATH_MAX];
	char zeile[256];
	FILE *inf_f,*lst_f;
	char *p;

	if ((inf_f = fopen(pfad,"r")) == NULL)
		return;
	sprintf(liste,"%s%s",config.infofiles,"gruppen.lst");
	if ((lst_f = fopen(liste,"w")) == NULL)
	{
		fclose(inf_f);
		return;
	}

	while (fgets(zeile,sizeof(zeile),inf_f))
	{
		if ((p = strpbrk(zeile,"\r\n")) != NULL)
			*p = 0;

		if (zeile[0] == 'G')
			fprintf(lst_f,"%s",zeile+1);
		else
		if (zeile[0] == 'U')
			fprintf(lst_f,",%s\r\n",zeile+1);
	}

	fclose(lst_f);
	fclose(inf_f);

} /* parse_itg */
