/* mark.c : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB3 (mausnet)
       (might be the wrong person except for offline)

    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
/*
	Offline - MausTausch-FrontEnd

	Nachrichten-Markierung
	(c) 1992-94 bei Frank Baumgart

	Version vom 21.11.1994
*/


#include "portab.h"
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <limits.h>

#include "system.h"
#include "mdb_conf.h"
#include "struktur.h"
#include "mdb_glob.h"
#include "mdb_intr.h"
#include "message.h"
#include "puffer.h"
#include "screen/io.h"
#include "fehler.h"

extern CONFIG config;



/*
	Nachrichten in einer Gruppe als zu lschen markieren bzw. demarkieren

	->				name		:	Gruppenname
					pm			:	PM oder Gruppe
					modus		:	TRUE = kill , FALSE = unkill
					zeit		:	nach Zeit (TRUE) oder Nummer (FALSE) suchen
					...		:	Start- und Endzeit bzw. Start- und Endnummer

	<-				TRUE		:	ok
					FALSE		:	Fehler aufgetreten

	Die Indexdatei darf vom Aufrufer nicht geffnet sein.

	Der Unterschied zwischen Gruppen und PM besteht darin, da bei PMs
	nur diejenigen Nachrichten als zu lschen markiert werden, die bereits 
	seit lngerem (config.tausch_pm) getauscht sind.

	Wird als Modus unkill gewhlt, werden auch bereits aufgrund des expire-
	Datums lschmarkierte Nachrichten restauriert.
*/
BOOLEAN mdb_markiere(const char *name,BOOLEAN pm,BOOLEAN modus,BOOLEAN zeit,...)
{
	va_list zeiger;
	time_t von_zeit = 0;			/* keep gcc happy */
	time_t bis_zeit = 0;			/*  "    "    "   */
	time_t jetzt;
	ULONG von_nr = 0;				/* keep gcc happy */
	ULONG bis_nr = 0;				/*  "    "    "   */
	KENNUNG kennung;
	INDEX ix;
	char ind_name[PATH_MAX];
	ULONG nummer = 0;				/* Nummer der aktuellen Nachricht */
	ULONG markiert = 0;			/* Anzahl (de-)markierter Nachrichten */
	ULONG anzahl;

	/* variable Parameter, Zeit- oder Nummernangabe ermitteln */
	va_start(zeiger,zeit);
	if (zeit)
	{
		von_zeit = va_arg(zeiger,time_t);
		bis_zeit = va_arg(zeiger,time_t);
	}
	else
	{
		von_nr   = va_arg(zeiger,ULONG);
		bis_nr   = va_arg(zeiger,ULONG);
	}
	va_end(zeiger);

	if (group_path(name,ind_name,pm) < 0)
	{
		fehler(NO_GROUP);
		return FALSE;
	}
	strcat(ind_name,"ind");

	/* Indexdatei ffnen */
	if (!i_open(ind_name))
		return FALSE;

	/* Kennung einlesen */
	i_read(&kennung,sizeof(KENNUNG));
	anzahl = kennung.anzahl;

	/* ab einer bestimmten Nummer ? */
	if (!zeit && von_nr)
	{
		/* dann diese, falls vorhanden, direkt anspringen */
		if (von_nr < anzahl)
		{
			i_goto((MESSAGE) von_nr);
			anzahl -= von_nr;
			nummer  = von_nr;
		}
	}

	/* aktuellen Zeitpunkt fr expire-Angaben merken */
	time(&jetzt);

	if (zeit)
	{
		mprintf("markiere vom %.24s ",ctime(&von_zeit));
		mprintf("bis %.24s\r\n",ctime(&bis_zeit));
	}
	else
		mprintf("markiere von #%lu bis #%lu\r\n",von_nr,bis_nr);

	while (anzahl--)
	{
		if (!i_read(&ix,sizeof(INDEX)))
			break;

		/* sind keine weiteren Matches mglich ? */
		if (!zeit && (nummer > bis_nr))
			break;

		/* trifft die Bedingung zu ? */
		if ((zeit && ix.zeit >= von_zeit && ix.zeit <= bis_zeit)
		|| (!zeit && nummer >= von_nr && nummer <= bis_nr)
		|| (ix.expire && ix.expire <= jetzt))
		{
			/* Msg. geschtzt ? */
			if (ix.flags & GESCHUETZT)
				continue;

			/* bereits markiert ? */
			if ((modus &&  (ix.flags & TOTAL))
			|| (!modus && !(ix.flags & TOTAL)))
				continue;

			if (modus == TRUE)
				ix.flags |= TOTAL;				/* kill */
			else
				ix.flags &= ~TOTAL;				/* unkill */

			/* PMs nur lschmarkieren, wenn seit X Tagen getauscht */
			if (!pm || !modus || (pm && config.tausch_pm && ix.getauscht
			&& (jetzt - ix.getauscht > config.tausch_pm * 86400L)))
			{
				i_back();
				(void) i_write(&ix,sizeof(INDEX));
				i_seek(i_tell());
				markiert++;
			}
		}

		nummer++;
	}

	/* wurden Nachrichten (de-)markiert ? */
	if (markiert)
	{
#ifdef TEXT_FRONTEND
		mprintf("markiert: %lu\r\n",markiert);
#endif
		/* dann Kennung aktualisieren */
		if (modus)
			kennung.putzen += (MESSAGE) markiert;
		else
			kennung.putzen -= (MESSAGE) markiert;
		kennung.changed = time(NULL);
		i_rewind();
		i_write(&kennung,sizeof(KENNUNG));
	}
	i_rewind();

	/*
		wurde die maximale Messagezahl in der Gruppe berschritten ?
		entweder bezogen auf die gruppenspezifische Einstellung, sofern vor-
		handen, ansonsten die globale Einstellung
	*/
	if (modus == TRUE
	&& ((kennung.max_msgs && (kennung.anzahl - kennung.putzen > kennung.max_msgs))
	||	(!kennung.max_msgs && (kennung.anzahl - kennung.putzen > config.max_msgs) && config.max_msgs)))
	{
		ULONG kill;

		mputs("zu viele Messages in der Gruppe -> markiere berzhlige");

		if (kennung.max_msgs)
			kill = (kennung.anzahl - kennung.putzen) - kennung.max_msgs;
		else
			kill = (kennung.anzahl - kennung.putzen) - config.max_msgs;

		mprintf("%lu Messages werden markiert\r\n",kill);
		markiert = 0;

		i_read(&kennung,sizeof(KENNUNG));
		anzahl = kennung.anzahl;

		while (kill && anzahl--)
		{
			if (!i_read(&ix,sizeof(INDEX)))
				break;

			if (!(ix.flags & TOTAL) && !(ix.flags & GESCHUETZT))
			{
				ix.flags |= TOTAL;
				i_back();
				(void) i_write(&ix,sizeof(INDEX));
				markiert++;
				kill--;
			}
		}

		if (markiert)
		{
			/* Kennung aktualisieren */
			kennung.putzen += markiert;
			kennung.changed = time(NULL);
			i_rewind();
			i_write(&kennung,sizeof(KENNUNG));
		}
	}

	i_close(TRUE);

	return TRUE;

} /* mdb_markiere */

/*
 * Markiere als gelesen, bis hoechsten noch max_kb in der Gruppe sind.
 */
BOOLEAN 
mdb_markiere_size (const char *name,BOOLEAN pm,size_t max_size)
{
	long total_size;
	long total_zahl;
	KENNUNG kennung;
	INDEX ix;
	char ind_name[PATH_MAX];
	ULONG markiert = 0;			/* Anzahl markierter Nachrichten */
	long i;
	size_t killed=0;

	if (max_size==0)
		return FALSE;

	if (group_path(name,ind_name,pm) < 0)
	{
		fehler(NO_GROUP);
		return FALSE;
	}

	strcat(ind_name,"ind");

	if (!group_info_msgs(name,pm,&total_zahl,&total_size))
		return FALSE; /* Hm? */

	/* Indexdatei ffnen */
	if (!i_open(ind_name))
		return FALSE; /* Hm? */

	/* Kennung einlesen */
	i_read(&kennung,sizeof(KENNUNG));

	/* erste Schleife bestimmt, wieviel schon als zu loeschen markiert ist */
	for (i=0;i<total_zahl;i++)
	{
		if (!i_read(&ix,sizeof(INDEX)))
			break;
		if (ix.flags & TOTAL)
			killed+=ix.len;
	}
	/* zweite Schleife loescht soviel wie noetig */
	i_rewind();
	/* Kennung einlesen */
	i_read(&kennung,sizeof(KENNUNG));
	for (i=0;i<total_zahl;i++)
	{
		if (total_size-killed<max_size)
			break; /* Alles getan */
		if (!i_read(&ix,sizeof(INDEX)))
			break;
		if (ix.flags & TOTAL)
			continue;
		killed+=ix.len;
		ix.flags |= TOTAL;				/* kill */
		i_back();
		(void) i_write(&ix,sizeof(INDEX));
		i_seek(i_tell());
		markiert++;
	}

	/* wurden Nachrichten (de-)markiert ? */
	if (markiert)
	{
#ifdef TEXT_FRONTEND
		mprintf("markiert: %lu\r\n",markiert);
#endif
		/* dann Kennung aktualisieren */
		kennung.putzen += (MESSAGE) markiert;
		kennung.changed = time(NULL);
		i_rewind();
		i_write(&kennung,sizeof(KENNUNG));
	}
	i_close(TRUE);

	return TRUE;

} /* mdb_markiere_size */
