/* message.c : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
       (might be the wrong person except for offline)

    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
/*
	Offline - MausTausch-FrontEnd

	Nachrichten-Verwaltung/Anzeige
	(c) 1992-95 bei Frank Baumgart

	begonnen am 24.10.1992
	Version vom 22.01.1995

	transparente Schicht zwischen Datenbank und Oberflche
*/


#include "portab.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <limits.h>
#include <unistd.h>
#include <sys/types.h>

#include "system.h"
#include "mdb_conf.h"
#include "mdb_glob.h"
#include "struktur.h"
#include "mdb/mdb_intr.h"
#include "screen/io.h"
#include "tool.h"
#include "fehler.h"
#include "message.h"
#include "missing/missing.h"

extern CONFIG config;

static int req_handle(pid_t pid);
static BOOLEAN test_handle(int gh,pid_t pid);
static BOOLEAN free_handle(int gh,pid_t pid);

static void dat_lesefehler(int gh);
/*static*/ void msg_split_header(int gh);
static void msg_kill_chain(int gh);


static pid_t g_handle[MAX_BROWSER];		/* Gruppen-Handles */

WORK *msg[MAX_BROWSER];



/*
	Handle-Verwaltung
	nchstes freies Handle liefern und belegen

	->			pid	:	Proze ID des Anforderers

	<-			-1		:	kein freies Handle mehr
			 	>= 0	:	Gruppen-Handle
*/
static int req_handle(pid_t pid)
{
	int i;

	for (i = 0; i < MAX_BROWSER; i++)
		if (!g_handle[i])
		{
			if ((msg[i] = (WORK *) calloc(1,sizeof(WORK))) == NULL)
				return -1;
			g_handle[i] = pid;

			return i;
		}

	return -1;

} /* req_handle */



/*
	Handle-Verwaltung
	Handle auf Gltigkeit testen

	->			gh		:	Gruppen-Handle
				pid	:	Proze ID oder 0

	<-			TRUE	:	ok
				FALSE	:	Fehler, z.B. Handle nicht belegt oder falsche PID

	Wird als pid eine 0 bergeben, findet keine berprfung der pid statt
*/
static BOOLEAN test_handle(int gh,pid_t pid)
{
	if ((gh < 0) || (gh >= MAX_BROWSER) || (pid && !g_handle[gh]))
	{
		mputs("ungltiges Handle");
		return FALSE;
	}

	if (pid && (pid != g_handle[gh]))
	{
		mputs("falsches Handle");
		return FALSE;
	}

	return TRUE;

} /* test_handle */



/*
	Handle-Verwaltung
	Handle freigeben

	->			gh		:	Gruppen-Handle

	<-			TRUE	:	ok
				FALSE	:	Fehler, z.B. Handle nicht belegt
*/
static BOOLEAN free_handle(int gh,pid_t pid)
{
	if (!test_handle(gh,pid))
		return FALSE;

	free(msg[gh]);
	g_handle[gh] = 0;

	return TRUE;

} /* free_handle */



#if !defined(CLIENT) || 1	/* ### s. Anmerkung zu group_path() im Client */
/*
	Gruppe ffnen

	->			pid		:	Proze ID des Aufrufers
				name		:	Name der Gruppe/User
				pm			:	PM (TRUE) oder Gruppe (FALSE)
				rw			:	nur lesen (FALSE) oder lesen und schreiben (TRUE)

	<-			-1			:	Fehler aufgetreten
				>= 0		:	Gruppen-Handle

	Im Boxmodus, insb. MultiPort, mu streng darauf geachtet werden, wann die 
	DB wirklich (im interaktiven Betrieb oder bei Befehlen wie z.B. KILL) 
	verndert werden mu. Sollte bereits ein anderer Proze Schreibzugriff auf 
	die gewnschte Gruppe besitzen, wird sie trotzdem geffnet, aber ohne 
	Schreibzugriff. (still to do)
	Schreibzugriffe globaler DB-Funktionen wie Import, Check oder Putz werden 
	derzeit noch gar nicht weiter bercksichtigt. Falls eine Gruppe bereits 
	mit Schreibzugriff geffnet ist und z.B. ein Import gestartet wird, mu 
	vor der Durchfhrung des Imports dem Message-Modul der Schreibzugriff 
	entzogen und spter wiedergegeben werden. Whrend der Abarbeitung globaler 
	Funktionen ist ja auch kein weiterer Zugriff via Message-Modul mglich.
	(bedingt durch die Serialisierung, gilt aber erst, wenn das message-Modul 
	ebenfalls via TCP/IP statt NFS luft), so da keine weitergehenden Synchro-
	nisationsmechanismen notwendig sind. (abschnittweises file-locking etc.)
*/
int group_open(pid_t pid,const char *name,BOOLEAN pm,BOOLEAN rw)
{
	char dat_name[PATH_MAX];
	char ind_name[PATH_MAX];
	int gh;
	WORK *mp;

	/* Pfade von Index- und Datendatei ermitteln */
	if (group_path(name,dat_name,pm) < 0)
		return -1;
	strcpy(ind_name,dat_name);
	strcat(ind_name,"ind");
	strcat(dat_name,"dat");

	if ((gh = req_handle(pid)) < 0)
		return -1;

	mp = msg[gh];


	if (rw)
		group_lock(name,pm,TRUE);

	/*
		zwei Puffer lokal einrichten: Datenpuffer und Anzeige(Umbruch)Puffer
		eigentlich ist der Anzeigepuffer Angelegenheit des clients
		clever wre es, in der KENNUNG die Gre der grten Nachricht zu 
		merken und einen Puffer von exakt dieser Gre anzufordern.
		Allerdings geht dann der Zugriff auf nach dem ffnen importierte, 
		grere Nachrichten nicht.
	*/
	if ((mp->puffer = (char *) malloc(MSG_SIZE)) == NULL)
	{
		free_handle(gh,pid);
		return -1;
	}
	mp->puffer_size = MSG_SIZE;

	if ((mp->anzeige = (char *) malloc(MSG_SIZE)) == NULL)
	{
		free(mp->puffer);
		free_handle(gh,pid);
		return -1;
	}

	/* Indexdatei zum Lesen (und ggf. Schreiben) ffnen */
#if 0		/* vgl. Bemerkung zum ffnen der Datendatei */
	if ((mp->ind_f = fopen(ind_name,(rw) ? BIN_RW_OPEN : BINREAD)) == NULL)
#else
	if ((mp->ind_f = fopen(ind_name,BIN_RW_OPEN)) == NULL)
#endif
	{
		free(mp->anzeige);
		free(mp->puffer);
		free_handle(gh,pid);
		fehler(IND_OPEN);
		return -1;
	}
	/* Indexdatei berhaupt gltig ? */
	if (filesize(mp->ind_f) < sizeof(KENNUNG))
	{
		free(mp->anzeige);
		free(mp->puffer);
		fclose(mp->ind_f);
		free_handle(gh,pid);
		fehler(ILL_IND);
		return -1;
	}

	/* Indexdatei puffern */
	fseek(mp->ind_f,0,SEEK_CUR);
	pufferung(mp->ind_f,NULL,_IOFBF,2 * sizeof(INDEX));
/*	setbuf(mp->ind_f,NULL);*/

	/*
		Datendatei zum Lesen und Schreiben ffnen
		Schreibmodus deshalb, damit z.B. der import dieselbe Gruppe zum 
		Schreiben ffnen kann und kein "access denied" bekommt
	*/
#if 0
	if ((mp->dat_f = fopen(dat_name,(rw) ? BIN_RW_OPEN : BINREAD)) == NULL)
#else
	if ((mp->dat_f = fopen(dat_name,BIN_RW_OPEN)) == NULL)
#endif
	{
		free(mp->anzeige);
		free(mp->puffer);
		fclose(mp->ind_f);
		free_handle(gh,pid);
		fehler(DAT_OPEN);
		return -1;
	}
	/* Datendatei "leicht" puffern */
	pufferung(mp->dat_f,NULL,_IOFBF,4096);

	/* Header aktualisieren */
	lies_kennung(mp->ind_f,&mp->kennung);

	/*
		optional Pufferung fr Stichwortliste einrichten.
		Dabei wird erstmal nur Speicher fr die Zeiger reserviert und diese 
		bei Bedarf gefllt.
		Solange nur eine Gruppe gleichzeitig bearbeitet werden kann, ist dies 
		auch auf Systemen ohne virt. Speicher wie TOS und DOS gnstiger.
		Auerdem entsprche ein 64 KB-Block unter DOS max. 819 Msgs./Gruppe.
		Allerdings werden je Msg. i.A. 4 Bytes fr den Zeiger fest reserviert.
	*/
#if !defined(BOX)		/* nur, solange keine Inhaltsdarstellung */
	if (config.browser_cache)
	{
		mp->ic = (char **) calloc((size_t) mp->kennung.anzahl,sizeof(char *));
		mp->ic_free = config.browser_cache;
	}
	else
#endif
	{
		mp->ic = NULL;
		mp->ic_free = 0;
	}

	/* ### prfen, ob bereits jemand anders Schreibzugriff besitzt */
	mp->su = rw;
	mp->pm = pm;
	mp->ursprung = LEER;
	mp->offen = TRUE;
	mp->inhalt = FALSE;

	return gh;

} /* group_open */



/*
	Kennung einer Gruppe aktualisieren
*/
BOOLEAN group_flush(int gh,pid_t pid)
{
	WORK *mp = msg[gh];

	if (!test_handle(gh,pid))
		return FALSE;

	/* Kennung aktualisieren: Zeitpunkt der letzten Bearbeitung */
	mp->kennung.latest = time(NULL);

	if (mp->su)
	{
		long save;

		save = ftell(mp->ind_f);
		rewind(mp->ind_f);
		fwrite(&mp->kennung,sizeof(KENNUNG),1,mp->ind_f);
		fseek(mp->ind_f,save,SEEK_SET);
	}

	return TRUE;

} /* group_flush */



/*
	Gruppe schlieen

	->			gh		:	Gruppen-Handle
				pid	:	Proze ID des Aufrufers

	<-			TRUE	:	ok
				FALSE	:	Fehler aufgetreten
*/
BOOLEAN group_close(int gh,pid_t pid)
{
	MESSAGE i;
	WORK *mp = msg[gh];

	if (!test_handle(gh,pid))
		return FALSE;

	group_flush(gh,pid);

	/* Cachezeilen fr Stichwortliste freigeben */
	if (mp->ic)
	{
		for (i = 0; i < mp->kennung.anzahl; i++)
			if (mp->ic[i])
			{
				free(mp->ic[i]);
				mp->ic[i] = NULL;
			}
		free(mp->ic);
	}

	fclose(mp->ind_f);
	fclose(mp->dat_f);
	mp->offen = FALSE;
	free(mp->anzeige);
	free(mp->puffer);

	if (mp->su)
		group_unlock(mp->kennung.name,mp->kennung.pm);

	free_handle(gh,pid);

	return TRUE;

} /* group_close */
#endif



/*
	Kennung der aktuellen Gruppe liefern
*/
BOOLEAN index_read_kennung(int gh,KENNUNG *kennung)
{
	if (!test_handle(gh,0))
		return FALSE;

	*kennung = msg[gh]->kennung;

	return TRUE;

} /* index_read_kennung */


/*
	Kennung der aktuellen Gruppe schreiben
*/
BOOLEAN index_write_kennung(int gh,KENNUNG *kennung)
{
	if (!test_handle(gh,0))
		return FALSE;

	msg[gh]->kennung = *kennung;

	return TRUE;

} /* index_write_kennung */


static void dat_lesefehler(int gh)
{
#ifdef TEXT_FRONTEND
	WORK *mp = msg[gh];
#endif

	fehler(READ_ERR);

	if (!test_handle(gh,0))
		return;
#ifdef TEXT_FRONTEND
	fprintf(stderr,"index.pos = %ld\n",(long) mp->ix.pos);
	fprintf(stderr,"index.len = %ld\n",(long) mp->ix.len);
	fprintf(stderr,"dat-size  = %ld\n",filesize(mp->dat_f));
	fprintf(stderr,"ind-size  = %ld\n",filesize(mp->ind_f));
	taste();
#endif

} /* dat_lesefehler */



/*
	Nummer der aktuellen Message ermitteln
*/
MESSAGE msg_get_nr(int gh)
{
	if (!test_handle(gh,0))
		return FALSE;

	fflush(msg[gh]->ind_f);
	return (MESSAGE)((ftell(msg[gh]->ind_f)-sizeof(KENNUNG)) / sizeof(INDEX));

} /* msg_get_nr */



/*
	zum Vorgnger

	->			gh	:	Gruppen-Handle

	bei Erfolg werden aktuelle Indexstruktur sowie puffer-Inhalt verndert.
*/
BOOLEAN msg_prev(int gh)
{
	if (!test_handle(gh,0))
		return FALSE;

	/* gibt es einen Vorgnger ? */
	if (msg[gh]->ix.links == LEER)
		return FALSE;

	if (index_goto(gh,msg[gh]->ix.links))
		return msg_read(gh);
	else
		return FALSE;

} /* msg_prev */



/*
	zum Nachfolger

	->			gh	:	Gruppen-Handle

	bei Erfolg werden aktuelle Indexstruktur sowie puffer-Inhalt verndert.
*/
BOOLEAN msg_next(int gh)
{
	if (!test_handle(gh,0))
		return FALSE;

	/* gibt es einen Nachfolger ? */
	if (msg[gh]->ix.rechts == LEER)
		return FALSE;

	if (index_goto(gh,msg[gh]->ix.rechts))
		return msg_read(gh);
	else
		return FALSE;

} /* msg_next */



/*
	zur kommentierten Nachricht

	->			gh	:	Gruppen-Handle

	bei Erfolg werden aktuelle Indexstruktur sowie puffer-Inhalt verndert.
*/
BOOLEAN msg_up(int gh)
{
	if (!test_handle(gh,0))
		return FALSE;

	/* gibt es eine kommentierte Nachricht ? */
	if (msg[gh]->ix.oben == LEER)
		return FALSE;

	if (index_goto(gh,msg[gh]->ix.oben))
		return msg_read(gh);
	else
		return FALSE;

} /* msg_up */



/*
	zum Kommentar

	->			gh	:	Gruppen-Handle

	bei Erfolg werden aktuelle Indexstruktur sowie puffer-Inhalt verndert.
*/
BOOLEAN msg_down(int gh)
{
	if (!test_handle(gh,0))
		return FALSE;

	/* gibt es einen Kommentar ? */
	if (msg[gh]->ix.unten == LEER)
		return FALSE;

	if (index_goto(gh,msg[gh]->ix.unten))
		return msg_read(gh);
	else
		return FALSE;

} /* msg_down */



/*
	Inhalt einer Inhaltscachezeile updaten
*/
static void msg_string_update(int gh,MESSAGE nr,ULONG flags)
{
	WORK *mp = msg[gh];
	int c;

	if (!mp->ic || !mp->ic[nr])
		return;

	if (flags & (TOTAL | TEIL))
		c = 'D';
	else
	if (flags & GESCHUETZT)
		c = '*';
	else
	if (mp->ix.getauscht)
		c = 'T';
	else
		c = ' ';

	mp->ic[nr][0] = c;

} /* msg_string_update */



/*
	Flag(s) setzen

	->			gh		:	Gruppen-Handle
				flags	:	zu setzende Flags, verodert

	<-			TRUE	:	Flag(s) gesetzt
				FALSE	:	Fehler beim Schreiben oder Flag(s) waren bereits gesetzt

	Indexstruktur wird im RAM und in der Datei verndert.
*/
BOOLEAN msg_set_flag(int gh,ULONG flags)
{
	WORK *mp = msg[gh];
	ULONG old;

	if (!test_handle(gh,0))
		return FALSE;

	old = mp->ix.flags;

	if ((flags & GELESEN) && !(old & GELESEN))
		mp->kennung.neue--;

	if ((flags & TOTAL) && !(old & TOTAL) && !(old & GESCHUETZT))
		mp->kennung.putzen++;
	else
		flags &= ~TOTAL;

	if ((old | flags) != old)
	{
		mp->ix.flags |= flags;
		msg_string_update(gh,msg_get_nr(gh),mp->ix.flags);

		return index_write(gh);
	}

	return FALSE;

} /* msg_set_flag */



/*
	Flag(s) lschen

	->			gh		:	Gruppen-Handle
				flags	:	zu lschende Flags, verodert

	Indexstruktur wird im RAM und in der Datei verndert.
*/
BOOLEAN msg_clr_flag(int gh,ULONG flags)
{
	WORK *mp = msg[gh];
	ULONG old;

	if (!test_handle(gh,0))
		return FALSE;

	old = mp->ix.flags;

	if ((flags & GELESEN) && (old & GELESEN))
		mp->kennung.neue++;
	if ((flags & TOTAL) && (old & TOTAL))
		mp->kennung.putzen--;

	if ((old & ~flags) != old)
	{
		mp->ix.flags &= ~flags;
		msg_string_update(gh,msg_get_nr(gh),mp->ix.flags);

		return index_write(gh);
	}

	return FALSE;

} /* msg_clr_flag */



/*
	Flags ermitteln

	->			gh	:	Gruppen-Handle

	<-			flags aus der Indexstruktur
*/
ULONG msg_get_flag(int gh)
{
	if (!test_handle(gh,0))
		return 0;

	return msg[gh]->ix.flags;

} /* msg_get_flag */



/*
	Messagestatus ermitteln

	->			gh	:	Gruppen-Handle

	<-			status aus der Indexstruktur
*/
UWORD msg_get_status(int gh)
{
	if (!test_handle(gh,0))
		return 0;

	return msg[gh]->ix.status & 0x7fff;

} /* msg_get_status */



/*
	Messagestatus setzen

	->			gh			:	Gruppen-Handle
				status	:	neuer Messagestatus

	Diese Funktion erzeugt auch das Flag, das anzeigt, ob sich der 
	Messagestatus verndert hat.
	Spezialfall: wenn status gleich GEAENDERT ist, wird das gendert-
	Flag zurckgesetzt (gelscht)
*/
void msg_set_status(int gh,UWORD status)
{
	WORK *mp = msg[gh];

	if (!test_handle(gh,0))
		return;

	/* GEAENDERT-Status lschen, falls gesetzt */
	if ((mp->ix.status & GEAENDERT) && (status == GEAENDERT))
	{
		mp->ix.status &= ~GEAENDERT;
		index_write(gh);
		mp->kennung.status_neu--;
	}
	else
	if ((mp->ix.status & 0x7fff) != status)
	{
		mp->ix.status = status | GEAENDERT;
		mp->ix.bearbeit = time(NULL);
		index_write(gh);
		mp->kennung.status_neu++;
	}

} /* msg_set_status */



/*
	aktuelle Indexposition merken
	nicht verschachtelbar

	Zweck: Aufrufer mu sich nicht um aktuelle Indexposition und fseek kmmern
*/
BOOLEAN index_push(int gh)
{
	if (!test_handle(gh,0))
		return FALSE;

	msg[gh]->save_nr = msg_get_nr(gh);

	return TRUE;

} /* index_push */



/*
	gemerkte Indexposition anfahren
	nicht verschachtelbar
*/
BOOLEAN index_pop(int gh)
{
	if (!test_handle(gh,0))
		return FALSE;

	index_goto(gh,msg[gh]->save_nr);

	return TRUE;

} /* index_pop */



/*
	aktuellen Index einlesen

	aktuelle Dateiposition bleibt unverndert
*/
BOOLEAN index_read(int gh)
{
	if (!test_handle(gh,0))
		return FALSE;

	if (fread(&msg[gh]->ix,sizeof(INDEX),1,msg[gh]->ind_f))
	{
		/* ### workaround fr alten Import-Bug: Eingabezeit wurde nicht gesetzt */
		if (!msg[gh]->ix.eingabe)
			msg[gh]->ix.eingabe = msg[gh]->ix.zeit;

		return index_prev(gh);
	}
	else
		return FALSE;

} /* index_read */



/*
	aktuellen Index schreiben

	<-			TRUE	:	ok
				FALSE	:	kein Schreibzugriff erlaubt oder Schreiben fehlerhaft

	aktuelle Dateiposition bleibt unverndert
*/
BOOLEAN index_write(int gh)
{
	if (!test_handle(gh,0))
		return FALSE;

	if (!msg[gh]->su)
		return FALSE;

	/* vor und nach jedem Schreibzugriff read/write Wechsel */
	fseek(msg[gh]->ind_f,0,SEEK_CUR);
	if (fwrite(&msg[gh]->ix,sizeof(INDEX),1,msg[gh]->ind_f))
		return index_prev(gh);
	else
		return FALSE;

} /* index_read */



/*
	zum vorherigen Index
*/
BOOLEAN index_prev(int gh)
{
	if (!test_handle(gh,0))
		return FALSE;

	fflush(msg[gh]->ind_f);
	if (ftell(msg[gh]->ind_f) > sizeof(KENNUNG))
	{
		if (fseek(msg[gh]->ind_f, - (long)(sizeof(INDEX)),SEEK_CUR))
			return FALSE;
		else
			return TRUE;
	}
	else
		return FALSE;

} /* index_prev */



/*
	zum nchsten Index
*/
BOOLEAN index_next(int gh)
{
	if (!test_handle(gh,0))
		return FALSE;

	if((msg_get_nr(gh) >= msg[gh]->kennung.anzahl)
	|| feof(msg[gh]->ind_f)
	|| fseek(msg[gh]->ind_f,(long)(sizeof(INDEX)),SEEK_CUR))
		return FALSE;
	else
		return TRUE;

} /* index_next */



/*
	Index der kommentierten Nachricht anfahren
*/
BOOLEAN index_up(int gh)
{
	if (!test_handle(gh,0))
		return FALSE;

	/* gibt es eine kommentierte Nachricht ? */
	if (msg[gh]->ix.oben == LEER)
		return FALSE;

	if (!index_goto(gh,msg[gh]->ix.oben))
		return FALSE;
	else
		return TRUE;

} /* index_up */



/*
	Indexposition anfahren

	->			gh	:	Gruppen-Handle
				i	:	Nummer der Message
*/
BOOLEAN index_goto(int gh,MESSAGE i)
{
	WORK *mp = msg[gh];
	long pos;

	if (!test_handle(gh,0))
		return FALSE;

	pos = sizeof(KENNUNG) + (long)(i) * sizeof(INDEX);

	fflush(mp->ind_f);
	if (ftell(mp->ind_f) == pos)
		return TRUE;
	else
	if (i >= mp->kennung.anzahl)
		return FALSE;
	else
		return !fseek(mp->ind_f,pos,SEEK_SET);

} /* index_goto */



/*
	Indexposition der nchsten Nachricht ab Referenzzeitpunkt anfahren
	Suche beginnt bei erster Mail der Gruppe

	->			gh		:	Gruppen-Handle
				ref	:	Referenzzeitpunkt

	<-			TRUE	:	ok
				FALSE	:	es ex. keine neuere Nachricht bzgl. ref

	Sowohl INDEX als auch Position in der Indexdatei werden verndert
*/
BOOLEAN index_goto_new(int gh,time_t ref)
{
	int i = 0;
	MESSAGE start = 0;	/* Position der erstmglichen neuen Nachricht */
	time_t *impref;
	WORK *mp = msg[gh];

	if (!test_handle(gh,0))
		return FALSE;

	/* letzte nderung vor Referenzzeitpunkt ? */
	if (mp->kennung.changed < ref)
		return FALSE;

	/* Positions-Informationen in Kennung ausnutzen */
	impref = mp->kennung.imported;
	while ((i < 10) && (*impref && *impref++ < ref))
		start = mp->kennung.imp_anf[i++];

	/*
		Position der frhestmglichen neuen Nachricht anfahren.
		Das bedeutet, da VOR dieser Nachricht mit Sicherheit keine liegen, 
		die neuer sind als das Export-Referenzdatum. Es mu sich aber nicht 
		zwangslufig um die erste neue Nachricht handeln.
	*/
	if (!index_goto(gh,start))
		return FALSE;

	do
	{
		if (!index_read(gh))
			return FALSE;

		if (mp->ix.zeit < ref)
		{
			if (!index_next(gh))
				return FALSE;
		}
		else
			return TRUE;
	}
	while (1);

} /* index_goto_new */



/*
	Ursprungsnachricht eines Verkettungsbaumes einlesen

	->			gh		:	Gruppen-Handle

	<-			TRUE	:	ok
				FALSE	:	keine kommentierte Nachricht bzw. Fehler

	Bei Erfolg wird die Indexstruktur im RAM verndert
*/
BOOLEAN index_goto_top(int gh)
{
	if (!test_handle(gh,0))
		return FALSE;

	if (!index_up(gh))
		return FALSE;

	while (index_up(gh))
		index_read(gh);

	return TRUE;

} /* index_goto_top */



/*
	Headerzeilen fr Autor, Betreff und ID trennen
*/
/*static*/ void msg_split_header(int gh)
{
	WORK *mp = msg[gh];
	char *p;

	if (!test_handle(gh,0))
		return;

	/* Offsets initialisieren */
	mp->offset.group  = mp->kennung.name;
	mp->offset.id     = mp->puffer + 1;
	mp->offset.author = mp->puffer + mp->ix.t_author;
	mp->offset.topic  = mp->puffer + mp->ix.t_topic;
	mp->offset.text   = mp->puffer + mp->ix.t_text;
	mp->offset.name   = strstr(mp->puffer,"\r\nN");		/* ### fails 2nd call */
	if (mp->offset.name)
	{
		mp->offset.name += 3;
		if ((p = strpbrk(mp->offset.name,"\r\n")) != NULL)
			*p = 0;
	}
	if (mp->offset.id)
		if ((p = strpbrk(mp->offset.id,"\r\n")) != NULL)
			*p = 0;
	if (mp->offset.author)
		if ((p = strpbrk(mp->offset.author,"\r\n")) != NULL)
			*p = 0;
	if (mp->offset.topic)
		if ((p = strpbrk(mp->offset.topic,"\r\n")) != NULL)
			*p = 0;

} /* msg_split_header */



/*
	Nachricht einlesen

	->			gh		:	Gruppen-Handle

	<-			TRUE	:	ok
				FALSE	:	keine Msg. bzw. Fehler

	Sollte die einzulesende Nachricht grer als der zur Verfgung stehende 
	Speicher sein, wird sie ggf. nur teilweise eingelesen und dennoch TRUE 
	zurckgeliefert.
	Bei XP-Referenzen wird temporr die Gruppe, die die Nachricht mit dem 
	body enthlt, geffnet und nach der ID durchsucht.
*/
BOOLEAN msg_read(int gh)
{
	WORK *mp = msg[gh];

	if (!test_handle(gh,0))
		return FALSE;

	/* Indexdatensatz einlesen */
	if (!index_read(gh))
		return FALSE;

	/* Nachricht grer als bisheriger Puffer? */
	if (mp->ix.len > mp->puffer_size)
	{
		char *tmp;

		tmp = (char *) realloc(mp->puffer,mp->ix.len +1);
		if (!tmp)
			fehler(NO_MEMORY);
		else
		{
			mp->puffer = tmp;
			mp->puffer_size = mp->ix.len +1;
		}
	}
	/*
		es scheint, als fehle hier die evtl. ebenfalls notwendige 
		Erweiterung des Anzeige-Puffers. Dieser ist aber allein 
		Angelegenheit des clients! (jedenfalls spter)
	*/

	/* Datensatzposition anfahren */
	fflush(mp->ind_f);
	if (ftell(mp->dat_f) != mp->ix.pos
	&&  fseek(mp->dat_f,mp->ix.pos,SEEK_SET))
	{
		perror("msg_read");
		dat_lesefehler(gh);
		return FALSE;
	}

	/* zwischen vollst. Nachricht und XP-Referenz unterscheiden */
	if (mp->ix.flags & XP_REF)
	{
		/* Name der den body enthaltende Gruppe */
		char name[MAX_CPG];
		int xh;

		puts("msg_read: XP-body wird abgerufen");
#if 0
		GLOBAL_REF ref;
		if (!msgid_find(id,&ref))
			return FALSE;
		if (!gruppen_nr2name(ref.grp_nr,mp->pm,name))
			return FALSE;
		if ((xh = group_open(getpid(),name,mp->pm,FALSE)) < 0)
			return FALSE;
		if (!index_goto(xh,ref.msg_nr) || !msg_read(xh))
		{
			group_close(getpid(),xh);
			return FALSE;
		}
puts("kopiere header/body");
		strcpy(mp->puffer,msg[xh].puffer);
		group_close(getpid(),xh);
#endif
taste();
		return FALSE;
	}
	else
	if (!fread(mp->puffer,mp->ix.len,1,mp->dat_f))
	{
		perror("msg_read");
		dat_lesefehler(gh);
		return FALSE;
	}

	/* handelt es sich um eine neue Nachricht ? */
	if (!(mp->ix.flags & GELESEN))
#ifdef BOX
		if (mp->pm)
#endif
		/* gelesene Nachrichten markieren */
		msg_set_flag(gh,GELESEN);

	/* Ende des Messagestrings setzen */
	mp->puffer[mp->ix.len] = 0;

	mp->offset.text = mp->puffer + mp->ix.t_text;

	/* kodierte Nachricht (PM) dekodieren */
	if (mp->pm && mp->ix.flags & CODIERT)
		decode((UBYTE *)mp->offset.text,0,0,mp->ix.len - mp->ix.t_text);

	return TRUE;

} /* msg_read */



/*
	Header einer Nachricht einlesen

	->			gh		:	Gruppen-Handle

	<-			TRUE	:	ok
				FALSE	:	keine Msg. bzw. Fehler

	- wie msg_read, aber aus Geschwindigkeitsgrnden wird nur der 
	  Header eingelesen, ntzlich bei Inhaltsanzeige
*/
BOOLEAN msg_read_header(int gh)
{
	WORK *mp = msg[gh];

	if (!test_handle(gh,0))
		return FALSE;

	/* Indexdatensatz einlesen */
	if (!index_read(gh))
		return FALSE;

	/* Datensatz einlesen */
	fflush(mp->ind_f);
	if (mp->ix.flags & XP_REF)
	{
		puts("msg_read_header: XP-header");
		return FALSE;
	}
	else
	{
		if ((ftell(mp->dat_f) != mp->ix.pos)
		&&  (fseek(mp->dat_f,mp->ix.pos,SEEK_SET)))
			return FALSE;
		if (!fread(mp->puffer,mp->ix.t_text,1,mp->dat_f))
		{
			perror("msg_read_header");
			dat_lesefehler(gh);
			return FALSE;
		}
	}

	/* Ende des Headerstrings setzen */
	mp->puffer[mp->ix.t_text] = 0;

	return TRUE;

} /* msg_read_header */



/*
	Index der aktuellen Nachricht als String zurckliefern

	->			gh		:	Gruppen-Handle
				maske	:	Maske, die auf Autor bzw. Zeile passen mu bzw. NULL
				len	:	Spaltenzahl (60 - 200)
			<-	s		:	Rckgabestring, Format s.u.

	<-			TRUE	:	ok
				FALSE	:	keine (gem maske passende) Msg. bzw. allg. Fehler

	Der Rckgabestring ist fr Inhaltsbersichten gedacht und sieht z.B. so aus:

	"  607 <> 01.07.94  Frank Baumgart @ PB     (  7) Kodierung von PWs im Mausnet"

	Diese Funktion liest selbst die erforderlichen Headerinformationen ein.
	maske darf keine Wildcards enthalten und mu in Kleinschrift vorliegen, 
	um die case-insensitiven Vergleiche mit Autor und Thema zu verschnellern.
	Der Vergleich bezieht den vollstndigen Autornamen bzw. Thema mit ein, 
	nicht nur den im Resultatstring sichtbaren Teil.
*/
BOOLEAN msg_index_string(int gh,const char *maske,int len,char *s)
{
	WORK *mp = msg[gh];
	MESSAGE nr;					/* akt. Messagenummer */
	char datum[12];
	unsigned char *t;			/* unsigned wg. Umlauten */
	int rest;
	int c;

	if (!test_handle(gh,0))
		return FALSE;

	nr = msg_get_nr(gh);

	if (!maske && mp->ic && mp->ic[nr])
	{
		strcpy(s,mp->ic[nr]);
		return TRUE;
	}

	/* Messageheader einlesen */
	msg_read_header(gh);
	msg_split_header(gh);

	/* ggf. auf Maske prfen */
	if (maske)
	{
		char tmp[MAX_HDR];

		strcpy(tmp,mp->offset.author);
		strlwr(tmp);
		if (!strstr(tmp,maske))
		{
			strcpy(tmp,mp->offset.topic);
			strlwr(tmp);
			if (!strstr(tmp,maske))
				return FALSE;
		}
	}

	if (len < 60)
		len = 60;
	else
	if (len > 200)
		len = 200;

	strftime(datum,11,"%d.%m.%y",localtime(&mp->ix.eingabe));

	if (mp->ix.flags & TOTAL)
		c = 'D';
	else
/*	if (mp->ix.flags & TEIL)
		c = 'T';
	else*/
	if (mp->ix.flags & GESCHUETZT)
		c = '*';
	else
	if (mp->ix.flags & GELESEN)
		c = 'G';
	else
	if (mp->ix.status & ZURUECKGESTELLT)
		c = 'Z';
	else
	if (mp->ix.status & BEANTWORTET)
		c = 'B';
	else
	if (mp->ix.status & WEITERGEGEBEN)
		c = 'W';
	else
	if (mp->ix.status & MAUSNET)
		c = 'M';
	else
	if (mp->ix.status & ANGEKOMMEN)
		c = 'A';
	else
	if (mp->ix.status & GATEWAY)
		c = 'Y';
	else
	if (mp->ix.status & KOPIERT)
		c = 'K';
	else
	if (mp->ix.status & GETAUSCHT)
		c = 'T';
	else
	if (mp->ix.status & GELESEN)
		c = 'G';
	else
		c = ' ';

	/* auf Autor (2/5) und Betreff aufzuteilen (3/5) */
	rest = len - 28;

	sprintf(s,"%c %4ld %c%c %8s %c%-*.*s (%3u) %-*.*s",
	        c,
	        (long) nr+1,
	        (mp->ix.oben  != LEER || mp->ix.links  != LEER) ? '<' : ' ',
	        (mp->ix.unten != LEER || mp->ix.rechts != LEER) ? '>' : ' ',
	        datum,
	        (mp->pm && mp->ix.flags & YOURMAIL) ? '>' : ' ',
	        (rest*2)/5,(rest*2)/5,
	        (mp->offset.name) ? mp->offset.name : mp->offset.author,
	        (unsigned)mp->ix.zeilen,
	        (rest*3)/5,(rest*3)/5,
	        mp->offset.topic);

	/* es gibt Idioten, die Cntrl-Zeichen im Betreff unterbringen -> filtern */
	t = (unsigned char *) s;
	while (*t)
	{
#if 0
		if (iscntrl(*t))			/* leider von LC_CTYPE/LANG abhngig */
#else
		if (*t && *t < ' ')
#endif
			*t = ' ';
		t++;
	}

	if (mp->ic && mp->ic_free && !mp->ic[nr])
	{
		if ((mp->ic[nr] = (char *) malloc((size_t)len)) != NULL)
		{
			strncpy(mp->ic[nr],s,(size_t)(len-1));
			mp->ic[nr][len-1] = '\0';
			mp->ic_free--;
		}
	}

	return TRUE;

} /* msg_index_string */



/*
	Index der angegebenen Nachricht als String zurckliefern

	->			gh		:	Gruppen-Handle
				nr		:	Nummer der Mail (ab 0)
				len	:	Spaltenzahl
			<-	s		:	Rckgabestring, Format s.u.

	<-			TRUE	:	ok
				FALSE	:	Fehler

	Rckgabestring s. msg_index_string()
	Falls der String bereits gepuffert ist, wird len nicht beachtet.
*/
BOOLEAN msg_index_string_nr(int gh,MESSAGE nr,int len,char *s)
{
	WORK *mp = msg[gh];
	BOOLEAN ok;
	MESSAGE old;

	if (!test_handle(gh,0))
		return FALSE;

	nr = msg_get_nr(gh);

	if (mp->ic && mp->ic[nr])
	{
		strcpy(s,mp->ic[nr]);
		return TRUE;
	}

	old = msg_get_nr(gh);
	if (!index_goto(gh,nr))
		return FALSE;
	ok = msg_index_string(gh,NULL,len,s);
	index_goto(gh,old);
	msg_read(gh);

	return ok;

} /* msg_index_string_nr */



/*
	nchste den Bedingungen entsprechende Nachricht einlesen

	->			gh		:	Gruppenhandle
				skip	:	Flags, deren Nachrichten berlesen werden sollen

	<-			TRUE	:	ok
				FALSE	:	Fehler bzw. keine neue/passende Nachricht
*/
BOOLEAN msg_next_new(int gh,ULONG skip_flags)
{
	WORK *mp = msg[gh];
	MESSAGE save;

	if (!test_handle(gh,0))
		return FALSE;

	save = msg_get_nr(gh);

	/* Index einlesen */
	while (1)
	{
		if (!index_next(gh) || !index_read(gh))
		{
			index_goto(gh,save);
			index_read(gh);
			msg_read(gh);

			return FALSE;
		}

		/* gelesene/(teil)zulschende Nachrichten berspringen ? */
#if defined(BOX)
		if (mp->ix.flags & TOTAL)
#else
		if (mp->ix.flags & skip_flags)
#endif
			continue;

		/* Nr. dieser Mail als Ursprungsnachricht merken */
		mp->ursprung = (MESSAGE) ((ftell(mp->ind_f)
							- sizeof(KENNUNG)) / sizeof(INDEX));

		return msg_read(gh);
	}
} /* msg_next_new */



/*
	Messagetext umbrechen

	->			quelle	:	umzubrechender Text, mit \0 abgeschlossen
				ziel		:	Ziel fr umgebrochene Textzeilen, jeweils mit \0 abg.
				spalten	:	Anzahl Textspalten (minimal 40)

	<-			Anzahl der Zeilen NACH dem Umbruch

	Merkmale:
	- durch den Umbruch stimmt die Anzahl der resultierenden Zeilen nicht 
	  zwangslufig mit dem Index-Eintrag (Ausgangszeilen) berein.
	- es kann nicht in sito (am Platze) umgebrochen werden, daher wird ein 
	  weiterer Puffer bentigt.
	- der Vorteil besteht darin, den Originaltext jederzeit neu umbrechen zu 
	  knnen (z.B. bei Vernderung der Fensterbreite) ohne dazu die Nachricht 
	  neu einlesen zu mssen.
	- Text-Attribute werden auf dieser Ebene nicht beachtet

	todo:
	Position der letzten Tearline merken, um nicht mitten im Message-
	text abzuschneiden, falls irgendein DAU dort Tearlines verwendet.

	Client(!)-Funktion
*/
long msg_umbruch(const char *quelle,char *ziel,int spalten)
{
#define WORTTRENNER " ,.\t;-:"

	long zeilen = 0;

	if (quelle == ziel)
		return 0;

	while (*quelle)
	{
		const char *p,*last;
		size_t len,l_rest;
		size_t l_quote;
		char quote[20];

		if (*quelle == ':' || *quelle == '>')
			quelle++;

		if (!config.show_footer
		&& (*quelle == '-')
		&& (!strncmp(quelle,"---\r\n",5) || !strncmp(quelle,"--\r\n",4)))
			return zeilen;

		if (!quelle[0] || !quelle[1])
			break;

		len = strcspn(quelle,"\r\n");

		/* Leerzeile ? */
		if (!len)
		{
			*ziel++ = 0;
			quelle++;
			if (*quelle == '\r' || *quelle == '\n')
				quelle++;
			zeilen++;
			continue;
		}

		/* pat die Zeile komplett ? */
		if (len <= spalten)
		{
			strncpy(ziel,quelle,len);
			ziel += len;
			*ziel++ = 0;
			quelle += len +1;
			if (*quelle == '\r' || *quelle == '\n')
				quelle++;
			zeilen++;
			continue;
		}

		p = quelle;
		last = quelle;

		/* ggf. Quote in Puffer bernehmen */
		l_quote = 0;
		*quote = 0;
#if 0
		if (isupper(*quelle))			/* Quote-Initialen beginnen gro */
#endif
		{
			char *q = strchr(p,'>');
			if (q && q-p < 4)
			{
				l_quote = q-p+1;
				strncpy(quote,p,l_quote);	/* strncat */
				quote[l_quote] = 0;
				p += l_quote;				/* */
				last += l_quote;			/* */
				while (*p == ' ')			/* direkt folgende Leerzeichen entfernen */
				{ p++; last++; }
			}
		}

		l_rest = strcspn(p,"\r\n");
*ziel = 0;

		/*
			wortweise umbrechen
		*/
		do
		{
			char *leer_pos;

			/* falls Zeilenrest inkl. Quote komplett pat, sofort ausgeben */
			if (p - last + l_rest + l_quote < spalten)
			{
				p += l_rest;
				break;
			}

			/* letztes passendes Leerzeichen ermitteln */
			while (((leer_pos = strchr(p,' ')) != NULL)
				&& (leer_pos - last + l_quote < spalten))
			{
				l_rest -= leer_pos-p;
				p += leer_pos-p;
				p++; l_rest--;
			}

			/* Worttrenner suchen */
			len = strcspn(p,WORTTRENNER "\r\n");
			if (p - last + len + l_quote < spalten)
			{
				/* wortweise vortasten */
				p += len;
				l_rest -= len;
				if (*p)  { p++; l_rest--; }
			}
			else if (len > spalten)
			{
				/* Zeile zu lang fr Umbruch -> bis Maximum ausgeben */
				strncpy(ziel,last,(size_t)spalten);
				ziel += spalten;
				*ziel++ = 0;
				p += spalten;
				l_rest -= spalten;
				last = p;
zeilen++;
			}
			else
			{
				/* max. Spaltenzahl erreicht -> Ausgabe */
				if (l_quote)
					strcat(ziel,quote),ziel += l_quote;
				strncpy(ziel,last,(size_t)(p-last));
				ziel[p-last] = 0;
				ziel += p-last;
				*ziel++ = 0;
				while (*p == ' ')  { p++; l_rest--; }
				last = p;
#if 0
				p++;
				l_rest--;
#endif
zeilen++;
			}
		}
		while (*p && (*p != '\r') && (*p != '\n'));
/*zeilen++;*/

		/* ggf. Zeilenrest ausgeben */
		if (last != p)
		{
			if (l_quote)
				strcpy(ziel,quote),ziel += l_quote;
			strncpy(ziel,last,(size_t)(p-last));
			ziel[p-last] = 0;
			ziel += p-last;
/* *(ziel+1) = 0;*/
#if 0
*ziel++ = 0;	/* ### fhrt _auer am Ende der Msg._ zu berflssigen Leerzeilen */
			zeilen++;
#endif
#if 1
			if (*(p+1))		/* ### */
			{
				*ziel=0; /* Uwe */
				zeilen++;	/* Endzeile ??? */
			}
#endif
		}

		quelle = p+1;
	}

	return zeilen;

} /* msg_umbruch */



/*
	Message in Datei schreiben

	->			gh		:	Gruppen-Handle
				datei	:	Dateiname
				zeilen:	Anzahl Zeilen (nach dem Umbruch)

	Client(!)-Funktion
*/
BOOLEAN msg_write(int gh,const char *datei,long zeilen)
{
	FILE *save;
	char *pos;				/* Zeiger auf umgebrochenen Messagetext */

	if (!test_handle(gh,0))
		return FALSE;

	if (!*datei)
		return FALSE;

	if ((save = fopen(datei,BINWRITE)) == NULL)
	{
		fehler(NO_CREATE);
		return FALSE;
	}

	pos = msg[gh]->anzeige;

	while (zeilen--)
	{
		fputs(pos,save);
		fputs("\r\n",save);
		pos += strlen(pos)+1;
	}
	fclose(save);

	return TRUE;

} /* msg_write */



/*
	Message im Outfile-Format in Datei schreiben

	->			gh		:	Gruppen-Handle
				datei	:	Dateiname
				titel	:	Titel-Zeile(n) bzw. NULL, im Outfileformat

	Die Nachricht wird in ihrer ursprnglichen Form gespeichert.
	Die Nachricht wird an eine ggf. bestehende Datei angehngt.
	Zwischen Header und Nachrichtentext wird die Titelzeile(n) 
	gespeichert.
	Anwendung z.B.: "Kopiert aus Postfach" (Box)
*/
BOOLEAN msg_export_raw(int gh,const char *datei,const char *titel)
{
	WORK *mp = msg[gh];
	FILE *exp_f;

	if (!test_handle(gh,0))
		return FALSE;

	if (!*datei)
		return FALSE;

	if ((exp_f = fopen(datei,BINAPPEND)) == NULL)
		return FALSE;

	/* Header schreiben */
	fwrite(mp->puffer,mp->ix.t_text,1,exp_f);

	/* Titelzeile(n) anhngen */
	if (titel && *titel)
		fputs(titel,exp_f);

	/* Nachrichtentext schreiben */
	fwrite(mp->puffer + mp->ix.t_text,mp->ix.len - mp->ix.t_text,1,exp_f);
	fclose(exp_f);

	return TRUE;

} /* msg_export_raw */



#if !defined(SERVER)
/*
	Quotekrzel zu einem Benutzernamen ermitteln

	->			autor	:	Benutzername
			<-	quote	:	Namenskrzel

	Beispiel:
	(Kurzquote)
	"Frank Baumgart"				ergibt "FB>"
	"Marc van Woerkom"			ergibt "MvW>"
	"Frithard Meyer-zu-Uptrup"	ergibt "FMzU>"
	(Langquote)
	"Frithard Meyer-zu-Uptrup" ergibt "Frithard"

	Es werden max. 10 Zeichen zusammengestellt, falls ein Vorname lnger 
	sein sollte, wird stattdessen das Krzelquote zurckgeliefert.
*/
static void msg_get_initials(const char *autor,char *quote)
{
	char tmp[MAX_HDR];
	char *pos;
	size_t len = 0;

	if (!autor || !quote)
		return;

	len = strcspn(autor," -.");
	if (config.initials == 2 && (len < 10))
	{
		if (len > 9)
			return;
		strncpy(quote,autor,len);
		quote[len++] = '>';
		quote[len] = 0;
	}
	else
	{
		/* Adresse und Leerzeichen entfernen */
		strncpy(tmp,autor,MAX_HDR);
		tmp[MAX_HDR-1] = 0;
		if ((pos = strchr(tmp,'@')) != NULL)
			*pos = 0;
		while (((len = strlen(tmp)) != 0) && tmp[len-1] == ' ')
			tmp[len-1] = 0;

		len = 0;
		pos = tmp;
		*quote++ = *pos;
		while (((pos = strpbrk(pos," -")) != NULL) && (len++ < 9))
			*quote++ = *(++pos);
		*quote++ = '>';
		*quote   = 0;
	}
} /* msg_get_initials */



/*
	Message als Quote in eine Datei schreiben
	von CE

	->			gh		:	Gruppen-Handle
				file	:	stream fr Datei, wird nicht geschlossen

	Die Nachricht wird nicht mehr umgebrochen sondern in der aktuellen 
	Darstellung gespeichert
	Da z.Z. kein quote-Umbruch funktioniert, werden Leerzeichen am Ende 
	der Zeilen entfernt.

	Client(!)-Funktion
*/
void msg_writequote(int gh,FILE *file)
{
	char quote[10];
	char *pos;
	long zeilen;
	size_t len;

	if (!test_handle(gh,0))
		return;

/*	zeilen = msg_info(gh);	### */
	zeilen = msg[gh]->ix.zeilen;

	if (config.initials)
		msg_get_initials(msg[gh]->offset.author,quote);

	fputs(CRLF,file);				/* Leerzeile hinter "Kommentar zu ..." */
	pos = msg[gh]->anzeige;		/* ### -> puffer */
	while (zeilen--)
	{
		len = strlen(pos);

		/* ggf. Leerzeichen am Zeilenende entfernen, s.o. */
		if (len && (pos[len-1] == ' '))
			pos[--len] = 0;

		if (len)
		{
			if (config.initials)
				fwrite(quote,strlen(quote),1,file);
			else
				fputc('>',file);
			fwrite(pos,len,1,file);
		}
		fputs(CRLF,file);
		pos += len + 1;
	}
} /* msg_writequote */



/*
	Message drucken

	->			text	:	Zeiger auf umgebrochenen Messagetext
				laenge:	Lnge des Textes in Byte (besser: Zeilen)

	<-			TRUE	:	ok
				FALSE	:	im CNF "Printer" nicht gesetzt
							keine Shell vorhanden
							Kommando fehlgeschlagen

	In der Printer-Variablen bedeutet %f die Nachrichtendatei.

	Client(!)-Funktion
*/
BOOLEAN msg_print(char *text,size_t laenge)
{
	char cmd[256];
	char *tmp_n;			/* Zeiger auf Namen der temporren Datei */
	FILE *tmp_f;			/* temporre Datei */
	char *p;
	BOOLEAN ok;

	if (!*config.print)
	{
		fehler(NO_PRINT);
		return FALSE;
	}

	/* Messagedaten in temporre Datei schreiben */
	tmp_n = tmpnam(NULL);
	tmp_f = fopen(tmp_n,BINWRITE);
	fwrite(text,1,laenge,tmp_f);
	fclose(tmp_f);

	/* Kommando zusammensetzen */
	strcpy(cmd,config.print);
	if ((p = strstr(config.print,"%f")) != NULL)
	{
		cmd[p-config.print] = 0;
		strcat(cmd,tmp_n);
		strcat(cmd,p+2);
	}

	/* Shell vorhanden ? */
	if (!shell(NULL))
	{
		fehler(NO_SHELL);
		return FALSE;
	}

	/* Druckkommando an die Shell absetzen */
	ok = (shell(cmd) == 0);

	/* temporre Datei wieder lschen */
	unlink(tmp_n);

	return ok;

} /* msg_print */
#endif



/*
	Alle folgenden Messages mit demselben Autor bzw. Thema
	oder gesamte folgende Verkettung als zu lschen markieren

	->			gh				:	Gruppen-Handle
				art			:	AUTOR, TITEL oder VERKETTUNG
				all			:	Flag, ob in der gesamten Gruppe
				vergleich	:	Vergleichsstring fr Autor/Titel
*/
void msg_kill(int gh,WORD art,BOOLEAN all,const char *vergleich)
{
	INDEX aktuell;					/* Index der aktuellen Message */
	long save_ind,save_dat;
	char zeile[MAX_HDR];
	WORK *mp = msg[gh];

	if (!test_handle(gh,0))
		return;

	save_ind = ftell(mp->ind_f);
	save_dat = ftell(mp->dat_f);

	if (art == VERKETTUNG)
	{
		if (all && index_read(gh))
		{
			/* in der Verkettung so weit wie mglich nach oben gehen */
			while (mp->ix.oben != LEER)
			{
				index_goto(gh,mp->ix.oben);
				index_read(gh);
			}
		}
		/*
			smtliche Nachfolger lschen;
			Gruppe-Handle modul-lokal speichern, um in der rekursiven Funktion 
			auf dessen bergabe als Parameter (Stack !) verzichten zu knnen
		*/
		msg_kill_chain(gh);
	}
	else
	if (art == AUTOR || art == TITEL)
	{
		BOOLEAN gefunden;
		size_t vgl_len;		/* Lnge des Vergleichsstrings */

		/* aktuellen Indexdatensatz lesen */
		index_read(gh);
		aktuell = mp->ix;

		/* in der gesamten Gruppe oder nur ab aktueller Message ? */
		if (all)  index_goto(gh,0);

		vgl_len = strcspn(vergleich,"\r\n");

		while (index_read(gh))
		{
			/* der Vergleich findet erstmal ber CRCs statt */
			if (!((art == AUTOR && mp->ix.author_crc == aktuell.author_crc)
			||    (art == TITEL && mp->ix.topic_crc  == aktuell.topic_crc)))
			{
				if (index_next(gh))
					continue;
				else
					break;
			}

			/* mit Texteintrag aus den Headerdaten vergleichen */
			if (art == AUTOR)
				fseek(mp->dat_f,mp->ix.pos + mp->ix.t_author,SEEK_SET);
			else /* TITEL */
				fseek(mp->dat_f,mp->ix.pos + mp->ix.t_topic,SEEK_SET);
			fgets(zeile,MAX_HDR,mp->dat_f);
			if (strcspn(zeile,"\r\n") == vgl_len)
				gefunden = !strncmp(vergleich,zeile,vgl_len);
			else
				gefunden = FALSE;

			if (gefunden && !(msg_get_flag(gh) & TOTAL))
				(void) msg_set_flag(gh,TOTAL | GELESEN);

			if (!index_next(gh))
				break;
		}
	}

	fseek(mp->dat_f,save_dat,SEEK_SET);
	fseek(mp->ind_f,save_ind,SEEK_SET);

	index_read(gh);

} /* msg_kill */



/*
	Tiefensuche
	zur Lschung der Verkettung

	soll ebenfalls die _rckwrtige_ Verkettung gelscht werden, kann
	man vorher so weit nach oben gehen, bis sich kein Vater mehr findet
*/
static void msg_kill_chain(int gh)
{
	WORK *mp = msg[gh];

	/* Indexdatensatz einlesen */
	if (!index_read(gh))
		return;

	/* Lsch- und Gelesen-Flags setzen */
	if (!(mp->ix.flags & TOTAL))
		(void) msg_set_flag(gh,TOTAL | GELESEN);

	/* Tiefensuche, also erst die Nachfolger */
	if (mp->ix.unten != LEER)
	{
		MESSAGE save_pos;

		save_pos = msg_get_nr(gh);
		index_goto(gh,mp->ix.unten);
		msg_kill_chain(gh);
		index_goto(gh,save_pos);
		index_read(gh);
	}

	/* jetzt die Nachrichten derselben Ebene */
	if (mp->ix.rechts != LEER)
	{
		MESSAGE save_pos;

		save_pos = msg_get_nr(gh);
		index_goto(gh,mp->ix.rechts);
		msg_kill_chain(gh);
		index_goto(gh,save_pos);
		index_read(gh);
	}
} /* msg_kill_chain */



/*
	effizienter Suchalgorithmus nach Boyer-Moore

	->			text	:	Zeiger auf zu durchsuchenden Text
				wort	:	zu suchender Text

	<-			0 bzw. Position des gefundenen Wortes
*/
static int boyermoore(const char *text,const char *wort)
{
	size_t delta[256];
	size_t textlen = strlen(text)-1;
	size_t wortlen = strlen(wort)-1;
	size_t w,t;
	int i;

	if (textlen == 0 || wortlen == 0 || wortlen > textlen)
		return 0;

	for (i = 0; i < 256; i++)
		delta[i] = wortlen;
	for (i = 0; i < wortlen; i++)
		delta[(int) wort[i]] = wortlen-i;

	w = t = wortlen;

	do
	{
		if (wort[w] == text[t])
		{
			w--;
			t--;
		}
		else
		{
			size_t dt = delta[(int) text[t]];

			if (wortlen-w+1 > dt)
				t += dt + (wortlen-w+1);
			else
				t += dt;

			w = wortlen;
		}
	}
	while (w > 0 && t <= textlen);

	if (w)
		return 0;
	else
		return (int) t+1;

} /* boyermoore */



/*
	Zeichenkette in der aktuellen Gruppe suchen

	->			gh			:	Gruppen-Handle
				maske		:	zu suchender String
				header	:	TRUE = auch Header durchsuchen, FALSE = nur Text
				vorwaerts:	TRUE = vorwrts , FALSE = rckwrts

	<-			Nummer der Message bzw. LEER

	Bei Mierfolg bleiben sowohl Dateipositionen als auch 
	Messagepufferinhalt unverndert.
	Bei Erfolg wird der aktuelle Pufferinhalt bernommen und die 
	Dateipositionen entsprechen der gefundenen Nachricht.

	Suchverfahren: Boyer-Moore
*/
MESSAGE msg_suche_string(int gh,const char *maske,BOOLEAN header,BOOLEAN vorwaerts)
{
	MESSAGE alte_msg;
	WORK *mp = msg[gh];

	if (!test_handle(gh,0))
		return LEER;

	/* sicherheitshalber abfragen */
	if (!mp->ind_f || !mp->dat_f)
		return LEER;

	alte_msg = msg_get_nr(gh);

	while (1)
	{
		const char *start;

		if (vorwaerts && !index_next(gh))
			break;
		else
		if (!vorwaerts && !index_prev(gh))
			break;

		/* Messagedaten einlesen */
		if (!index_read(gh) || !msg_read(gh))
			break;

#if 0
		mp->puffer[mp->ix.len] = '\0';
#endif
		start = (header) ? mp->puffer : mp->puffer + mp->ix.t_text;

		if (boyermoore(start,maske))
			return msg_get_nr(gh);
	}

	/* wiedereinlesen des bei Funktionsaufruf vorhandenen Messagetextes */
	index_goto(gh,alte_msg);
	index_read(gh);
	msg_read(gh);

	mprintf("\a");

	return LEER;

} /* msg_suche_string */



/*
	gesetzten Merker (Autor oder Thema) vor- oder rckwrts suchen

	->			klemm		:	Art (AUTOR oder TITEL) usw.
				vorwaerts:	vorwrts (TRUE) oder rckwrts suchen

	<-			TRUE		:	gefunden
				FALSE		:	nicht gefunden

	Achtung:
	1. diese Funktion kann die Position im ind-stream verndern
	2. es findet keine Stringberprfung statt sondern nur ein CRC-Check
*/
BOOLEAN msg_suche_merker(int gh,KLEMM *klemm,BOOLEAN vorwaerts)
{
	MESSAGE save_pos;
	BOOLEAN found = FALSE;
	WORK *mp = msg[gh];

	if (!test_handle(gh,0))
		return FALSE;

	if (!klemm->art)
		return FALSE;

	save_pos = msg_get_nr(gh);

	/*
		bei rckwrtiger Suche jetzt einen Datensatz zurck, denn sonst 
		stimmt der aktuelle bereits berein, analog bei Vorwrtssuche zur 
		nchsten Nachricht
	*/
	if (vorwaerts)
	{
		if (save_pos >= (mp->kennung.anzahl - 1))
			return FALSE;

		index_next(gh);
	}
	else
	{
		if (!save_pos)
			return FALSE;

		index_prev(gh);
	}

	while ((vorwaerts) ? (!feof(mp->ind_f)) : (msg_get_nr(gh) > 0))
	{
		if (!index_read(gh))
			break;
		if((klemm->art == AUTOR && mp->ix.author_crc == klemm->crc)
		|| (klemm->art == TITEL && mp->ix.topic_crc  == klemm->crc))
		{
			found = TRUE;
			break;
		}
		if (vorwaerts)
			index_next(gh);
		else
			index_prev(gh);
	}

	if (!found)
	{
		index_goto(gh,save_pos);
		index_read(gh);
		mprintf("\a");
	}

	return found;

} /* msg_suche_merker */



/*
	akt. Nachricht rot13 (de-)kodieren
*/
void msg_rot13(int gh)
{
	char *p = msg[gh]->puffer + msg[gh]->ix.t_text;
	long len = msg[gh]->ix.len - msg[gh]->ix.t_text;
	int c;

	if (!test_handle(gh,0))
		return;

	while (len--)
	{
		c = *p;
		if (c >= 'A' && c <= 'Z')
			c = ((c - 52) % 26) + 'A';
		else
		if (c >= 'a' && c <= 'z')
			c = ((c - 84) % 26) + 'a';
		*p++ = c;
	}

} /* msg_rot13 */
