/* message.h : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
       (might be the wrong person except for offline)

    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
#if !defined(MSG_STRUCT)
#define MSG_STRUCT

#include <sys/types.h>

#include "struktur.h"


/*
	Flags fr einzelne Nachrichten
*/
/* #define GELESEN			0x1 */
#define TEIL				0x8
#define TOTAL				0x10

#define XP_REF				0x20			/* Msg. ist eine crossposting-Referenz */

#define GUEK				0x200			/* Gruppen-bergreifender Kommentar */
#define URGENT				0x400
#define GESCHUETZT		0x800			/* vor Lschen geschtzt */

#define CODIERT			0x1000		/* Msg. ist codiert (PM) */

#define V_LOKAL			0x2000L
#define V_MAUSNET			0x4000L
#define V_NET				0x8000L
#define V_REGION			0x10000L

#define XP					0x20000L		/* Msg. ist ein crossposting */
#define YOURMAIL			0x40000L		/* selbst erstellte PM */


/*
	sonstiges
*/
#define DEFEKT				0x80000000L


/*
	Bearbeitungsstatus
*/
#define GELESEN			1
#define ZURUECKGESTELLT	2
#define BEANTWORTET		3
#define WEITERGEGEBEN	4
#define MAUSNET			5
#define ANGEKOMMEN		6
#define GATEWAY			7
#define KOPIERT			8
#define GETAUSCHT			9
#define GEAENDERT			0x8000


#define AUTOR				1
#define TITEL				2
#define VERKETTUNG		3

#define SEL_ALL			0
#define SEL_AUTOR			1
#define SEL_THEMA			2
#define SEL_TEXT			3


/*
	max. gleichzeitig zu ffnende Gruppen
*/
#define MAX_BROWSER		256


/*
	zum Merken von Autor bzw. Thema
*/
typedef struct
{
	int art;							/* AUTOR/TITEL/INHALT */
	CRC crc;							/* crc darber */
	char name[80];					/* Name des Autors/Titels/Inhalts selbst */

} KLEMM;



/*
	Startadressen der einzelnen Eintrge innerhalb einer Message
*/
typedef struct
{
	char *id;
	char *author;					/* enthlt die Absenderadresse */
	char *topic;
	char *group;
	char *text;
	char *name;						/* enthlt den Namen, falls ungleich author */

} OFFSET;



/*
	ffentliche und interne Daten fr jede im Browser geffnete Gruppe
*/
typedef struct
{
	/* public: */
	KENNUNG kennung;				/* Header der Datenbank */
	OFFSET offset;					/* Zeiger auf Headereintrge */
	INDEX ix;						/* Indexstruktur */
	MESSAGE ursprung;				/* Kommentarverkettungseinstieg */
	BOOLEAN inhalt;				/* Inhaltsdarstellung */

	/* mixed: */
	BOOLEAN su;						/* Flag, ob su/Sysop/Brettbearbeiter/Owner */
	/* server */
	char *puffer;					/* Zeiger auf Messagepuffer */
	size_t puffer_size;			/* Gre in Bytes */
	/* client */
	char *anzeige;					/* Zeiger auf (umgebrochenen) Anzeigepuffer */

	/* private: */
	GRUPPE gruppe;					/* Gruppennummer */
	BOOLEAN pm;						/* Postfach oder Gruppe */
	BOOLEAN offen;					/* Flag, ob Gruppe geffnet */
	MESSAGE save_nr;				/* Msg.nr. fr index_push/pop() */
	char **ic;						/* Inhaltszeilen-Cache */
	long ic_free;					/* max. zu reservierende Cachezeilen */
	FILE *ind_f;					/* streams fr Index- */
	FILE *dat_f;					/* und Datendatei */

} WORK;



/*
	Funktionsprototypen zur Nachrichten-Verwaltung/Anzeige
*/

int     group_open(pid_t pid,const char *gruppenname,BOOLEAN pm,BOOLEAN rw);
BOOLEAN group_flush(int gh,pid_t pid);
BOOLEAN group_close(int gh,pid_t pid);
BOOLEAN group_info(const char *gruppenname,BOOLEAN pm,KENNUNG *kennung);
BOOLEAN group_info_msgs(const char *name,BOOLEAN pm,long *anzahl,long *groesse);

BOOLEAN index_read_kennung(int gh,KENNUNG *kennung);
BOOLEAN index_write_kennung(int gh,KENNUNG *kennung);

MESSAGE msg_get_nr(int gh);

BOOLEAN index_push(int gh);
BOOLEAN index_pop(int gh);

BOOLEAN index_read(int gh);
BOOLEAN index_write(int gh);
BOOLEAN index_prev(int gh);
BOOLEAN index_next(int gh);
BOOLEAN index_up(int gh);
BOOLEAN index_goto(int gh,MESSAGE i);
BOOLEAN index_goto_new(int gh,time_t ref);
BOOLEAN index_goto_top(int gh);

BOOLEAN msg_read_header(int gh);
BOOLEAN msg_index_string(int gh,const char *maske,int len,char *s);
BOOLEAN msg_index_string_nr(int gh,MESSAGE nr,int len,char *s);

BOOLEAN msg_read(int gh);
BOOLEAN msg_prev(int gh);
BOOLEAN msg_next(int gh);
BOOLEAN msg_up  (int gh);
BOOLEAN msg_down(int gh);
BOOLEAN msg_next_new(int gh,ULONG skip_flags);

BOOLEAN msg_set_flag(int gh,ULONG flags);
BOOLEAN msg_clr_flag(int gh,ULONG flags);
ULONG   msg_get_flag(int gh);

UWORD   msg_get_status(int gh);
void    msg_set_status(int gh,UWORD status);

void    msg_kill(int gh,WORD art,BOOLEAN all,const char *vergleich);
BOOLEAN msg_write(int gh,const char *datei,long spalten);
void    msg_writequote(int gh,FILE *file);
BOOLEAN msg_export_raw(int gh,const char *datei,const char *titel);
BOOLEAN msg_print(char *text,size_t laenge);
void    msg_rot13(int gh);

MESSAGE msg_suche_string(int gh,const char *maske,BOOLEAN header,BOOLEAN vorwaerts);
BOOLEAN msg_suche_merker(int gh,KLEMM *klemm,BOOLEAN vorwaerts);

/* Client-Funktionen */
long msg_umbruch(const char *quelle,char *ziel,int spalten);

#endif
