/* msg_serv.c : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
       (might be the wrong person except for offline)

    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
/*
	Offline - MausTausch-FrontEnd

	Verwaltungs- und Synchronisationsebene des message-Moduls
	(c) 1992-94 bei Frank Baumgart

	begonnen am 09.08.1994
	Version vom 09.08.1994
*/


#include "portab.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "system.h"
#include "mdb_glob.h"
#include "message.h"
#include "mdb/mdb_intr.h"		/* wg. group_path() */
#include "tool.h"



/*
	Kennung einer Gruppe holen

	->			name		:	Name der Gruppe/User
				pm			:	PM (TRUE) oder Gruppe (FALSE)
				kennung	:	Puffer fr Kennung oder NULL

	<-			TRUE		:	ok
				FALSE		:	Fehler aufgetreten bzw. Indexdatei fehlt/fehlerhaft

	Diese Funktion kann und soll auch, insb. von nicht-internen Aufrufern, 
	dazu verwendet werden, die Existenz einer Gruppe zu berprfen, da 
	group_path() nur fr interne Anwendungen mit Zugriff auf interne Daten-
	strukturen gedacht ist.
*/
BOOLEAN group_info(const char *name,BOOLEAN pm,KENNUNG *kennung)
{
	char ind_name[PATH_MAX];
	FILE *ind_f;

	/* Namen von Index- und Datendatei ermitteln */
	if (group_path(name,ind_name,pm) < 0)
		return FALSE;

	if (kennung)
	{
		strcat(ind_name,"ind");

		/* Indexdatei zum Lesen und Schreiben ffnen */
		if ((ind_f = fopen(ind_name,BIN_RW_OPEN)) == NULL)
			return FALSE;

		/* enthlt die Gruppe berhaupt Nachrichten ? */
		if (filesize(ind_f) < sizeof(KENNUNG))
		{
			fclose(ind_f);
			return FALSE;
		}

		lies_kennung(ind_f,kennung);
		fclose(ind_f);
	}

	return TRUE;

} /* group_info */



/*
	Anzahl Mails und Gre einer Gruppe ermitteln

	->			name		:	Gruppenname
				pm			:	TRUE = PM, FALSE = Gruppe
			<-	anzahl	:	Rckgabe der Anzahl Mails, falls ungleich NULL
			<-	groesse	:	Gre der Gruppe/Postfach in Bytes, "

	<-			TRUE		:	Gruppe/Postfach existiert
				FALSE		:	ex. nicht

	Da Index- und Datendatei nicht geffnet werden brauchen, ist diese 
	Funktion sehr schnell und eignet sich insb. fr die Liste im Bretteditor 
*/
BOOLEAN group_info_msgs(const char *name,BOOLEAN pm,long *anzahl,long *groesse)
{
	char ind_name[PATH_MAX];
	char dat_name[PATH_MAX];
	struct stat buf;

	if (group_path(name,dat_name,pm) < 0)
		return FALSE;
	strcpy(ind_name,dat_name);
	strcat(ind_name,"ind");
	strcat(dat_name,"dat");

	if (anzahl)
		if (!stat(ind_name,&buf))
			*anzahl = (buf.st_size - sizeof(KENNUNG)) / sizeof(INDEX);
		else
			*anzahl = 0;

	if (groesse)
		if (!stat(dat_name,&buf))
			*groesse = buf.st_size;
		else
			*groesse = 0;

	return TRUE;

} /* group_info_msgs */



#if !defined(CLIENT)
/*
	Datenbankgre etc. ermitteln

	->		<-	user		:	Anzahl User-Datenbanken
			<-	gruppen	:	Anzahl Gruppen-Datenbanken
			<-	mails		:	Anzahl Mails insgesamt
			<-	bytes		:	Anzahl Bytes insgesamt
			<-	overall	:	Anzahl berhaupt importierter Mails bzw. NULL
			<-	neue		:	Anzahl neue (ungelesene) Mails bzw. NULL
			<-	putzen	:	Anzahl zu putzende Mails bzw. NULL

	<-			TRUE		:	ok
				FALSE		:	(kann derzeit nicht vorkommen)

	Da Index- und Datendatei nicht geffnet werden brauchen, falls man nicht 
	gerade an der Anzahl jeher importierter bzw. ungelesener/zu putzender Mails 
	interessiert ist (letzteres eh nur im FE sinnvoll) ist diese Funktion sehr 
	schnell und eignet sich insb. fr die Liste im Bretteditor.
	### eigentlich ist diese Funktion in diesem Modul nicht ganz richtig 
	aufgehoben, da sie nichts mit der Msg.Server-Ebene zu tun hat.
	Ansonsten ist sie wg. ihres globalen Charakters im MDB-Server am besten 
	aufgehoben.
*/
BOOLEAN
mdb_info(ULONG *user,ULONG *gruppen,ULONG *mails,ULONG *bytes,ULONG *overall,
         ULONG *neue,ULONG *putzen)
{
	char ind_name[PATH_MAX];
	char dat_name[PATH_MAX];
	struct stat buf;
	KENNUNG kennung;
	char name[MAX_CPG];
	BOOLEAN pm;

	*user = *gruppen = *mails = *bytes = 0;
	if (overall)
		*overall = 0;
	if (neue)
		*neue = 0;
	if (putzen)
		*putzen = 0;

	for (pm = 0; pm <= 1; pm++)
	{
		open_grplst(pm);
		while (read_grplst(name))
		{
			if (group_path(name,dat_name,pm) < 0)
				continue;
			strcpy(ind_name,dat_name);
			strcat(ind_name,"ind");
			strcat(dat_name,"dat");

			if (mails && !stat(ind_name,&buf))
				*mails += (buf.st_size - sizeof(KENNUNG)) / sizeof(INDEX);

			if (bytes && !stat(dat_name,&buf))
				*bytes += buf.st_size;

			/*
				wenn spezielle Daten aus der Kennung bentigt werden, fhrt kein 
				Weg am relativ zeitaufwendigen ffnen/Lesen der Indexdatei vorbei
			*/
			if ((overall || neue || putzen) && group_info(name,pm,&kennung))
			{
				if (overall)
					*overall += kennung.overall;
				if (neue)
					*neue += kennung.neue;
				if (putzen)
					*putzen += kennung.putzen;
			}

			(pm) ? (*user)++ : (*gruppen)++;
		}
		close_grplst();
	}

	return TRUE;

} /* mdb_info */
#endif
