/* adressen.c : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB3 (mausnet)
       (might be the wrong person except for offline)

    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
/*
	Offline - MausTausch-FrontEnd

	Adre-Auswahl
	(c) 1992-95 bei Frank Baumgart

	begonnen am 18.10.1994
	Version vom 18.10.1994
*/


#include "portab.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "system.h"
#include "mdb/mdb_intr.h"
#include "frontend/frontend.h"
#include "io.h"
#include "mdb_conf.h"
#include "fehler.h"
#include "missing/missing.h"

extern CONFIG config;


#ifdef TEXT_FRONTEND

static SEL_EINTRAG *adresse;
static size_t anzahl;

static BOOLEAN read_adresses(void);
static void forget_adresses(void);



static int srt_name(const void *a,const void *b);

static int srt_name(const void *a,const void *b)
{
	return stricmp((*(const SEL_EINTRAG *)a).text,(*(const SEL_EINTRAG *)b).text);

} /* srt_name */



static BOOLEAN read_adresses(void)
{
	char datei[PATH_MAX];
	char name[80];
	FILE *adr_f;
	char *p;

	strcpy(datei,config.messages);
	strcat(datei,"adressen.lst");
	if ((adr_f = fopen(datei,"r")) == NULL)
		return FALSE;
	anzahl = 0;
	while (fgets(name,sizeof(name),adr_f))
	{
		if ((p = strpbrk(name,"#;\r\n\t")) != NULL)
			*p = 0;
		if (strlen(name))
			anzahl++;
	}
	if ((adresse = (SEL_EINTRAG *) calloc(anzahl,sizeof(SEL_EINTRAG))) == NULL)
	{
		fclose(adr_f);
		fehler(NO_MEMORY);
		return FALSE;
	}
	rewind(adr_f);
	anzahl = 0;
	while (fgets(name,sizeof(name),adr_f))
	{
		if (*name == ';' || *name == '#')
			continue;
		if ((p = strpbrk(name,"\r\n")) != NULL)
			*p = 0;
		if (!strlen(name))
			continue;
		p = strchr(name,'\t');
		while (p && *p == '\t')
			*p++ = 0;
		adresse[anzahl].text = strdup(name);
		adresse[anzahl].info = p ? strdup(p) : NULL;
		adresse[anzahl].fett = FALSE;
		anzahl++;
	}
	fclose(adr_f);

	qsort(adresse,anzahl,sizeof(SEL_EINTRAG),srt_name);

	return TRUE;

} /* read_adresses */



static void forget_adresses(void)
{
	while (anzahl)
	{
		anzahl--;
		if (adresse[anzahl].info)
			free(adresse[anzahl].info);
		free(adresse[anzahl].text);
	}
	free(adresse);

} /* forget_adresses */



/*
	Auswahl/Eingabe einer Adresse

	->		<-	name	:	Eingabe, Puffer mu mind. 80 Bytes gro sein

	<-			TRUE	:	Eingabe ok/abgeschlossen
				FALSE	:	Eingabe falsch/abgebrochen
*/
BOOLEAN input_adress(char *name)
{
	static int last = 0;
	int tmp;

	if (!read_adresses())
		return FALSE;

	*name = 0;
	tmp = uni_eingabe("Empfnger:",INPUT_USER,adresse,anzahl,last,TRUE,name);
	forget_adresses();
	if (tmp == -1)
		return FALSE;
	if (tmp >= 0)
		last = tmp;

	if (!*name)
		return FALSE;
	else
		return TRUE;

} /* input_adress */

#endif /* TEXT_FRONTEND */



BOOLEAN insert_adress(const char *name)
{
	char datei[PATH_MAX];
	FILE *adr_f;
	char eintrag[80];

	strcpy(datei,config.messages);
	strcat(datei,"adressen.lst");

	if ((adr_f = fopen(datei,"r")) == NULL)
		return FALSE;
	while (fgets(eintrag,75,adr_f))
	{
		char *p;
		size_t len;

		if ((p = strpbrk(eintrag,"#,\r\n")) != NULL)
			*p = 0;
		len = strlen(eintrag);
		while (len && (isspace(eintrag[--len])))
			;
		if (!stricmp(eintrag,name))
		{
			mputs("doppelter Eintrag");
			do_beep();
			sleep(2);
			fclose(adr_f);
			return TRUE;
		}
	}
	fclose(adr_f);

	if ((adr_f = fopen(datei,"a")) == NULL)
		return FALSE;
	fputs(name,adr_f);
	fclose(adr_f);

	return TRUE;

} /* insert_adress */
