/* io.c : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB3 (mausnet)
       (might be the wrong person except for offline)

    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
/*
	Offline - MausTausch-FrontEnd

	IO mit termcap-Handling
	(c) 1992-95 bei Frank Baumgart

	begonnen am  x.08.1992
	Version vom 20.11.1994
*/

#include "portab.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <signal.h>
#include <unistd.h>

#if defined(__GO32__) || defined(__BORLANDC__) || defined(EMX) || defined(DJGPP) || defined(__WATCOMC__)
#include <conio.h>
#elif __TOS__ && !defined(__MINT__) && !defined(__UNIX__)
#include <ext.h>
#else
#include <fcntl.h>
#if defined(__MINT__)
#include <ioctl.h>
#include <termios.h>
#include <sys/termios.h>
#endif
#endif

#include "system.h"
#include "io.h"
#include "mdb_conf.h"

#if defined(BOX) && !defined(SERVER)
#include "modem.h"
#include "user.h"
extern char *vt_pos_S(int y,int x);
extern USER user;
extern UWORD anz_lineW;
extern int user_out_fd;
#endif

#include "termcap.h"
/*
	Zugriff auf diese Variable ist wg. Offline-interner termcap-Eintrge 
	notwendig, da sie in termcap.lib static ist, mu ein eigenes termcap 
	Modul verwendet werden (tgetent nochmal genau ansehen!)
*/
extern char *term_entry;

extern CONFIG config;				/* globale Einstellungen (Verkettung...) */
extern BOOLEAN interaktiv;



int bs_zeilen,bs_spalten;			/* Bildschirmmae in Zeichen */
int vt100 = FALSE;					/* Flag, ob v100 capabilities (Fenster) */


#if defined(TEXT_FRONTEND) && !defined(SERVER)

static void io_exit(void);

#ifdef __UNIX__
static struct termios cons;

static BOOLEAN ignore_stop = FALSE;
static BOOLEAN ignore_cont = FALSE;
#endif


/*
	stdin auf RAW und NOECHO setzen,
	Signal-handler fr SIGSTOP, SIGCONT und SIGWINCH installieren
*/
void io_init(void)
{
#ifdef __UNIX__
	struct termios console;

	if (!isatty(STDIN_FILENO))
		return;

	tcgetattr(STDIN_FILENO,&cons);
	console = cons;
	console.c_cflag &= ~CSIZE;
	console.c_cflag |= CS8;
	console.c_lflag &= ~ICANON;
	console.c_lflag &= ~ECHO;
	console.c_cc[VMIN]  = 1;
	console.c_cc[VTIME] = 0;

	atexit(io_exit);

	/* handler fr SIGSTOP und SIGCONT installieren */
	if (signal(SIGSTOP,io_exit) == SIG_IGN)
		ignore_stop = TRUE;
	if (signal(SIGCONT,io_init) == SIG_IGN)
		ignore_cont = TRUE;
#if 0
	if (signal(SIGWINCH,get_screen) == SIG_IGN)
		;
#endif

	tcsetattr(STDIN_FILENO,TCSADRAIN,&console);
#endif

} /* io_init */



/*
	stdin auf Ursprungseinstellung zurcksetzen
*/
static void io_exit(void)
{
#ifdef __UNIX__
	tcsetattr(STDIN_FILENO,TCSADRAIN,&cons);
#endif
} /* io_exit */



/*
	Termcap-Eintrag fr bestimmte Taste ermitteln und speichern
	i.W. fr Cursorsteuerung der Kommentarverkettung gedacht
*/
static void get_key_entry(const char *entry,const char **save)
{
	/* der Speicher wird von tgetstr bereitgestellt */
	*save = tgetstr(entry,NULL);

} /* get_key_entry */

#endif /* TEXT_FRONTEND */



/*
	Tastatureingabe ohne Echo
*/
int taste(void)
{
#if defined(SERVER)
	/* Server machen keine interaktiven Eingaben */
	return 0;

#elif defined(BOX)
	return mbchar();

#else
	if (!interaktiv)
		return 0;

	fflush(stdout);

#ifdef HAS_GETCH
	{
		int key;

		key = getch();
		if (!key)
		{
			/* Sondertasten auswerten und Offline-einheitliche Codes zuordnen */
			key = getch();
			if (key == 72)
				return C_UP;
			if (key == 80)
				return C_DOWN;
			if (key == 75)
				return C_LEFT;
			if (key == 77)
				return C_RIGHT;
			if (key == 71)
				return C_HOME;
			if (key == 98)
				return C_HELP;
			if (key == 55)
				return C_PRINT;
			if (key == 73)
				return C_PGUP;
			if (key == 81)
				return C_PGDN;
			if (key == 79)
				return C_END;
			if (key == 97)
				return C_UNDO;

			return 0;
		}

		return key;
	}

#else
	{
#ifdef TEXT_FRONTEND
		static BOOLEAN key_init = FALSE;
		static const char *c_up,*c_down,*c_left,*c_right;
		static const char *c_pgdn,*c_pgup,*c_home,*c_center;
		static const char *c_f1,*c_f2,*c_f3,*c_f4,*c_f5,*c_f6;
		static const char *c_f7,*c_f8,*c_f9,*c_f10; /* ,*c_f11,*c_f12;*/

		static struct
		{
			const char *cap;
			const char **save;
			int code;
		}
		key[] =
		{
			{ "ku",	&c_up,		C_UP },
			{ "kd",	&c_down,		C_DOWN },
			{ "kl",	&c_left,		C_LEFT },
			{ "kr",	&c_right,	C_RIGHT },
			{ "sf",	&c_pgdn,		C_PGDN },
			{ "sb",  &c_pgup, 	C_PGUP },
			{ "ho",	&c_home,		C_HOME },
			{ "K3",	&c_center,	C_CENTER },

			{ "k1",	&c_f1,		C_F1 },
			{ "k2",	&c_f2,		C_F2 },
			{ "k3",	&c_f3,		C_F3 },
			{ "k4",	&c_f4,		C_F4 },
			{ "k5",	&c_f5,		C_F5 },
			{ "k6",	&c_f6,		C_F6 },
			{ "k7",	&c_f7,		C_F7 },
			{ "k8",	&c_f8,		C_F8 },
			{ "k9",	&c_f9,		C_F9 },
			{ "k0",	&c_f10,		C_F10 },

			{ NULL,	NULL,			0 },
		};

		/* Codes fr Cursor/Sondertasten ermitteln */
		if (!key_init)
		{
			int i = 0;

			while (key[i].cap)
			{
				get_key_entry(key[i].cap,key[i].save);
				i++;
			}

			/*
				leider stimmen termcap-Eintrag und String, der von den 
				Sondertasten zurckgeliefert wird, nicht berein, allerdings 
				gibt es zwei Modi: application und ???
			*/
			c_up		= "\033[A";
			c_down	= "\033[B";
			c_left	= "\033[D";
			c_right	= "\033[C";
			c_pgdn	= "\033[6~";
			c_pgup	= "\033[5~";
/*			c_home	= "\033[\0";
			c_end		= "\033[\0";*/

			key_init = TRUE;
		}
#endif

		{
			static int tty = -1;
			char zeichen;

			if (tty == -1)
				tty = isatty(STDIN_FILENO);

			if (!tty)
				return fgetc(STDIN_FILENO);

			read(STDIN_FILENO,&zeichen,1);

#if 1 /* war: ifdef TERMCAP */
			if (zeichen == '\033')
			{
				char string[20];
				int anzahl;
				int match;

				memset(string,0,20);
				string[0] = zeichen;

				read(STDIN_FILENO,string+1,1);

				/* Zeichen bis zu eindeutiger bereinstimmung lesen */
				do
				{
					int i = 0;

					read(STDIN_FILENO,string + strlen(string),1);

					anzahl = 0;
					match = 0;

					while (i < (sizeof(key)/sizeof(key[0])) -1)
					{
						if (!*key[i].save)
						{
							i++;
							continue;
						}

						if (!strncmp(*key[i].save +1,string+1,strlen(string)-1))
						{
							match = i;
							if (++anzahl > 1)
								break;
						}
						i++;
					}
				}
				while ((anzahl > 1) && (strlen(string) < 19));

				if (anzahl == 1)
					return key[match].code;
			}
#endif
			return zeichen;
		}
	}
#endif
#endif

} /* taste */



#ifndef SERVER

/*
	Warnton ausgeben
*/
void do_beep(void)
{
#ifdef BOX
	write(user_out_fd,"\a",1);
#else
	mprintf("\a");
#endif
} /* do_beep */



/*
	Bildschirmmae ermitteln
*/
void get_screen(void)
{
#ifdef BOX
	bs_zeilen = anz_lineW;
	if (bs_zeilen < 10)
		bs_zeilen = 10;
	bs_spalten = user.bs_spalten;
	if (!bs_spalten || bs_spalten < 60)
		bs_spalten = 80;
#endif

#ifdef TEXT_FRONTEND
	/* Anzahl Zeilen */
	if (getenv("LINES"))
		bs_zeilen = atoi(getenv("LINES"));
	else
	if (getenv("ROWS"))
		bs_zeilen = atoi(getenv("ROWS"));
	else
		bs_zeilen = 25;

	/* Anzahl Spalten */
	if (getenv("COLUMNS"))
		bs_spalten = atoi(getenv("COLUMNS"));
	else
		bs_spalten = 80;
#endif

} /* get_screen */

#endif /* !SERVER */



/*
	variable Ausgabefunktionen fr folgende Zwecke:
	- Text-FrontEnd
	- FrontEnd mit eigener Ausgabe
	- Box, direkt
	- Box-Server
	- Box-Client
*/



/*
	meta-puts
*/
int mputs(const char *text)
{
#if defined(TEXT_FRONTEND)
	return puts(text);

#elif defined(SERVER)
	/* WENN schon Ausgaben, dann hchstens auf stdout (besser: syslog) */
	/* die Ausgaben gehen jedenfalls der Box verloren */
	return puts(text);

#elif defined(BOX)
	static char tmp[256];
	strncpy(tmp,text,253);
	tmp[253] = 0;
	strcat(tmp,"\r\n");
	return mboutput(tmp);
#else
	return win_puts(text);
#endif
} /* mputs */



int mprintf(const char *format,...)
{
#if defined(BOX)
	static char tmp[500];
#endif
	va_list param;
	int zeichen;

	va_start(param,format);

#if defined(TEXT_FRONTEND)
	zeichen = vprintf(format,param);

#elif defined(SERVER)
	/* Server-Ausgaben derzeit auf stdout; besser: syslog */
	zeichen = vprintf(format,param);

#elif defined(BOX)
	zeichen = vsprintf(tmp,format,param);
	zeichen = mboutput(tmp);
#else
	zeichen = win_printf(format,param);
#endif

	va_end(param);

	return zeichen;

} /* mprintf */



void mgets(char *puffer,size_t len)
{
	char *pos;

#if defined(TEXT_FRONTEND)
	char *p;
	size_t n;
	int c;

	if (*puffer)
	{
		mprintf("%s",puffer);
		n = strlen(puffer);
		p = puffer + n;
	}
	else
	{
		n = 0;
		p = puffer;
	}

	while (n < len)
	{
		c = taste();

		/* backspace oder delete ? */
		if ((c == '\b') || (c == 127))
		{
			if (n)
			{
				*--p = 0;
				putchar('\b');
				putchar(' ');
				putchar('\b');
				c = 0;
				n--;
			}
		}
		else
		if (c == '\r' || c == '\n')
			break;
		else
		if (c == ('u' & 0x1f))
		{
			p = puffer;
			*p = 0;
			while (n)
			{
				putchar('\b');
				putchar(' ');
				putchar('\b');
				n--;
			}
			continue;
		}
		else
		{
			*p++ = c;
			*p = 0;
			n++;
		}
		putchar(c);
	}
#elif defined(SERVER)
	/* server machen keine interaktiven Eingaben */
#elif defined(BOX)
	mbquestion((int) len,0,TRUE,"",puffer,puffer);
#else
	win_gets(puffer,len);
#endif

	if ((pos = strpbrk(puffer,"\r\n")) != NULL)
		*pos = 0;

} /* mgets */



#if defined(TEXT_FRONTEND)

/*
	default-Termcaps fr vt52 und vt100
*/
static char default_termcap_vt52[] =
{
/*
"do=^J:al=\\EL:dl=\\EM:"
*/
"vt52:"
"up=\\EA:do=\\EB:nd=\\EC:le=\\ED:cl=\\EE:ho=\\EH:sr=\\EI:cd=\\EJ:ce=\\EK:"
"cm=\\EY%+ %+ :co#80:li#25:"
"so=\\Ep:se=\\Eq:"
"mr=\\Ep:me=\\Eq:am:bs:"
"km:kR=\\E#8:kr=\\E#M:ku=\\E#H:rc=\\Ek:sc=\\Ej:"
"us=\\Eb1:ue=\\Eb?:"
};

static char default_termcap_vt100[] =
{
"vt100:\\\n"
"do=^J:bl=^G:co#80:li#25:cl=\\E[;H\\E[2J:le=^H:bs:"
"am:cm=\\E[%i%d;%dH:nd=\\E[C:up=\\E[A:ce=\\E[K:cd=\\E[J:"
"so=\\E[7m:se=\\E[m:he=\\E[m:us=\\E[4m:ue=\\E[m:mr=\\E[7m:"
"mb=\\E[5m:me=\\E[m:md=\\E[1m:"
"ku=\\EOA:kd=\\EOB:kr=\\EOC:kl=\\EOD:kb=^H:ho=\\E[H:k1=\\EOP:k2=\\EOQ:"
"ta=^I:pt:sr=5\\EM:vt#3:xn:sc=\\E7:rc=\\E8:cs=\\E[%i%d;%dr:"
"al=\\E[1L:dl=\\E[1M:\n"
};

static char default_termcap_ansi[] =
{
"ansi:\\\n"
"do=^J:bl=^G:co#80:li#25:cl=\\E[;H\\E[2J:le=^H:bs:"
"am:cm=\\E[%i%d;%dH:nd=\\E[C:up=\\E[A:ce=\\E[K:cd=\\E[J:"
"so=\\E[7m:se=\\E[m:he=\\E[m:us=\\E[4m:ue=\\E[m:mr=\\E[7m:"
"mb=\\E[5m:me=\\E[m:"
"ku=\\EOA:kd=\\EOB:kr=\\EOC:kl=\\EOD:kb=^H:ho=\\E[H:k1=\\EOP:k2=\\EOQ:"
"ta=^I:pt:sr=5\\EM:vt#3:xn:sc=\\E[s:rc=\\E[u:cs=\\E[%i%d;%dr:"
"al=\\E[1L:dl=\\E[1M:\n"
};



/*
	Termcap initialisieren

	Zeiger auf Termcapeintrag des aktuellen Terminals setzen
	(im Environment mit TERM angegeben)
	Ist kein Terminaltyp angegeben, wird vt52 verwendet.
	Kann keine termcap gefunden werden, wird die eingebaute vt52-
	termcap verwendet.
*/
void init_termcap(void)
{
	char default_terminal[] = "vt52";
	char *term;

	if (*config.termtyp)
		term = config.termtyp;
	else
		term = getenv("TERM");
	if (!term)
		term = default_terminal;

	if (!strcmp(term,"ansi"))
	{
		term_entry = default_termcap_ansi;
		return;
	}
	else
	if (!strcmp(term,"vt100"))
	{
		term_entry = default_termcap_vt100;
		vt100 = TRUE;
		return;
	}
	else
	if (!strcmp(term,"vt52"))
	{
		term_entry = default_termcap_vt52;
		return;
	}

	if (((long) tgetent(NULL,term)) <= 1L)
	{
		if (((long) tgetent(NULL,"vt52")) <= 1L)
		{
			/* kein Eintrag - Angabe in Konfigurationsdatei untersuchen */
			if (!strcmp(config.termtyp,"vt100"))
			{
				term_entry = default_termcap_vt100;
				vt100 = TRUE;
			}
			else
			if (!strcmp(config.termtyp,"vt52"))
				term_entry = default_termcap_vt52;
			else
			if (!strcmp(config.termtyp,"ansi"))
				term_entry = default_termcap_ansi;
		}
	}
} /* init_termcap */
#endif /* TEXT_FRONTEND */



#if (defined(TEXT_FRONTEND) || defined(BOX)) && !defined(SERVER)
void tprint(const char *cap)
{
	char *x = tgetstr(cap,NULL);
	char *y;

	if (x)
	{
		for (y = x; *y; y++)
			putchar(*y);
		fflush(stdout);
		free(x);
	}
}



/*
	Cursorposition anfahren
*/
void gotoxy(int x,int y)
{
#ifdef TEXT_FRONTEND
	static char *cm = NULL;

	if (!cm)
		cm = tgetstr("cm",NULL);
	if (!cm)
		return;

	printf("%s",tgoto(cm,x,y));
#endif
#ifdef BOX
	mprintf("%s",vt_pos_S(y+1,x));
#endif
} /* gotoxy */

#endif
