/*
 * QUARKOUTFILE - beseitigt einige Unschoenheiten in Quark Outfiles.
 * Copyright (C) 1993,1994,1995,1996  Uwe Ohse
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * For questions contact:
 * e-mail: (prefered)
 * uwe@tirka.gun.de
 * Uwe_Ohse@mausdu3.gun.de
 * Uwe Ohse @ DU3 (inside MausNet)
 */

/*
 * quarkoutfile - ein Hack, um die durch die Quarkdatenbank bedingten
 *
 * Hacker's Guide: Dieses Programm ist ein Hack, entstanden aus einer Laune,
 * geschrieben von jemandem, der 100%ig wusste, was im Quarktausch so
 * ablaeuft.
 */

#include "tomsysdep.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <limits.h>
#include <errno.h>
#include <ctype.h>
#include <getopt.h>

#include "long-options.h"
#include "quarkoutfile.h"
#include "libquark.h"
#include "dynhash.h"
#include "crc32.h"


/* dies ist problematisch fr stdin/stdout auf msloss. */
#define CRLF "\r\n"

#ifndef HAVE_TIMEZONE_VER
/* MET */
#define timezone 3600 
#endif

#ifdef HAVE_MMAP
#  include <sys/types.h>
#  include <sys/mman.h>
#  include <sys/stat.h>
#  include <fcntl.h>
#  include <unistd.h>
#endif

/* Laenge einer Headerzeile - reicht auch fuer MAUS 9 */
#define HEADLEN	512

/* maximale Laenge einer Textzeile - das reicht allemale */
#define LINESIZE 32767

/* fuer klinisch tote Systeme ohne verlaessliches argc/argv */
const char *program_name = "quarkoutfile";

#ifdef HAVE_MMAP
caddr_t mmaddr;
size_t mmsize;
int is_mmapped;
#endif

/* Liste zur Unterbringung der Kopfzeilen */
typedef struct headerline_t {
	unsigned char *line;
	struct headerline_t *next;
} headerline_t;

headerline_t *kanker;				/* Pointer auf erste Kopfzeile */
headerline_t *klast;				/* Pointer auf letzte Kopfzeile */
/* unbekannte Headerzeilen */
headerline_t *uanker;
headerline_t *ulast;

/* 
 * Alle chars sind unsigned, weil ein paar daemliche C-Librarys meinen,
 * sie duerften sonst Zeichen wegwerfen oder ruinieren. Das betraf 
 * auch (aeltere) Linux Libs.
 */

static unsigned char l_mausid[HEADLEN];		/* # */
static unsigned char l_mausref[HEADLEN];	/* - */
static unsigned char l_from[HEADLEN];	/* V */
static unsigned char l_to[HEADLEN];		/* A */
static unsigned char l_subject[HEADLEN];	/* W */
static unsigned char l_date[HEADLEN];	/* E */
static unsigned char l_groups[HEADLEN];		/* G */
static unsigned char l_status[HEADLEN];		/* B */
static unsigned char l_realref[HEADLEN];	/* R */
static unsigned char l_organization[HEADLEN];	/* O */
static unsigned char l_name[HEADLEN];	/* N */
static unsigned char l_distribution[HEADLEN];	/* D */
static unsigned char l_realid[HEADLEN];		/* I */
static unsigned char l_gateway[HEADLEN];	/* Y */
static unsigned char l_followup_to[HEADLEN];	/* F */
static unsigned char l_reply_to[HEADLEN];	/* T */
static unsigned char l_sender[HEADLEN];	/* S */
static unsigned char l_mime[HEADLEN];	/* M */
static unsigned char l_xflags[HEADLEN];	/* X */

#define NO_KOPF 0
#define SAVE_KOPF 1
#define STRIP_KOPF 2
static int savekopf_flag;	/* >-Zeilen erzeugen? */

static short showdots = 0;		/* Verlaufsanzeige */
static FILE *dotsfile;			/* Punkte auf stdout (Default) */
static unsigned int columns = 80;	/* Zeichen pro BS-Zeile */
static BOOLEAN gconv2long;		/* Gruppennamen Kurz->Lang */
static BOOLEAN gconv2short;		/* Gruppennamen Lang->Kurz */
static BOOLEAN long2short_ids;	/* Lange IDs in Kurze wandeln */
static BOOLEAN short2long_ids;	/* und kurze in lange */
static char *flame_mausprogrammierer;
static char *bouncefile;
const char *mailer_daemon = NULL;
const char *sysop_adresse = NULL;
const char *mysite = NULL;
char *hide_mouse = NULL;

	/* Nachrichten mit nichtkonvertierbaren IDs wegwerfen? */
static BOOLEAN discard_false_ids;
static FILE *probsfile;								/* Problemmeldungen in welches File? */
static unsigned char *filter_lines = NULL;	/* Zeilen mit diesen Kennungen weg */
static unsigned char *filter_status = NULL;	/* diese Status wegwerfen */
static BOOLEAN filter_b_in_pms;	/* B-Zeilen in PMs wegwerfen */
static BOOLEAN undomainize = FALSE;		/* MAUS/Gatewaybug bei Domainadressen */

void *id_handle = NULL;
void *reimport_handle = NULL;

/*
 * Nun doch noch: Prototypen.
 */

void clean_header (void);
BOOLEAN write_header (FILE * f, BOOLEAN was_ww);
void savekopf (unsigned char *s);
unsigned char *nextline (FILE * in);
void quarkoutfile (FILE * in, FILE * out);
void usage (int);
void saveunknown (unsigned char *s);
short convert_long_maus_id (char *shortidline, char *longidline);

static 
BOOLEAN is_importbox(const char *p)
{
	DYN_ENTRY such;
	DYN_ENTRY *e;

	such.crc = crc32ansi (p, strlen (p));
	such.u.nr = 0;
	e = hhsearch (reimport_handle, &such, DYN_FIND);
	return (e!=NULL);
}

void
clean_header (void)
{
	headerline_t *p;
	headerline_t *pnext;

	memset (l_mausid, 0, HEADLEN);
	memset (l_mausref, 0, HEADLEN);
	memset (l_from, 0, HEADLEN);
	memset (l_to, 0, HEADLEN);
	memset (l_subject, 0, HEADLEN);
	memset (l_date, 0, HEADLEN);
	memset (l_groups, 0, HEADLEN);
	memset (l_status, 0, HEADLEN);
	memset (l_realid, 0, HEADLEN);
	memset (l_realref, 0, HEADLEN);
	memset (l_organization, 0, HEADLEN);
	memset (l_name, 0, HEADLEN);
	memset (l_distribution, 0, HEADLEN);
	memset (l_gateway, 0, HEADLEN);
	memset (l_followup_to, 0, HEADLEN);
	memset (l_reply_to, 0, HEADLEN);
	memset (l_sender, 0, HEADLEN);
	memset (l_mime, 0, HEADLEN);
	memset (l_xflags, 0, HEADLEN);

	p = kanker;
	while (p) {
		pnext = p->next;
		free (p->line);
		free (p);
		p = pnext;
	}
	kanker = NULL;
	klast = NULL;

	p = uanker;
	while (p) {
		pnext = p->next;
		free (p->line);
		free (p);
		p = pnext;
	}
	uanker = NULL;
	ulast = NULL;
}

#define MAUSDE ".maus.de"
#define MAUSDE_LEN (sizeof(MAUSDE)-1)

/* Lange Message-IDs in den #- und Minuszeilen konvertieren in kurze 
 * in den #- und Minuszeilen sowie lange in den I/R-Zeilen.
 * 
 * Die Routine ist noch hlicher als ntig, weil sie auch mit 
 * Gnter Stcks kaputten PM-IDs umgehen kann.
 */
short
convert_long_maus_id (char *shortidline, char *longidline)
{
	char *mausde;
	char *at;
	char *dot;
	char *p;
	char c;

	at = strchr (shortidline, '@');
	if (!at)
		return 1;				/* Spezialnachricht o.ae. */

	if (longidline)
		strcpy (longidline, shortidline);

	dot = strchr (shortidline, '.');
	if (dot && dot < at)
	{
		/* xxx.yyy@zzz */
		p = shortidline;
		while (*p) {
			if (p==dot)
				break;
			if (!isdigit(*p))
				return 0;
			p++;
		}
		/* Prfen, ob nach dem Buchstaben nur noch Nummern kommen */
		c = toupper(dot[1]);
		if (!isupper (c))
			return 0;
		p=dot+2;
		while (isdigit(*p))
			p++;
		if (*p!='@')
			return 0;
	}
	else if (!dot)
	{
		/* das ist in jedem Fall nichts, was wir bersetzen knnen */
		return 0;
	}
	else
	{
		dot=NULL;
		/* Gnter Stcks ID-Form fr persnliche Nachrichten */
		/* xxxxx@yyy.zzz */
		c = shortidline[0];
		if (c!='P')
			return 0;
		c = toupper (c);
		/* Prfen, ob nach dem P nur noch Nummern kommen */
		p=shortidline+1;
		while (isdigit(*p))
			p++;
		if (*p!='@')
			return 0;
	}

	/*
	 * blah@du3.maus.de
	 * blah@du3.maus.de:2
	 */
	mausde = shortidline + strlen (shortidline) - MAUSDE_LEN;
	if (strcasecmp (mausde, MAUSDE)) {
		/* Knnte gesplittete Nachricht sein */
		char *q = strchr (mausde, ':');
		if (!q || ((size_t)(q - shortidline)) < MAUSDE_LEN)
			return 0;
		q -= MAUSDE_LEN;
		if (strncasecmp (q, MAUSDE, MAUSDE_LEN))
			return 0;
		mausde = q;
	}
	/* .maus.de muss mindestens zwei Zeichen nach dem @ kommen */
	if (mausde < at + 2)
		return 0;

	/* 
	 * Das ist eine lange MAUS-ID. Sie hat das Format:
	 *                 P10693@du.maus.de  GST-Format 
	 *    199406250722.a10693@du.maus.de
	 *                ^dot   ^at^mausde
	 * Zuerst mal in die Zeile fuer die lange ID kopieren, falls 
	 * erwuenscht. Sinnvollerweise wuerde das *immer* gemacht, aber
	 * MAUS laesst Fremdboxen zur Zeit keine I-Zeilen schicken.
	 */
	*mausde = '\0';				/* .maus.de weg */
	/*
	 * prfen ob das hinter dem @ wirklich ein Mauskrzel sein kann.
	 */
	if (!isalpha(at[1]))
		return 0;
	for (p=at+2;*p;p++)
	{
		if (!isalnum(*p) && *p!='-')
			return 0;
	}

	/* ->              P10693@du GST-Format */
	/* -> 199406250722.a10693@du */

	p = at;
	while (*p)
		*p++ = toupper (*p);

	/* ->              P10693@DU GST-Format */
	/* -> 199406250722.a10693@DU */

	if (dot)
		*dot = '\0';

	/* ->               P10693@DU GST-Format */
	/* -> 199406250722\0a10693@DU */

	*shortidline = c;
	/* ->               P10693@DU GST-Format */
	/* -> A99406250722.a10693@DU */

	if (dot)
		memmove (shortidline + 1, dot + 2, strlen (dot + 2) + 1);

	/* -> A10693@DU */

	return TRUE;				/* gelungen */
}

/*
 * Schreibt die Headerdaten in das File.
 * liefert TRUE, falls die Nachricht weiterhin geschrieben werden,
 * und FALSE, wenn sie Richtung /dev/null wandern soll.
 */
BOOLEAN
write_header (FILE * f, BOOLEAN was_ww)
{
	/* Falls keine lange ID vorhanden ist und die Nachricht aus dem
	 * MausNet ist, dann eine lange ID erzeugen.
	 */
	int is_status = 0;

	if (!l_realid[0] && l_date[0] && strchr (l_mausid, '@')) {
		char *p;

		p = strchr (l_mausid, '@');
		if (p && !strchr (p, '.')) {
			sprintf (l_realid, "%s.%s.maus.de", l_date, l_mausid);
			p = l_realid;
			while (*p)
				*p++ = tolower (*p);
		}
	}
	if (short2long_ids) {
		static char last_id[HEADLEN];
		unsigned char *ap = strchr (l_realid, '@');

		if (ap)
			ap = strrchr (ap, ':');
		if (ap) {
			if (!strncmp (last_id, l_realid, ap - l_realid))
				return TRUE;
		}
		strcpy (last_id, l_realid);
		ap=strrchr(l_realid,':');
		if (ap && strcmp(ap,":1")==0) {
			/* erwinsgate-Fehler */
			*ap=0;
		}
	}
#define OUTPUT(x,y) do { \
	if (y[0] && (!filter_lines || !strchr(filter_lines, x))) \
		fprintf(f,"%c%s"CRLF,x,y); \
	} while (0)
#define KLUDGE(x,c,y) do { \
	if (y[0] && (!filter_lines || !strchr(filter_lines, c))) \
		fprintf(f,":%s: %s"CRLF,x,y); \
	} while (0)

	if (long2short_ids) {
		/* Zur Vergabe temporrer IDs */
		static int lfd_nr = 0;
		unsigned char real_id_copy[HEADLEN];
		*real_id_copy = '\0';

		if (lfd_nr == 0)
			lfd_nr = time (NULL);
		/* 
		 * Lange IDs in kurze wandeln: Ein primitiver USE2MAUS, der
		 * mit brutaler Gewalt aus langen IDs in der "#"-Zeile kurze
		 * macht und die lange in die I-Zeile kopiert.
		 */
		if (!convert_long_maus_id (l_mausid, l_realid)) {
			grpconv_t *gc;
			time_t ti = time (NULL);

			/* konnte nicht konvertiert werden? */
			fprintf (probsfile, "%s: %s", l_mausid, ctime (&ti));
			fprintf (probsfile, "  kann ID nicht in kurze konvertieren\n");
			/* Statusmeldungen ... */
			if (!*l_groups && !*l_from && *l_status) {
				return FALSE;
			}
			if (discard_false_ids && *l_groups
				&& ((gc = find_shortname (l_groups)) == NULL || !gc->import_flag)) {
				fprintf (probsfile, "  discarded, Gruppe %s\n", l_groups);
				return FALSE;
			}
			if (*l_groups)
				fprintf (probsfile, "  not discarded, Gruppe %s\n", l_groups);
			else
				fprintf (probsfile, "  not discarded, Empfnger %s\n", l_to);
			/* damit 20 Zeilen tiefer eine temporre ID eingesetzt wird */
			*l_mausid=0;
		} else if (*l_groups && reimport_handle) {
			/* MAUS-Reimport-Probleme ... */

			/* kurze MAUS-ID -> entweder wir, oder Boxen unter uns, oder 
			   reimport in import-Gruppen */
			/* "wir" ist einfach, "Boxen unter uns" ist nur einfach wenn 
			 * man von einem sternfrmigen Netz ausgeht, was natrlich 
			 * ziemlich unsinnig ist */
			grpconv_t *gc;
			time_t ti = time (NULL);
			/* ID-Analyse */
			char *p=strchr(l_mausid,'@');
			if (!p) {
				fprintf(probsfile,"%s/%s: %s no @ in mausid",
					l_mausid,l_realid, ctime(&ti));
				return FALSE;
			}
			p++; /* == boxkuerzel */

			if (!is_importbox(p)
				&& ((gc = find_shortname (l_groups)) == NULL || !gc->import_flag)) {
				fprintf (probsfile, "%s: %s", l_mausid, ctime (&ti));
				fprintf (probsfile, "  keine Importbox, keine Importgruppe %s\n",
					l_groups);
				return FALSE;
			}
		}
		/*
		 * Alles, was hier ankommt, soll importiert werden.
		 */

		/* Statusmeldungen? */
		if (!*l_groups && !*l_from && *l_status) {
			strcpy (real_id_copy, l_realid);
			*l_realid = 0;
			l_status[1] = 0;
			is_status = 1;
			if (filter_status && strchr(filter_status,*l_status))
				return FALSE;
		}
		/*
		 * wenn keine MAUS-ID vorhanden ist, dann eine temporre ID
		 * einsetzen.
		 */
		else if (!*l_mausid) {
			sprintf (l_mausid, "%d", lfd_nr++);
		}
		if (*l_mausref && !convert_long_maus_id (l_mausref, l_realref)) {
			/* Das ist normal */
			strcpy (l_realref, l_mausref);
			*l_mausref = '\0';
		}
		/*
		 * MAUS kann nicht mehrere Nachrichten mit derselben ID 
		 * transportieren.
		 */
		if (strchr (l_mausid, '@') && !is_status) {
			DYN_ENTRY such;
			DYN_ENTRY *e;
			char *q;
			q = strrchr (l_realid, ':');
			if (q && !strchr (q, '.'))
				*q = '\0';
			else
				q = NULL;
			such.crc = crc32ansi (l_realid, strlen (l_realid));
			if (q)
				*q = ':';
			such.u.nr = 0;
			e = hhsearch (id_handle, &such, DYN_FIND);
			if (e)
				/* die hatten wir schon mal ... */
				sprintf (l_mausid, "%d", lfd_nr++);
			else				/* ID als schon exportiert merken */
				hhsearch (id_handle, &such, DYN_ENTER);
		} else {				/* ID als schon exportiert merken */
			DYN_ENTRY such;
			such.crc = crc32ansi (l_realid, strlen (l_realid));
			such.u.nr = 0;
			hhsearch (id_handle, &such, DYN_ENTER);
		}
		OUTPUT ('#', l_mausid);
		OUTPUT ('I', l_realid);
		if (bouncefile)
			add_bounceinfo (l_mausid,
							(*l_realid) ? l_realid : real_id_copy,
							l_groups, l_to, l_from);
		OUTPUT ('-', l_mausref);
		OUTPUT ('R', l_realref);
		OUTPUT ('E', l_date);
	} else if (short2long_ids && strchr (l_mausid, '@')) {
		/* 
		 * Kurze IDs verwerfen, um lange einzusetzen.
		 * Eine lange ID haben wir schon!
		 */
		if (!strstr (l_realid, MAUSDE) && !*l_gateway) {
			char *p = strchr (l_mausid, '@');

			sprintf (l_gateway, "Gateway @ %s", p + 1);
		}
		strcpy (l_mausid, l_realid);
		*l_realid = '\0';
		strcpy (l_mausref, l_realref);
		*l_realref = '\0';
		OUTPUT ('#', l_mausid);
		OUTPUT ('-', l_mausref);
		OUTPUT ('E', l_date);
	} else {
		OUTPUT ('#', l_mausid);
		OUTPUT ('I', l_realid);
		OUTPUT ('-', l_mausref);
		OUTPUT ('R', l_realref);
		OUTPUT ('E', l_date);
	}

	OUTPUT ('D', l_distribution);
	OUTPUT ('Y', l_gateway);
	OUTPUT ('O', l_organization);
	OUTPUT ('M', l_mime);
	OUTPUT ('X', l_xflags);
	if (undomainize) {
		char *at;

		at = strchr (l_from, '@');
		if (at) {
			const char *dom;
			char *p = at + 1;

			while (isspace (*p))
				p++;
			dom = get_boxkuerzel (p);
			if (dom)
				strcpy (at + 1, dom);
		}
		at = strchr (l_to, '@');
		if (at) {
			const char *dom;
			char *p = at + 1;

			while (isspace (*p))
				p++;
			dom = get_boxkuerzel (p);
			if (dom)
				strcpy (at + 1, dom);
		}
	}
	OUTPUT ('V', l_from);

	OUTPUT ('N', l_name);
	OUTPUT ('A', l_to);
	OUTPUT ('W', l_subject);
	OUTPUT ('F', l_followup_to);
	OUTPUT ('T', l_reply_to);
	OUTPUT ('S', l_sender);
	OUTPUT ('G', l_groups);
	if (!filter_b_in_pms || !*l_to)
		OUTPUT ('B', l_status);
	if (!*l_groups && *l_to 
		&& (!filter_lines || !strchr(filter_lines,'U'))) {
		char buf1[128];
		struct tm *tm;
		time_t t;
		char c = timezone > 0 ? '+' : '-';
		long t2=timezone > 0 ? timezone : -timezone;
		t=time(NULL);
		tm=localtime(&t);
		
		strftime(buf1,sizeof(buf1),"%d %b %Y %H:%M:%S",tm);
		fprintf(f,
				"UUReceived: by %s (quarkoutfile) via mausnet "
				"with maustausch for %s; %s %c%02d%02d" CRLF,
				mysite, l_to, buf1, c,
				(int) (t2/3600), 
				(int) ((t2/60) % 60));
	}
	if (uanker) {
		headerline_t *p;

		p = uanker;
		while (p) {
			char c=p->line[0];
			char *s=p->line+1;
			OUTPUT (c,s);
			p = p->next;
		}
	}
	if (savekopf_flag && kanker) {
		headerline_t *p;

		p = kanker;
		while (p) {
			OUTPUT ('>', p->line);
			p = p->next;
		}
		fprintf (f, ">" CRLF);
	}
	if (was_ww && (*l_mausref || *l_realref)) {
		fprintf (f, ":Kommentar zu %s" CRLF,
				 l_mausref ? l_mausref : l_realref);
	}
#undef OUTPUT
#undef KLUDGE
	return TRUE;
}

/*
 * unbekannte Zeile sichern
 */
void
saveunknown (unsigned char *s)
{
	headerline_t *p;
	/* Keine leeren Kopfzeilen */
	if (!*s)
		return;

	p = XMALLOC (sizeof (headerline_t));
	p->line=XSTRDUP(s);

	p->next = NULL;
	if (!ulast) {
		uanker = p;
		ulast = p;
	} else {
		ulast->next = p;
		ulast = p;
	}
}

/*
 * Kopfzeile sichern
 */
void
savekopf (unsigned char *s)
{
	headerline_t *p;
	/* Keine leeren Kopfzeilen */
	if (!*s)
		return;

	p = XMALLOC (sizeof (headerline_t));
	p->line=XSTRDUP(s);

	p->next = NULL;
	if (!klast) {
		kanker = p;
		klast = p;
	} else {
		klast->next = p;
		klast = p;
	}
}

unsigned char *inputbuffer;

unsigned char *
nextline (FILE * in)
{
	unsigned char *p;

#ifdef HAVE_MMAP
	if (is_mmapped) {
		unsigned char *ret;

		ret = inputbuffer;
		if (!ret)
			return NULL;
	rescan:
		p = strpbrk (inputbuffer, "\r\n");
		if (!p) {
			/* Moeglicherweise Muell in der Zeile */
			/* ASCII-Null *in* der Zeile kommt sehr, sehr schlecht */
			while (1) {
				p = inputbuffer + strlen (inputbuffer);
				if (p != (unsigned char *) (mmaddr + mmsize - 1))
					*p = ' ';
				else
					break;
			}
			p = strpbrk (inputbuffer, "\r\n");
		}
		if (p) {
			if (*p == '\r') {
				if (p[1]=='\n') {
					*p++ = '\0';
					*p++ = '\0';
				} else {
					*p=' ';
					goto rescan;
				}
			} else
				*p++ = '\0';
			if (p <= ((unsigned char *) mmaddr + mmsize - 1))
				inputbuffer = p;
			else
				inputbuffer = NULL;
		} else
			inputbuffer = NULL;
		return ret;
	}
#endif
	if (!fgets (inputbuffer, LINESIZE, in))
		return NULL;
	p=inputbuffer+strlen(inputbuffer)-1;
	if (*p=='\n') {
		*p--=0;
		if (*p=='\r')
			*p--=0;
	}
	return inputbuffer;
}

/*
 *    Die Funktion, die die ganze Arbeit erledigt
 */
void
quarkoutfile (FILE * in, FILE * out)
{
	unsigned char *line;
	BOOLEAN is_header;			/* Sind wir im Header einer Nachricht? */
	enum {
		KEIN_SPECIAL,			/* in normaler Nachricht */
		SOME_SPECIAL,			/* in irgendeiner Spezialnachricht */
		ITG,					/* im ITG */
		ITS,					/* im ITS */
		HEAD,					/* im HEAD */
		LOG						/* im LOG */
	} in_special = KEIN_SPECIAL;
	int logcmdwritten;			/* :#CMD in #LOG eingefuegt? */
	int killdoublePdoubleC;		/* naechste :#-Zeile loeschen */
	int nachrichtennr;			/* Nummer der aktuellen Nachricht */
	BOOLEAN work_on_it = TRUE;	/* Diese Nachricht sichern? */
	BOOLEAN got_greater = FALSE;
	BOOLEAN got_ww = FALSE;

#ifdef HAVE_MMAP
	if (is_mmapped)
		inputbuffer = (unsigned char *) mmaddr;
	else
#endif
		inputbuffer = XMALLOC (LINESIZE);

	clean_header ();
	is_header = FALSE;
	logcmdwritten = 0;
	killdoublePdoubleC = 0;
	nachrichtennr = 0;

	/* Solange es noch Daten zu lesen gibt */
	while ((line = nextline (in)) != NULL) {
		if (!work_on_it) {
			if (*line != '#')
				continue;
			work_on_it = TRUE;
		}
		/* Sind wir in einer Spezialnachricht (LOG oder Ixx)? */
		if (in_special != KEIN_SPECIAL && line[0]!='#') {
			if (in_special == ITS && hide_mouse) {
				static BOOLEAN as_fremdbox=FALSE;
				if (line[1]=='*') {
					as_fremdbox=mask_as_fremdbox(line+2);
					if (as_fremdbox)
						fprintf(out,"#CNF" CRLF);
				}
				if (line[1]=='#' && as_fremdbox) {
					if (line[2]!='-') {
						fprintf(out,":#-%s" CRLF,line+2);
						continue;
					}
				}
				if (as_fremdbox)
					fprintf(out,"%s" CRLF,line);
				continue;
			}
			/* zur Konvertierung der
			 * IDs im LOG. Das gestaltet sich schwieriger, da die
			 * Quark 1.x die langen IDs nicht im #LOG zurueckmeldet
			 * (da sie sie auch gar nicht zur MAUS schicken darf und
			 * von daher nicht weiss, welche ID die MAUS tatsaechlich
			 * dafuer einsetzt).
			 */
			if (in_special == LOG && short2long_ids) {
				static char last_gleich[HEADLEN] = "";
				static char last_id[HEADLEN] = "";
				char c = line[1];

				if (c == '#') {
					strcpy (last_id, line + 2);
				} else if (c == '"' || c == '$')
					*last_id = 0;

				/* dann im LOG IDs konvertieren */
				if (c == '"' || c == '#' || c == '$') {
					if (*last_gleich) {
						fprintf (out, "%s" CRLF, last_gleich);
						*last_gleich = '\0';
					}
				}
				if (!strncmp (line, ":=", 2)) {
					/* sicherheitshalber aufbewahren */
					strcpy (last_gleich, line);
					continue;
				}
				if (!strncmp (line, ":!=", 3)) {
					fprintf (out, ":=%s" CRLF, line + 3);
					*last_gleich = '\0';
					continue;
				}
				if (!strncmp (line, ":?", 2) && bouncefile
					&& *last_id) {
					merke_bounce (last_id, line + 2);
					*last_id=0; /* um weitere ?-Zeilen abzufangen */
				}
			}
			/* nun zum Fehlen der #CMD-Bestaetigung in der Quark. 
			 * Braindamage (c) Uwe Ohse.
			 */
			if (in_special == LOG && !logcmdwritten) {
				if (!strcmp (line, ":#CMD")) {
					fprintf (out, "%s" CRLF, line);
					logcmdwritten = 1;
					continue;
				}
				if (line[1] == '\"' || line[1] == '$') {
					fprintf (out, ":#CMD" CRLF);
					fprintf (out, ":#" CRLF);
					fprintf (out, "%s" CRLF, line);
					logcmdwritten = 1;
					killdoublePdoubleC = 1;		/* muss :# wegwerfen */
					continue;
				}
			}
			if (in_special == LOG && killdoublePdoubleC &&
				!strcmp (line, ":#")) {
				/* genau diese Zeile loeschen! */
				continue;
			}
			if (in_special == ITG && (gconv2short || gconv2long)) {
				/* Leerzeichen am Ende der G-Zeilen loeschen */
				if (*line == ':' && line[1] == 'G') {
					/* da war mal ein Bug in der Quark */
					strtrim (line);

					if (gconv2short) {
						grpconv_t *gc;

						gc = find_longname (line + 1);
						if (gc) {
							fprintf (out, ":G%s" CRLF, gc->shortname);
							continue;
						}
					} else if (gconv2long) {
						grpconv_t *gc;

						gc = find_shortname (line + 1);
						if (gc) {
							fprintf (out, ":G%s" CRLF, gc->longname);
							continue;
						}
					}
				}
				fprintf (out, "%s" CRLF, line);
				continue;
			}
			fprintf (out, "%s" CRLF, line);
			continue;
		}
		/*
		 *    Der eigentliche Parser:
		 */
		switch (line[0]) {
		case '#':				/* ID-Zeile, Beginn einer neuen Nachricht */
			/*
			 *    Was muss getan werden?
			 *  - Wenn der Header noch nicht gesichert wurde (was definitiv
			 *    bei Statusnachrichten so ist), dann den Header wegschreiben.
			 *  - die Headerdaten muessen geloescht werden.
			 *  - Bei Spezialnachrichten (Ixx,Jxx,LOG etc) Kopiermodus
			 *    einschalten.
			 *  - Wir muessen uns merken, dass wir wieder im Header sind.
			 *  - und natuerlich die ID kopieren!
			 */

			/* Falls noch im Header (warum auch immer), dann Header sichern */
			if (is_header && in_special==KEIN_SPECIAL)
				(void) write_header (out, got_ww);
			work_on_it = TRUE;
			/* alte Headerdaten loeschen */
			clean_header ();

			in_special = KEIN_SPECIAL;
			got_greater = FALSE;
			got_ww = FALSE;

			/* Verlaufsanzeige */
			if (showdots && nachrichtennr % showdots == 0) {
				static int dotcount = 0;

				fputc ('.', dotsfile);
				if (++dotcount % columns == 0)
					fprintf (dotsfile, CRLF);
				fflush (dotsfile);
			}
			/* wenn die ID keinen @ enthaelt */
			if (!strchr (line + 1, '@')) {
				/*
				 *    dann ist es eine Spezialmitteilung
				 */
				in_special = SOME_SPECIAL;
				if (!strcmp (line, "#ITG"))
					in_special = ITG;
				if (!strcmp (line, "#HEAD"))
					in_special = HEAD;
				if (!strcmp (line, "#LOG"))
					in_special = LOG;
				if (!strcmp (line, "#ITS"))
					in_special = ITS;
				killdoublePdoubleC = 0;
				logcmdwritten = 0;
				/* die "#"-Zeile (Fileende) nicht 
				 * ausgeben - vorher mssen wir noch
				 * die Bounces erzeugen 
				 */
				if (line[1])
					fprintf (out, "%s" CRLF, line);
				continue;
			}
			/* ID kopieren */
			strcpy (l_mausid, line + 1);
			/* wir sind im Header einer Nachricht */
			is_header = TRUE;
			break;
		case '-':				/* Verkettungs-ID */
			strcpy (l_mausref, line + 1);
			break;
		case 'V':				/* Von */
			strcpy (l_from, line + 1);
			break;
		case 'A':				/* an */
			strcpy (l_to, line + 1);
			break;
		case 'Y':				/* gateway */
			strcpy (l_gateway, line + 1);
			break;
		case 'W':				/* Wegen */
			strcpy (l_subject, line + 1);
			break;
		case 'E':				/* Eingabezeit */
			strcpy (l_date, line + 1);
			break;
		case 'G':				/* Gruppe */
			/*
			 *  Die Gruppennamen bei Quarks sind etwas laenger (16 Zeichen),
			 *  und das bringt einige Frontends in leichte Probleme (z.B.
			 *  MAUTAU (considered dead) und auch fruehere Versionen von
			 *  Catputz 2.0x (heutige sind clean).
			 */
			if (gconv2long) {
				grpconv_t *gc;

				gc = find_shortname (line + 1);
				if (gc)
					strcpy (l_groups, gc->longname);
				else
					strcpy (l_groups, line + 1);
			} else if (gconv2short) {
				grpconv_t *gc;

				gc = find_longname (line + 1);
				if (gc)
					strcpy (l_groups, gc->shortname);
				else
					strcpy (l_groups, line + 1);
			} else
				strcpy (l_groups, line + 1);
			break;
		case 'B':				/* Bearbeitungsstatus */
			strcpy (l_status, line + 1);
			break;
		case 'I':				/* ID - die echte ID aus Fremdnetzen */
			strcpy (l_realid, line + 1);
			break;
		case 'R':				/* echte References aus Fremdnetzen */
			strcpy (l_realref, line + 1);
			break;
		case 'N':				/* Realname aus Fremdnetzen */
			strcpy (l_name, line + 1);
			break;
		case 'O':				/* Organization aus Fremdnetzen */
			strcpy (l_organization, line + 1);
			break;
		case 'M':				/* MIME */
			strcpy (l_mime, line + 1);
			break;
		case 'X':				/* MIME */
			strcpy (l_xflags, line + 1);
			break;
		case 'F':				/* followup-to */
			strcpy (l_followup_to, line + 1);
			break;
		case 'S':				/* sender */
			strcpy (l_sender, line + 1);
			break;
		case 'T':				/* reply-to */
			strcpy (l_reply_to, line + 1);
			break;
		case 'D':				/* Distribution */
			/* D-Zeile kann am Ende einer Nachricht *sehr* ntzlich sein. */
			if (!is_header)
				fprintf (out, "%s" CRLF, line);
			else
				strcpy (l_distribution, line + 1);
			break;
		case '>':				/* Kopfzeilen */
			/*
			 *  Die Quark lieferte diese Zeilen vor Version 1.6 nicht.
			 *  In Quark 1.6 sind sie allerdings nicht brauchbarer als die
			 *  normalen Headerzeilen, deshalb ignoriert Quarkoutfile sie.
			 */
			got_greater = TRUE;
			if (savekopf_flag == STRIP_KOPF) {
				/* >Von : ...
				 * > Von: ... 
				 * 012345
				 */
				if (strlen (line + 1) < 5 || line[5] != ':')
					savekopf (line + 1);
				else {
					char buf[5];
					char *p;

					memcpy (buf, line + 1, 4);
					buf[4] = '\0';
					p = buf + 3;
					while (isspace (*p) && p > buf)
						*p-- = '\0';
					p = buf;
					while (isspace (*p))
						p++;
					if (strcasecmp (p, "An")
						&& strcasecmp (p, "Von")
						&& strcasecmp (p, "MId")
						&& strcasecmp (p, "RId")
						&& strcasecmp (p, "Box")
						&& strcasecmp (p, "Name")
						&& strcasecmp (p, "Wg.")) {
						savekopf (line + 1);
					}
				}
			} else if (savekopf_flag)
				savekopf (line + 1);
			break;
		case ':':				/* Textzeilen */
			if (!is_header) {
				fprintf (out, "%s" CRLF, line);
				break;
			}
			work_on_it = write_header (out, got_ww);
			is_header=FALSE;
			if (work_on_it)
				fprintf (out, "%s" CRLF, line);
			break;
		default:
			saveunknown (line);
			break;
		}						/* switch */
	}							/* while */
	if (showdots && nachrichtennr > showdots)
		fputc ('\n', dotsfile);
	if (bouncefile && short2long_ids)
		generate_bounces (out, flame_mausprogrammierer);
	fprintf (out, "#" CRLF);
}

void
usage (int ecode)
{
	FILE *f;
	if (ecode==0)
		f=stdout;
	else
		f=stderr;
	fprintf (f, "Benutzung: quarkoutfile [Optionen]\n");
	fprintf (f,
"  Erlaubte Optionen:\n"
"    -0, --discard-false-ids News mit IDs !~ /maus.de/ loeschen\n"
"    -b, --bufsize BUFSIZE   Puffer fr Ein- und Ausgabe in KB\n"
"    -B, --bounceinfo FILE   Bounceinformationsdatenbank in FILE\n"
"    -d, --dots MSGS         Einen Punkt pro MSGS Nachrichten ausgeben\n"
"    -D, --undomanize ITS    Langformen d. Adressen d. Boxen in ITS kuerzen\r\n"
"    -f, --filter-lines STR  Mit Zeichen aus STR beginnende Zeilen filtern\n"
"    -F, --flame ADDR        Bei Bounces Report an ADDR senden\n"
"    -g, --groups2short FILE Entsprechend FILE Gruppennamen von lang nach\n"
"                            kurz wandeln (lang Quark, kurz MAUS)\n"
"    -G, --groups2long FILE  Entsprechend FILE Gruppennamen von kurz nach\n"
"                            lang wandeln (kurz MAUS, lang Quark)\n"
"    -h, --help              dieser Text\n"
"    -H, --hide-mouses KURZ  ITS so in etliche CNFs aufsplitten da nur noch\n"
"                            die Boxen nicht unter KURZ brigbleiben und als\n"
"                            `Fremdboxen' angesehen werden\n"
"    -i, --infile INFILE     Input aus INFILE statt von stdin\n"
"    -k, --stripkopf         berflssige Kopfzeilen (>) entfernen\n"
"    -K, --savekopf          Kopfzeilen unverndert durchreichen\n"
"                            ohne -k und -K werden sie weggeworfen\n"
"    -l, --long2short-ids    Lange IDs in kurze wandeln (Quark->MAUS\n"
"    -L, --short2long-ids    Kurze IDs in lange wandeln (MAUS->Quark\n"
"    -m, --mysite SITE       eigenes Boxkrzel angeben (fr Fehlermails)\n"
"    -M, --mailer-daemon X@Z Absender fr Fehler/Bouncemails\n"
"    -o, --outfile OUTFILE   Ausgabe nach OUTFILE schreiben\n"
"    -p, --problemsfile FILE ID-Konvertierungsprobleme nach FILE schreiben\n"
"    -P, --filter-b-in-mail  B-Zeilen in Mails entfernen (Quark -> MAUS)\n"
"    -S, --sysop ADRESSE     Summarischen Report an ADRESSE senden\n"
"    -t, --filter-status STR Alle Statusnachrichten in STR filtern\n"
"    -v, --version           Versionsnummer ausgeben\n"
		);
	exit (ecode);
}

/*****************************************************************************
 * Hauptprogramm: Kommandozeilenauswertung und Dateihandling.
 *****************************************************************************
 */
extern char *optarg;
int
main (int argc, char **argv)
{
	char *in_name = NULL;
	char *out_name = NULL;
	char *grp_name = NULL;
	FILE *in = stdin;			/* Default */
	FILE *out = stdout;			/* Default */

	int default_buffer = 1;		/* Defaultpuffer verwenden */
	long buffersize = 0;		/* Groess des Puffers in K */
	size_t calc_buffersize = 0;	/* berechnete wirkliche Puffergroesse */
	char tmpfname[PATH_MAX];
	char *probsfname = NULL;
	char *itb_name = NULL;
	int in_place = 0;
	char *p;

	struct option opts[] =
	{
		{"discard-false-ids", no_argument, NULL, '0'},
		{"bufsize", required_argument, NULL, 'b'},
		{"bounceinfo", required_argument, NULL, 'B'},
		{"dots", required_argument, NULL, 'd'},
		{"undomainize", required_argument, NULL, 'D'},
		{"filter-lines", required_argument, NULL, 'f'},
		{"flame", required_argument, NULL, 'F'},
		{"groups2long", required_argument, NULL, 'G'},
		{"groups2short", required_argument, NULL, 'g'},
		{"hide-mouses", required_argument, NULL, 'H'},
		{"help", no_argument, NULL, 'h'},
		{"infile", required_argument, NULL, 'i'},
		{"its", required_argument, NULL, 'I'},
		{"itsfile", required_argument, NULL, 'I'},
		{"stripkopf", no_argument, NULL, 'k'},
		{"savekopf", no_argument, NULL, 'K'},
		{"long2short-ids", no_argument, NULL, 'l'},
		{"short2long-ids", no_argument, NULL, 'L'},
		{"mysite", required_argument, NULL, 'm'},
		{"mailer-daemon", required_argument, NULL, 'M'},
		{"outfile", required_argument, NULL, 'o'},
		{"problemfile", required_argument, NULL, 'p'},
		{"filter-b-in-mail", no_argument, NULL, 'P'},
		{"reimport-boxen", required_argument, NULL, 'r'},
		{"sysop", required_argument, NULL, 'S'},
		{"filter-status", required_argument, NULL, 't'},
		{"version", no_argument, NULL, 'v'},
		{NULL, 0, NULL, 0}
	};

	int option_index = 0;

	program_name = strrchr(argv[0],'/');
	if (program_name)
		program_name++;
	else
		program_name=argv[0];

 	parse_long_options (argc, argv, program_name, PACKAGE_VERSION, usage);

	dotsfile = stdout;
	probsfile = stderr;

	while (1) {
		int c;

		c = getopt_long (argc, argv,
						 "0b:B:d:D:f:F:g:G:hH:i:I:kKlLm:M:o:p:Pr:sS:t:Uv",
						 opts, &option_index);
		if (c == -1)
			break;
		switch (c) {
		case '0':
			discard_false_ids = TRUE;
			break;
		case 'b':
			default_buffer = 0;
			buffersize = atol (optarg);
			if (buffersize < 0)
				error(2,0, "Puffergroesse muss > 0 sein");
			break;
		case 'B':
			bouncefile = optarg;
			break;
		case 'd':
			showdots = atol (optarg);
			if (showdots < 0)
				error (2,0,"Msgs/Punkt muss > 0 sein");
			break;
		case 'D':
			undomainize = TRUE;
			itb_name = optarg;
			break;
		case 'f':
			filter_lines = optarg;
			break;
		case 'F':
			flame_mausprogrammierer = optarg;
			break;
		case 'g':
			gconv2long = FALSE;
			gconv2short = TRUE;
			grp_name = optarg;
			break;
		case 'G':
			gconv2long = TRUE;
			gconv2short = FALSE;
			grp_name = optarg;
			break;
		case 'H':
			hide_mouse = optarg;
			break;
		case 'h':
			usage (0);
			break;
		case 'i':
			in_name = optarg;
			break;
		case 'I':
			itb_name = optarg;
			break;
		case 'k':
			savekopf_flag = STRIP_KOPF;
			break;
		case 'K':
			savekopf_flag = SAVE_KOPF;
			break;
		case 'l':
			long2short_ids = TRUE;
			short2long_ids = FALSE;
			id_handle = hhcreate (1000);
			XNULLCHECK (id_handle);
			break;
		case 'L':
			long2short_ids = FALSE;
			short2long_ids = TRUE;
			break;
		case 'M':
			mailer_daemon = optarg;
			break;
		case 'm':
			mysite = optarg;
			break;
		case 'o':
			out_name = optarg;
			break;
		case 'p':
			probsfname = optarg;
			break;
		case 'r':
			/* CSV liste der Boxen, fr die Import erlaubt ist */
			{
				const char *q;
				const char *s=optarg;
				if (!reimport_handle) {
					/* so "-r DU3 -r L2" works */
					reimport_handle = hhcreate (1000);
					XNULLCHECK (reimport_handle);
				}
				while (s) {
					DYN_ENTRY such;
					q=strchr(s,',');
					if (!q)
						such.crc = crc32ansi (s, strlen (s));
					else
						such.crc = crc32ansi (s, q-s);
					such.u.nr = 0;
					hhsearch (reimport_handle, &such, DYN_ENTER);
					if (q)
						s=q+1;
					else
						s=NULL;
				}
			}
			break;
		case 'P':
			filter_b_in_pms = TRUE;
			break;
		case 'S':
			sysop_adresse = optarg;
			break;
		case 't':
			filter_status = optarg;
			break;
		case 'v':
			fprintf (stderr, "%s version %s\n", program_name, VERSION);
			break;				/* kein exit! */
		case '?':
		default:
			error (0,0,"unbekannte Option %s",optarg);
			usage(2);
			break;
		}
	}							/* while */

	if (!mysite) {
		fprintf (stderr, "--mysite not given. bad default used\n");
		mysite = "misconfigured.site";
	}
	if (!mailer_daemon) {
		mailer_daemon = concat ("mailer-daemon@", mysite, NULL);
		XNULLCHECK (mailer_daemon);
	}
	if (!sysop_adresse) {
		sysop_adresse = concat ("sysop@", mysite, NULL);
		XNULLCHECK (sysop_adresse);
	}
	p = getenv ("COLUMNS");
	if (p) {
		columns = strtol (p, NULL, 10);
		if (columns <= 0)
			columns = 80;
	}
	if (grp_name)
		lade_grplist (grp_name);
	if (itb_name)
		lade_boxlist (itb_name, mysite, hide_mouse);
	if (bouncefile && short2long_ids)
		load_bouncefiles (bouncefile);
	else if (bouncefile && long2short_ids)
		create_bouncefile (bouncefile);

	if (in_name && out_name && !strcmp (in_name, out_name) && strcmp (in_name, "-")) {
		in_place = 1;
		/* temporaerdatei im Verzeichnis des Infiles anlegen, das 
		 * spart die Zeit zum Kopieren
		 */
		/* Ach, ist das schoen, wie vielseitig doch die Dateisysteme 
		 * auf der Welt sind. Nur schade, dass noch nirgendwo das 
		 * kleine e als Trennzeichen verwendet wird :-)
		 */
#if defined(__TOS__) || defined(__DOS__) || defined(__OS2__)
		p = strrchr (in_name, '\\');
#else
		p = strrchr (in_name, '/');
#endif
		if (p) {
			memcpy (tmpfname, in_name, p - in_name + 1);
			strcat (tmpfname, "qoutfile.tmp");
		} else
			strcpy (tmpfname, "qoutfile.tmp");
	}
	if (!default_buffer) {
		calc_buffersize = (size_t) buffersize *1024;

		/* Fuer Systeme mit size_t == 2 Byte - absolut braindamaged */
		if (sizeof (size_t) == sizeof (short)) {
			if (buffersize * 1024 > USHRT_MAX) {
				fprintf (stderr,
						 "%s: Puffer zu gross. Verwende Maximum (%ld).\n",
						 program_name, (long) (USHRT_MAX / 1024));
				calc_buffersize = USHRT_MAX / 1024;
			}
		}
	}
	if (in_name && strcmp (in_name, "-")) {
#ifdef HAVE_MMAP
		int fd;
		struct stat st;

		fd = open (in_name, O_RDONLY);
		if (fd < 0)
			error(1,errno, "cannot open %s for reading", in_name);
		if (fstat (fd, &st)) 
			error(1,errno, "cannot fstat %s", in_name);
		mmaddr = mmap (0, st.st_size,
					   PROT_READ | PROT_WRITE, MAP_PRIVATE, fd, 0);
		if (mmaddr == (caddr_t) - 1) {
			error(0,errno, "cannot mmap %s", in_name);
			in = fdopen (fd, "r");
			if (!in)
				error(1,errno, "cannot fdopen %s for reading", in_name);
			mmaddr = (caddr_t) - 1;
		} else {
			mmsize = st.st_size;
			/* dirty trick: String abschliessen mit \0. Ist eigentlich Schmutz,
			 * aber als letztes Zeichen *sollte* immer ein \n kommen, und
			 * darauf kann quarkoutfile auch verzichten.
			 */
			((char *) mmaddr)[mmsize - 1] = '\0';
			close (fd);
			is_mmapped = 1;
		}
#else
		in = fopen (in_name, "rb");
		if (!in)
			error(1,errno, "cannot open %s for reading", in_name);
#endif
	}
#ifdef HAVE_MMAP
	else {
		struct stat st;

		mmaddr = (caddr_t) - 1;
		/* Versuchen wir mal, stdin zu mmappen */
		if (fstat (0, &st) == 0 && st.st_size)
			mmaddr = mmap (0, st.st_size, PROT_READ | PROT_WRITE, MAP_PRIVATE, 0, 0);
		if (mmaddr != (caddr_t) - 1) {
			mmsize = st.st_size;
			/* dirty trick: String abschliessen mit \0. Ist eigentlich Schmutz,
			 * aber als letztes Zeichen *sollte* immer ein \n kommen, und
			 * darauf kann quarkoutfile auch verzichten.
			 */
			((char *) mmaddr)[mmsize - 1] = '\0';
			close (0);
			is_mmapped = 1;
		} else
			mmaddr = (caddr_t) - 1;
	}
#endif

	if (out_name && !in_place && strcmp (out_name, "-")) {
		out = fopen (out_name, "wb");
		if (!out)
			error(1,errno, "cannot open %s for writing", out_name);
	}
	if (probsfname) {
		probsfile = fopen (probsfname, "w");
		if (!probsfile)
			error(1,errno, "cannot open %s for writing", probsfname);
		/* Die Datei zeilenweise puffern. */
		setvbuf (probsfile, NULL, _IOLBF, 0);
	}
	if (in_place) {
		out = fopen (tmpfname, "wb");
		if (!out)
			error(1,errno, "cannot open %s", tmpfname);
	}
	if (!default_buffer) {
		if (buffersize == 0) {
#ifdef HAVE_MMAP
			if (!is_mmapped)
#endif
				setvbuf (in, NULL, _IONBF, 0);	/* ungepuffert */
			setvbuf (out, NULL, _IONBF, 0);		/* ungepuffert */
		} else {
			/*
			 * Linux libc (V 4.5.19) laesst bei setvbuf(f,NULL,...)
			 * fast alles unveraendert. Was soll das?
			 * Immerhin ist das schon besser als das Verhalten von
			 * libc (Version 4.4.4), die hat da naemlich auf ungepuffert
			 * geschaltet ...
			 */
			char *bin = NULL;
			char *bout = malloc (calc_buffersize);

#ifdef HAVE_MMAP
			if (!is_mmapped)
				bin = malloc (calc_buffersize);
#endif

			if (bin && setvbuf (in, bin, _IOFBF, calc_buffersize)) {
				/* Fehlschlag, aber das stoert nicht die Funktion des Tools */
				fprintf (stderr, "%s: couldn't allocate outputbuffer\n",
						 program_name);
			}
			if (bout && setvbuf (out, bout, _IOFBF, calc_buffersize)) {
				/* Fehlschlag, aber das stoert nicht die Funktion des Tools */
				fprintf (stderr, "%s: couldn't allocate outputbuffer\n",
						 program_name);
			}
		}
	}
	if (showdots && out == stdout) {
		dotsfile = stderr;
	}
	quarkoutfile (in, out);
	if (long2short_ids)
		close_bouncefile ();
#ifdef HAVE_MMAP
	if (is_mmapped)
		munmap (mmaddr, mmsize);
#endif

	if (grp_name)
		speichere_grplist (grp_name);


	if (in_place) {
		/* GEMDOS, langsam und dumm, kann geoeffnete Dateien nicht
		 * fehlerfrei umbenennen 
		 */
		fclose (out);
		if (rename (tmpfname, in_name)) {
			/* ? Warum hat das nicht geklappt? */
			error(1,errno,"kann temporaerdatei nicht umbenennen");
		}
	}
	return (0);					/* return, not exit, for pure's sake */
}
