/* localroute.c : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
       Snail Mail (don't expect me to answer):
             Uwe Ohse
             Drosselstrae 2
             47055 Duisburg
             Germany
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
       (might be the wrong person except for offline)

    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/

/***************************************************************************
 * lokale Router fr Bangpath, Prozent-Hack, Aliasse, Vertreter und 
 * CHEF Gruppe sowie natrlich Usernamen.
 ***************************************************************************
 */

#include "config.h"

#include "portab.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <syslog.h>

#include "prototyp.h"
#include "user.h"
#include "brett.h"
#include "tompaths.h"
#include "box_conf.h"
#include "routing.h"
#include "route_int.h"
#include "address_hash.h"

static int r_einsetzen (r_adr_t * adr, const char *neu_adresse, r_a_work_t * work);
static int
r_einsetzen (r_adr_t * adr, const char *neu_adresse, r_a_work_t * work)
{
	char *p;

   /*
    * cleanup zuerst - was da auch immer drin steht, es wird nun 
    * nicht mehr bentigt.
    */
	if (work->local)
	{
		free (work->local);
		work->local = NULL;
	}

	if (work->domain)
	{
		free (work->domain);
		work->domain = NULL;
	}

	p = strchr (neu_adresse, '@');
	if (p)
	{
	   /* 
	    * user@kiste
	    * ^neu_adresse
	    *     ^p
	    */
		char *s;

		s = strdup (neu_adresse);
		if (!s)
		{
			R_PRINTF (0, ("r_einsetzen: out of memory\r\n"));
			adr->errorflags |= R_E_TEMP;
			adr->errorflags |= R_E_INTERN;
			adr->dsn=r_dsn_out_of_memory;
			return R_A_ERROR;
		}

		p = strchr (s, '@');
		*p++ = 0;
		strtrim (s);
		strtrim (p);

		work->local = strdup (s);
		if (!work->local)
		{
			free (s);
			R_PRINTF (0, ("r_einsetzen: out of memory\r\n"));
			adr->errorflags |= R_E_TEMP;
			adr->errorflags |= R_E_INTERN;
			adr->dsn=r_dsn_out_of_memory;
			return R_A_ERROR;
		}

		work->domain = strdup (p);
		free (s);
		if (!work->domain)
		{
			R_PRINTF (0, ("r_einsetzen: out of memory\r\n"));
			adr->errorflags |= R_E_TEMP;
			adr->errorflags |= R_E_INTERN;
			adr->dsn=r_dsn_out_of_memory;
			return R_A_ERROR;
		}
	}
	else
	{
		work->local = strdup (neu_adresse);
		if (!work->local)
		{
			R_PRINTF (0, ("r_einsetzen: out of memory\r\n"));
			adr->errorflags |= R_E_TEMP;
			adr->errorflags |= R_E_INTERN;
			adr->dsn=r_dsn_out_of_memory;
			return R_A_ERROR;
		}
		strtrim (work->local);
	}
	return R_A_DONE;
}

int
r_user_aufloeser (r_adr_t * adr, r_a_work_t * work)
{
	USERNR unr=0,*unra;
	USER u;
	int ret;
	BOOLEAN real = FALSE;
	const char *uname = work->local;

	if (!strnicmp (uname, "real-", 5))
	{
		real = TRUE;
		uname += 5;
	}

	if (isdigit(*uname) && strstr(uname,"-expire-")) {
		int r;
		const char *err_d=NULL; /* make compiler happy */
		const char *err_e=NULL; /* make compiler happy */
		BOOLEAN do_err=FALSE;
		r=address_hash_parse(uname,&unr);
		switch (r) {
		case 0: /* matchte nicht ... */
			break;
		case 1: /* aha, gefunden */
			/* unr ist nun != 0 */
			break;
		case -1: /* crippled */
			err_d="Adresse verstmmelt";
			err_e="Address crippled";
			do_err=TRUE;
			unr=0;
			break;
		case -2: /* unparsable date */
			err_d="Adresse nicht parsbar";
			err_e="Address unparsable";
			unr=0;
			do_err=TRUE;
			break;
		case -3: /* unknown user */
			unr=0;
			break; /* fall through */
		case -4: /* forged */
			err_d="Adresse geflscht";
			err_e="Address forged";
			do_err=TRUE;
			break;
		case -5: /* expired */
			err_d="Gltigkeit der Adresse abgelaufen";
			err_e="Address expired";
			do_err=TRUE;
			break;
		}
		if (do_err) {
			/* print to the screen ... */
			uostr_t infotext=UOSTR_INIT;
			uostr_xdup_cstrmulti(&infotext,"r_user_aufloeser: ",err_d,": ",uname,"\r\n",NULL);
			uostr_x0(&infotext);
			R_PRINTF (2, ("%s",infotext.data));
			uostr_fcut(&infotext);
			if (ub_read_user(unr ? unr : strtoul(uname,NULL,10),&u,FALSE)) {
				/* generate logstr, which wil be used for error message */
				uostr_t newaddress=UOSTR_INIT;
				r_writecomment (R_LOG_ERROR, adr, "Name %s ungltig", uname);
				if (u.address_hash_expire && u.address_hash_secret) {
					uostr_xdup_cstrmulti(&newaddress,
						address_hash_create(&u,time(NULL)+u.address_hash_expire),
						"@",r_domainize(MYDOMAINPART),NULL);
				} else {
					char tmp[UDAT_NAME_LEN*2+1];
					char *t;
					ub_canonical(u.name,tmp);
					for (t=tmp;*t;t++) {
						if (*t==' ')
							*t='_';
					}
					uostr_xdup_cstrmulti(&newaddress,
						tmp,"@",r_domainize(MYDOMAINPART),NULL);
				}
				uostr_x0(&newaddress);
				uostr_xdup_cstrmulti(&infotext,
					"Die Nachricht konnte nicht zugestellt werden. Grund:\n"
					"  ", uname, ": ",err_d,"\n\n",
					"Gltige Adresse: ",newaddress.data, "\n\n"
					"Could not deliver the message. Reason:\n"
					"  ", uname, ": ",err_e,"\n\n"
					"Please try: ",newaddress.data, "\n",
					NULL);
				uostr_x0(&infotext);
				/* need to copy string 'cause of uostr leakcheck ... */
				adr->infotext=XSTRDUP(infotext.data);
				uostr_freedata(&infotext);
				uostr_freedata(&newaddress);
				return R_A_ERROR;
			} /* else fall through to "unknown user" */
			uostr_freedata(&infotext);
		}
	}

	/* unr ist != 0 wenn wir eben eine kurzzeitig eindeutig Adresse
	 * erfolgreich auflsten 
	 */
	if (!unr) {
		unra=ub_find_username_multi (uname,FALSE);
		if (unra && unra[1]!=0) {
			/* uneindeutig. Das wird auch kaum besser wenn wir "real-"
			 * weglassen.
			 */
			R_PRINTF (2, ("r_user_aufloeser: Adresse %s nicht eindeutig\r\n", 
						  uname));
			r_writecomment (R_LOG_ERROR, adr,
							"Adresse %s nicht eindeutig", uname);
			adr->errorflags |= R_E_ADR;
			adr->dsn=r_get_dsn("5.1.4 (address %s ambiguous)",uname);
			free(unra);
			return R_A_ERROR;
		}
		if (unra) {
			unr=unra[0];
			free(unra);
		} else {
			unr=0;
		}
	}

	if (!unr)
	{
		BOOLEAN notfound = TRUE;

		if (real)
		{
			uname -= 5;
			real = FALSE;
			unra=ub_find_username_multi(uname,FALSE);
			if (unra && unra[1]) {
				/* uneindeutig. Das wird auch kaum besser wenn wir "real-"
				 * weglassen.
				 */
				R_PRINTF (2, ("r_user_aufloeser: Adresse %s nicht eindeutig\r\n", 
							  uname));
				r_writecomment (R_LOG_ERROR, adr,
								"Adresse %s nicht eindeutig", uname);
				adr->errorflags |= R_E_ADR;
				adr->dsn=r_get_dsn("5.1.4 (address %s ambiguous)",uname);
				free(unra);
				return R_A_ERROR;
			}
			if (unra) {
				unr=unra[0];
				free(unra);
			} else
				unr=0;
			if (unr)
			{
				R_PRINTF (3, ("r_user_aufloeser: real- gefunden\r\n"));
				r_writecomment (R_LOG_INFO, adr, "real- gefunden");
				notfound = FALSE;
			}
		}
		if (notfound)
		{
			R_PRINTF (2, ("r_user_aufloeser: Benutzer %s unbekannt\r\n", uname));
			r_writecomment (R_LOG_ERROR, adr,
							"Benutzer %s unbekannt", uname);
			adr->errorflags |= R_E_ADR;
			adr->dsn=r_get_dsn("5.1.1 (user %s unknown)",uname);
			return R_A_ERROR;
		}
	}
	if (!ub_read_user (unr, &u, FALSE))
	{
	   /* Should not happen */
		R_PRINTF (2, ("r_user_aufloeser: Benutzer %s unlesbar\r\n", uname));
		r_writecomment (R_LOG_ERROR, adr,
						"Benutzer %s unlesbar", uname);
		syslog (LOG_ERR, "Usernummer %ld gefunden, Daten nicht!", (long) unr);
		adr->errorflags |= R_E_INTERN;
		adr->dsn=r_get_dsn(
			"5.1.1 (user %s unknown, maybe just have been deleted)",
			uname);
		return R_A_ERROR;
	}

	if (!real && u.vertreter[0])
	{
		ret = r_einsetzen (adr, u.vertreter, work);
		if (ret != R_A_ERROR)
		{
		   /*
		    * damit niemand als Vertreter
		    * "echo sysop >tom/system/alias.cnf" 
		    * oder hnlich schlimme Dinge anstellen kann, wird es hier etwas
		    * komplizierter.
		    */
			adr->sec_bezahlt = u.bezahlt;
			adr->bezahlt |= u.bezahlt;
			adr->sec_level = u.level;
			if (u.level > adr->level)
				adr->level = u.level;
			adr->owner = u.nummer;
			adr->flags |= R_F_NEWENVELOPE;
		}

		return ret;
	}

   /* Schluendlich haben wir also einen Namen bekommen */
	ret = r_einsetzen (adr, u.name, work);
	adr->user_gate_nr = unr;
	adr->flags |= R_F_LOKAL;
	adr->transport = r_t_maustausch;

	return ret;
}

int
r_useralias_aufloeser (r_adr_t * adr, r_a_work_t * work)
{
	r_alias_t *al;
	char *kopie;
	size_t kopielen;

	if (!r_useralias_anker)
		return R_A_NOTDONE;

   /* Arbeitskopie des Usernamens anlegen */
	kopie = malloc (strlen (work->local) * 2 + 1);
	if (!kopie)
	{
		R_PRINTF (0, ("r_useralias_aufloeser: out of memory\r\n"));
		r_writecomment (R_LOG_FATAL, adr, "out of memory");
		adr->errorflags |= R_E_TEMP;
		adr->errorflags |= R_E_INTERN;
		adr->dsn=r_dsn_out_of_memory;
		return R_A_ERROR;
	}
	ub_canonical (work->local, kopie);
	kopielen=strlen(kopie);

   /* Aliasse auflsen */
	al = r_useralias_anker;
	while (al)
	{
		if (al->fromlen==kopielen && 0==stricmp (al->from, kopie))
		{
		   /* Alias gefunden */
			char *p;
			int ret;

			free (kopie);

			p = strpbrk (al->to, "\t ");
			if (p)
				*p = 0;

			ret = r_einsetzen (adr, al->to, work);

			adr->flags |= R_F_NEWENVELOPE;

			if (p)
				*p++ = ' ';

			if (ret != R_A_DONE)
				return ret;

			while (p && *p)
			{
				while (isspace (*p))
					p++;
				if (*p)
				{
					r_adr_t *n = r_alloc ();
					char *q;

					if (!n)
					{
						r_free (n);
						R_PRINTF (0, ("r_useralias_aufloeser: out of memory\r\n"));
						r_writecomment (R_LOG_FATAL, adr, "out of memory");
						adr->errorflags |= R_E_TEMP;
						adr->errorflags |= R_E_INTERN;
						adr->dsn=r_dsn_out_of_memory;
						return R_A_ERROR;
					}
					q = strpbrk (p, "\t ");
					if (q)
						*q = 0;
					n->inputadr = strdup (p);
					if (adr->erroradr)
						n->erroradr = strdup (adr->erroradr);
					else
						n->erroradr = NULL;
					if (!n->inputadr || (!n->erroradr && adr->erroradr))
					{
						r_free (n);
						R_PRINTF (0, ("r_useralias_aufloeser: out of memory\r\n"));
						r_writecomment (R_LOG_FATAL, adr, "out of memory");
						adr->errorflags |= R_E_TEMP;
						adr->errorflags |= R_E_INTERN;
						adr->dsn=r_dsn_out_of_memory;
						return R_A_ERROR;
					}
					n->loopcount = adr->loopcount + 1;
					n->bezahlt = adr->bezahlt;
					n->sec_bezahlt = adr->sec_bezahlt;
					n->level = adr->level;
					n->sec_level = adr->sec_level;
					n->owner = adr->owner;

					n->flags = adr->flags & R_F_TAUSCHLOG;
					/* n->flags |= R_F_NEUE_ID; */
					n->flags |= R_F_NEWENVELOPE;

					n->next = adr->next;
					adr->next = n;

					if (q)
					{
						*q = ' ';
						p = q + 1;
					}
					else
						p = 0;
				}
			}					/* while */
			return R_A_DONE;

		}
		al = al->next;
	}
	free (kopie);
	return R_A_NOTDONE;
}

const char *
routing_suche_alias (const char *uname)
{
	r_alias_t *al;
	char work[1024];
	size_t worklen;

	if (!r_useralias_anker)
		return NULL;

   /* Um lokale Aliasse nicht nach oben reichen zu mssen */

   /* Arbeitskopie des Usernamens anlegen */
	ub_canonical (uname, work);
	worklen=strlen(work);

   /* Aliasse auflsen */
	al = r_useralias_anker;
	while (al)
	{
		if (al->fromlen==worklen && !stricmp (al->from, work))
		{
			return al->to;
		}
		al = al->next;
	}
	return NULL;
}


int
r_chef_aufloeser (r_adr_t * adr, r_a_work_t * work)
{
	char *local;
	char *p;
	BRETTNR bnr;
	LISTUSER *lu;
	BRETT *b;
	const char *uname;

	if (strnicmp (work->local, "CHEF ", 5))
		return R_A_NOTDONE;

	local = malloc (strlen (work->local) * 2 + 1);
	if (!local)
	{
		R_PRINTF (0, ("AC: out of memory\r\n"));
		r_writecomment (R_LOG_FATAL, adr, "out of memory");
		adr->errorflags |= R_E_TEMP;
		adr->errorflags |= R_E_INTERN;
		adr->dsn=r_dsn_out_of_memory;
		return R_A_ERROR;
	}

	ub_canonical (work->local, local);

	p = local + 5;
	strtrim (p);

	bnr=brettnr(p);
	if (!bnr || !br_ind[bnr].brett)
	{
		free (local);
		r_writecomment (R_LOG_ERROR, adr, "Gruppe %s unbekannt", p);
		R_PRINTF (2, ("AC: Gruppe unbekannt: %s\r\n", p));
		adr->errorflags |= R_E_ADR;
		adr->dsn=r_get_dsn("5.1.1 (newsgroup %s unknown)",
						   p);
		return R_A_ERROR;
	}
	b=br_ind[bnr].brett;

	if (b->grp_chef == 0 || (lu = ub_find_usernummer (b->grp_chef)) == NULL)
	{
	   /* 
	    * Kein Gruppenchef eingetragen oder Benutzer nicht mehr 
	    * vorhanden. Motto: Jede Gruppe hat einen Chef zu haben.
	    */
		if (box_conf.default_chef)
			uname = box_conf.default_chef;
		else
			uname = "sysop";
		r_writecomment (R_LOG_NOTICE, adr,
						"Kein Chef vorhanden, %s eingesetzt", uname);
		R_PRINTF (3, ("AC: Kein Chef fr %s\r\n", p));
	}
	else
		uname = lu->name;

	free (local);
   /* adr->flags|=R_F_NEWENVELOPE; ?? */
	return r_einsetzen (adr, uname, work);
}

/*
 * !-Aufloeser:
 *
 * host!user -> user@host
 *
 * Parameter:
 *   adr:  Adressstruktur. Dummyparameter
 *   work: wird ersetzt, fa
 * verndert die bergebene Adressstruktur.
 *
 * Verfahren: das am weitesten links stehende Ausrufezeichen wird 
 * gesucht, der Teil davor als neuer Domainpart verwendet.
 */
int
r_bang_aufloeser (r_adr_t * adr, r_a_work_t * work)
{
	char *localpart = work->local;
	char *neu;
	char *p;
	int ret;

	p = strchr (localpart, '!');
	if (!p)
		return R_A_NOTDONE;

   /*
    * Bangadressen kommen nur aus der uucp-Welt, und da kommen definitiv
    * keine Spaces vor.
    * Man knnte das nochmal berdenken.
    */
	if (strpbrk (localpart, " \t"))
	{
		r_writecomment (R_LOG_NOTICE, adr,
						"Bangadressen mit Spaces nicht erlaubt");
		R_PRINTF (2, ("A!: Bangadresse mit Whitespace?\r\n"));
		adr->dsn=r_get_dsn("5.1.3 (bang address must not contain whitespace)");
		return R_A_ERROR;
	}

	neu = malloc (strlen (localpart) + 1);	/* lnger wird es garantiert nicht */
	if (!neu)
	{
		r_writecomment (R_LOG_FATAL, adr, "out of memory");
		R_PRINTF (0, ("A!: out of memory\r\n"));
		adr->errorflags |= R_E_TEMP;
		adr->errorflags |= R_E_INTERN;
		adr->dsn=r_dsn_out_of_memory;
		return R_A_ERROR;
	}

	*p++ = 0;
	strtrim (p);
	strtrim (localpart);
	sprintf (neu, "%s@%s", p, localpart);
	ret = r_einsetzen (adr, neu, work);
	adr->flags |= R_F_NEWENVELOPE;
	free (neu);
	return ret;
}


/*
 * %-Aufloeser:
 *
 * user%host -> user@host
 *
 * verndert die bergebene Adressstruktur.
 *
 * Verfahren: das am weitesten rechts stehende Prozentzeichen wird 
 * durch ein `at' ersetzt.
 */
int
r_percent_aufloeser (r_adr_t * adr, r_a_work_t * work)
{
	int ret;
	char *p;
	char *localpart = work->local;

   /* Das Magic dieses Auflsers */
	p = strrchr (localpart, '%');
	if (!p)
		return R_A_NOTDONE;

   /* Wenn dem % nichts folgt oder das % am Anfang steht, ist die
    * Adresse mehr Mll als alles andere.
    */
	if (!p[1] || p == localpart)
	{
		R_PRINTF (2, ("A%%: %%-Zeichen fehlplaziert\r\n"));
		r_writecomment (R_LOG_INFO, adr, "%%-Zeichen fehlplaziert?");
		adr->dsn=r_get_dsn("5.1.3 (misplaced %%)");
		return R_A_NOTDONE;
	}

   /* Inplace ersetzen. Das geht nur, wenn r_einsetzen String 
    * nicht free()d, deshalb ... 
    */
	work->local = NULL;
	*p = '@';
	ret = r_einsetzen (adr, localpart, work);
	adr->flags |= R_F_NEWENVELOPE;
    free(localpart);
	return ret;
}
