/* router/rewrite.c: Umschreiben von Domains, Aliasse fr Domains */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
*/

#include "tomsysdep.h"

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <syslog.h>
#include <ctype.h>
#include <limits.h>
#include <fnmatch.h>

#include "prototyp.h"
#include "tompaths.h"
#include "modem.h"
#include "box_conf.h"
#include "user.h"
#include "routing.h"
#include "route_int.h"
#include "prototyp.h"


/*
 * Domainparts umschreiben. 
 * Beispiele:                          Rewriter-List-Entry
 * AC    -> AC.maus.de                 ac -> AC.maus.de
 * tirka -> tirka.gun.de               tirka -> tirka.gun.de
 * willy.host.domain -> host.domain    *.host.domain -> host.domain
 * tirka.gun.de -> LEER                tirka.gun.de  -> NULL
 *
 * Wenn die Tabelle u.a.
 *        *tirka -> tirka
 *        tirka  -> tirka.gun.de
 * enthlt, wird eine Domain "ftp.tirka" erst in tirka und dann tirka.gun.de
 * ungewandelt.
 */

static BOOLEAN
r_find_alias(const char *domain, r_alias_t *r, char **fundstueck)
{
	char such[512];
	size_t suchlen;
	r_hcanonical(domain,such);
	suchlen=strlen(such);

	while (r)
	{
		BOOLEAN found=FALSE;
		if (*r->from=='*' || r->from[r->fromlen-1]=='*') {
#ifdef FNM_CASEFOLD
			found=(0==fnmatch(r->from,such,FNM_CASEFOLD));
#else
			char buf[512];
			entumlauten(r->from,buf);
			strlwr(buf);
			found=(0==fnmatch(buf,such,0));
#endif
		}
		else if (r->fromlen==suchlen)
			found=!strcmp(such,r->from);
		if (found)
		{
			*fundstueck=r->to;
			return TRUE;
		}
		r=r->next;
	}
	return FALSE;
}

/* 
 * "Umschreiben" von Domainpart. Wenn XY nicht nur wie AB behandelt
 * werden soll, sondern auch als AB heraus gehen soll, dann ist 
 * der rewriter gefragt.
 */
int 
r_rewriter_aufloeser(r_adr_t *adr, r_a_work_t *work)
{
	char *ergebnis;
	if (!work->domain)
		return R_A_NOTDONE;
	if (!r_find_alias(work->domain,r_rewrite_anker,&ergebnis))
		return R_A_NOTDONE;
	free(work->domain);
	work->domain=strdup(ergebnis);
	if (!work->domain)
	{
		R_PRINTF(0,("r_rewriter_aufloeser: out of memory\r\n"));
		r_writecomment(R_LOG_FATAL,adr,"out of memory");
		adr->errorflags |= R_E_TEMP;
		adr->errorflags |= R_E_INTERN;
		adr->dsn=r_dsn_out_of_memory;
		return R_A_ERROR;
	}
	adr->flags |= R_F_NEWENVELOPE;
	return R_A_DONE;
}


int 
r_hostalias_aufloeser(r_adr_t *adr, r_a_work_t *work)
{
	char *ergebnis=NULL;
	if (!work->domain)
		return R_A_NOTDONE;
	if (!r_find_alias(work->domain,r_hostalias_anker,&ergebnis))
		return R_A_NOTDONE;
	free(work->domain);
	work->domain=strdup(ergebnis);
	if (!work->domain)
	{
		R_PRINTF(0,("r_hostalias_aufloeser: out of memory\r\n"));
		r_writecomment(R_LOG_FATAL,adr,"out of memory");
		adr->errorflags |= R_E_TEMP;
		adr->errorflags |= R_E_INTERN;
		adr->dsn=r_dsn_out_of_memory;
		return R_A_ERROR;
	}
	return R_A_DONE;
}

char *
r_domainize(const char *idomain)
{
	char *ergebnis=NULL;

	if (!r_init_done)
		r_do_init();

	while (TRUE)
	{
		char *alterg=ergebnis;
		const char *such=ergebnis ? ergebnis : idomain;
		if (r_find_alias(such,r_rewrite_anker,&ergebnis)
			|| r_find_alias(such,r_hostalias_anker,&ergebnis))
		{
			if (alterg)
				free(alterg);
			ergebnis=strdup(ergebnis);
			if (!ergebnis)
				return NULL;
		}
		else
			break;
	}

	if (!ergebnis)
		ergebnis=strdup(idomain);
	
	return ergebnis;
}

/*
 * Das sollte lieber ber den Adressparser geschehen ...
 * dazu mte der aber erst auf die Benutzung eines Stacks 
 * umgeschrieben werden.
 */
char *
r_convert_to_inetadr(const char *iname)
{
	char *ergebnis=NULL;
	char *kopie;
	char *at;
	char *p;
	char *domain;

	if (!r_init_done)
		r_do_init();

	at=strrchr(iname,'@');
	if (!at)
	{
		p=strrchr(iname,'%');
		if (p)
		{
			char *q=strdup(iname);
			if (!q)
				return q;
			q[p-iname]='@';
			kopie=q;
		}
		else
		{
			p=strchr(iname,'!');
			if (!p)
			{
				/* ah ja - keine Domain in irgendeiner Form vorhanden */
				kopie=concat(iname,"@",box_conf.domainname,NULL);
				if (!kopie)
					return NULL;
			}
			else
			{
				char *work=strdup(iname);
				size_t offset;
				if (!work)
					return NULL;
				offset=p-iname;
				p=work+offset;
				*p++=0;
				strtrim(work);
				strtrim(p);
				kopie=concat(p,"@",work,NULL);
				free(work);
				if (!kopie)
					return NULL;
			}
		}
	}
	else
	{
		kopie=strdup(iname);
		if (!kopie)
			return NULL;
	}
	at=strrchr(kopie,'@');

	/*
	 * So, eine @-Adresse haben wir also in kopie.
	 */
	*at++=0;
	strtrim(kopie);
	strtrim(at);
	/*
	 * Domainpart vervollstndigen
	 */
	domain=r_domainize(at);
	if (!domain)
	{
		free(kopie);
		return NULL;
	}
	/*
	 * Localpart irgendwie in eine Internetadresse umwandeln.
	 */
	{
		char *neu;
		char *local;
		char *akt;
		char *writepos;
		neu=malloc(strlen(kopie)+1);
		if (!neu)
		{
			free(kopie);
			return NULL;
		}
		akt=kopie;
		writepos=neu;
		while(akt && *akt)
		{
			char *next;
			char c=0;
			next=strpbrk(akt,"%@!");
			if (next)
			{
				c=*next;
				*next++=0;
			}
			strtrim(akt);
			p=akt;
			while (*p)
			{
				if (*p==' ' || *p=='\t')
				{
					if (p[1]!=' ' && p[1]!='\t')
						*writepos++='_';
				}
				else
					*writepos++=*p;
				p++;
			}
			*writepos++=c;
			akt=next;
		}
		/* So, berflssige Spaces sind beseitigt, innere nach _ gewandelt.
		 * nun entumlauten 
		 */
		local=entumlauten1(neu);
		free(neu);
		if (!local)
		{
			free(domain);
			free(kopie);
			return NULL;
		}
		ergebnis=concat(local,"@",domain,NULL);
		free(local);
		free(domain);
		free(kopie);
		return ergebnis;
	}
}

/*
 * Nachrichten an bestimmte Domains bouncen.
 */
int 
r_domainbounce_aufloeser(r_adr_t *adr, r_a_work_t *work)
{
	char *ergebnis;
	if (!work->domain)
		return R_A_NOTDONE;
	/* Fidoadressen sind nicht betroffen */
	if (strchr(work->domain,' '))
		return R_A_NOTDONE;
	if (!r_find_alias(work->domain,r_domainbounce_anker,&ergebnis))
		return R_A_NOTDONE;

	R_PRINTF(0,("r_domainbounce_aufloeser: %s: %s\r\n",
				work->domain,ergebnis ? ergebnis : "nicht erlaubt"));
	r_writecomment(R_LOG_FATAL,adr,"%s: %s", 
				   work->domain,ergebnis ? ergebnis : "nicht erlaubt");
	adr->errorflags |= R_E_ADR;
	adr->dsn=r_get_dsn("5.7.0 (%s: %s)",
					   work->domain,ergebnis ? ergebnis : "nicht erlaubt");
	return R_A_ERROR;
}
