/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
/*
 * error-deliver: Transport aller Fehlernachrichten.
 * Wird nur aus dem Tausch benutzt, aber hier untergebracht, weil es
 * 100%ig Routercode *ist*.
 */

#include "tomsysdep.h"

#include <portab.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <dirent.h>
#include <ctype.h>
#include <syslog.h>
#include <time.h>

#include "prototyp.h"
#include "box_conf.h"
#include "user.h"
#include "routing.h"
#include "route_int.h"
#include "msg.h"
#include "tompaths.h"


#ifndef LINEEND
#define LINEEND "\r\n"
#endif

static BOOLEAN r_t_error_deliver(message_t *head, r_adr_t *adr);
static BOOLEAN r_t_error_flush(struct r_transport_t *);

r_transport_t *r_t_error;

void
r_t_error_define(char *str)
{
	r_transport_t *n;

	if (0!=stricmp(str,"error"))
	{
		syslog(LOG_WARNING,"error_define called for unknown function %s",str);
		return;
	}
	n=XMALLOC(sizeof(r_transport_t));
	strlwr(str);
	n->name=XSTRDUP(str);
	n->deliver=r_t_error_deliver;
	n->flush=r_t_error_flush;
	r_t_error=n;
	r_t_chain_transport(n);
}


/*
 * Die zentrale Fehlertransportroutine. 
 */
static BOOLEAN
r_t_error_deliver(message_t *head, r_adr_t *adr)
{
	r_adr_t *akt;
	char fname[PATH_MAX];
	const char *md;
	char *from;
	const char *to;
	int i;
	FILE *f;

	md= box_conf.mailer_daemon ? box_conf.mailer_daemon : "MAILER-DAEMON";
	from =concat(md, " @ ",(MYDOMAINPART),NULL);

	for (akt=adr; akt; akt=akt->next)
	{
		if (!(akt->flags & R_F_TO_MASK))
			akt->flags=R_F_TO_SENDER;
	}

	sprintf(fname,"%sbounce.txt",tom_tmp_dir);

	for (i=0;i<3;i++)
	{
		long flag;

		if (i==0)
			flag=R_F_TO_SENDER;
		else if (i==1)
			flag=R_F_TO_OWNER;
		else
			flag=R_F_TO_POSTMASTER;

		for (akt=adr; akt; akt=akt->next)
		{
			const char *subj;
			r_adr_t *akt2;
			char *tmp=NULL;

			if (!akt->error)
				continue;

			/* richtiger Empfnger? */
			if (!(akt->flags & flag))
				continue;

			/* schon erledigt? */
			if (akt->flags & R_F_ZUGESTELLT)
				continue;

			f=fopen(fname,BINWRITE);
			if (!f)
			{
				const char *s=strerror(errno);
				/* 
				 * In diese Datei sollte die Fehlermeldung ...
				 */
				r_paniclog("kann bouncemeldung %s nicht erzeugen: %s",
					fname,s);
				msglog("  ?kann bouncemeldung nicht in %s schreiben: %s\r\n",
					fname,s);
				r_tauschuserlog(adr,":?kann nicht bouncen: %s\r\n",s);
				continue;
			}
			if (i==0)
				subj="Mail failed";
			else if (i==1)
				subj="Mail failed, sending to owner";
			else
				subj="Mail failed, sending to postmaster";

			fprintf(f,
"Ihre Nachricht %s konnte nicht an die folgenden Adressen zugestellt werden:\n"
"Your message %s wasn't delivered to:\n\n",
				head->msg_id, head->msg_id);

			/* Adresse und Fehlermeldung ausgeben */
			fprintf(f, "%s\n"
				"    %s\n",
				akt->inputadr,akt->logstr);
			akt->flags |= R_F_E_TRIED;

			if (akt->infotext) {
				fputs(akt->infotext,f);
			}

			/* Etwaige andere Fehlermeldungen an denselben Empfnger
			 * gleich mit erledigen
			 */
			for (akt2 = akt->next; akt2; akt2 = akt2->next)
			{
				if (!strcmp(akt2->resultadr,akt->resultadr))
				{
					/* Adresse und Fehlermeldung ausgeben */
					fprintf(f, "%s\n"
						"    %s\n",
						akt2->inputadr,akt2->logstr);
					akt2->flags |= R_F_E_TRIED;
				}
			}

			/*
			 * So, jetzt noch ein paar Infos:
			 */
			fprintf(f,"\n\n");
			fprintf(f,"Message-ID: %s\n",head->msg_id);
			fprintf(f,"Referenz:   %s\n",head->ref_id);
			fprintf(f,"Betreff:    %s\n",head->betreff);
			fprintf(f,"Datum:      %s\n",head->eingabezeit);
			fprintf(f,"From:       %s\n",head->absender);
			fprintf(f,"Sender:     %s\n",head->sender);
			fprintf(f,"Reply-To:   %s\n",head->reply_to);
			{
				simplelist_t *s;
				s=head->to;
				while(s)
				{
					fprintf(f,"To:         %s\n",s->data);
					s=s->next;
				}
			}
			fclose(f);

			tmp=NULL;
			if (i==0)
				to=adr->erroradr;
			else if (i==1)
			{
				tmp=concat("owner-",akt->inputadr,NULL);
				to=tmp;
			}
			else
				to="sysop";

			if (!to || !einfachnachricht(TRUE,9,TRUE,fname,
				'A', to,
				'V', (from ? from : md),
				'W', subj,
				NULL))
			{
				/* Arg - das ist aber berhaupt nicht schn! */
				for (akt2 = akt; akt2; akt2 = akt2->next)
				{
					if (akt2->flags & R_F_E_TRIED)
					{
						akt2->flags &= ~R_F_E_TRIED;
						akt2->flags |= R_F_TO_POSTMASTER;
					}
				}

				msglog("  ?kann bouncemeldung nicht senden\r\n");
				r_tauschuserlog(adr,":?kann Bouncemail nicht senden\r\n");
			}
			else 
			{
				/* OK, *das* hat mal geklappt */
				for (akt2 = akt; akt2; akt2 = akt2->next)
				{
					if (akt2->flags & R_F_E_TRIED)
					{
						akt2->flags &= ~R_F_E_TRIED;
						akt2->flags |= R_F_ZUGESTELLT;
					}
				}
			}
			if (i==1 && tmp)
				free(tmp);
			unlink(fname);
		} /* durch alle Adresse */
	} /* i von 0 bis 2, sprich sender/owner/postmaster */

	if (from)
		free(from);

	return TRUE;
}

static BOOLEAN
r_t_error_flush(struct r_transport_t *dummy TOM_ATTRIB_UNUSED)
{
	return TRUE;
}


