/* Deliver via internal routines ... */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
#include "tomsysdep.h"

#include <portab.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <dirent.h>
#include <ctype.h>
#include <errno.h>
#include <syslog.h>
#include <time.h>
#include <sys/stat.h>

#include "box_conf.h"
#include "prototyp.h"
#include "user.h"
#include "brett.h"
#include "modem.h"
#include "tompaths.h"
#include "routing.h"
#include "route_int.h"
#include "msg.h" /* write_msg fr #LOG */

#include "lockfile.h" /* libquark */

#ifndef LINEEND
#define LINEEND "\r\n"
#endif

static BOOLEAN r_t_intern_deliver(message_t *head, r_adr_t *adr);
static BOOLEAN r_t_intern_flush(struct r_transport_t *);

static BOOLEAN r_t_intern_deliver_one(message_t *head,r_adr_t *adr, const char *id);

static BOOLEAN intern_fn_misc(message_t *head, r_adr_t *adr, const char *buf, 
	const char *id, char **errtext);

/* einen internen transport definieren. Format der Zeile:
 * "intern bounce" */
void
r_t_intern_define(char *str)
{
	r_transport_t *n;
	if (0!=stricmp(str,"bounce")
		&& 0!=stricmp(str,"echo")
		&& 0!=stricmp(str,"notice")
		&& 0!=stricmp(str,"discard"))
	{
		syslog(LOG_WARNING,"intern_define called for unknown function %s",str);
		return;
	}
	n=XMALLOC(sizeof(r_transport_t));
	strlwr(str);
	n->name=XSTRDUP(str);
	n->deliver=r_t_intern_deliver;
	n->flush=r_t_intern_flush;
	r_t_chain_transport(n);
}

static BOOLEAN
intern_fn_misc(message_t *head, r_adr_t *adr, const char *buf, 
	const char *id, char **errtext)
{
	FILE *f;
	BOOLEAN ret;
	char fname[PATH_MAX];
	char *absender;
	const char *md;

	if (0==stricmp(adr->thint,"bounce"))
	{
		*errtext=strdup("Wie gewnscht gebounced");
		return FALSE;
	}

	if (0==stricmp(adr->thint,"discard"))
	{
		if (head->msgtyp==pers_msg || head->msgtyp==status_msg)
		{
			for ( ; adr; adr=adr->next)
			{
				if (!stricmp(adr->thint,"discard")
					&& TRANSPORT_ALLOWED(adr))
				{
					char idbuf[256];

					if (adr->flags & R_F_NEUE_ID)
					{
						make_msg_id(idbuf,TRUE, NULL, NULL);
						id=idbuf;
					}

					/* der eigentliche Transport */
					adr->flags |= R_F_ZUGESTELLT;

					/* Eintrge ins Messagelog */
					if (head->msgtyp==pers_msg)
					{
						msglog("  Tintern/discard\n");
						msglog("  =%s\n",id);
					}
					else
						/* Status */
						msglog("  =%s (Status, discarded)\n",id);

					/*
					 * Tauschlog
					 */
					r_tauschuserlog(adr, ":!discarded\r\n",id);
					r_tauschuserlog(adr, ":=%s\r\n",id);
				}
			}
		}
		else /* Gruppennachricht */
		{
			msglog("  =%s\n",head->msg_id);
			r_tauschuserlog(adr, ":!discarded\r\n",id);
			r_tauschuserlog(adr, ":=%s\r\n",head->msg_id);
		}
		return TRUE;
	}

	if (!adr->erroradr)
	{
		*errtext=strdup("Keine Antwortadresse");
		return FALSE;
	}

    md= box_conf.mailer_daemon ? box_conf.mailer_daemon : "MAILER-DAEMON";
	absender =concat(md, " @ ",(MYDOMAINPART),NULL);
	if (!absender)
	{
		*errtext=strdup("Out of memory");
		return FALSE;
	}

	sprintf(fname,"%s%misc%lu.rt",tom_tmp_dir,(ULONG) getpid());
	f=fopen(fname,"w");
	if (!f)
	{
		free(absender);
		*errtext=concat("cannot open ",fname,": ",strerror(errno),NULL);
		return FALSE;
	}

	fprintf(f,"Message-ID: %s\n",id);
	fprintf(f,"Referenz:   %s\n",head->ref_id);
	fprintf(f,"Betreff:    %s\n",head->betreff);
	fprintf(f,"Datum:      %s\n",head->eingabezeit);
	fprintf(f,"From:       %s\n",head->absender);
	fprintf(f,"Sender:     %s\n",head->sender);
	fprintf(f,"Reply-To:   %s\n",head->reply_to);
	{
		simplelist_t *s;
		s=head->to;
		while(s)
		{
			fprintf(f,"To:         %s\n",s->data);
			s=s->next;
		}
	}

	if (0==stricmp(adr->thint,"echo"))
	{
		const char *p;
		const char *nl;
		p=buf;
		do 
		{
			nl=strchr(p,'\n');
			if (!nl)
			{
				fwrite(p,strlen(p),1,f);
				p=nl;
			}
			else
			{
				fwrite(p,nl-p+1,1,f);
				p=nl+1;
			}

		} while(p && *p);
	}
	else
		fprintf(f,"\r\nThis is a notice\r\n");
	fclose(f);
	ret=einfachnachricht(TRUE, adr->level, adr->bezahlt,
		fname,
		'W', adr->thint,
		'V', absender,
		'A', adr->erroradr,
		0);
	free(absender);
	unlink(fname);
	return ret;
}

static BOOLEAN
r_t_intern_deliver_one(message_t *head,r_adr_t *adr, const char *id)
{
	FILE *in=NULL; /* keep gcc quiet */
	char *buffer=NULL;
	char *errtext=NULL;

	if (!adr->thint)
	{
		/* wenn wir keinen "transport hint" bekommen, wissen wir nicht,
		 * an welche Funktion die Nachricht tatschlich geht */
		r_paniclog("no transport hint in r_t_intern_deliver_one for: %s\n",
			adr->resultadr);
		r_tauschuserlog(adr,":?TSNH: kein transport hint ...\r\n");
		msglog("   ?TSNH: kein transport hint fr Nachricht an: %s\r\n",
			adr->resultadr);
		adr->errorflags|=R_E_INTERN;
		adr->flags|=R_F_TO_POSTMASTER;
		adr->flags|=R_F_TO_SENDER;
		adr->error=TRUE;
		return FALSE;
	}
	if (0!=stricmp(adr->thint,"echo")
		&& 0!=stricmp(adr->thint,"bounce")
		&& 0!=stricmp(adr->thint,"notice")
		&& 0!=stricmp(adr->thint,"discard"))
	{
		r_paniclog(
			"transport hint %s notexistant in r_t_intern_deliver_one for %s\n",
			(char *) adr->thint, adr->resultadr);
		r_tauschuserlog(adr,":?TSNH: illegaler transport hint ...\r\n");
		msglog("   ?TSNH: illegaler transport hint %s fr Nachricht an %s\r\n",
			(char *) adr->thint, adr->resultadr);
		adr->errorflags|=R_E_INTERN;
		adr->flags|=R_F_TO_POSTMASTER;
		adr->flags|=R_F_TO_SENDER;
		adr->error=TRUE;
	}

	if (head->fname)
	{
		/* Text aus der Datei lesen */
		long pos;
		in=fopen(head->fname,"r");

		if (!in)
		{
			const char *s=strerror(errno);
			r_paniclog("cannot open inputfile %s: %s\n",head->fname,s);
			r_tauschuserlog(adr,":?TSNH: kann Nachricht nicht ffnen: %s\r\n",s);
			msglog("   ?TSNH: kann Nachricht nicht ffnen: %s\r\n",s);
			adr->errorflags|=R_E_INTERN;
			adr->flags|=R_F_TO_POSTMASTER;
			adr->flags|=R_F_TO_SENDER;
			adr->error=TRUE;
			return FALSE;
		}
		pos=filesize(in);
		if (pos<0)
		{
			const char *s=strerror(errno);
			r_paniclog("cannot seek on inputfile %s: %s\n",head->fname,s);
			r_tauschuserlog(adr,":?TSNH: kann Nachrichtendatei nicht seeken: %s\r\n",s);
			msglog("   ?TSNH: kann Nachrichtendatei nicht seeken: %s\r\n",s);
			adr->errorflags|=R_E_INTERN;
			adr->flags|=R_F_TO_POSTMASTER;
			adr->flags|=R_F_TO_SENDER;
			adr->error=TRUE;
			fclose(in);
			return FALSE;
		}
		buffer=malloc(pos+1);
		if (!buffer)
		{
			r_paniclog("cannot malloc %ld bytes\n",pos);
			r_tauschuserlog(adr,":?cannot malloc %ld bytes\r\n",pos);
			msglog("   ?cannot malloc %ld bytes\r\n",pos);
			adr->errorflags|=R_E_INTERN;
			adr->flags|=R_F_TO_SENDER;
			adr->error=TRUE;
			fclose(in);
			return FALSE;
		}
		if (1!=fread(buffer,pos,1,in))
		{
			const char *s=strerror(errno);
			r_paniclog("cannot read %lu bytes from %s: %s\n",pos,
				head->fname,s);
			r_tauschuserlog(adr,":?cannot read %lu bytes from %s: %s\r\n",pos,
				head->fname,s);
			msglog("   ?cannot read %lu bytes from %s: %s\r\n",pos,
				head->fname,s);
			adr->errorflags|=R_E_INTERN;
			adr->flags|=R_F_TO_SENDER;
			adr->error=TRUE;
			fclose(in);
			return FALSE;
		}
		buffer[pos]=0;
		fclose(in);
	}
	else
		buffer=head->body.buf;

	if (!intern_fn_misc(head,adr,buffer,id,&errtext))
	{
		const char *s=errtext ? errtext : "unknown error";
		r_paniclog("function call %s failed: %s\n",
			(char *) adr->thint,s);
		r_tauschuserlog(adr,":?function call %s failed: %s\r\n",
			(char *) adr->thint,s);
		msglog("   ?function call %s failed: %s\r\n",
			(char *) adr->thint,s);
		adr->errorflags|=R_E_INTERN;
		adr->flags|=R_F_TO_SENDER;
		adr->error=TRUE;
		if (errtext)
			free(errtext);
		return FALSE;
	}

	/* cleanup */
	if (errtext)
		free(errtext);

	if (head->fname)
		free(buffer);

	msglog("  =%s\n",id);
	r_tauschuserlog(adr,":=%s\r\n",id);
	return TRUE;
}

/* 
 * Hierhin kommen nur vollstndige Nachrichten, keine Spezialnachrichten
 * und keine Statusmeldungen.
 */
BOOLEAN
r_t_intern_deliver(message_t *head, r_adr_t *adr)
{
	BOOLEAN ok=TRUE;

	if (head->msgtyp==pers_msg)
	{
		char idbuf[256];
		char *id;

		if (adr->flags & R_F_NEUE_ID)
		{
			make_msg_id(idbuf,TRUE, NULL, NULL);
			id=idbuf;
		}
		else
			id=head->msg_id;

		if (!r_t_intern_deliver_one(head,adr,id))
		{
			r_writecomment(R_LOG_ERROR,adr,"Transport gescheitert");
			adr->error=TRUE;
			adr->errorflags|=R_E_INTERN;
			ok=FALSE;
		}
		else
			adr->flags |= R_F_ZUGESTELLT;
	}
	else if (head->msgtyp==status_msg)
	{
		/* Hh .. */
		r_tauschuserlog(adr,":!not delivered to internal function\r\n");
		r_tauschuserlog(adr,":=%s\r\n",head->msg_id);
		msglog("  =%s (Status, not delivered through intern router)\r\n",
			head->msg_id);
	}
	return ok;
}

static BOOLEAN
r_t_intern_flush(struct r_transport_t *dummy TOM_ATTRIB_UNUSED)
{
	return TRUE; /* really a dummy */
}
