/* Tausch, Handling der Listen der bestellten Gruppen. */
/*
    Copyright (C) 1993 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
#include "tomsysdep.h"

#include <portab.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <limits.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <syslog.h>

#include "prototyp.h"
#include "modem.h"
#include "brett.h"
#include "msg.h"
#include "user.h"

#include "tausch.h"
#include "ta_int.h"
#include "ta_ext.h"

bestell_t *ta_blist;
TAUSCH_OPTS ta_opts;

/****************************** Prototypen *********************************/
/* Ascii Tauschinstallation aus der angegebenen Datei lesen. */
static void ta_inst_laden (const char *datei);

/* Ascii Tauschdaten schreiben.  */
static void ta_inst_sichern (const char *datei, const char *comment);

void
ta_mark_invalid (void)
{
	ta_opts.usernr = 0;
}

/*
 * Tauschinstallation laden.
 * Ldt die aktuelle Tauschinstallation.
 * Danach werden Zwangsbretter hinzugefgt.
 * Zum Schlu werden alle nicht erlaubten Gruppen wieder gestrichen.
 * Die Reihenfolge ist *wichtig*, werden die Zwangsbretter am Ende
 * dazugeladen, gibt es keine Mglichkeit, Zwangsgruppen nur fr
 * Leute ab Level x einzurichten.
 * Liest natrlich nur, wenn die Tauschinstallation noch nicht
 * geladen ist.
 * Es kann durchaus vorkommen, da auch schon nach dem ersten Laden
 * der Liste das changed-flag gesetzt ist!
 * Falls async_uebernehmen gesetzt ist, werden die Daten in 
 * tausch/usernummer.aconf in die Konfiguration bernommen. Dies
 * ist fr die Exportphase des Asynctauschs notwendig.
 */
void
ta_lade_installation (BOOLEAN async_uebernehmen)
{
	BRETTNR i;
	BOOLEAN first=FALSE;
	struct stat st;
	char fname[PATH_MAX];

	/* Falls ... */
	if (ta_opts.usernr == user.nummer)
		return;

	/* allgemeine Tauschkonfiguration behandeln */
	memset (&ta_opts, 0, sizeof (ta_opts)); /* ### Schweinekram */

	/* Erstmal alles auf abbestellt */
	if (!ta_blist)
		ta_blist=XMALLOC((max_brettnr+1) * sizeof(bestell_t));
	for (i=0;i<max_brettnr;i++)
		ta_blist[i]=GRP_UNBESTELLT;

	/* Nun Tauschinstallation von der Platte holen */
	sprintf (fname, "%s%ld.tau", box_conf.tausch_dir, user.nummer);
	if (stat (fname, &st) == 0)
		ta_inst_laden (fname);
	else
	{
		mboutput (" Bisher noch keine Gruppenliste installiert (siehe TAUSCH INST)\r\n\n");
		ta_opts.umlaut = TRUE;
		ta_opts.only_one_group = FALSE;
		ta_opts.quarkoutfile = FALSE;
		ta_opts.changed = TRUE; /* Wegen der Defaultbretter */
		ta_opts.last_ren = time(NULL);
		ta_opts.last_newstausch = time(NULL)-5*86400L;
		ta_opts.last_mailtausch = 0L;
		ta_opts.last_status = 0;
		ta_opts.maxlen=16000;
		ta_opts.pathexcludes=NULL;
		ta_opts.async_export=FALSE;
		first=TRUE;
	}
	/* Zwangsbretter dazumischen */
	for (i = 1; i < max_brettnr; i++)
	{
		if (!br_ind[i].brett)
			continue;
		/* Bestellt, aber Lesen nicht erlaubt? */
		if (ta_blist[i]!=GRP_UNBESTELLT)
		{
			if (!ta_check_readperm(i,0))
			{
				/* Dann herauswerfen */
				ta_blist[i] = GRP_UNBESTELLT;
				ta_opts.changed = TRUE;
			}
			continue;
		}
		/* Sie ist also nicht bestellt. Mssen wir sie anstellen? */
		if ( (br_ind[i].brett->flags & B_ZWANG && !user.pflicht_befreit)
			|| (br_ind[i].brett->flags & B_DEFAULT && first))
		{
			/* Zwangs/Defaultbrett, das wir anstellen sollten. Aber
			 * zuvor mu noch geprft werden, ob die Gruppe gelesen
			 * werden darf: Zwang Sysop.Info!
			 */
			if (ta_check_readperm(i,NULL))
			{
				ta_opts.changed = TRUE;
				ta_blist[i] = GRP_BESTELLT;
			}
		}
	}

	ta_opts.usernr=user.nummer;
}								/* ta_lade_installation */


/* 
   Tauschinstallation sichern
   (relativ ausfhrliche Beschreibung in ta_inst.h)
 */
void
ta_sichere_installation (BOOLEAN async_sichern)
{
	char fname[PATH_MAX];
	char un[UDAT_NAME_LEN*2];
	char *p;

	/* Wenn tauschinstallation nicht geladen, dann sind wir schon fertig */
	if (!ta_opts.usernr)
		return;

	/* Keine nderung? Kein Sichern ntig! */
	if (!ta_opts.changed)
		return;

	sprintf (fname, "%s%ld.tau", box_conf.tausch_dir, user.nummer);
	ta_inst_sichern (fname,NULL);
	ub_canonical(user.name,un);
	for (p=un;*p;p++) {
		if (isspace(*p))
			*p='_';
	}
	if (box_conf.user_dir) {
		sprintf (fname, "%s%s/subscriptions", box_conf.user_dir, un);
		ta_inst_sichern (fname,"Diese Datei _nicht_ veraendern. Sie wird nur wieder ueberschrieben");
	}

	ta_opts.changed = FALSE;

}								/* ta_sichere_installation */


BOOLEAN
ta_brett_bestellen (BRETTNR bnr, char *fehl)
{
	ta_lade_installation (FALSE);
	if (ta_blist[bnr]==GRP_UNBESTELLT)
	{
		ta_blist[bnr] = GRP_BESTELLT;
		ta_opts.changed = TRUE;
		return TRUE;
	}
	else
	{
		sprintf (fehl, "die Gruppe ist schon bestellt");
		return FALSE;
	}
}

BOOLEAN
ta_brett_abbestellen (BRETTNR bnr, char *fehl)
{
	ta_lade_installation (FALSE);
	if (ta_blist[bnr]==GRP_UNBESTELLT)
	{
		/* Tja, besser so fr den Tausch - schlecht fr Online ... */
		sprintf (fehl, "Die Gruppe bleibt abbestellt!");
		return FALSE; /* ? oder besser doch TRUE? */
	}
	else if ((br_ind[bnr].brett->flags & B_ZWANG) && !user.pflicht_befreit)
	{
		/* Tja, besser so fr den Tausch - schlecht fr Online ... */
		sprintf (fehl, "Pflichtbrett, nicht abbestellt!");
		return FALSE;
	}
	else
	{
		ta_blist[bnr] = GRP_UNBESTELLT;
		ta_opts.changed = TRUE;
		return TRUE;
	}
}


/* 
 * Exportliste fr die Datenbank schreiben.
 *
 * ->           datei   :   Name der zu erzeugenden ASCII-Datei
 *
 */
void
ta_schreibe_exportliste (char *datei, BOOLEAN binaer)
{
	FILE *exp_f;
	BRETTNR *sortiert;

	sortiert=b_sortierung();
	if (!sortiert)
	{
		mboutput("Speichermangel (b_sortierung), Abbruch!\r\n");
		return;
	}

	exp_f = xfopen (datei, BINWRITE);

	ta_lade_installation (FALSE);

	if (ta_passinfo.send_mail)
	{
		char buf[28];
		const char *p;
		struct tm *tm;
		tm=localtime(&ta_opts.last_mailtausch);
		if (tm)
		{
			strftime(buf,sizeof(buf)-1,",%d.%m.%Y,%H:%M:%S",tm);
			p=buf;
		}
		else
			p="";
		if (!binaer)
		{
			if (ta_passinfo.send_pa)
				fprintf (exp_f, "*%s,1.1.1980,00:00:00\r\n", user.name);
			else
				fprintf (exp_f, "*%s%s\r\n", user.name, p);
		}
		else
			fprintf (exp_f, "Programmfach%s\r\n", p);
	}

	if (ta_passinfo.send_news)
	{
		for ( ; *sortiert; sortiert++)
		{
			BRETTNR i;
			i=*sortiert;
			if (!br_ind[i].brett)
				continue;
			if ((binaer && br_ind[i].brett->art!=BINAER_BRETT)
				|| (!binaer && br_ind[i].brett->art!=TEXT_BRETT))
				continue;
			if (ta_blist[i]!=GRP_UNBESTELLT 
				&& !(br_ind[i].brett->flags & B_READ))
				ta_blist[i] = GRP_UNBESTELLT;
			if (ta_blist[i]==GRP_BESTELLT)
			{
				fputs(br_ind[i].brett->name,exp_f);
				fputs("\r\n",exp_f);
			}
		}
	}
	fclose (exp_f);
}								/* ta_schreibe_exportliste */


/*****************************************************************************
 * Interne Funktionen dieses Quelltextfiles    
 *****************************************************************************
 */

/* 
 * interne Tauschbrettliste als ASCII-Datei exportieren
 *
 * ->           datei   :   Name der zu erzeugenden ASCII-Datei
 *
 * Anwendung:
 * - Editieren mit ASCII-Editor 
 * - Statistiktools
 */
static void
ta_inst_sichern (const char *datei, const char *comment)
{
	char buf[20];
	FILE *exp_f;
	BRETTNR i;

	exp_f = xfopen (datei, "w");
	if (comment)
		fprintf (exp_f, "#%s\n", comment);

	time_t_to_mtzeit_sec(&ta_opts.last_newstausch,buf);
	fprintf (exp_f, "#NEWSTAUSCHDATUM=%s\n", buf);
	time_t_to_mtzeit_sec(&ta_opts.last_mailtausch,buf);
	fprintf (exp_f, "#MAILTAUSCHDATUM=%s\n", buf);
	time_t_to_mtzeit_sec(&ta_opts.last_ren,buf);
	fprintf (exp_f, "#RENDATUM=%s\n", buf);
	time_t_to_mtzeit_sec(&ta_opts.last_status,buf);
	fprintf (exp_f, "#STATUSDATUM=%s\n", buf);
	fprintf (exp_f, "#UB=%ld\n", (long) ta_opts.umbruch);
	fprintf (exp_f, "#USER %ld %s\n", (long) user.nummer,user.name);
	fprintf (exp_f, "#UTYP %s\n", 
		(user.typ==UTYP_USER) ? "user" 
		: (user.typ==UTYP_GATE) ? "gate" : "box");
	
	fprintf (exp_f, "#MAXLEN=%lu\n",ta_opts.maxlen);
	if (ta_opts.umlaut)
		fprintf (exp_f, "#UW\n");
	if (ta_opts.quarkoutfile)
		fprintf (exp_f, "#QUARKOUTFILE\n"); /* Quarkoutfile benutzen */
	if (ta_opts.only_one_group)
		fprintf (exp_f, "#OC\n"); /* ohne Crosspostings */
	if (ta_opts.async_export)
		fprintf (exp_f, "#ASYNC-EXPORT\n"); 
	if (ta_opts.pathexcludes)
		fprintf (exp_f, "#PATHEXCLUDES %s\n",ta_opts.pathexcludes); 

	for (i = 1; i < max_brettnr; i++)
	{
		char c;
		if (!br_ind[i].brett)
			continue;
		/* format: 
		 * C brettnr name.
		 */
		if (br_ind[i].brett->flags & B_ZWANG && !user.pflicht_befreit)
			c='P';
		else if (ta_blist[i]==GRP_BESTELLT)
			c='+';
		else if (ta_blist[i]==GRP_IHAVE)
			c='I';
		else
			continue;

		putc(c,exp_f);
		fputs(br_ind[i].brett->name,exp_f);
		putc('\n',exp_f);
	}
	fclose (exp_f);
}								/* ta_inst_sichern */

/* 
 * Asc-Datei in interne Tauschbrettliste importieren
 *
 * ->           datei   :   Name der ASCII-Datei
 *
 */
static void
ta_inst_laden (const char *datei)
{
	FILE *imp_f;
	BRETTNR i = 0;
	char buffer[1024];
	char *p;
	char *q;

	if ((imp_f = fopen (datei, "r")) == NULL)
		return;

	ta_opts.maxlen=16*1024; /* Default */

	while (fgets (buffer, sizeof (buffer) - 1, imp_f))
	{
		p = strpbrk (buffer, "\r\n");
		if (p)
			*p = '\0';
		if (*buffer=='#')
		{
			char *inhalt;
			inhalt=strchr(buffer,'=');
			if (inhalt)
			{
				/* Wertzuweisungen */
				*inhalt++='\0';
				if (!strcmp (buffer + 1, "DA"))
				{
					ta_opts.last_mailtausch = strtoul (inhalt, NULL, 10);
					ta_opts.last_newstausch=ta_opts.last_mailtausch;
				}
				else if (!strcmp (buffer + 1, "TAUSCHDATUM"))
				{
					mtzeit_to_time_t(inhalt, &ta_opts.last_mailtausch);
					ta_opts.last_newstausch=ta_opts.last_mailtausch;
				}
				else if (!strcmp (buffer + 1, "MAILTAUSCHDATUM"))
					mtzeit_to_time_t(inhalt, &ta_opts.last_mailtausch);
				else if (!strcmp (buffer + 1, "NEWSTAUSCHDATUM"))
					mtzeit_to_time_t(inhalt, &ta_opts.last_newstausch);
				else if (!strcmp (buffer + 1, "STATUSDATUM"))
					mtzeit_to_time_t(inhalt, &ta_opts.last_status);
				else if (!strcmp (buffer + 1, "REN"))
					ta_opts.last_ren = strtoul (inhalt, NULL, 10);
				else if (!strcmp (buffer + 1, "RENDATUM"))
					mtzeit_to_time_t(inhalt, &ta_opts.last_ren);
				else if (!strcmp (buffer + 1, "UB"))
					ta_opts.umbruch = strtol (inhalt, NULL, 10);
				else if (!strcmp (buffer + 1, "MAXLEN"))
					ta_opts.maxlen = strtoul (inhalt, NULL, 10);
				else if (!strcmp (buffer + 1, "PATHEXCLUDES"))
					ta_opts.pathexcludes=XSTRDUP(inhalt);
			}
			else
			{
				/* Flags */
				if (!strcmp (buffer + 1, "UW"))
					ta_opts.umlaut = TRUE;
				if (!strcmp (buffer + 1, "QUARKOUTFILE"))
					ta_opts.quarkoutfile = TRUE;
				if (!strcmp (buffer + 1, "OC"))
					ta_opts.only_one_group = TRUE;
				if (!strcmp (buffer + 1, "ASYNC-EXPORT"))
					ta_opts.async_export = TRUE;
			}
			continue;
		}
		p=buffer+1;
		while (isspace(*p))
			p++;
		if (*p) {
			q=p+1;
			while (isdigit(*q))
				q++;
			if (q!=p+1 && *q) {
				p=q+1;
				while(isspace(*p))
					p++;
			}
		} else
			continue; /* garbage line */
		i=(BRETTNR) brettnr(p);
		if (i<0 || i>=max_brettnr) 
			continue;
		if (!i) {
			int tries=5;
			while (tries--) {
				bb_rename_t *r=bb_find_rename_entry(p);
				if (!r) {
					i=-1;
					break;
				}
				p=r->new_name;
				i=brettnr(p);
				if (i>0 && i<=max_brettnr)
					break;
			}
			if (i<=0 || i>=max_brettnr)
				continue;
		}
		if (*buffer=='-')
			ta_blist[i]=GRP_UNBESTELLT;
		else if (*buffer=='I')
			ta_blist[i]=GRP_IHAVE;
		else
			ta_blist[i]=TRUE;
	}
	fclose (imp_f);
}								/* ta_inst_laden */
