/* ta_hilf.c : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
*/

#include <portab.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <dirent.h>

#include "tomsysdep.h"
#include "user.h"
#include "prototyp.h"
#ifdef OFFLINE2
#include "offline/inc/mdb_global.h"
#else
#include "offline/mdb_global.h"
#endif
#include "brett.h"
#include "modem.h"
#include "routing.h"
#include "cslib.h"

#include "tausch.h"
#include "ta_int.h"
#include "ta_ext.h"

#ifndef __GNUC__
ULONG
line_checksum (const char *s)
{
	size_t cnt = 0;
	ULONG csum = 0;

	while (*s)
	{
		csum += *(s++) * ++cnt;
	}
	return csum;
}								/* line_crc */
ULONG
line_checksum_n (const char *s, size_t n)
{
	size_t cnt = 0;
	ULONG csum = 0;

	while (n--)
	{
		csum += *(s++) * ++cnt;
	}
	return csum;
}								/* line_crc */
#endif

void
ta_umlautwandlung (BOOLEAN out_modeB, const char *to_change)
{
	char cmd[128+PATH_MAX*2];
	int error_occured;
	cstl_t *cs = NULL;
	BOOLEAN do_umlautwandlung=FALSE;
	BOOLEAN do_umbruch=FALSE;
	BOOLEAN do_split=FALSE;

	sprintf (cmd, "%s%s", box_conf.bin_dir, "cstrans");

	if (!fileexist (cmd))
	{
		mboutput ("  Programm fehlt\r\n");
		return;
	}
	strcat(cmd," ");

	if (ta_opts.umbruch && out_modeB)
	{
		sprintf (cmd+strlen(cmd), "-L %lu ", 
			(ULONG) ta_opts.umbruch);
		do_umbruch=TRUE;
	}
	if (ta_opts.maxlen && out_modeB)
	{
		sprintf (cmd+strlen(cmd), "-s %lu ", 
			(ULONG) (ta_opts.maxlen==16384 ? 16000 : ta_opts.maxlen));
		do_split=TRUE;
	}

	if (user.zeichensatz && ta_opts.umlaut)
	{
		cs = get_charset_from_nr (user.zeichensatz);
		if (!cs)
		{
			/* sollte nie passieren */
			mboutput ("    Zeichensatz nicht gefunden???\r\n");
			return;
		}
		if (cs->flags != CSF_NATIVE)
		{
			sprintf (cmd+strlen(cmd), "-%c %s ", 
				(out_modeB) ? 't' : 'f', cs->name);
			do_umlautwandlung=TRUE;
		}
	}
	strcat(cmd,to_change);

	if (!do_umlautwandlung && !do_split && !do_umbruch)
		return;
	mboutput ("  ");
	if (do_umlautwandlung)
		mboutput ("Umlautwandlung ");
	if (do_umbruch)
		mboutput ("Umbruch ");
	if (do_split)
		mboutput ("Splitten ");
	mboutput (" ...\r\n");

	if (!fileexist (to_change))
	{
		mboutput ("  ... keine Datei vorhanden!\r\n");
		return;
	}

	error_occured = call_extern(cmd,
		NULL,                 /* spezielles Arbeitsverzeichnis */
		time(NULL)+600,       /* killzeitpunkt,just for the case */
		FALSE,                /* ignore_exitcode */
		FALSE);               /* nicht unter uid des Users */

	if (error_occured)
		mbprintf (" Fehler: %d\r\n", error_occured);

}								/* tausch_umlautwandlung */

BOOLEAN
ta_add_rename_info(time_t zeit, const char *from, const char *to)
{
	char fname[PATH_MAX];
	FILE *f;
	sprintf(fname,"%srename.dat",box_conf.var_dir);
	f=fopen(fname,BINAPPEND);
	if (f)
	{
		fprintf(f,"%ld@%s@%s\n",zeit,from,to);
		fclose(f);
		return TRUE;
	}
	return FALSE;
}

BOOLEAN
ta_check_readperm (BRETTNR bnr, const char **err)
{
	BOOLEAN darf = TRUE;

	if (err) *err=NULL;
	if (!(br_ind[bnr].brett->flags & B_READ))
	{
		darf = FALSE;
		if (err)
		{
			if (br_ind[bnr].brett->flags & B_GEHEIM)
			{
				if (!b_tst_access (bnr, user.nummer))
					*err="geheime Gruppe. Woher kennen Sie die?";
			}
			if (!*err) *err="Lesezugriff verweigert";
		}
	}
	if (darf && br_ind[bnr].brett->flags & B_PASSWORT && level()<8)
	{
		darf = FALSE;
		if (err) *err="Passwortgeschtzte Gruppen sind nur online erreichbar";
	}
	if (darf && br_ind[bnr].brett->flags & B_LESMITGLIED)
	{
		if (!b_tst_access (bnr, user.nummer))
		{
			darf = FALSE;
			if (err) *err="Sie sind kein Mitglied. Wenden Sie sich an 'Chef Gruppe'";
		}
	}
	if (darf && br_ind[bnr].brett->flags & B_ACL)
	{
		char aclfname[PATH_MAX];

		sprintf (aclfname, "%s%s", brettpfad (bnr), "acl.lst");
		if (!acl_check (aclfname, user.name, level (), 'R', NULL))
		{
			darf = FALSE;
			if (err) *err="Sie sind nicht in die Zugriffskontrolliste eingetragen";
		}
	}
	if (level()==9)
		darf=TRUE;
	return darf;
}

FILE *
ta_open_out_block(const char *cmd)
{
	char fname[PATH_MAX];
	FILE *out;
	sprintf(fname, "%s" KOMMANDO_OUT_F , tom_tmp_dir );
	out=xfopen(fname,"a");
	ta_opts.out_block_counter++;
	fprintf(out,"#OUT%c%c%c%c%c\r\n",
		'A'+ta_opts.out_block_counter / 26*26*26*26,
		'A'+(ta_opts.out_block_counter / 26*26*26) % 26,
		'A'+(ta_opts.out_block_counter / 26*26) % 26,
		'A'+(ta_opts.out_block_counter / 26) % 26,
		'A'+(ta_opts.out_block_counter ) % 26
		);
	fprintf(out,":%s\r\n",cmd);
	return out;
}
