/* ta_imp.c : Nachrichtenimport, Hauptrouting */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/

#include "tomsysdep.h"

#include <portab.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <dirent.h>
#include <ctype.h>
#include <syslog.h>
#include <time.h>
#include <sys/stat.h>
#include <unistd.h>

#include "prototyp.h"
#include "user.h"
#include "brett.h"
#include "modem.h"
#include "tompaths.h"
#include "routing.h"
#include "msg.h" /* write_msg fr #LOG */

#include "tausch.h"
#include "ta_int.h"
#include "ta_ext.h"
#include "ta_info.h"

#ifndef LINEEND
#define LINEEND "\r\n"
#endif

#define MAX_KOPIEN 10

typedef enum 
{ 
	umode_normal,  /* kein unbekannter Block */
	umode_show,    /* "show" dem user. Hm. */
	umode_keep,    /* "keep", irgendwo ablegen */
	umode_trash,   /* "trash", weg damit */
	umode_panic    /* das mu man kennen */
} unknown_mode;


extern BRETTINFO *br_ind;
extern const char *tausch_processed_file;

/* **************************************************************************
 * Beschreibung der Funktionen mitsamt Prototypen
 ****************************************************************************
 */

#define ERRLOG(x) { \
	ta_msg_log("  ?%s\n",x); \
	ta_do_log_error(x); \
	}

static void ta_debug_output (message_t *head);

static BOOLEAN ta_getline(FILE *f, char **buffer);
static BOOLEAN ta_read_message(FILE *f, char **buffer, message_t *head, 
	BOOLEAN *errflag, unknown_mode *umode);

/*****************************************************************************
 * Ende der Prototypen
 *****************************************************************************
 */

#define TA_LINEBUFSIZE 16384
static char *linebuf;
static BOOLEAN zurueckgestellt;

static char *path_entry;
static BOOLEAN first_zuordnung;

static BOOLEAN 
ta_getline(FILE *f, char **buffer)
{
	char *p;
	if (zurueckgestellt)
	{
		zurueckgestellt=FALSE;
		return TRUE;
	}
	if (f)
	{
		if (fgets(linebuf,TA_LINEBUFSIZE,f))
		{
			p=linebuf;
			while (*p)
			{
				if (*p>0 && *p<32 && !isspace(*p))
					*p='.';
				else if (*p=='\r' || *p=='\n')
				{
					*p=0;
					return TRUE;
				}
				p++;
			}
			return TRUE;
		}
		return FALSE;
	}
	else
	{
		/* Aus "buffer" lesen */
		if (!**buffer)
		{
			*linebuf=0;
			return TRUE;
		}
		p=strpbrk(*buffer,"\r\n");
		if (p)
		{
			if (*p=='\r')
				*p++=0;
			if (*p=='\n')
				*p++=0;
			strcpy(linebuf,*buffer);
			*buffer=p;
		}
		else
		{
			strcpy(linebuf,*buffer);
			*buffer=*buffer+strlen(*buffer);
		}
		p=linebuf;
		while (*p)
		{
			if (*p>0 && *p<32 && !isspace(*p))
				*p='.';
			p++;
		}
		return TRUE;
	}
	
}

static BOOLEAN
work_dyn_headerline(char **hl, const char *bufline, char kennung, 
					BOOLEAN shorten)
{
	size_t l;
	if (*hl)
	{
		char buf[256];
		sprintf(buf,"%c-Zeile doppelt",kennung);
		ERRLOG(buf);
		return FALSE;
	}
	l=strlen(bufline);
	if (l>=HEADLEN)
	{
		if (!shorten)
		{
			char buf[256];
			sprintf(buf,"%c-Zeile zu lang",kennung);
			ERRLOG(buf);
			return FALSE;
		}
		*hl=XMALLOC(HEADLEN);
		memcpy(*hl,bufline,HEADLEN - 1);
		(*hl)[HEADLEN-1]='\0';
		return TRUE;
	}
	*hl=XSTRDUP(bufline);
	return TRUE;
}

static BOOLEAN
work_headerline(char *hl, const char *bufline, char kennung, 
	BOOLEAN shorten)
{
	if (*hl)
	{
		char buf[256];
		sprintf(buf,"%c-Zeile doppelt",kennung);
		ERRLOG(buf);
		return FALSE;
	}
	if (strlen(bufline)>=HEADLEN)
	{
		if (!shorten)
		{
			char buf[256];
			sprintf(buf,"%c-Zeile zu lang",kennung);
			ERRLOG(buf);
			return FALSE;
		}
		strncpy(hl,bufline,HEADLEN-1);
		hl[HEADLEN]='\0';
		return TRUE;
	}
	strcpy(hl,bufline);

	return TRUE;
}

static BOOLEAN
append_line_to_list(simplelist_t **list, const char *line)
{
	simplelist_t *entry;
	simplelist_t *last;
	for(last=*list;last && last->next;last=last->next)
		/* nothing */;
	entry=simple_alloc(line);
	if (!entry)
	{
		ERRLOG("out of virtual memory");
		return FALSE;
	}
	entry->next=NULL;
	if (last)
		last->next=entry;
	else
		*list=entry;
	return TRUE;
}

static
#if (defined(__GNUC__) && defined(__OPTIMIZE__))
  inline
#endif
void
add_line_to_body(message_t *head,char *loc_linebuf)
{
	size_t l=strlen(loc_linebuf);
	if (l+head->body.physlen+3>head->body.bufsize)
	{
		size_t s=16*1024;
		while (s<l+3)
			s*=2;
		if (!head->body.buf)
		{
#ifdef MSG_MMAP
			head->body.buf=large_malloc(s);
#else
			head->body.buf=XMALLOC(s);
#endif
			*head->body.buf='\0';
			head->body.next=head->body.buf;
		}
		else
		{
#ifdef MSG_MMAP
			head->body.buf=large_realloc(head->body.buf,head->body.bufsize+s);
#else
			head->body.buf=XREALLOC(head->body.buf,head->body.bufsize+s);
#endif
			if (!head->body.bufsize)
				*head->body.buf='\0';
			head->body.next=head->body.buf+strlen(head->body.buf);
		}
		head->body.bufsize+=s;
	}
	if (l)
		memcpy(head->body.next,loc_linebuf,l);
	head->body.next+=l;
	*head->body.next=0; /* Stringende */
	head->body.len+=l-1; /* wegen :, der zhlt nicht auf die 16K */
	head->body.physlen+=l;
	strcpy(head->body.next,LINEEND);
	head->body.next+=sizeof(LINEEND)-1; /* -1 wg. "\r\n\0" */
	head->body.len+=1; /* CR/LF zhlt nur als ein Zeichen */
	head->body.physlen+=sizeof(LINEEND)-1;
}

static BOOLEAN
insert_line_in_list(simplelist_t **list, const char *line)
{
	simplelist_t *entry;
	entry=simple_alloc(line);
	if (!entry)
	{
		ERRLOG("out of virtual memory");
		return FALSE;
	}
	entry->next=*list;
	*list=entry;
	return TRUE;
}


static BOOLEAN 
ta_read_message(FILE *f, char **buffer, message_t *head, BOOLEAN *errflag, unknown_mode *umode)
{
	BOOLEAN doublecross;
	BOOLEAN star;
	char zwi[256];

	doublecross=FALSE;
	star=FALSE;
	*umode=umode_normal;

	*errflag=FALSE;

	/* alten Messageheader lschen */
	msg_clean_message(head);
	head->msgtyp=false_msg;

	while (ta_getline(f,buffer))
	{
#if 0
		if (!*linebuf)
			break;
#endif
		switch (*linebuf)
		{
		case '*':
			if (star)
			{
				zurueckgestellt=TRUE;
				return TRUE;
			}
			star=TRUE;
			ta_passinfo.send_star=TRUE;
			ta_do_log (":*%s" LINEEND, linebuf+1);
			ta_msg_log("%s\n", linebuf);

			switch(linebuf[1])
			{
			case 0: /* letzte Nachricht */
				return FALSE;
			case 'P': 
				head->msgtyp=pers_msg;
				break;
			case 'G': 
				head->msgtyp=gruppen_msg;
				break;
			case 'R': 
				head->msgtyp=ren_msg;
				break;
			case 'N': 
				head->msgtyp=nocem_msg;
				break;
			case 'H': 
				head->msgtyp=head_msg;
				break;
			case 'L': 
				head->msgtyp=log_msg;
				break;
			case 'C': 
				head->msgtyp=cmd_msg;
				break;
			case 'S': 
				head->msgtyp=status_msg;
				break;
			case 'B': 
				head->msgtyp=cnf_msg;
				break;
			case 'I': 
				head->msgtyp=info_msg;
				break;
			case 'O': 
				head->msgtyp=out_msg;
				break;
			default: /* typ unbekannt */
				switch(linebuf[1])
				{
				case 'S':
					*umode=umode_show;
					break;
				case 'K':
					*umode=umode_keep;
					break;
				case 0:
				case 'T':
					*umode=umode_trash;
					break;
				default:
				case 'P':
					*umode=umode_panic;
					break;
				}
				break;
			}
			break;
		case '#':
			if (doublecross)
			{
				zurueckgestellt=TRUE;
				return TRUE;
			}
			doublecross=TRUE;

			ta_do_log (":#%s" LINEEND, linebuf+1);
			ta_msg_log("%s\n", linebuf);

			if (!strcmp(linebuf,"#"))
				return FALSE; /* letzte Nachricht */

			strcpy(head->msg_id,linebuf+1);
			break;
		case '-': 
			/* Wegen alten kurzen Message-IDs in den Frontend-Datenbanken
			 * etwas umstndlich (Problem: R knnte vor - gekommen sein).
			 */
			if (!linebuf[1]) {
				ERRLOG("Illegale leere Minuszeile"); 
				*errflag=TRUE;
			} else if (user.typ!=UTYP_USER || !strchr(head->ref_id,'.')) {
				if (user.typ!=UTYP_USER && *head->ref_id)
				{
					ERRLOG("Minuszeile doppelt");
					*errflag=TRUE;
				}
				strcpy(head->ref_id,linebuf+1);
			}
			break;
#define SIMPEL(kenn,headername,shorten) \
case kenn: \
	*errflag|=!work_headerline(head->headername,linebuf+1,kenn,shorten); \
	break
#define DYN_SIMPEL(kenn,headername,shorten) \
case kenn: \
	*errflag|=!work_dyn_headerline(&head->headername,linebuf+1,kenn,shorten); \
	break;

		SIMPEL('V',absender,TRUE);
		SIMPEL('S',sender,TRUE);
		SIMPEL('T',reply_to,TRUE);
		SIMPEL('W',betreff,TRUE);
		SIMPEL('E',eingabezeit,FALSE);
		SIMPEL('B',bearbeitungsstatus,FALSE);
		SIMPEL('O',organisation,TRUE);
		SIMPEL('N',realname,TRUE);
		DYN_SIMPEL('Y',gateway_zeile,TRUE);
		DYN_SIMPEL('M',mimeinfo,FALSE);
		DYN_SIMPEL('p',path_zeile,TRUE);
		DYN_SIMPEL('X',flags_zeile,TRUE);
#undef SIMPEL
#if 0
	bad idea (tm)
			if (!*errflag && !strchr(head->empfaenger,'@'))
			{
				strcat(head->empfaenger," @ ");
				strcat(head->empfaenger,box_conf.domainname);
			}
#endif
			break;
		case 'D':
			head->distribution_byte=linebuf[1];
			if (!linebuf[1] || linebuf[2])
			{
				ERRLOG("Illegale Distribution");
				*errflag=TRUE;
			}
			break;
		case 'G': 
			*errflag|=!append_line_to_list(&head->gruppen,linebuf+1); 
			break;
		case 'F': 
			*errflag|=!append_line_to_list(&head->followup,linebuf+1); 
			break;
		case 'I':
			if (user.typ==UTYP_GATE)
			{
				/* Gateway drfen in der #-Zeile einen Index angeben
				 * und in der I-Zeile die echte ID nachliefern. 
				 */
				if (strchr(head->msg_id,'@'))
				{
					ERRLOG("Echte ID in # *und* I verboten"); 
					*errflag=TRUE;
				} else if (!linebuf[1]) {
					ERRLOG("Illegale leere I-Zeile"); 
					*errflag=TRUE;
				} else
					strcpy(head->msg_id,linebuf+1);
				break;
			}
			/* FALLTHROUGH */
		case 'R':
			if (*linebuf=='R' && user.typ==UTYP_USER 
				&& !strchr(head->ref_id,'.'))
			{
				/* Temporrfix fr Umstellung - die Frontends haben am
				 * Anfang die langen References in der R-Zeile.
				 */
				strcpy(head->ref_id,linebuf+1);
			}
			else
			{
				/* Tom don't needs them, and doesn't really like them
				 * either.
				 */
				sprintf(zwi,"%c-Zeile weder erlaubt noch notwendig",*linebuf);

				ERRLOG(zwi);
				*errflag=TRUE;
			}
			break;

		case 'A':
		case 'K': 
			*errflag|=!insert_line_in_list(&head->to,linebuf+1); 
			if (*linebuf=='K' && ta_getline(f,NULL))
			{ 
				/* Das verlt sich darauf, da insert_line_in_list
				 * neue Elemente vorne in die Liste einfgt 
				 */
				if (*linebuf=='>')
				{
					/* Failure doesn't matter? */
					head->to->xstring=strdup(linebuf+1);
				}
				else
					zurueckgestellt=TRUE;
			}
			break;
		case '>':
			*errflag|=!append_line_to_list(&head->comments,linebuf+1); 
			break;
		case '!':
			*errflag|=!append_line_to_list(&head->uo_comments,linebuf+1); 
			break;
		case ':':
			add_line_to_body(head,linebuf);
			break;
		default:
			break;
		} /* switch */
	} /* while */
	if (doublecross)
		return TRUE;
	return FALSE;
}


BOOLEAN
ta_infile_einsortieren (const char *infilename)
{
	char id_fname[PATH_MAX];
	char immer_fname[PATH_MAX];
	const char *aliasname;
	time_t t;
	FILE *f;
	FILE *id_f;
	message_t *head;
	BOOLEAN errflag;
	BOOLEAN panic=FALSE;
	unknown_mode umode;
	const char *preprozessor=NULL;

	if (user.typ!=UTYP_USER && ta_opts.quarkoutfile)
	{
        int error_occured;
        char buf[PATH_MAX * 4];
        char buf2[PATH_MAX];
		sprintf(buf,"%s%lu.qin",box_conf.tausch_dir,(ULONG) user.nummer);
		if (fileexist(buf)) {
			sprintf(buf,"%s%lu.qin %s",box_conf.tausch_dir,(ULONG) user.nummer,
				infilename);
			error_occured = call_extern (buf,
								 NULL,  /* spezielles Arbeitsverzeichnis */
								 akt_anruf.soft_limit + 600, /* killzeitpunkt */
								 FALSE,     /* ignore_exitcode */
								 FALSE);    /* nicht unter uid des Users */
			if (error_occured)
			{
				mbprintf (" Quarkoutfilescript beendete sich mit Exitcode %d\r\n",
						  error_occured);
				syslog (LOG_ERR,"Quarkoutfilescript: Exitcode %d\r\n", error_occured);
				return FALSE;
			}
		} else {
			sprintf(buf,
				"%s/quarkoutfile -k -p %s%s.quarkoutfile -L --mysite %s -i %s -o %s",
				box_conf.bin_dir, box_conf.log_dir, datum_pro_S(),
				box_conf.mausnet_kennung,
				infilename,infilename);
			sprintf(buf2,"%s%lu.grp",box_conf.tausch_dir,(ULONG) user.nummer);

			if (fileexist(buf2))
			{
				strcat(buf," -G ");
				strcat(buf,buf2);
			}

			mboutput("Quarkoutfile aufrufen\r\n");

			error_occured = call_extern (buf,
								 NULL,  /* spezielles Arbeitsverzeichnis */
								 akt_anruf.soft_limit + 600, /* killzeitpunkt */
								 FALSE,     /* ignore_exitcode */
								 FALSE);    /* nicht unter uid des Users */
			if (error_occured)
			{
				mbprintf (" Quarkoutfile beendete sich mit Exitcode %d\r\n",
						  error_occured);
				syslog (LOG_ERR,"Quarkoutfile: Exitcode %d\r\n", error_occured);
				return FALSE;
			}
		}

	}

	if (user.typ==UTYP_USER)
		preprozessor=box_conf.tausch_in_user_preprozessor;
	else if (user.typ==UTYP_GATE)
		preprozessor=box_conf.tausch_in_gate_preprozessor;
	else
		preprozessor=box_conf.tausch_in_box_preprozessor;
	if (preprozessor && *preprozessor)
	{
		int orror_occured;
		char buf[PATH_MAX*4];
		tausch_processed_file=infilename;
		if (!do_interpret(preprozessor,buf,sizeof(buf)))
		{
			mbprintf(
				" Interpretation des Preprozessorkommandos gescheitert\r\n");
			syslog(LOG_ERR,
				" Interpretation des Preprozessorkommandos gescheitert\r\n");
			tausch_processed_file=NULL;
		}
		else
		{
			struct stat st1;
			struct stat st2;
			tausch_processed_file=NULL;
			stat(infilename,&st1);
			mboutput(" Preprozessor ...\r\n");
			orror_occured = call_extern(buf,
				NULL,                 /* spezielles Arbeitsverzeichnis */
				akt_anruf.soft_limit+600,/* killzeitpunkt,just for the case */
				FALSE,                /* ignore_exitcode */
				FALSE);               /* nicht unter uid des Users */
			if (orror_occured)
			{
				mbprintf(" Preprozessor beendete sich mit Exitcode %d\r\n",
					orror_occured);
				stat(infilename,&st2);
				if (st1.st_mtime != st2.st_mtime )
				{
					ta_userdupecheck_end();
					return FALSE;
				}
			}
		}
	}

	f=fopen(infilename,"r");
	if (!f)
	{
		mboutput(" kann Infile nicht lesen\r\n");
		ta_userdupecheck_end();
		return FALSE;
	}

	if (user.typ!=UTYP_USER)
	{
		const char *zwi;
		zwi=get_path_entry(user.nummer);
		if (!zwi)
		{
			mboutput(
" Nachrichtenrouter nicht korrekt konfiguriert, kein Patheintrag vorhanden\r\n"
				);
			ta_userdupecheck_end();
			fclose(f);
			return FALSE;
		}
		/* um von nderungen in der Routingdatenbank verschont zu bleiben! */
		path_entry=XSTRDUP(zwi);
	}
	else
		path_entry=NULL;

	first_zuordnung=TRUE;

	sprintf (id_fname, "%s" NOEXPID_LST_F, tom_tmp_dir);
	id_f = xfopen (id_fname, BINAPPEND);

	ta_open_log();

	linebuf=XMALLOC(TA_LINEBUFSIZE);
	zurueckgestellt=FALSE;
	if (user.vertreter[0])
	{
		/* Nur um dem User auf den Grund zu stoen, warum er keine PMs bekommt?
		 * Nein, weit gefehlt - auch um ihm eine Chance zu geben zu erkennen
		 * da der Sysop ein Fiesling ist und seine PMs umleitet.
		 */
		ta_do_log (":!Sie haben einen Vertreter angegeben: %s" LINEEND,
			user.vertreter);
	}
	aliasname=routing_suche_alias(user.name);
	if (aliasname)
	{
		/* same reason here */
		ta_do_log (":!Es existiert ein Alias fr Sie auf %s" LINEEND,
			aliasname);
	}

	head=msg_alloc();

	/* Nchster #CMD ist der erste (passinfo initialisieren) */
	ta_opts.first_cmd_block=TRUE;

	while (ta_read_message(f,NULL,head,&errflag,&umode))
	{
		BOOLEAN bounce_it=FALSE;
		r_adr_t *adr=NULL;

		head->importzeit=time(NULL);
		switch(umode)
		{
		case umode_normal:
			/* bekannter Nachrichtentyp, nur mit "*"-Zeile */
			break;
		case umode_panic:
			ERRLOG("Panicflag gesetzt. Abbruch!");
			panic=TRUE;
			break;
		case umode_trash:
			ERRLOG("trashed");
			continue;
			break;
		case umode_keep:
			ERRLOG("trashed, not keeped");
			continue;
			break;
		case umode_show:
			if (user.typ==UTYP_USER)
			{
				ERRLOG("schreiben sie den Sysop doch selbst an!");
				continue;
			}
			/* wat nu? */
			if (!append_line_to_list(&head->to,"sysop"))
			{
				ERRLOG("trashed, no memory, no show");
				errflag=TRUE;
				continue;
			}
			head->msgtyp=pers_msg;
			sprintf(head->betreff,
				"unbekannter Block im Infile von %s",user.name);
			sprintf(head->absender,"Quark");
			make_msg_id(head->msg_id,TRUE,NULL,NULL);
			time_t_to_mtzeit_sec(&head->importzeit,head->eingabezeit);
			head->eingabezeit_tt=head->importzeit;
			break;
		}
		if (panic)
			break;
		if (errflag)
		{
			ta_msg_log("  ?garbage\n");
		}
		else
		{
			errflag=!ta_artcheck(head,&adr,&first_zuordnung,
				path_entry,&bounce_it);
			if (logout_modus)
				panic=TRUE;
		}
		if (head->msgtyp==pers_msg)
		{
			ta_debug_output(head);
			/* Nachrichten transportieren, mit Ausnahme der Fehler */
			r_transport(head,adr,FALSE);
			/* Nachrichten transportieren, diesmal die Fehler */
			r_transport(head,adr,TRUE);

			r_free_chain(adr);

#if 0
			/* wird nun von ta_artcheck() erledigt. */
			/* Kopie an den User? */
			if (ta_passinfo.send_pe)
			{
				/* Im Gegensatz zur MAUS mu hier eine echte Kopie erzeugt
				 * werden. Der zustzliche Routeraufruf ist verhltnismig
				 * teuer, aber falls das mal problematisch wird, knnte man
				 * den Router auch "einfach" bergehen und direkt den
				 * Transportmechanismus anwerfen.
				 * Oder in das Outfile fr den User schreiben, aber das
				 * haben wir u.U. nicht verfgbar (--infile ohne --outfile).
				 *
				 * Noch ein Problem: Wenn Kopien an andere Benutzer
				 * versandt worden sie (K-Zeilen oder mehrere A-Zeilen),
				 * dann bekommt der Benutzer keine Kopien davon. Der
				 * tiefere Grund, auer Programmiererfaulheit, ist, da
				 * nur geraten werden kann, welche ID die Kopien bekommen 
				 * haben. Das Problem erledigt sich allerdings, sobald
				 * alle Kopien dieselbe ID erhalten.
				 */
				adr=r_alloc();
				adr->inputadr=concat(user.name,"@", MYDOMAINPART, NULL);
				if (!adr->inputadr)
					r_free_chain(adr);
				else
				{
					r_router(head,adr);
					r_transport(head,adr,FALSE);
					r_free_chain(adr);
				}
			}
#endif
			/* Next Step: Nun mssen wir einen "beantwortet"-Status
			 * setzen, wenn die Nachricht transportiert werden 
			 * konnte.
			 */
			if (head->ref_id)
			{
				r_adr_t *akt;
				BOOLEAN one_sent=FALSE;
				for (akt=adr;akt;akt=akt->next)
				{
					if (akt->error)
						one_sent=TRUE;
				}
				if (one_sent)
				{
					r_st_set_status(head->ref_id,
						head->absender,
						NULL,
						"B",
						head->eingabezeit_tt);
				}
			}
		}
		else if (!errflag)
		{
			if (head->msgtyp==status_msg)
			{
				ta_work_status(head);
			}
			else if (head->msgtyp==gruppen_msg)
			{
				ta_debug_output(head);
				fprintf(id_f,"%s\n",head->msg_id);
				r_transport(head,NULL, FALSE);
			}
			else
			{
				/* Nur wenn nicht leer ... */
				if (head->body.buf && *head->body.buf)
					ta_work_special(head);
			}
			continue;
		}
	} /* while nicht Ende der Datei */

	fclose (f);
	fclose (id_f);

	sprintf(immer_fname,"%s%lu.immer",box_conf.tausch_dir,user.nummer);
	f=fopen(immer_fname,"r");
	if (f) {
		if (!linebuf)
			linebuf=XMALLOC(TA_LINEBUFSIZE);
		while (fgets(linebuf,TA_LINEBUFSIZE,f)) {
			char *p=strrchr(linebuf,'\r');
			if (p)
				*p=0;
			ta_cmd_work(linebuf);
		}
		fclose(f);
	}

	msg_free(head);

	free(linebuf);
	if (path_entry)
		free(path_entry);

	if (!panic)
	{

		time(&t);
		msglog("T Einsortieren %s",ctime(&t));
		mboutput (" Daten werden importiert\r\n");
		r_flush();
	}

	time(&t);
	ta_do_log_end();
	msglog("T Ende %s",ctime(&t));
	msglog_end();

	mboutput ("\r\n");
	ta_info_sync();

	return !panic;
}

static void
ta_debug_output (message_t *head)
{
	size_t l;
	simplelist_t *akt;
	if (box_conf.tausch_debug_level==0 
		|| (!akt_connection.is_console && box_conf.tausch_debug_level==1))
		return;
	printf ("-------------\n");
	printf ("ID:          %-60.60s\n", head->msg_id);
	printf ("Komm_ID      %-60.60s\n", head->ref_id);
	printf ("Von:         %-60.60s\n", head->absender);
	printf ("Sender:      %-60.60s\n", head->sender);
	printf ("Reply-To:    %-60.60s\n", head->reply_to);
	printf ("An:          ");
	l=13;
	for(akt=head->to; akt; akt=akt->next)
	{
		size_t len;
		len=strlen(akt->data);
		if (l+len>75)
		{
			printf("\n            ");
			l=13;
		}
		l+=len+1;
		printf("%s ",akt->data);
	}
	printf("\n");
	printf ("Wegen:       %-60.60s\n", head->betreff);
	printf ("Eingabezeit: %-60.60s\n", head->eingabezeit);
	printf ("Gruppe:      ");
	l=13;
	for(akt=head->gruppen; akt; akt=akt->next)
	{
		size_t len;
		len=strlen(akt->data);
		if (l+len>75)
		{
			printf("\n            ");
			l=13;
		}
		l+=len+1;
		printf("%s ",akt->data);
	}
	printf("\n");
	printf ("Followup-To: ");
	l=13;
	for(akt=head->followup; akt; akt=akt->next)
	{
		size_t len;
		len=strlen(akt->data);
		if (l+len>75)
		{
			printf("\n            ");
			l=13;
		}
		l+=len+1;
		printf("%s ",akt->data);
	}
	printf("\n");
	printf ("Status:      %-60.60s\n", head->bearbeitungsstatus);
	printf ("Organisation:%-60.60s\n", head->organisation);
	printf ("Realname:    %-60.60s\n", head->realname);
	if (head->gateway_zeile)
		printf ("Gateway:     %-60.60s\n", head->gateway_zeile);
	if (head->path_zeile)
		printf ("Path:        %-60.60s\n", head->path_zeile);
	if (head->mimeinfo)
		printf ("MIME:        %-60.60s\n", head->mimeinfo);
	if (head->distribution_byte)
		printf ("Dist:        %c\n", head->distribution_byte);
	printf ("Text:        %ld Bytes\n", (long) head->body.len);
}

