/* tausch.c : still no description */
/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
*/
#include "tomsysdep.h"

#include <portab.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <unistd.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <dirent.h>

#include "prototyp.h"
#include "brett.h"
#include "user.h"
#include "modem.h"
#include "msg.h"
#include "terminal.h"
#include "aktdata.h"
#include "tompaths.h"

#include "packer.h"

#include "lockfile.h" /* libquark */

#include "tausch.h"
#include "ta_int.h"
#include "ta_ext.h"

extern USER user;
extern BOOLEAN mehr_disabled;

const char *tausch_processed_file;

TA_PASSINFO ta_passinfo;

	/* Zeiger auf Variable auf Speicherbereich mit dem Namen des
	 * Spoolverzeichnisses des Users. Der Speicherbereich selbst ist
	 * auf dem Stack und wird in ta_run_tausch angelegt.
	 */
static char *userspooldir;

typedef struct leftover_struct {
	USERNR nr;
	struct leftover_struct *next;
} leftover_t;

/* Dringender Hinweis fr Alle, die nicht verstehen, warum die Listen-
 * elemente beim Import nicht schon gelscht werden: Das ginge schief,
 * sobald ta_run_infiles luft, denn die Funktion luft ber die Liste
 * und verarbeitet sie selbst. Sie, bzw. die von ihr aufgerufenen 
 * Funktionen, arbeiten problemlos weiter, wenn jemand in der Liste
 * eingetragen ist, aber kein Infile von ihm mehr zurckgelassen ist.
 * So don't worry.
 */
static leftover_t *infile_anker;

static BOOLEAN ta_read_incoming_dir (char *itsname);
static BOOLEAN ta_read_passinfo_23 (void);
static BOOLEAN ta_read_passinfo_34 (char *kennung, char *endung);

static BOOLEAN istail (const char *s, const char *t);
static BOOLEAN suche_infile_txt (const char *infilename, char *infiletxtname,
								 char *pkenn, char *pend);
static BOOLEAN is_infile_txt(const char *infilename);
static BOOLEAN ta_phase_empfangen (const char *infilename,
								   char *workfilename);
static BOOLEAN work_on_infile (char *infilename);
BOOLEAN ta_phase_einsortieren (const char *workfilename);
static BOOLEAN ta_phase_exportieren (char *workfilename);
static BOOLEAN ta_phase_senden (const char *workfilename, const char *outfilename);
static BOOLEAN ta_async_copy(const char *fname, BOOLEAN to_tmp);

static BOOLEAN 
ta_async_copy(const char *fname, BOOLEAN to_tmp)
{
	char from[PATH_MAX];
	char to[PATH_MAX];
	if (to_tmp)
	{
		sprintf(to,"%s%s", tom_tmp_dir, fname);
		sprintf(from,"%s%s",  userspooldir, fname);
	}
	else
	{
		sprintf(from,"%s%s", tom_tmp_dir, fname);
		sprintf(to,"%s%s", userspooldir, fname);
	}
	if (fileexist(from) && !file_copy(from,to))
	{
		mboutput(" kann Datei nicht kopieren! Abbruch!\r\n");
		syslog(LOG_ERR,"file_copy %s %s: %s",from,to, strerror(errno));
		return FALSE;
	}
	return TRUE;
}

void
ta_init_passinfo (TA_PASSINFO * i)
{
	i->echo_news = TRUE;
	i->echo_mail = FALSE;
	i->status = TA_KEIN_STATUS;
	i->maxbytes = ULONG_MAX;
	i->send_news = FALSE;
	i->send_mail = FALSE;
	i->send_ren = FALSE;
	i->send_pe = FALSE;

	i->send_crosspostings = TRUE;

	if (i->pkenn)
	{
		free (i->pkenn);
		i->pkenn = NULL;
	}
	if (i->pend)
	{
		free (i->pend);
		i->pend = NULL;
	}

	i->protokoll = 0;
}

static BOOLEAN
istail (const char *s, const char *t)
{
	size_t ls;
	size_t lt;

	ls = strlen (s);
	lt = strlen (t);
	if (ls > lt && !stricmp (s + ls - lt, t))
		return TRUE;
	return FALSE;
}

/*
 * Ist diese Datei mit dem wahrscheinlich vllig gegen die Konvention
 * verstoenden Namen 'infilename' ein infile.txt? 
 * Auf den Namen alleine kann man sich nicht verlassen, zuviele Frontends
 * meinen, beliebige Namen mitschicken zu drfen.
 */
static BOOLEAN
is_infile_txt(const char *infilename)
{
	FILE *f;
	char buf[1024];
	BOOLEAN ok=FALSE;
	char *p;

	f = fopen (infilename, BINREAD);
	if (!f)
		return FALSE;
	if (fgets (buf, sizeof (buf), f))
	{
		BOOLEAN got_cr=FALSE;
		BOOLEAN got_lf=FALSE;
		char *at;
		ok = TRUE;
		buf[sizeof(buf)-1]='\0';

		if (*buf != '#')
			ok = FALSE;
		p=buf;
		while (*p && ok)
		{
			unsigned char c=(unsigned char) *p;
			if (c=='\r')
			{	
				if (got_cr)
					ok=FALSE;
				else
					got_cr=TRUE;
			}
			else if (c=='\n')
			{	
				if (got_lf)
					ok=FALSE;
				else
					got_lf=TRUE;
			}
			else if (c<=32|| c>126)
				ok=FALSE;
			p++;
		}
		at=strrchr(buf,'@');
		if (at) {
			*at++=0;
			if (strpbrk(at,"()<>,;:\\\" \t"))
				ok=FALSE;
		} 
		/* gewisse Spezialzeichen nur mit Escaping erlaubt */
		p=strpbrk(buf,"()<>@,;:\" \t");
		if (p && p[-1]!='\\')
			ok = FALSE;
		if (!got_lf)
			ok = FALSE;
	}
	fclose (f);
	return ok;
}

static BOOLEAN
suche_infile_txt (const char *infilename, char *infiletxtname,
				  char *pkenn, char *pend)
{
	int found = 0;
	size_t l;
	char fullpath[PATH_MAX];
	char *ibasename;

	l = strlen (infilename);
	if (pkenn)
		*pkenn = '\0';
	if (pend)
		*pend = '\0';
	if (istail (infilename, ".txt")
		|| istail(infilename,"anbox") || istail(infilename,"vonbox"))
	{
		if (!strstr(infilename,tom_utmp_dir))
		{
			const char *p;
			p=strrchr(infilename,'/');
			if (!p)
				p=infilename;
			else
				p++;
			sprintf(infiletxtname,"%s%s",tom_utmp_dir,p);
			file_copy(infilename,infiletxtname);
		}
		else
			strcpy (infiletxtname, infilename);
		return TRUE;
	}
	/* 
	 * Textfile, anders benannt?
	 */
	if (is_infile_txt(infilename))
	{
		if (!strstr(infiletxtname,tom_tmp_dir))
		{
			sprintf(infiletxtname,"%sinfile.txt",tom_tmp_dir);
			file_copy(infilename,infiletxtname);
		}
		else
			strcpy (infiletxtname, infilename);
		return TRUE;
	}

#ifdef HAVE_DOSDRIVES
	if (infilename[2]==':')
#else
	if (*infilename=='/')
#endif
		strcpy(fullpath,infilename);
	else
	{
		getcwd(fullpath,sizeof(fullpath));
		strcat(fullpath,"/");
		strcat(fullpath,infilename);
	}
#ifdef HAVE_DOSDRIVES
	{
		const char *grrr=strrchr(fullpath,'/');
		if (grrr)
			ibasename=grrr;
		else
			ibasename=strrchr(fullpath,'\\')+1;
	}	
#else
	ibasename=strrchr(fullpath,'/')+1;
#endif

	if (packer_unpack (fullpath, tom_tmp_dir, TRUE, pkenn, pend))
	{
		/* der Packer hat etwas ausgepackt */
		DIR *d;
		struct dirent *e;
		int durchlauf=0;

#define THISTMP (durchlauf==1 ? tom_tmp_dir : tom_utmp_dir)
second_try:
		durchlauf++;
		d = opendir (THISTMP);
		if (!d)
			return FALSE;

		while ((e = readdir (d)) != NULL)
		{
			if (*e->d_name == '.')
				continue;
			if (!stricmp (e->d_name, ibasename))
				continue;		/* muss das Archiv selbst sein */
			if (!strnicmp (e->d_name, "infile.", 7)
				|| !stricmp (e->d_name, "infile")
				|| !strnicmp (e->d_name, "outfile.", 8)
				|| !stricmp (e->d_name, "outfile")
				|| !strnicmp (e->d_name, "anbox.", 6)
				|| !stricmp (e->d_name, "anbox")
				|| !stricmp (e->d_name, "asi")
				|| !strnicmp (e->d_name, "asi.",4)
				|| !strnicmp (e->d_name, "vonbox.", 7)
				|| !stricmp (e->d_name, "vonbox"))
			{
				/* mit Endung */
				char buf[PATH_MAX];
				sprintf(buf,"%s%s",THISTMP,e->d_name);
				if (is_infile_txt(buf))
				{
					found++;
					strcpy(infiletxtname,buf);
				}
			}
		}
		closedir (d);
		if (found > 1)
		{
			mbprintf ("Zuviele Dateien gefunden - kann Infile nicht erkennen!\r\n");
			return FALSE;
		}
		else if (found == 1)
			return TRUE;

		/* zweiter Versuch, unter Vernachlssigung der Filenamen */
		d = opendir (THISTMP);
		if (!d)
			return FALSE;

		while ((e = readdir (d)) != NULL)
		{
			if (*e->d_name == '.')
				continue;
			if (!stricmp (e->d_name, ibasename))
				continue;		/* muss das Archiv selbst sein */
			{
				/* mit Endung */
				char buf[PATH_MAX];
				sprintf(buf,"%s%s",THISTMP,e->d_name);
				if (is_infile_txt(buf))
				{
					found++;
					strcpy(infiletxtname,buf);
				}
			}
		}
		closedir (d);
		if (found > 1)
		{
			mbprintf ("Zuviele Dateien gefunden - kann Infile nicht erkennen!\r\n");
			return FALSE;
		}
		else if (found == 1)
			return TRUE;
		if (durchlauf==1)
			goto second_try;
	}
	mbprintf ("Kein Infile gefunden!\r\n");
	return FALSE;
}


static BOOLEAN
work_on_infile (char *infilename)
{
	BOOLEAN ret;
	BOOLEAN undupe=FALSE;

	if (user.typ==UTYP_USER)
	{
		if (!ta_userdupecheck_init ())
			return FALSE;
		undupe=TRUE;
	}
	if (ta_opts.umlaut)
	{
		ta_umlautwandlung (FALSE, infilename);
	}

	ret = ta_infile_einsortieren (infilename);

	if (ret && undupe)
		ta_userdupecheck_end ();

	ta_cmd_end ();
	return ret;
}

static BOOLEAN
ta_phase_empfangen (const char *infilename, char *workfilename)
{
	char realfile[PATH_MAX];
	/* Infile empfangen, falls ntig */
	if (!infilename)
	{
		if (!ta_infile_empfangen (realfile))
			return FALSE;
	}
	else
		strcpy (realfile, infilename);
	if (workfilename)
		strcpy(workfilename,realfile);
	return TRUE;
}

static BOOLEAN
ta_read_incoming_dir (char *itsname)
{
	char pad[PATH_MAX];
	char *p;
	DIR *d;
	struct dirent *e;

	sprintf (pad, "%sincoming" , userspooldir);
	d=opendir(pad);
	if (!d)
		return FALSE;
	while ((e=readdir(d))!=NULL)
	{
		p=e->d_name;
		if (*p=='.')
			continue;
		sprintf(itsname,"%s/%s",pad,p);
		closedir(d);
		return TRUE;
	}
	closedir(d);
	return FALSE;
}

/*
 * sortiert die Datei workfilename ein (nicht wenn async!), und alle
 * anderen liegengebliebenen Infiles.
 */
BOOLEAN
ta_phase_einsortieren (const char *workfilename)
{
	char pkenn[128]="\0";
	char pend[128]="\0";
	char infiletxtname[PATH_MAX];
	char packedname[PATH_MAX];
	char *p;

	strcpy(packedname,workfilename);
	if (!suche_infile_txt (packedname, infiletxtname, pkenn, pend))
	{
		return FALSE;
	}
	p=strrchr(infiletxtname,'/');
	if (!p)
		p=infiletxtname;
	else
		p++;
	if (!strnicmp(p,"asi",3)) {
		char aspoold[PATH_MAX];
		char tmpaspoold[PATH_MAX];
		char *ep;
		int en=0;
		sprintf (aspoold,"%sasync_i", box_conf.spool_dir);
		mkdir(aspoold,S_IRWXU);
		sprintf (aspoold,"%sasync_i/%lu", box_conf.spool_dir,(long) user.nummer);
		ep=aspoold+strlen(aspoold);
		do {
			sprintf(ep,".%d.txt",++en);
		} while(fileexist(aspoold));
		sprintf (tmpaspoold,"%sasync_i/_%lu.%d.txt", 
			box_conf.spool_dir, (long) user.nummer, en);
		if (!file_move(infiletxtname,tmpaspoold)) {
			mboutput(
				" Tausch, Async: Kann async-infile.txt nicht verschieben!\r\n");
			syslog(LOG_ERR,
				"Tausch, Async: Kann async-infile.txt nicht verschieben!\r\n");
			syslog(LOG_ERR," nach %s: %s",tmpaspoold, strerror(errno));
			unlink(infiletxtname);
			return FALSE;
		}
		if (!file_move(tmpaspoold,aspoold)) {
			mboutput(
				" Tausch, Async: Kann async-infile.txt nicht verschieben!\r\n");
			syslog(LOG_ERR,
				"Tausch, Async: Kann async-infile.txt nicht verschieben!\r\n");
			syslog(LOG_ERR," nach %s: %s",aspoold, strerror(errno));
			unlink(tmpaspoold);
			return FALSE;
		}
		return TRUE;
	}
	/* 
	 * Falls _later.cmd und/oder _tausch.log gesichert waren, dann
	 * dann mssen sie in den tmp-Ordner gebracht werden, damit 
	 * neue Informationen angehngt werden knnen.
	 */
	if (!ta_async_copy(LATER_CMD_F,TRUE)
		|| !ta_async_copy(TAUSCH_LOG_F,TRUE)
		|| !ta_async_copy(NOEXPID_LST_F,TRUE)
		|| !ta_async_copy(KOMMANDO_OUT_F,TRUE))
		return FALSE;

	/* eventuell alte ta_passinfo_23 lesen */
	if (!ta_read_passinfo_23 ())
		return FALSE;

	if (!work_on_infile (infiletxtname))
	{
		return FALSE;
	}

	if (*pkenn)
	{
		ta_passinfo.pkenn = strdup (pkenn);
		ta_passinfo.pend = strdup (pend);
		if (!ta_passinfo.pkenn || !ta_passinfo.pend)
		{
			speicherproblem (FALSE, "ta_phase_einsortieren");
			return FALSE;
		}
	}

	/* Asyncinfos sichern */

	if (!ta_async_copy(LATER_CMD_F,FALSE)
		|| !ta_async_copy(TAUSCH_LOG_F,FALSE)
		|| !ta_async_copy(NOEXPID_LST_F,FALSE)
		|| !ta_async_copy(KOMMANDO_OUT_F,FALSE))
		return FALSE;
	{
		FILE *f;

		/* Spoolverzeichnis anlegen - es besteht die Moeglichkeit, dass
		 * dies der erste Versuch ist, dort etwas hinzuschreiben.
		 */
		char asyncname[PATH_MAX];
		sprintf (asyncname, "%s" PASSINFO_23, userspooldir);
		f = xfopen (asyncname, BINWRITE);
		if (ta_passinfo.echo_news)
			fprintf (f, "echo news\n");
		if (ta_passinfo.echo_mail)
			fprintf (f, "echo mail\n");
		if (ta_passinfo.status == TA_STATUS_ALLE)
			fprintf (f, "status alle\n");
		if (ta_passinfo.status == TA_STATUS_NEUE)
			fprintf (f, "status neue\n");
		if (ta_passinfo.maxbytes > 0)
			fprintf (f, "maxbytes %ld\n", (long) ta_passinfo.maxbytes);
		if (ta_passinfo.send_news)
			fprintf (f, "send news\n");
		if (ta_passinfo.send_mail)
			fprintf (f, "send mail\n");
		if (!ta_passinfo.send_crosspostings)
			fprintf (f, "ohne crosspostings\n");
		if (ta_passinfo.protokoll)
			fprintf (f, "protokoll %c\n",ta_passinfo.protokoll);
		if (ta_passinfo.pkenn)
		{
			fprintf (f, "packer %s\n", ta_passinfo.pkenn);
			fprintf (f, "endung %s\n", ta_passinfo.pend);
		}
		fclose (f);
	}
	return TRUE;
}

static BOOLEAN
ta_read_passinfo_23 (void)
{
	char asyncname[PATH_MAX];
	char buf[1024];
	char *p;
	FILE *f;

	sprintf (asyncname, "%s" PASSINFO_23 , userspooldir);
	f = fopen (asyncname, BINREAD);
	if (!f)
		return TRUE; /* dafaults gesetzt */
	while (fgets (buf, sizeof (buf), f))
	{
		p = strpbrk (buf, "\r\n");
		if (p)
			*p = '\0';
		strtrim (buf);
		if (!stricmp (buf, "echo news"))
			ta_passinfo.echo_news = TRUE;
		else if (!stricmp (buf, "echo mail"))
			ta_passinfo.echo_mail = TRUE;
		else if (!stricmp (buf, "status alle"))
			ta_passinfo.status = TA_STATUS_ALLE;
		else if (!stricmp (buf, "status neue"))
			ta_passinfo.status = TA_STATUS_NEUE;
		else if (!strnicmp (buf, "maxbytes", 8))
			ta_passinfo.maxbytes = strtol (buf + 8, NULL, 10);
		else if (!stricmp (buf, "send any"))
		{
			ta_passinfo.send_news = TRUE;
			ta_passinfo.send_mail = TRUE;
		}
		else if (!stricmp (buf, "send mail"))
			ta_passinfo.send_mail = TRUE;
		else if (!stricmp (buf, "send news"))
			ta_passinfo.send_news = TRUE;
		else if (!stricmp (buf, "ohne crosspostings"))
			ta_passinfo.send_crosspostings = FALSE;
		else if (!strnicmp (buf, "protokoll ", 10))
			ta_passinfo.protokoll=buf[10];
		else if (!strnicmp (buf, "packer", 6))
		{
			ta_passinfo.pkenn = strdup (buf + 6);
			if (!ta_passinfo.pkenn)
			{
				fclose (f);
				speicherproblem (FALSE, "ta_phase_exportieren");
				return FALSE;
			}
			strtrim (ta_passinfo.pkenn);
		}
		else if (!strnicmp (buf, "endung", 6))
		{
			ta_passinfo.pend = strdup (buf + 6);
			if (!ta_passinfo.pend)
			{
				fclose (f);
				speicherproblem (FALSE, "ta_phase_exportieren");
				return FALSE;
			}
			strtrim (ta_passinfo.pend);
		}
	}							/* while */
	fclose (f);
	return TRUE;
	/* Async-Infos sind gelesen */
}

static BOOLEAN
ta_read_passinfo_34 (char *kennung, char *endung)
{
	char buf[1024];
	char asyncname[PATH_MAX];
	char *p;
	FILE *f;

	sprintf (asyncname, "%s" PASSINFO_34, userspooldir);
	f = fopen (asyncname, BINREAD);
	if (!f)
		return FALSE;
	*endung = '\0';
	*kennung='\0';
	while (fgets (buf, sizeof (buf), f))
	{
		p = strpbrk (buf, "\r\n");
		if (p)
			*p = '\0';
		strtrim (buf);
		if (!strnicmp (buf, "endung ", 7))
		{
			strtrim (buf + 7);
			strcpy (endung, buf + 7);
		}
		else if (!strnicmp (buf, "kennung ", 8))
		{
			strtrim (buf + 8);
			strcpy (kennung, buf + 8);
		}
	}
	fclose (f);
	return TRUE;
}

/*
 * Erstellt ein Outfile und trgt den Namen in workfilename ein, falls
 * workfilename nicht NULL ist.
 */
static BOOLEAN
ta_phase_exportieren (char *workfilename)
{
	char txtname[PATH_MAX];	/* ungepacktes, zu erstellendes OF */
	char packedname[PATH_MAX];
	char asyncname[PATH_MAX];
	FILE *f;
	FILE *outfile=NULL;
	char *to_send;
	char endung[128]="\0";
	char kennung[128]="\0";

	/* Steuerinfos fr die Erstellung des Outfiles laden */
	if (!ta_read_passinfo_23 ())
		return FALSE;

	/* 
	 * jetzt den Namen des temporren Outfiles festlegen, genauer gesagt
	 * den Namen der Datei, in die gleich exportiert wird.
	 */
	sprintf (txtname, "%soutfile.txt", tom_tmp_dir);

	/* 
	 * information ueber Packer holen 
	 */
	(void) ta_read_passinfo_34 (kennung, endung);
	/* 
	 * outfile zusammenstellen.
	 */
	/* _later.cmd und _tausch.log an den Ort des Geschehens bringen */
	if (!ta_async_copy(LATER_CMD_F,TRUE)
		|| !ta_async_copy(TAUSCH_LOG_F,TRUE)
		|| !ta_async_copy(KOMMANDO_OUT_F,TRUE)
		|| !ta_async_copy(NOEXPID_LST_F,TRUE))
		return FALSE;
	if (!ta_erstelle_outfile (outfile,txtname))
		return FALSE;
	/*
	 * Outfile packen, bei Bedarf.
	 */
	if (ta_passinfo.pkenn)
	{
		char *p;
		char *basep;

		basep=stpcpy (packedname, tom_tmp_dir);
		strcat (basep, "outfile");
		strcat (basep, ta_passinfo.pend);
		/* Einpacken ohne Pfadangabe */
		p = strrchr (txtname, '/');
		if (p)
			p++;
		else
			p = txtname;

		mboutput("Packeraufruf\r\n");
		if (!packer_pack (basep, tom_tmp_dir, ta_passinfo.pkenn, p))
			return FALSE;
		to_send = packedname;
		if (!fileexist(to_send))
		{
			/* Grr - gzip outfile.txt -> outfile.txt.gz */
			strcpy (packedname, tom_tmp_dir);
			basep=packedname+strlen(packedname);
			strcat (basep, "outfile.txt");
			strcat (basep, ta_passinfo.pend);
		}
	}
	else
		to_send = txtname;

	/* Reste beseitigen (auf jeden Fall, egal was kommt, denn entweder 
	 * ist async_in gesetzt gewesen (dann sind die Dateien oben 
	 * ins tmp kopiert worden), oder vorher lief ein Einsortieren, 
	 * bei dem genau dasselbe geschah 
	 */
	{
		char del[PATH_MAX];
		sprintf (del, "%s" LATER_CMD_F, userspooldir);
		unlink(del);
		sprintf (del, "%s" TAUSCH_LOG_F, userspooldir);
		unlink(del);
		sprintf (del, "%s" NOEXPID_LST_F, userspooldir);
		unlink(del);
		sprintf (del, "%s" KOMMANDO_OUT_F, userspooldir);
		unlink(del);
#if 0
		sprintf (del, "%s" PASSINFO_23, userspooldir);
		unlink(del);
#endif
	}

	sprintf (asyncname, "%s" PASSINFO_34 , userspooldir);
	f = xfopen (asyncname, BINWRITE);
	if (ta_passinfo.pkenn)
	{
		fprintf (f, "endung %s\n", ta_passinfo.pend);
		fprintf (f, "kennung %s\n", ta_passinfo.pkenn);
	}
	fclose (f);
	if (ta_passinfo.pkenn)
		sprintf (asyncname, "%soutfile%s", userspooldir,ta_passinfo.pend);
	else
		sprintf (asyncname, "%soutfile.txt", userspooldir);
	strcpy(workfilename,asyncname);
	if (!file_move(to_send,asyncname))
	{
		mboutput(" Tausch, Phase 3, Async: Kann Outfile nicht verschieben!\r\n");
		syslog(LOG_ERR,
			"Tausch, Phase 3, Async: Kann Outfile nicht verschieben!\r\n");
		syslog(LOG_ERR," nach %s: %s",asyncname, strerror(errno));
		return FALSE;
	}
	return TRUE;
}

static BOOLEAN
ta_phase_senden (const char *workfilename,
				 const char *outfilename)
{
	/*
	 * Outfile senden.
	 */
	if (outfilename)
	{
		/* lokal "senden" == copy */
		if (!file_copy (workfilename, outfilename))
			return FALSE;
	}
	else
	{
		char saugfile[PATH_MAX];

		sprintf(saugfile,"%s%ld/%s",box_conf.spool_dir,user.nummer,SAUGCNTRL);
		if (fileexist(saugfile))
		{
			FILE *f;
			size_t size=PATH_MAX*2;
			size_t len=0;
			char *filenames=XMALLOC(size);
			char buf[PATH_MAX*2];
			char *p;

			strcpy(filenames,workfilename);
			len=strlen(workfilename);

			f=xfopen(saugfile,"r");
			while (fgets(buf,sizeof(buf),f))
			{
				char ziel[PATH_MAX];
				p=strpbrk(buf,"\r\n");
				if (p)
					*p=0;
				p=strpbrk(buf,"\t ");
				if (p)
					*p++=0;
				strtrim(p);
				strtrim(buf);
				sprintf(ziel,"%s%s",tom_utmp_dir,p);
				if (fileexist(ziel))
				{
					mbprintf("Download von %s: Datei schon vorhanden\r\n",p);
				}
				else if (
#ifdef HAVE_SYMLINK
					symlink (buf, ziel) &&
#endif
#ifdef HAVE_LINK
					link (buf, ziel) &&
#endif
					!file_copy (buf, ziel))
				{
					syslog (LOG_NOTICE, "(sym)link/copy to %s: %s",
						ziel, strerror (errno));
					mbprintf("Download von %s unmoeglich: %s\r\n",
						p,strerror(errno));
				}
				else
				{
					size_t l_name=strlen(ziel);
					if (l_name+1+len>=size)
					{
						size*=2;
						filenames=XREALLOC(filenames,size);
					}
					len+=sprintf(filenames+len," %s",ziel);
				}
			} /* while */
			unlink(saugfile);
			if (!ta_sende_outfile (filenames,ta_passinfo.protokoll))
			{
				free(filenames);
				return FALSE;
			}
			free(filenames);
		}
		else
		{
			if (!ta_sende_outfile (workfilename,ta_passinfo.protokoll))
				return FALSE;
		}
	}
	return TRUE;
}

BOOLEAN
ta_run_tausch (enum tausch_modus modus, const char *username,
			   char *infilename, char *outfilename, char *mtzeit)
{
	USER swapuser;
	BOOLEAN swapped = FALSE;
	USER u;
	USERNR unr;
	BOOLEAN ok = TRUE;
	BOOLEAN old_disabled;
	time_t vor_export_zeit;

	char spooldirname[PATH_MAX];
	char workfilename[PATH_MAX];
	char lockfname[PATH_MAX];

	time_t seit=(time_t) 0;

	/* Diverse Scripts warten auf MausTausch */
	mboutput("MausTausch startet\r\n");

	*workfilename = '\0';

	if (mtzeit)
	{
		if (!mtzeit_to_time_t(mtzeit,&seit))
		{
			mbprintf("kann Uhrzeit %s nicht wandeln\r\n",mtzeit);
			return FALSE;
		}
	}
	if (username)
	{
		if (!ub_find_username (&unr, username))
		{
			syslog(LOG_ERR,"kann Benutzer nicht finden: %s",username);
			mbprintf("%s ist leider unbekannt\r\n",username);
			return FALSE;
		}
		sprintf(lockfname,"%s%lu.LCK",box_conf.spool_dir,unr);
	}
	else
		sprintf(lockfname,"%s%ld.LCK",box_conf.spool_dir,user.nummer);

	if (lock_loop(30,5,0,lockfname,NULL))
	{
		mboutput(" Das Spoolverzeichnis ist zur Zeit gesperrt!\r\n");
		syslog(LOG_NOTICE,"Tausch: kein Lock auf %s",lockfname);
		return FALSE;
	}

	if (username)
	{
		if (!ub_read_user (unr, &u, TRUE))
		{
			lock_delete(lockfname);
			return FALSE;
		}
		swapuser = user;
		user = u;
		b_set_flags (&user);
		swapped = TRUE;
	}

	/* von jetzt an keine returns mehr, wegen geswapptem User! */

	if (box_conf.user_dir) {
		char utmp[UDAT_NAME_LEN*2+1];
		char *up;
		uostr_t dn={0};
		struct stat st;
		ub_canonical_underscore(user.name,utmp);
		uostr_xdup_cstrmulti(&dn,box_conf.user_dir,utmp,NULL);
		uostr_x0(&dn);
		if (0!=stat(dn.data,&st)) {
			mkdir(dn.data,0700);
			/* sptere Routinen sollten sowieso entweder einen Fehler
			 * bemerken oder das Verzeichnis gar nicht erst brauchen!
			 */
		}
		uostr_freedata(&dn);
	}

	akt_anruf.mitten_im_tausch=TRUE;

	old_disabled=mehr_disabled;
	mehr_disabled=TRUE;

	sprintf (spooldirname,"%s%ld", box_conf.spool_dir, (long) user.nummer);
	mkdir(spooldirname, S_IRWXU);
	strcat(spooldirname,"/incoming");
	mkdir(spooldirname, S_IRWXU);

#ifdef HAVE_DOSDRIVES
	{
		char *grrr=strrchr(spooldirname,'\\');
		if (grrr)
			*grrr=0;
		else
			*(strrchr(spooldirname,'/')) = '\0';
	}
#else
	*(strrchr(spooldirname,'/'))='\0';
#endif
	strcat(spooldirname, "/" );

	userspooldir=spooldirname; /* modulstatische Variable setzen */

	rm_dir_content (tom_tmp_dir);
	rm_dir_content (tom_utmp_dir);
	ta_lade_installation (FALSE);
	ta_init_passinfo (&ta_passinfo);


	switch (modus)
	{
	case ta_modus_normal:
	case ta_modus_inout:
		/*
		 * einen ganz normalen Tausch durchfhren 
		 */
		set_status_line ("Tausch");
		strcpy (akt_anruf.zuletzt, "TAU");

		if (modus == ta_modus_inout)
		{
			if (!infilename)
			{
				mbprintf ("Kein Infilename angegeben. Abbruch\r\n");
				ok=FALSE;
			}
			if (!outfilename)
			{
				mbprintf ("Kein outfilename angegeben. Abbruch\r\n");
				ok=FALSE;
			}
		}

		if (ok && !ta_phase_empfangen (infilename, workfilename))
			ok=FALSE;
		/* workfilename enthlt jetzt den Namen des Infile.txt */

		if (ok && !ta_phase_einsortieren (workfilename))
			ok=FALSE;

		if (ok && mtzeit)
		{
			ta_opts.last_newstausch=seit;
			ta_opts.last_mailtausch=seit;
		}
		{
			/* 
			 * warten bis eine Sekunde vorbei ist, um zu vermeiden, da
			 * die selbst geschriebenen Nachrichten doppelt kommen.
			 * Dies ist ein Hack, aber eine bessere Lsung fllt mir
			 * im Moment nicht ein. Kostenpunkt: eine halbe Sekunde
			 * im Schnitt (prziser: 0.55).
			 */
			 time_t t1=time(NULL);
			 do
			 {
			 	msleep(100);
			 } while (time(NULL)-t1==0);
		}
		vor_export_zeit=time(NULL);

		if (ok && !ta_phase_exportieren (workfilename))
			ok=FALSE;
		/* In workfilename steht jetzt der Name der zu sendenden Datei */

		if (ok && !ta_phase_senden (workfilename, outfilename)) 
			ok=FALSE;
		else
			unlink(workfilename);

		if (ok)
		{
			if (ta_passinfo.send_mail)
				ta_opts.last_mailtausch = vor_export_zeit;
			if (ta_passinfo.send_news)
				ta_opts.last_newstausch = vor_export_zeit;
			if (ta_passinfo.status!=TA_KEIN_STATUS)
				ta_opts.last_status=vor_export_zeit;
			ta_opts.changed = TRUE;
			akt_anruf.hat_neue_pms=FALSE;
		}

		break;
	case ta_modus_in:
		set_status_line ("Tausch: Infile bearbeiten");
		strcpy (akt_anruf.zuletzt, "TIN");

		if (!infilename)
		{
			mbprintf ("Kein Infilename angegeben. Abbruch\r\n");
			ok=FALSE;
		}

		if (ok && !ta_phase_empfangen (infilename, workfilename))
			ok=FALSE;

		if (ok && !ta_phase_einsortieren (workfilename))
			ok=FALSE;

		break;
	case ta_modus_out:
		set_status_line ("Tausch: Outfile bearbeiten");
		strcpy (akt_anruf.zuletzt, "TOU");

		if (!outfilename)
		{
			mbprintf ("Kein Outfilename angegeben. Abbruch\r\n");
			ok=FALSE;
		}

		if (ok && mtzeit)
		{
			ta_opts.last_newstausch=seit;
			ta_opts.last_mailtausch=seit;
		}

		vor_export_zeit=time(NULL);


		if (ok && !ta_phase_exportieren (workfilename))
			ok=FALSE;

		/* In workfilename steht jetzt der Name der zu sendenden Datei */
		if (ok && !ta_phase_senden (workfilename, outfilename))
			ok=FALSE;

		if (ok)
		{
			if (ta_passinfo.send_mail)
				ta_opts.last_mailtausch = vor_export_zeit;
			if (ta_passinfo.send_news)
				ta_opts.last_newstausch = vor_export_zeit;
			if (ta_passinfo.status!=TA_KEIN_STATUS)
				ta_opts.last_status=vor_export_zeit;
			ta_opts.changed = TRUE;
		}

		break;
	}

	/*
	 * An dieser Stelle ist der eigentliche Tausch OK.
	 * Also *jetzt* die Tauschinstallation sichern.
	 */
	if (ok)
	{
		ta_sichere_installation (FALSE);
		akt_anruf.tausch_status = TRISTATE_OK;
		mboutput("Tausch erfolgreich abgeschlossen.\r\n");
	}
	else
	{
		akt_anruf.tausch_status = TRISTATE_FAILURE;
		mboutput("Tausch nicht erfolgreich.\r\n");
	}

	if (!logout_modus)
		strcpy (akt_anruf.zuletzt, "BOX");
	lock_delete(lockfname);

	mehr_disabled=old_disabled;
	akt_anruf.mitten_im_tausch=FALSE;

	if (swapped)
	{
		/* Erstmal Tauschinstallation fr ungltig erklren. */
		ta_mark_invalid ();
		/* Dann wieder auf den alten User wechseln */
		ub_write_user(&user,TRUE);
		user = swapuser;
		b_set_flags (&user);
	}
	/* und nun aufrumen - das Risiko beseitigen */
	rm_dir_content (tom_utmp_dir);
	rm_dir_content (tom_tmp_dir);
	return ok;
}

BOOLEAN
ta_run_infiles(BOOLEAN only_known)
{
	char aspoold[PATH_MAX];
	char *ep;
	DIR *d;
	struct dirent *e;
	
	sprintf (aspoold,"%sasync_i", box_conf.spool_dir);
	ep=aspoold+strlen(aspoold);
	*ep='/';
    strcpy(ep+1,"LCK..import");
    if (lock_create(aspoold,NULL))
    	return FALSE;
    *ep=0;

	d=opendir(aspoold);
	if (!d)
	{
    	strcpy(ep+1,"LCK..import");
		return FALSE;
	}

	while ((e=readdir(d))!=NULL)
	{
		USERNR unr;
		char *p;
		LISTUSER *lu;

		p=e->d_name;
		if (*p=='.')
			continue;
		unr=strtol(p,&p,10);
		if (!unr)
			continue;
		if (*p!='.')
			continue;
		lu=ub_find_usernummer(unr);
		if (!lu)
			continue;
		sprintf(ep,"/%s",e->d_name);
		if (ta_run_tausch (ta_modus_in,lu->name,aspoold, NULL, NULL))
			unlink(aspoold);
	}
	closedir(d);
	*ep='/';
    strcpy(ep+1,"LCK..import");
    unlink(aspoold);
	return TRUE;
}
