#ifndef TAUSCH_H
#define TAUSCH_H

/*
 * globale Tauschdaten. Einzige Headerdatei, die von Modulen auerhalb
 * des Tauschs included werden sollte.
 */

/***************************************************************************
 *
 * Funktion zur Durchfhrung des Tauschs
 *
 ***************************************************************************
 */
enum tausch_modus
{
	ta_modus_normal,        /* empfangen, einsortieren, exportieren, senden */
	ta_modus_inout,         /* wie normal, aber ohne Downloads etc */
	ta_modus_in,      		/* infile empfangen und einsortieren */
	ta_modus_out     		/* outfile erstellen und senden */
};

/* einen Tausch durchfhren (oder, je nach Modus, Teile davon */
BOOLEAN ta_run_tausch(enum tausch_modus modus, const char *username, 
	char *infname, char *outfname, char *mtzeit);
/* alle noch ausstehenden Infiles verarbeiten. Wenn only_known gesetzt
 * ist, werden nur die seit dem letzten Boxstart angefallenen Infiles
 * bearbeitet (die merkt sich die Box in einer Liste), sonst werden
 * alle User geprft, ob noch ein Infile herumliegt.
 */
BOOLEAN ta_run_infiles(BOOLEAN only_known);

void tausch_brett(const char *para);

	/*
	 * Das sortiert nur eine PM ein, ohne irgendetwas sonst
	 * zu tun. Keine Tauschinstallation, etc.
	 * Achtung: nur eine Nachricht!
	 */
#include "routing.h"
BOOLEAN ta_mail_einsortieren (r_adr_t *to_adr, char *buffer);

 
/***************************************************************************
 *
 * Rund um die Tauschinstallation
 *
 ***************************************************************************
 */

/*
 * Tauschinstallation laden, und eventuell Informationen aus der ersten
 * Phase des Async-Tauschs bernehmen.
 */
void ta_lade_installation(BOOLEAN async_uebernehmen);

/*
 * Tauschinstallation sichern, eventuell Informationen fr die 
 * zweite Phase des Async-Tauschs sichern.
 * Das "changed"-Flag wird zurckgesetzt.
 */
void ta_sichere_installation(BOOLEAN close_flag);

/*
 * Interne Daten als ungltig deklarieren
 */
void ta_mark_invalid(void);

/*
 * Exportliste fr die Datenbank schreiben.
 *
 * ->           datei   :   Name der zu erzeugenden ASCII-Datei
 *
 * setzt das Exportdatum ntigenfalls ein.
 */
void ta_schreibe_exportliste(char *datei, BOOLEAN binaer);

/*
	Brett bestellen. 
	Parameter:
	- Brettnr des zu bestellenden Brettes
	- Tauschzeitpunkt (0 unerlaubt, 1 globale Tauschzeit bernehmen,
	  sonst neues Tauschdatum fr das Brett).
	- Zeiger auf ausreichend groen Puffer fr eine Fehlermeldung.
	  128 Byte sind ausreichend.
	Die Funktion ldt die Tauschinstalltion, falls das notwendig ist,
	und setzt auch das changed-Flag.
*/
BOOLEAN ta_brett_bestellen(BRETTNR,char *);

/*
	Brett abbestellen. Anmerkungen wie oben
*/
BOOLEAN ta_brett_abbestellen(BRETTNR,char *);


	/* 
	 * Nach to, falls gegeben, sonst via Zmodem oder Consolendownload.
	 * Darf nicht aufgerufen werden, falls beide Dateien identisch
	 * sind.
	 */
BOOLEAN ta_sende_outfile(const char *from, char protokoll);
BOOLEAN ta_erstelle_outfile(FILE *f, const char *fname);
BOOLEAN ta_infile_einsortieren(const char *infilename);
BOOLEAN ta_online_varinfofile (const char *fname, const char *kurz);


	/*
	 * Umbenennungsinformationen sichern.
	 */
BOOLEAN ta_add_rename_info(time_t zeit, const char *from, const char *to);

	/*
	 * Tauschinstallationsroutine
	 */
void ta_do_inst(void);

	/*
	 * Nur zur Benutzung durch den Nachrichtenrouter!
	 */
void ta_do_log (const char *format,...);
void ta_do_log1 (const char *str);



#endif
