#include "tomsysdep.h"
#include "libquark.h"
#include <stdlib.h>
#include <stdio.h>
#include <dirent.h>
#include <string.h>
#include <fnmatch.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <syslog.h>
#include "argument.h"
#include "box_conf.h"
#include "brett.h"
#include "long-options.h"

const char *program_name;

struct sarray {
	char **base;
	int size;
	int zahl;
};

static void
add_to_sarray(struct sarray *a, const char *s)
{
	if (!a->base) {
		a->base=XMALLOC(10*sizeof(char *));
		a->size=10;
		a->zahl=0;
	}
	a->base[a->zahl++]=XSTRDUP(s);
}

struct group {
	char *name;
	int wc;
	struct group *next;
};

struct user {
	char *name;
	char *tmpfname;
	char *finalfname;
	struct group *groups;
	struct group *pathexcludes;
	struct user *next;
	int fd;
} *uanker;

static struct user *
allocate_user(const char *name)
{
	struct user *u=XMALLOC(sizeof(struct user));
	u->groups=NULL;
	u->pathexcludes=NULL;
	u->tmpfname=NULL;
	u->finalfname=NULL;
	u->name=XSTRDUP(name);
	u->next=uanker;
	uanker=u;
	return u;
}

static void work_users(const char *dirname);

static void
add_group(char *g, struct user *u)
{	
	struct group *gp;
	gp=XMALLOC(sizeof(struct group));
	gp->name=entumlauten1(g);
	XNULLCHECK(gp->name);
	strlwr(gp->name);
	if (strpbrk(g,"?[]*"))
		gp->wc=1;
	else
		gp->wc=0;
	gp->next=u->groups;
	u->groups=gp;
}

static void
add_pathexclude(char *g, struct user *u)
{	
	struct group *gp;
	gp=XMALLOC(sizeof(struct group));
	gp->name=entumlauten1(g);
	XNULLCHECK(gp->name);
	strlwr(gp->name);
	if (strpbrk(g,"?[]*"))
		gp->wc=1;
	else
		gp->wc=0;
	gp->next=u->pathexcludes;
	u->pathexcludes=gp;
}

static void
usage (int ecode) {
	exit(ecode);
}

static void
verteile_nachricht(char *buf,size_t len, struct sarray *groups, struct sarray *path)
{
	struct user *u;
	for (u=uanker;u;u=u->next) {
		int i;
		int found=0;
		struct group *g=NULL;
		for (i=0;i<groups->zahl;i++) {
			for (g=u->groups;g;g=g->next) {
				if (0==fnmatch(g->name,groups->base[i],0)) {
					found=1;
					break;
				}
			}
			if (found)
				break;
		}
		if (found)
			printf("send mes to %s, %s ~ %s\n",u->name, g->name,groups->base[i]);
		else
			printf("do not send mes to %s\n",u->name);
		/* don't mess with stdio - our messages are large. And doing lot's of copies doesn't help */
		if (u->fd!=-1 && found) {
			if (u->pathexcludes) {
				for (i=0;i<path->zahl;i++) {
					for (g=u->pathexcludes;g;g=g->next) {
						if (0==strcmp(g->name,path->base[i])) {
							found=0;
							break;
						}
					}
					if (!found)
						break;
				}
			}
			if (found)
				write(u->fd,buf,len);
			else
				printf("do not send mes to %s: path %s\n",u->name, path->base[i]);
		}
	}
	/* fwrite(buf,len,1,stdout); */
}

static void
verteiler(void)
{
	char *buf=NULL;
	size_t size=0;
	size_t used=0;
	char *lstart=NULL;
	struct sarray groups={0};
	struct sarray path={0};
	size=16384;
	buf=XMALLOC(size);

	do {
		int c;
		c=getc(stdin);
		if (c==EOF) 
			break;
		if (c=='#' && used>4) {
			verteile_nachricht(buf,used,&groups,&path);
			used=0;
			while (groups.zahl--) {
				free(groups.base[groups.zahl]);
			}
			groups.zahl=0; /* in case of crippled messages */
			while (path.zahl--) {
				free(path.base[path.zahl]);
			}
			path.zahl=0; /* in case of crippled messages */
		}
		if (size-used<2048) {
			size*=2;
			buf=XREALLOC(buf,size);
		}
		lstart=buf+used;
		buf[used++]=c;
		/* get rest of line */
		while(1) {
			size_t got;
			if (size-used<2048) {
				size*=2;
				buf=XREALLOC(buf,size);
			}
			if (!fgets(buf+used,size-used,stdin))
				break;
			got=strlen(buf+used);
			used+=got;
			if (buf[used-1]=='\n') {
				break;
			}
			/* line too long */
		}
		/* ok, got the line */
		if (c=='G' && buf+used-lstart<=256)  {
			char gbuf[512];
			char *p;
			int d;
			p=buf+used-1;
			if (*p=='\n')	
				p--;
			if (*p=='\r')	
				p--;
			d=p[1];
			p[1]=0;
			bb_canonical(lstart+1,gbuf);
			p[1]=d;
			add_to_sarray(&groups,gbuf);
		}
		if (c=='p' && buf+used-lstart<=2048)  {
			char gbuf[4096];
			char *p;
			int d;
			p=buf+used-1;
			if (*p=='\n')	
				p--;
			if (*p=='\r')	
				p--;
			d=p[1];
			p[1]=0;
			entumlauten(lstart+1,gbuf);
			p[1]=d;
			/* ! separatoren entfernen */
			{
				char *s1,*s2;
				for (s1=gbuf,s2=NULL;s1 && *s1;s1=s2?s2+1:NULL) {
					s2=strchr(s1,'!');
					if (s2)
						*s2=0;
					add_to_sarray(&path,s1);
				}
			}
		}
		if (c=='#' && used<4) /* '#\r\n' */
			break;
	} while(1);
	if (used>4) /* catch single '#' */
		verteile_nachricht(buf,used,&groups,&path);
}

int 
main(int argc, char **argv)
{
    if (argv[0]) {
        program_name=strrchr(argv[0],'/');
        if (program_name)
            program_name++;
        else
            program_name=argv[0];
    } else
        program_name="tombatcher";


	parse_long_options (argc, argv, program_name, PACKAGE_VERSION, usage);

	if (!read_quark_configuration(
			argument_cnffile? argument_cnffile : SYSCONFDIR "/quark.cnf")) {
		fprintf(stderr,"Auswertung von %s fehlgeschlagen\n",
			argument_cnffile? argument_cnffile : SYSCONFDIR "/quark.cnf");
		exit(1);
	}
	work_users(box_conf.tausch_dir);
	verteiler();
	{
		struct user *u;
		for (u=uanker;u;u=u->next) {
			struct stat st;
			if (u->fd==-1)
				continue;
			if (fstat(u->fd,&st)==0) {
				if (st.st_size) {
					rename(u->tmpfname,u->finalfname);
				} else {
					unlink(u->tmpfname);
				}
			} else
				unlink(u->tmpfname);
			close(u->fd);
		}
	}
	exit(0);
}

#define XMKDIR(newdir) do {                                                  \
        if (mkdir(newdir,0700)==-1 && errno!=EEXIST) {                       \
            int err=errno;                                                \
            fprintf(stderr,"kann %s nicht anlegen: %s\r\n",           \
                newdir,strerror(err));                                     \
            syslog(LOG_ERR,"kann %s nicht anlegen: %s\r\n",           \
                newdir,strerror(err));                                     \
			exit(1);													\
        } 																\
        } while(0)

void
work_users(const char *dirname)
{
	DIR *dir;
	struct dirent *e;
	dir=opendir(dirname);
	if (!dir) {
		int err=errno;
		fprintf(stderr,"kann %s nicht ffnen: %s\n",dirname,strerror(err));
		syslog(LOG_ERR,"kann %s nicht ffnen: %s",dirname,strerror(err));
		exit(1);
	}
	if (chdir(dirname)) {
		int err=errno;
		fprintf(stderr,"kann nicht nach %s wechseln: %s\n",dirname,strerror(err));
		syslog(LOG_ERR,"kann nicht nach %s wechseln: %s",dirname,strerror(err));
		exit(1);
	}
	while (NULL!=(e=readdir(dir))) {
		size_t l;
		if (*e->d_name=='.')
			continue;
		l=strlen(e->d_name);

		if (l>4 && 0==strcmp(e->d_name+l-4,".tau")) {
			FILE *f;
			char buf[1024];
			char *p;
			int unused=0;
			struct user *u=NULL;

			f=fopen(e->d_name,"r");
			if (!f) {
				int err=errno;
				fprintf(stderr,"kann %s nicht lesen: %s\n",dirname,strerror(err));
				syslog(LOG_ERR,"kann %s nicht lesen: %s",dirname,strerror(err));
				continue;
			}
			while (fgets(buf,sizeof(buf),f)) {
				if ((p = strpbrk (buf, "\r\n")) != NULL)
					*p = '\0';
				if (*buf=='#') {
					if (0==strncmp(buf,"#NEWSTAUSCHDATUM=",17)) {
						time_t t;
						if (mtzeit_to_time_t(buf+17,&t)) {
							if (t>time(NULL)+86400*180 && t!=0)
								unused=1;
							fprintf(stderr,"t=%lu,now=%lu\n",t,time(NULL));
						}
					}
					if (0==strncmp(buf,"#USER ",6)) {
						p=buf+6;
						while (isspace(*p))
							p++;
						while (isdigit(*p))
							p++;
						while (isspace(*p))
							p++;
						u=allocate_user(p);
					}
					if (0==strncasecmp(buf,"#PATHEXCLUDE ",13)) {
						char *q;
						p=buf+13;
						for (p=buf+13,q=NULL;p && *p;p=q?q+1:NULL) {
							q=strpbrk(p,",!");
							if (q)
								*q=0;
							if (*p) {
								add_pathexclude(p,u);
							}
						}
					}
				} else if (*buf=='-') {
					continue; 
				} else {
					char *q;
					p=buf+2;
					while (isdigit(*p)) /* u.U. Gruppennummer berspringen */
						p++;
					if (isspace(*p)) {
						q=p+1;
						while (isspace(*q))
							q++;
					} else
						q=buf+2;
					/* */
					if (u)
						add_group(q,u);
				}
			}
			fclose(f);
			if (u && u->groups && !unused) {
				char fname[PATH_MAX];
				int num;
				char *up;
				int try=0;
			  retry:
				p=stpcpy(fname,box_conf.user_dir);
				ub_canonical(u->name,p);
				while (*p) {
					if (isspace(*p))
						*p='_';
					p++;
				}
				up=p;
				p=stpcpy(p,"/outgoing");
				p+=sprintf(p,"/mtnews.%lu.", (unsigned long) getpid());
				num=0;
				while (1) {
					sprintf(p,"%d.tmp",num++);
					u->fd=open(fname,O_WRONLY | O_EXCL | O_CREAT,0644);
					if (u->fd==-1 && errno==ENOENT && try==0) {
						try++;
						*up=0;
						XMKDIR(fname); /* user dir angelegen */
						strcpy(up,"/outgoing");
						XMKDIR(fname);
						p=stpcpy(up,"/Maildir");
						XMKDIR(fname);
						strcpy(p,"/new");
						XMKDIR(fname);
						strcpy(p,"/cur");
						XMKDIR(fname);
						strcpy(p,"/tmp");
						XMKDIR(fname);
						goto retry;
					}
					if (-1==u->fd && errno!=EEXIST) {
						fprintf(stderr,"cannot open %s: %s\n",fname,strerror(errno));
						exit(1);
					}
					if (u->fd!=-1) {
						u->tmpfname=XSTRDUP(fname);
						u->finalfname=XSTRDUP(fname);
						u->finalfname[strlen(fname)-4]=0;
						break;
					}
				}
			} else if (u)
				u->fd=-1;
		}
	}
	closedir(dir);
}

