/*
    Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
    Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
    Contact3: MausNet Newsgroups Quark & Q-Sysops.
*/
#include "tomsysdep.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <portab.h>
#include <errno.h>
#include <unistd.h>
#include <ctype.h>

#include "box_conf.h"
#include "prototyp.h" /* entumlauten */
#include "userbase.h"
#include "ub_int.h"
#include "modem.h"

#ifdef TOM_CACHE_STATISTICS
#  ifndef TOM_ATTRIB_CONSTRUCTOR
#    undef TOM_CACHE_STATISTICS
#  endif
#endif
#ifdef TOM_CACHE_STATISTICS
#include <syslog.h>
#endif

/*
 * Es werden sehr hufig dieselben Namen gesucht (Name des Servers fr
 * Infofiles, eines Gateways usw). Daher wird ein kleiner Cache verwendet.
 *
 * Wichtig: Nicht gefundene Namen drfen nicht in den Cache wandern (auch
 * nicht mit Usernummer 0). Beim Anlegen von Usern wrde das fr Probleme
 * sorgen.
 * Es fehlt eine Behandlung fr das Lschen eines Users (but who cares -
 * it doesn't matter).
 */

#ifdef TOM_CACHE_STATISTICS
static unsigned long u_cache_hit;
static unsigned long u_cache_miss;
static unsigned long u_cache_reorder;

static void u_cache_print(void) TOM_ATTRIB_DESTRUCTOR;
static void
u_cache_print(void)
{
	if (u_cache_hit || u_cache_miss)
		syslog(LOG_DEBUG,"ub_find_user: %lu,%lu,%lu",
			u_cache_hit, u_cache_miss, u_cache_reorder);
}

#define ADD_COUNTER(x) x++

#else
#define ADD_COUNTER(x)
#endif


#define CACHESIZE 10
struct u_cache
{
	char *canon;
	USERNR unr;
};
static struct u_cache u_cache[CACHESIZE];


#define CACHE_IT(str,nr) \
do { \
	free(u_cache[CACHESIZE-1].canon); \
	memmove(u_cache+1,u_cache,(CACHESIZE-1)*sizeof(struct u_cache)); \
	u_cache[0].canon=strdup(str); /* Fehler macht nichts */ \
	u_cache[0].unr=nr; \
} while(0)

#define PUT_ERGEBNIS(nr) \
do { \
	int ii; \
	for (ii=0;ii<array_used;ii++) if (array[ii]==nr) break; \
	if (array_used==array_len-1) {\
		array_len*=2; array=XREALLOC(array,sizeof(USERNR)*array_len); \
	} \
	array[array_used++]=nr; \
	array[array_used]=0; \
} while(0)
		 
		  
/* Usernummer zu einem Usernamen ermitteln

	->		<-	nr		:	Speicher fr Usernummer (kann NULL sein)
				s		:	Username

	<-			TRUE	:	User vorhanden
				FALSE	:	User nicht vorhanden

	******92, Uwe, erste Fassung 
	01.12.93, Uwe, kommt nun auch mit berflssigen Leerzeichen und
	          Umlauten in Usernamen klar.
	22.12.93, Uwe, Bezeichner gendert, besser dokumentiert.
*/
USERNR *ub_find_username_multi(const char *s, BOOLEAN only_one)
{
	int array_len;
	int array_used;
	USERNR *array;
	static USERNR static_usernr;

	LISTUSER *akt;
		/* 
			temporrbuffer mit Doppelfunktion:
			- erst Puffer fr von berflssigen Leerzeichen bereinigte
			  Form des bergebenen Usernamens (die dann noch entumlautet 
			  wird)
			- spter Puffer fr entumlautete, zu vergleichende Namen
			  aus der Liste.
		*/
	char tempbuf[UDAT_NAME_LEN*3];
		/*	
			Puffer fr endgltige Form des Usernamens, ohne Umlaute und
			berflssige Leerzeichen. Nach dieser Version kann gesucht
			werden.
		*/
	char suchname[UDAT_NAME_LEN*3];
	size_t suchnamelen;
	const char *op; /* Pointer auf _o_riginal */
	char *kp; /* Zeiger auf _k_opie */
	struct u_cache *cache_ptr;
	int i;
	BOOLEAN need_wildcard=FALSE;
	char *may_retry_wildcard=NULL;

	array_len=10;
	array=XMALLOC(array_len*sizeof(USERNR));
	array[0]=0;
	array_used=0;

	op = s;
	kp = tempbuf;

	/* Eventuelle Leerzeichen am Anfang bergehen */
	while (isspace(*op))
		op++;

	/* pathologischer Fall */
	if (!*op)
		return NULL;

	/* Kopierschleife, killt doppelte Leerzeichen */
	while (*op)
	{
		if (!isspace(*op) || !isspace(op[1]))
			*(kp++)=*op;
		op++;
	}

	/* 
		Jetzt hngt eventuell ein berflssiges Leerzeichen am Namen,
		auerdem fehlt noch ein \0 am Ende der Kopie
	*/
	*(kp--)='\0';
	if (isspace(*kp))
		*kp='\0';

	ub_canonical(tempbuf,suchname);
	suchnamelen=strlen(suchname);

	/* 
	 * Im Cache suchen. 
	 */
	for (cache_ptr=u_cache,i=0;i<CACHESIZE;i++) {
		if (cache_ptr->canon && *cache_ptr->canon==*suchname) {
			if (0==strcmp(cache_ptr->canon,suchname)) {
				ADD_COUNTER(u_cache_hit);
				if (i!=0) {
					struct u_cache h=*cache_ptr;
					memmove(u_cache+1,u_cache,9*sizeof(struct u_cache));
					free(u_cache[0].canon);
					u_cache[0].canon=strdup(suchname);
					u_cache[0].unr=h.unr;
					ADD_COUNTER(u_cache_reorder);
				}
				if (u_cache[0].unr==0) {
					free(array);
					return NULL;
				}
				if (only_one) {
					free(array);
					static_usernr=u_cache[0].unr;
					return &static_usernr;
				}
				PUT_ERGEBNIS(u_cache[0].unr);
				/* Wir knnen zurckkehren: Der Cache wird nur gefllt,
				 * wenn genau *ein* Name pat. 
				 */ 
				return array;
			}
		}
	}

	ADD_COUNTER(u_cache_miss);

	/* Falls keine Wildcards vorkommen, zuerst die schnellere Variante
	 * ohne Wildcardvergleich versuchen 
	 */
	if (!strchr(suchname,'.'))
	{
		/* Wenn ein Space im Namen vorkommt, dann mit dem vollstndigen
		 * Namen des Users vergleichen. Das spart einen Durchlauf durch
		 * die Liste.
		 */
		char *space;
		BOOLEAN complete=FALSE;
		space=strchr(suchname,' ');

		if (space || box_conf.allow_one_part_usernames)
		{
			akt = ub_anker;
			while(akt)
			{
				/* 
				 * Schn wre es, wenn man sich berflssige Umlautwandlungen
				 * ersparen knnte. Das geht allerdings schief, wenn der 
				 * Benutzer "Albert Oel" sendet und der Empfnger "Albert l"
				 * heit. Tja - umlautwandlung immer.
				 */
				ub_canonical(akt->name,tempbuf);
		
				if (!strcmp(suchname,tempbuf))
				{
					PUT_ERGEBNIS(akt->nummer);
					complete=TRUE;
				}
				akt = akt->next;
			}
		}

		/* Na schn, den Namen selbst haben wir nicht gefunden */
		/*
		 * Mgliche Alternativschreibweisen fr Frithard Meyer Zu Uptrup,
		 * zuerst ohne Wildcards:
		 *  Frithard Uptrup
		 *  Frithard Meyer Uptrup
		 *  Frithard Zu Uptrup
		 *  Frithard
		 *  Uptrup
		 *  Zu Uptrup
		 *  Meyer Zu Uptrup
		 * Also ist erlaubt:
		 * - nur das erste Wort
		 * - weglassen am Anfang
		 * - weglassen von Teilen in der Mitte
		 */
		 if (!complete)
		 {
			char *p;
			akt=ub_anker;

			while(akt)
			{
				ub_canonical(akt->name,tempbuf);
				/* Varianten mit nur einem Wort zuerst */
				if (!space)
				{
					p=strstr(tempbuf,suchname);
					if (p==tempbuf)
					{
						/* Form: "Frithard" */
						if (!tempbuf[suchnamelen] 
							|| isspace(tempbuf[suchnamelen]))
							PUT_ERGEBNIS(akt->nummer);
					}
					else if (p && isspace(p[-1]) && p[suchnamelen]=='\0')
					{
						/* Form: "Uptrup" */
						PUT_ERGEBNIS(akt->nummer);
					}
				}
				else
				{
					/* Mglicherweise:
					 *  Frithard Uptrup
					 *  Frithard Meyer Uptrup
					 *  Frithard Zu Uptrup
					 *  Zu Uptrup
					 *  Meyer Zu Uptrup
					 */
					p=strstr(tempbuf,suchname);
					if (p && p[suchnamelen]=='\0')
					{
						/* das war:
						 *  Zu Uptrup
						 *  Meyer Zu Uptrup
						 */
						PUT_ERGEBNIS(akt->nummer);
					}
					else
					{
						/* Mglicherweise:
						 *  Frithard Uptrup
						 *  Frithard Meyer Uptrup
						 *  Frithard Zu Uptrup
						 */
						char *ps=suchname;
						char *pt=tempbuf;
						while (*ps==*pt)
						{
							ps++;
							pt++;
						}
						if (ps!=suchname && ps[-1]==' ')
						{
							p=strstr(pt,ps);
							if (p && p[strlen(ps)]==' ')
								PUT_ERGEBNIS(akt->nummer);
						}
					}
				}
				akt=akt->next;
			} /* while akt */
			
			/* Wenn Leerzeichen oder Name lnger als 4 Zeichen.
			 * "UO" -> "U* O*"
			 * "SYSOP" -> nein
			 */
			if (suchnamelen <=4 && !space)
			{
				char *pt=tempbuf;
				char *ps=suchname;
				strcpy(tempbuf,suchname);
				need_wildcard=TRUE;
				while (*pt)
				{
					*ps++=*pt++;
					*ps++='*';
					if (*pt)
						*ps++=' ';
				}
				*ps='\0';
			}
		}
	} /* wenn keine Wildcard darin */
	else
	{
		/* "F.U." und "F*U*" -> "F* U*" */
		char *pt=tempbuf;
		char *ps=suchname;

		strcpy(tempbuf,suchname);
		may_retry_wildcard=XSTRDUP(tempbuf);
		need_wildcard=TRUE;

		while (*pt)
		{
			if ((*pt=='.' || *pt=='*'))
			{
				*ps++='*';
				if (pt[1]!=' ' && pt[1])
					*ps++=' ';
				pt++;
			}
			else
				*ps++=*pt++;
		}
		*ps='\0';
	}

	if (need_wildcard)
	{
		/*
		 * So, jetzt haben wir eine Wildcardvariante, so a la:
		 * "F* U*" oder "F* M* Z* U*"
		 */
	retry_wildcard:
		akt=ub_anker;
		while(akt)
		{
			ub_canonical(akt->name,tempbuf);
			if (wildcard(tempbuf,suchname))
				PUT_ERGEBNIS(akt->nummer);
			akt = akt->next;
		}
		if (array_used==0 && may_retry_wildcard) {
			/* "F.U." und "F*U*" -> "F*U*" */
			char *pt=may_retry_wildcard;
			char *ps=suchname;

			while (*pt)
			{
				if ((*pt=='.' || *pt=='*'))
				{
					*ps++='*';
					pt++;
				}
				else
					*ps++=*pt++;
			}
			*ps=0;
			free (may_retry_wildcard);
			may_retry_wildcard=NULL;
			goto retry_wildcard;
		}
	}

	if (only_one)
	{
		static_usernr=*array;
		CACHE_IT(suchname,array[0]);
		free(array);
		if (array_used!=1)
			return NULL;
		return &static_usernr;
	}

	if (array_used==1)
		CACHE_IT(suchname,array[0]);
	if (array_used==0)
	{
		free(array);
		array=NULL;
	}
	return array;

} /* ub_find_username */

