/*
 * Copyright (C) 1994 1995 1996 Frank Baumgart, Uwe Ohse
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * Contact: uwe@tirka.gun.de, Uwe Ohse @ DU3 (mausnet)
 * Contact2: godot@uni-paderborn.de, Frank Baumgart @ PB (mausnet)
 * Contact3: MausNet Newsgroups Quark & Q-Sysops.
 */
#include "tomsysdep.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <portab.h>
#include <errno.h>
#include <ctype.h>
#include <unistd.h>
#include <limits.h>
#include <syslog.h>
#include <sys/types.h>			/* off_t */

#include "prototyp.h"
#include "userbase.h"
#include "ub_int.h"
#include "filelock.h"
#include "str_ulong.h"

/*
 * Userdatensatz schreiben und optional Schreibsperre aufheben
 * 
 * ->           u           :   Userstruktur
 * unlock   :   Schreibsperre aufheben
 * 
 * <-           TRUE        :   Erfolg
 * FALSE        :   Fehler (z.B. keine Schreibberechtigung)
 */
BOOLEAN
ub_write_user (USER * u, BOOLEAN unlock)
{
	ub_fd_control_t *li = NULL;
	char *name_dup;

	if (!u || !u->name || !*u->name) {
		fprintf (stderr, "ub_write_user: no username for user\n");
		syslog (LOG_ERR, "ub_write_user: no username for user");
		return FALSE;
	}
	if (u->magic != UB_MAGIC) {
		fprintf (stderr, "ub_write_user: wrong magic for user # %lu\n", u->nummer);
		syslog (LOG_ERR, "ub_write_user: wrong magic for user # %lu", u->nummer);
		return FALSE;
	}
	li = ub_fd_anker;
	while (li && li->unr != u->nummer)
		li = li->next;

	if (!li) {
		/* Was? Wie? */
		fprintf (stderr, "ub_write_user: file not open for user # %lu\n", u->nummer);
		syslog (LOG_ERR, "ub_write_user: file not open for user # %lu", u->nummer);
		return FALSE;
	}
	if (-1 == lseek (li->fd, 0, SEEK_SET)) {
		int e = errno;
		fprintf (stderr, "ub_write_user: cannot lseek for user # %lu: %s\n", li->unr, strerror (e));
		syslog (LOG_ERR, "ub_write_user: cannot lseek for user # %lu: %s", li->unr, strerror (e));
		return FALSE;
	}
	name_dup = strdup (u->name);
	if (!name_dup) {
		fprintf (stderr, "ub_write_user: out of memory for user # %lu\n", li->unr);
		syslog (LOG_ERR, "ub_write_user: out of memory for user # %lu", li->unr);
		return FALSE;
	}
	if (-1 == xwrite (li->fd, u, sizeof (USER))) {
		int e = errno;
		fprintf (stderr, "ub_write_user: cannot write for user # %lu: %s\n",
				 li->unr, strerror (e));
		syslog (LOG_ERR, "ub_write_user: cannot write for user # %lu: %s",
				li->unr, strerror (e));
		return FALSE;
	}
	if (unlock) {
		ub_fd_control_t *last;

		close (li->fd);
		for (last = ub_fd_anker; last && last->next != li; last = last->next)
			/* nothing */ ;
		if (last)
			last->next = li->next;
		else
			ub_fd_anker = li->next;
		free (li);
	}
	/* nun die Daten in der Liste an eventuelle nderungen anpassen */
	{
		LISTUSER *l2;
		BOOLEAN need_rewrite = FALSE;
		l2 = ub_anker;
		while (l2 && l2->nummer != u->nummer)
			l2 = l2->next;
		if (l2) {
			if (0 != strcmp (l2->name, name_dup))
				need_rewrite = TRUE;
			free (l2->name);
			l2->name = name_dup;
		} else
			free (name_dup);
		if (need_rewrite) {
			/* falls ein Benutzername gendert worden ist mu auch
			 * die globale Liste angepat werden. 
			 * das kann nicht ber die prozesslokale Liste geschehen
			 * weil mittlerweile wieder neue User da sein knnten ...
			 */
			FILE *f=NULL, *f2=NULL;
			BOOLEAN ret=FALSE;
			char buf[1024];
			uostr_t fname={0};
			uostr_t fname_backup={0};
			uostr_xdup_cstrmulti(&fname,ub_basepath,USERDATNAME,NULL);
			uostr_x0(&fname);
			uostr_xdup_cstrmulti(&fname_backup,fname.data,".bak",NULL);
			uostr_x0(&fname_backup);

			/* Liste der User ffnen und locken */
			f = fopen (fname.data, "r+");
			if (!f) {
				int e = errno;
				fprintf (stderr, "ub_write_user: cannot open %s: %s\n",
						 fname.data, strerror (e));
				syslog (LOG_ERR, "ub_write_user: cannot open %s: %s\n",
						fname.data, strerror (e));
				ret=TRUE;	/* we *did* write the user */
				goto cleanup;
			}
			if (fd_writew_lock (fileno (f), 0, SEEK_SET, 1) != 0) {
				int e = errno;
				fprintf (stderr, "ub_write_user: cannot lock %s: %s\n",
						 fname.data, strerror (e));
				syslog (LOG_ERR, "ub_write_user: cannot lock %s: %s\n",
						fname.data, strerror (e));
				ret=TRUE;	/* we *did* write the user */
				goto cleanup;
			}
			/* Backup anlegen - davon mssen wir gleich lesen! */
			if (!file_copy (fname.data, fname_backup.data)) {
				int e = errno;
				fprintf (stderr, "ub_write_user: cannot backup %s: %s\n",
						 fname.data, strerror (e));
				syslog (LOG_ERR, "ub_write_user: cannot backup %s: %s\n",
						fname.data, strerror (e));
				unlink (fname_backup.data);
				ret=TRUE;	/* we *did* write the user */
				goto cleanup;
			}
			f2 = fopen (fname_backup.data, "r");
			if (!f2) {
				int e = errno;
				fprintf (stderr, "ub_write_user: cannot open %s: %s\n",
						 fname_backup.data, strerror (e));
				syslog (LOG_ERR, "ub_write_user: cannot open %s: %s\n",
						fname_backup.data, strerror (e));
				unlink (fname_backup.data);
				ret=TRUE;
				goto cleanup;
			}
			while (fgets (buf, sizeof (buf), f2)) {
				if (strtol (buf, NULL, 10) != u->nummer) {
					if (EOF == fputs (buf, f))
						break;
				} else {
					if (EOF==fprintf (f,"%lu:%s\n", u->nummer, u->name))
						break;
				}
			}
			if (fflush (f) || ferror (f) || fclose (f)) {
				char tbuf[STR_ULONG];
				uostr_t fname_backup2={0};
				int e = errno;
				fprintf (stderr, "ub_create_user: cannot write %s: %s\n",
						 fname.data, strerror (e));
				fprintf (stderr, 
					"ub_create_user: user database most possibly corrupt\n");
				syslog (LOG_CRIT, "ub_create_user: cannot write %s: %s",
						fname.data, strerror (e));
				syslog (LOG_CRIT, 
					"ub_create_user: user database most possibly corrupt\n");
				/*
				 * wir duerfen nicht das Backup umbenennen in die richtige, 
				 * nun zerstrte Datei. Da stecken mehr Raceconditions drin 
				 * als gut ist!
				 * 
				 * wir duerfen aber auch nicht die korrupte Datei lschen, 
				 * dann kommt nmlich niemand mehr in die Box, also auch nicht 
				 * der Sysopknecht, der das wieder beheben muss ...
				 * der kommt aber unter Umstnden sowieso nicht 'rein ...
				 */
				str_ulong(tbuf,time(NULL));
				uostr_xdup_cstrmulti(&fname_backup2,fname,".",tbuf,NULL);
				rename (fname_backup.data, fname_backup2.data);
				uostr_freedata(&fname_backup2);
				ret=FALSE;
				goto cleanup;
			}					/* failed fflush, fclose or fputs */
			f=NULL;
		  cleanup:
		  	if (f)
				fclose(f);
			if (f2)
				fclose(f2);
			uostr_freedata(&fname);
			uostr_freedata(&fname_backup);
		      
		}						/* need_rewrite */
	}

	return TRUE;
}								/* ub_write_user */
