/* x_misc.c: misc. functions specially for the extension languages */
/*
    Copyright (C) 1998 Uwe Ohse

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact: uwe@ohse.de, Uwe Ohse @ DU3 (mausnet)
    Contact2: MausNet Newsgroup Quark
*/

#include "tomsysdep.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <syslog.h>
#include <unistd.h>
#include "tomsysdep.h"

#include "prototyp.h"
#include "aktdata.h"
#include "befehle.h"
#include "brett.h"
#include "user.h"
#include "io.h"
#include "modem.h"
#include "terminal.h"
#include "routing.h"
#include "tompaths.h"
#include "box_conf.h"
#include "x_functions.h"
#include "ta_info.h"

extern int logout_modus;
extern BYTE modemloginB;
extern int col_ix[];
extern int user_in_fd;
extern BRETTART postfach_art;
extern int keine_ausgaben;
extern USER user;

#define X_TMPBUFSIZE 1024
static char *x_tmpbuf=NULL;
#define SETUP_TMPBUF do { if (!x_tmpbuf) { x_tmpbuf=XMALLOC(X_TMPBUFSIZE+1); } } while (0)

static const char *got_exception=NULL;
void 
x_clear_exception(void)
{
	got_exception=NULL;
	
}

void 
x_throw(const char *s)
{
	got_exception=s;
}

const char * 
x_got_exception(void)
{
	return got_exception;
}

const char *
tom_getvar (const char *name)
{
	if (!strchr (name, '%') && !strchr (name, '$') && !strchr (name, '\\')) {
		if (!strcmp (name, "logout_modus"))
			return logout_modus ? "0" : "1";

		if (!strcmp (name, "postfach_art")) {
			if (postfach_art == TEXT_BRETT)
				return "PM";
			else
				return "PP";
		} else if (!strcmp (name, "porttype")) {
			if (akt_connection.is_modem)
				return "modem";
			else if (akt_connection.is_console)
				return "console";
			else if (akt_connection.is_tty)
				return "tty";
			else if (akt_connection.peer)
				return "peer";
			else
				return "raw";
		}
	}
	SETUP_TMPBUF;
	if (!do_interpret (name, x_tmpbuf, X_TMPBUFSIZE)) {
		x_throw ("unverstaendliche Variable");
		return NULL;
	}
	return x_tmpbuf;
}

int 
tom_puts(const char *s, const char *s2)
{
	if (0==strcmp(s,"-nonewline")) {
		if (mboutput(s2)<0)
			return 0;
	} else {
		if (s2) {
			x_throw("tom_puts: illegaler 2. Parameter");
			return 0;
		}
		if (mboutput(s)<0)
			return 0;
		if (mboutput("\r\n")<0)
			return 0;
	}
	return 1;
}

int 
tom_putfehler(const char *s)
{
	mbfehler(s);
	return 0;
}

char
tom_getchar (void)
{
	return mbchar();
}

char *
tom_gets(void)
{
	SETUP_TMPBUF;
	mbquestion(X_TMPBUFSIZE,0,FALSE,NULL,NULL,x_tmpbuf);
	return x_tmpbuf;
}

void
tom_befehl_quiet(const char *s)
{
	char *d=XSTRDUP(s);	
	keine_ausgaben++;
	befehl_ausfuehrung(d);
	free(d);
	keine_ausgaben--;
}

void
tom_befehl(const char *s)
{
	char *d=XSTRDUP(s);	
	befehl_ausfuehrung(d);
	free(d);
}

void
tom_befehl_umgelenkt(const char *s, const char *datei)
{
	FILE *f;
	char pfad[PATH_MAX];
	char *d;
	if (!datei)
		datei="/dev/null";
	sprintf(pfad,"%stmp.bat",tom_tmp_dir);
	f=xfopen(pfad,"w");
	fprintf(f,"%s\n",s);
	fclose(f);
	akt_anruf.batch_status=TRISTATE_OK;
	befehls_umlenkung(datei);
	d=XSTRDUP(s);
	befehl_ausfuehrung(d);
	free(d);
	befehls_umlenkung(NULL);
	unlink(pfad);
}

time_t
tom_time(void)
{
	return time(NULL);
}

char *
tom_strftime(const char *format, time_t tim)
{
	struct tm *tm;
	SETUP_TMPBUF;
	tm=localtime(&tim); 
	if (!tm) {
		x_throw("localtime lieferte NULL");
		return NULL;
	}
	strftime(x_tmpbuf,X_TMPBUFSIZE, format, tm);
	return x_tmpbuf;
}

time_t
tom_mtzeit_to_time_t(const char *s)
{
	time_t t;
	if (!mtzeit_to_time_t(s,&t)) {
		x_throw("mtzeit_to_time_t fehlgeschlagen");
		return -1;
	}
	return t;
}

const char *
tom_time_t_to_mtzeit(time_t t)
{
	SETUP_TMPBUF;
	if (!time_t_to_mtzeit(&t,x_tmpbuf)) {
		x_throw("time_t_to_mtzeit fehlgeschlagen");
		return "";
	}
	return x_tmpbuf;
}

const char *
tom_vt(const char *A)
{
	if (!strcmp(A,"off"))
		return vt_S(ATTR_OFF);
	else if (!strcmp(A,"invers"))
		return vt_S(ATTR_INVERS);
	else if (!strcmp(A,"fett"))
		return vt_S(ATTR_FETT);
	else if (!strcmp(A,"del_scr"))
		return vt_S(DEL_SCR);
	else if (!strcmp(A,"del_scr_rest"))
		return vt_S(DEL_SCR_REST);
	else
	{
		int c=-1;
		if (!strcmp(A,"schwarz"))
			c=C_SCHWARZ;
		else if (!strcmp(A,"rot"))
			c=C_ROT;
		else if (!strcmp(A,"gruen") || !strcmp(A,"grn"))
			c=C_GRUEN;
		else if (!strcmp(A,"gelb"))
			c=C_GELB;
		else if (!strcmp(A,"blau"))
			c=C_BLAU;
		else if (!strcmp(A,"pink"))
			c=C_PINK;
		else if (!strcmp(A,"trkis") || !strcmp(A,"tuerkis"))
			c=C_TUERKIS;
		else if (!strcmp(A,"weiss") || !strcmp(A,"wei"))
			c=C_WEISS;
		else {
			x_throw("unbekanntes Terminalattribut");
			return NULL;
		}
		return vt_S(col_ix[c]);
	}
}

int 
tom_charavail(void)
{
	return io_charavail(user_in_fd);
}

/* tom_exec [-user] [-workdir PATH] [-killzeit TIME_T] COMMAND */
int 
tom_exec(const char *p1, const char *p2, const char *p3,
	const char *p4, const char *p5, const char *p6) 
{
	const char *a[7];
	int i;
	const char *wd=NULL;
	time_t killzeit=0;
	BOOLEAN as_user=FALSE;
	a[0]=p1;
	a[1]=p2;
	a[2]=p3;
	a[3]=p4;
	a[4]=p5;
	a[5]=p6;
	a[6]=0;
	for (i=0;i<7;i++) {
		if (!a[i]) {
			x_throw("tom_exec: parameter unverstaendlich");
			return -1;
		}
		if (0==strcmp(a[i],"-user")) 
			as_user=1;
		else if (0==strcmp(a[i],"-workdir")) {
			if (!a[i+1]) {
				x_throw("tom_exec: parameter unverstaendlich");
				return -1;
			}
			wd=a[i+1];
			i++;
		} else if (0==strcmp(a[i],"-killzeit")) {
			if (!a[i+1]) {
				x_throw("tom_exec: parameter unverstaendlich");
				return -1;
			}
			killzeit=strtol(a[i+1],NULL,10);
			i++;
		} else
			break;
	}
	if (!a[i]) {
		x_throw("tom_exec: parameter unverstaendlich");
		return -1;
	}
	if (a[i+1]) {
		x_throw("tom_exec: zu viele Parameter");
		return -1;
	}
	return call_extern(a[i],
		wd && *wd ? wd : NULL,
		killzeit,FALSE,as_user);
}

		/* strings/(U)LONG/UWORD/BOOLEAN setzen bzw. lesen */
#define CmpSetString(arg)	if (!strcmp(var,#arg)) \
								strncpy(up->##arg,val,sizeof(up->##arg)); \
							else
#define CmpSetLong(arg)		if (!strcmp(var,#arg)) \
								up->##arg = atol(val); \
							else
#define CmpSetUword(arg)	if (!strcmp(var,#arg)) \
								up->##arg = (UWORD) strtoul(val,NULL,10); \
							else
#define CmpSetFlag(arg)		if (!strcmp(var,#arg)) \
								up->##arg = (atoi(val)) ? 1 : 0; \
							else
#define CmpSetChar(arg)		if (!strcmp(var,#arg)) \
								up->##arg = *val; \
							else
#define CmpSetMFlag(name,setval) if (!strcmp(var,#name)) { \
								if (strtol(val,NULL,10)) up->flags|=(setval); \
								else up->flags&=~(setval); \
							} else

const char *
tom_user_set(const char *un, const char *var, const char *val)
{
	USER u;
	USER *up=&user;
	USERNR unr;
	if (!ub_find_username(&unr,un) || !unr) {
		x_throw("Benutzer unbekannt");
		return NULL;
	}
	if (unr!=user.nummer) {
		if (!ub_read_user(unr,&u,TRUE)) {
			x_throw("Benutzerdaten unlesbar");
			return NULL;
		}
		up=&u;
	}
	CmpSetString(name)
	CmpSetString(ansage)
	CmpSetString(telefon)
	CmpSetString(plz)
	CmpSetString(ort)
	CmpSetString(strasse)
	CmpSetString(computer)
	CmpSetString(vertreter)
		
	CmpSetLong(restzeit)
	CmpSetLong(anrufe)
	CmpSetLong(geschriebene_mails)
	CmpSetLong(uploads)
	CmpSetLong(ul_bytes)
	CmpSetLong(downloads)
	CmpSetLong(dl_bytes)
	CmpSetLong(txt_bytes)
	CmpSetLong(bezahlt_bis)
	CmpSetLong(passwortdatum)

	CmpSetUword(level)
	CmpSetUword(geburtsjahr)
	CmpSetUword(geburtsmonat)
	CmpSetUword(geburtstag)
	CmpSetUword(mehr)
	CmpSetUword(bs_zeilen)
	CmpSetUword(bs_spalten)
	CmpSetUword(terminal_typ)
	CmpSetUword(computertyp)
	CmpSetUword(menue_stil)
	CmpSetUword(adresse_frei)
	CmpSetUword(auflisten)

	CmpSetFlag(auto_status)
	CmpSetFlag(glocke)
	CmpSetFlag(gast_upload)
	CmpSetFlag(autohilfe)
	CmpSetFlag(autobefehl)
	CmpSetFlag(protokoll)
	CmpSetFlag(pflicht_befreit)
	CmpSetFlag(auto_grp_create)
	CmpSetChar(default_protokoll)

	CmpSetMFlag(flag_auto_resize,UB_FLAG_AUTO_RESIZE)
	CmpSetMFlag(flag_readline,UB_FLAG_READLINE)
	CmpSetMFlag(flag_promi,UB_FLAG_PROMI)

	if (0==strcmp(var,"typ")) {
		if (0==strcmp(val,"user"))
			up->typ=UTYP_USER;
		else if (0==strcmp(val,"box"))
			up->typ=UTYP_BOX;
		else if (0==strcmp(val,"gate"))
			up->typ=UTYP_GATE;
		else {
			x_throw("tom_user set: unbekannter Benutzertyp");
			return NULL;
		}
	} else if (0==strcmp(var,"geschlecht")) {
		if (0==strcmp(val,"neutrum"))
			up->geschlecht=NEUTRUM;
		else if (0==strcmp(val,"mann"))
			up->geschlecht=MAENNLICH;
		else if (0==strcmp(val,"frau"))
			up->geschlecht=WEIBLICH;
		else {
			x_throw("tom_user set: unbekanntes Geschlecht");
			return NULL;
		}
	} else if (0==strcmp(var,"box_stil")) {
		if (0==strcmp(val,"quark"))
			up->box_stil=QUARK_STYLE;
		else if (0==strcmp(val,"maus"))
			up->box_stil=MAUS_STYLE;
		else if (0==strcmp(val,"dialog"))
			up->box_stil=DIALOG_STYLE;
		else {
			x_throw("tom_user set: unbekannter Boxstil");
			return NULL;
		}
	} else if (0==strcmp(var,"uhrmodus")) {
		if (0==strcmp(val,"aus"))
			up->uhrmodus=UHR_AUS;
		else if (0==strcmp(val,"rest"))
			up->uhrmodus=UHR_REST;
		else if (0==strcmp(val,"normal"))
			up->uhrmodus=UHR_NORMAL;
		else if (0==strcmp(val,"zeit"))
			up->uhrmodus=UHR_ZEIT;
		else {
			x_throw("tom_user set: unbekannter Uhrmodus");
			return NULL;
		}
	} else {
		x_throw("tom_user set: unbekannte Variable");
		return NULL;
	}
	/* Zahlerflag anpassen */
	if (0==strcmp(var,"bezahlt_bis")) {
		if (up->bezahlt_bis < time(NULL))
			up->bezahlt=FALSE;
		else
			up->bezahlt=TRUE;
	}

	if (up != &user) {
		if (!ub_write_user(up,TRUE)) {
			x_throw("kann Userdaten nicht sichern");
			return NULL;
		}
	}
	return x_tmpbuf;
}

#define RetString(arg)	if (!strcmp(var,#arg)) \
							strcpy(x_tmpbuf,up->##arg); \
						else
#define RetLong(arg)	if (!strcmp(var,#arg)) \
							sprintf(x_tmpbuf,"%lu",(ULONG) up->##arg); \
						else
#define RetUword(arg)	if (!strcmp(var,#arg)) \
							sprintf(x_tmpbuf,"%u",up->##arg); \
						else
#define RetFlag(arg)	if (!strcmp(var,#arg)) \
							strcpy(x_tmpbuf,(up->##arg) ? "1" : "0"); \
						else
#define RetChar(arg)	if (!strcmp(var,#arg)) {\
							x_tmpbuf[0]=up->##arg; \
							x_tmpbuf[1]=0; } \
						else
#define RetMFlag(name,val) if (!strcmp(var,#name)) { \
							x_tmpbuf[1]=0; \
							if (up->flags & (val)) x_tmpbuf[0]='1'; \
							else x_tmpbuf[0]='0'; \
							} else

const char *
tom_user_get(const char *un, const char *var)
{
	USER u;
	USER *up=&user;
	USERNR unr;
	if (!ub_find_username(&unr,un) || !unr) {
		x_throw("Benutzer unbekannt");
		return NULL;
	}
	if (unr!=user.nummer) {
		if (!ub_read_user(unr,&u,FALSE)) {
			x_throw("Benutzerdaten unlesbar");
			return NULL;
		}
		up=&u;
	}

	RetString(name)
	RetString(passwort)
	RetString(ansage)
	RetString(telefon)
	RetString(plz)
	RetString(ort)
	RetString(strasse)
	RetString(computer)
	RetString(vertreter)
		
	RetLong(restzeit)
	RetLong(anrufe)
	RetLong(geschriebene_mails)
	RetLong(uploads)
	RetLong(ul_bytes)
	RetLong(downloads)
	RetLong(dl_bytes)
	RetLong(txt_bytes)
	RetLong(nummer)
	RetLong(passwortdatum)
	RetLong(bezahlt_bis)

	RetUword(level)
	RetUword(geburtsjahr)
	RetUword(geburtsmonat)
	RetUword(geburtstag)
	RetUword(mehr)
	RetUword(bs_zeilen)
	RetUword(bs_spalten)
	RetUword(terminal_typ)
	RetUword(computertyp)
	RetUword(menue_stil)
	RetUword(adresse_frei)
	RetUword(auflisten)

	RetFlag(bezahlt)
	RetFlag(auto_status)
	RetFlag(glocke)
	RetFlag(gast_upload)
	RetFlag(autohilfe)
	RetFlag(autobefehl)
	RetFlag(protokoll)
	RetFlag(pflicht_befreit)
	RetFlag(auto_grp_create)

		RetMFlag(flag_readline,UB_FLAG_READLINE)
		RetMFlag(flag_auto_resize,UB_FLAG_AUTO_RESIZE)
		RetMFlag(flag_promi,UB_FLAG_PROMI)

	RetChar(default_protokoll)

	/* hier wird's schwieriger ... */
	if (0==strcmp(var,"typ")) {
		if (up->typ==UTYP_USER) {
			return "user";
		} else if (up->typ==UTYP_BOX) {
			return "box";
		} else {
			return "gate";
		}
	} else if (0==strcmp(var,"geschlecht")) {
		if (up->geschlecht==MAENNLICH) {
			return "mann";	
		} else if (up->geschlecht==WEIBLICH) {
			return "frau";	
		} else {
			return "neutrum";	
		}
	} else if (0==strcmp(var,"box_stil")) {
		if (up->box_stil==QUARK_STYLE) {
			return "quark";	
		} else if (up->box_stil==MAUS_STYLE) {
			return "maus";	
		} else {
			return "dialog";	
		}
	} else if (0==strcmp(var,"uhrmodus")) {
		if (up->uhrmodus==UHR_AUS) {
			return "aus";	
		} else if (up->uhrmodus==UHR_REST) {
			return "rest";	
		} else if (up->uhrmodus==UHR_NORMAL) {
			return "normal";	
		} else {
			return "zeit";	
		}
	} else {
		x_throw("tom_user_get: unknown variable");
		return NULL;
	}
	return "";
}
const char *
tom_user_find(const char *name)
{
	USERNR nr;
	if (!ub_find_username(&nr,name)) {
		return NULL;
	} else {
		LISTUSER *lu=ub_find_usernummer(nr);
		if (!lu)
			return NULL;
		return lu->name;
	}
}

/* i was totally stupid when i defined that interface:
	tom_user [-user USERNAME] set VAR VALUE
	tom_user [-user USERNAME] get VAR
	tom_user find NAME
*/
const char * 
tom_user(const char *p1,const char *p2, const char *p3, const char *p4, const char *p5)
{
	const char *un=user.name;
	if (!p1 || !p2) {
		x_throw("tom_user: can't parse arguments");
		return NULL;
	}
	if (!strcmp(p1,"-user")) {
		un=p2;
		if (!p3 || !p4) {
			x_throw("tom_user: can't parse arguments");
			return NULL;
		}
		if (!strcmp(p3,"set")) {
			if (!p5) {
				x_throw("tom_user: can't parse arguments");
				return NULL;
			}
			return tom_user_set(un,p4,p5);
		} else if (!strcmp(p3,"get")) {
			return tom_user_get(un,p4);
		}
		x_throw("tom_user: can't parse arguments");
		return NULL;
	}
	if (!strcmp(p1,"find")) {
		return tom_user_find(p2);
	}
	if (!strcmp(p1,"set")) {
		if (!p3) {
			x_throw("tom_user: can't parse arguments");
			return NULL;
		}
		return tom_user_set(un,p2,p3);
	} 
	if (!strcmp(p1,"get")) {
		return tom_user_get(un,p2);
	}
	x_throw("tom_user: can't parse arguments");
	return NULL;
}
	
/* TOMTCL_FUNCTION_DOKU([[
tom_r_vrfy ADRESSE LEVEL {$|-}
	Prft ob ADRESSE von Benutzer mit Level LEVEL und Zahlerstatus bzw.
	ohne Zahlerstatus angeschrieben werden kann.
]])
TOMTCL_FUNCTION(tom_r_vrfy,ttcl_r_vrfy)
*/
int
tom_r_vrfy(const char *adr, int level, char zahler)
{
	char *comment = NULL;
	if (!r_vrfy (adr, &comment, (UWORD) level, zahler=='$'))
	{
		free (comment);
		return 0;
	}
	else
	{
		free (comment);
		return 1;
	}
}

/* TOMTCL_FUNCTION_DOKU([[
tom_userliste
    Liefert eine Liste aller Benutzer zurck.
]])
TOMTCL_FUNCTION(tom_userliste,ttcl_userliste)
*/
const char *
tom_walkuser(int first)
{
	static LISTUSER *akt=NULL,*t;
	if (first)
		akt=ub_get_first();
	if (!akt) {
		return NULL;
	}
	t=akt;
	akt=akt->next;
	return t->name;
}

BOOLEAN
tom_edit(const char *para1, const char *para2)
{
	if (para2) {
		if (0!=strcmp(para1,"-backup")) {
			x_throw("illegale Parameter");
			return 0;
		}
		return !call_editor(para2,TRUE);
	}
	return !call_editor(para1,FALSE);
}

int
tom_select(char **argv) 
{
	int zahl;
	int c;
	int argc;
	int i;
	struct mbselect_t *sel;
	for (argc=0;argv[argc];argc++)
		;;
	if (argc&1) {
		x_throw("illegale Anzahl Parameterfuer tom_select");
		return 0;
	}
	zahl=argc/2;
	sel=malloc(sizeof(struct mbselect_t) *(zahl+1));
	if (!sel)
	{
		x_throw("out of memory in tom_select");
		return 0;
	}
    for (i=0;i<zahl;i++)
    {
        sel[i].key=*argv[i*2];
        if (!argv[i*2+1] || !*argv[i*2+1])
            sel[i].output=NULL;
        else
            sel[i].output=argv[i*2+1];
    }
    sel[zahl].key=0;

    c=mbselect(sel);

    free(sel);
	return c;
}

char 
tom_transfer_protokollauswahl(const char *p1, const char *p2)
{
	BOOLEAN all=FALSE;
	BOOLEAN space=FALSE;
	if (p1) {
		if (!strcmp(p1,"-all"))
			all=TRUE;
		else if (!strcmp(p1,"-space"))
			space=TRUE;
		else {
			x_throw("illegaler Parameter");
			return 0;
		}
	}
	if (p2) {
		if (!strcmp(p2,"-all"))
			all=TRUE;
		else if (!strcmp(p2,"-space"))
			space=TRUE;
		else {
			x_throw("illegaler Parameter");
			return 0;
		}
	}
	return transfer_protokollauswahl(all,space);
}

char
tom_transfer_receive(char c)
{
	char d=c;
	if (0==transfer_receive(&d,NULL,NULL))
		return d;
	return 0;
}

int 
tom_transfer_send (char art, const char *name)
{
	char d=art;
	return transfer_send(&d,name);
}

BOOLEAN
tom_display_file(const char *p1, const char *p2)
{
	if (p1 && !strcmp(p1,"-interpret")) {
		if (!p2) {
			x_throw("illegaler Parameter");
			return FALSE;
		}
		return zeige_interpret_datei(p2);
	} 
	if (p2) {
		x_throw("illegaler Parameter");
		return FALSE;
	}
	return zeige_datei(p1);
}

int 
tom_exec_piped(const char *fname)
{
	return !call_piped(fname,FALSE);
}

BOOLEAN 
tom_acl(const char *p1,const char *p2,const char *p3,
	const char *p4,const char *p5,const char *p6,const char *p7)
{
	const char *a[9];
	int i;
	int level=0;
	char type='R';
	const char *username=NULL;
	const char *err=NULL;

	a[0]=p1;
	a[1]=p2;
	a[2]=p3;
	a[3]=p4;
	a[4]=p5;
	a[5]=p6;
	a[6]=p7;
	a[7]=NULL;
	a[8]=NULL; /* einer zu viel? siehe invalidate */
	for (i=0;a[i];i++) {
		if (0==strcmp(a[i],"-user")) {
			if (!a[i+1]) {x_throw("illegale Parameter"); return 0; }
			username=a[i+1];
			i++;
		} else if (0==strcmp(a[i],"-level")) {
			if (!a[i+1]) {x_throw("illegale Parameter"); return 0; }
			level=strtol(a[i+1],NULL,10);
			i++;
		} else if (0==strcmp(a[i],"-type")) {
			if (!a[i+1]) {x_throw("illegale Parameter"); return 0; }
			type=*(a[i+1]);
			i++;
		} else if (0==strcmp(a[i],"-invalidate")) {
			if (!a[i+1]||a[i+2]) {x_throw("illegale Parameter"); return 0; }
			acl_invalidate_cachefile(a[i+1]);
			return TRUE;
		} else
			break;
	}
	if (!a[i] || a[i+1]) {x_throw("illegale Parameter"); return 0; }
	return acl_check(a[i], username, level, type, &err);
}

int
tom_info_level(const char *name)
{
	infofile_t *found;
	found=ta_find_infofile(name);
	if (!found) { x_throw("unbekanntes Infofile"); return 9; }
	return (int) found->level;
}

const char *
tom_info_text(const char *name)
{
	infofile_t *found;
	found=ta_find_infofile(name);
	if (!found) { x_throw("unbekanntes Infofile"); return NULL; }
	return found->text;
}
char
tom_info_key(const char *name, const char *submenu)
{
	infofile_t *found;
	found=ta_find_infofile(name);
	if (!found) { x_throw("unbekanntes Infofile"); return 0; }
	if (0!=strcasecmp(found->group,submenu)) {
		x_throw("Infofile nicht im Menue"); 
		return 0; 
	}
	return found->zeichen;
}

const char *
tom_info_menu(const char *name)
{
	infofile_t *found;
	found=ta_find_infofile(name);
	if (!found) { x_throw("unbekanntes Infofile"); return NULL; }
	return found->group;
}

int
tom_infomenu_level(const char *name)
{
	infofilemenu_t *akt;
	for (akt=ta_get_infofilemenu_anker();akt;akt=akt->next)
	{
		if (0==strcasecmp(akt->kuerzel,name))
			return akt->minlevel;
	}
	x_throw("unbekanntes Men");
	return 9;
}

const char *
tom_infomenu_title(const char *name)
{
	infofilemenu_t *akt;
	for (akt=ta_get_infofilemenu_anker();akt;akt=akt->next)
	{
		if (0==strcasecmp(akt->kuerzel,name))
			return akt->beschreibung;
	}
	x_throw("unbekanntes Men");
	return NULL;
}
const char *
tom_infomenu_key(const char *name, const char *submenue)
{
	infofilemenu_t *akt;
	for (akt=ta_get_infofilemenu_anker();akt;akt=akt->next)
	{
		if (0==strcasecmp(akt->kuerzel,name)) {
			simplelist_t *sim;
			for (sim=akt->submenues;sim;sim=sim->next)
			{
				if (0==strcasecmp(submenue,sim->data))
				{
					SETUP_TMPBUF;
					x_tmpbuf[1]=0;
					if (sim->xstring)
						x_tmpbuf[0]=*sim->xstring;
					else
						x_tmpbuf[0]=0;
					return x_tmpbuf;
				}
			}
			x_throw("unbekanntes Submen");
			return NULL;
		}
	}
	x_throw("unbekanntes Men");
	return NULL;
}
const char *
tom_infomenu_files(const char *name)
{
	char *p;
	infofile_t *i=ta_get_info_anker();
	SETUP_TMPBUF;
	p=x_tmpbuf;

	for (; i; i=i->next)
	{
		if (i->group && 0==strcasecmp(i->group,name)) {
			size_t l;
			l=strlen(i->name);
			if (l+2+(p-x_tmpbuf) >= X_TMPBUFSIZE) {
				x_throw("string too long");
				return NULL;
			}
			if (p!=x_tmpbuf)
				*p++=' ';
			memcpy(p,i->name,l);
			p+=l;
		}
	}
	*p=0;
	return x_tmpbuf;
}

const char *
tom_infomenu_submenus(const char *name)
{
	infofilemenu_t *akt;
	simplelist_t *sim;
	char *p;
	SETUP_TMPBUF;
	p=x_tmpbuf;
	for (akt=ta_get_infofilemenu_anker();akt;akt=akt->next)
	{
		if (0==strcasecmp(akt->kuerzel,name))
			break;
	}
	if (!akt) {
			x_throw("unbekanntes Infofile");
			return NULL;
	}
	for (sim=akt->submenues;sim;sim=sim->next)
	{
		size_t l;
		l=strlen(sim->data);
		if (l+2+(p-x_tmpbuf) >= X_TMPBUFSIZE) {
			x_throw("string too long");
			return NULL;
		}
		if (p!=x_tmpbuf)
			*p++=' ';
		memcpy(p,sim->data,l);
		p+=l;
	}
	*p=0;
	return x_tmpbuf;
}

BOOLEAN
tom_einfachnachricht(BOOLEAN mail,int level,  BOOLEAN bezahlt, const char *fname,
	char **argv) 
{
	int i;
	int ret;
	int argc;
	struct veinfachpara *vpara=NULL;

	for (argc=0;argv[argc];argc++)
		;;
	if (argc&1) {
		x_throw("illegale Anzahl Parameter fuer tom_einfachnachricht");
		return FALSE;
	}
	vpara=XMALLOC((argc/2 + 1)*sizeof(struct veinfachpara));
	i=0;
	while (argc)
	{
		vpara[i].kenn=*argv[0];
		vpara[i].str=argv[1];
		i++;
		argc-=2;
		argv+=2;
	}
	vpara[i].kenn=0;
	vpara[i].str=NULL;
	ret=veinfachnachricht(mail, level, bezahlt, fname, vpara);
	free(vpara);
	return ret;
}
	

