/* Erweiterungssprachenschnittstelle fr TOM -- hi emacs, dies ist -*- C -*- */

#include "tomsysdep.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <syslog.h>
#include <unistd.h>
#include <ctype.h>


#include "tomsysdep.h"

#include "prototyp.h"
#include "aktdata.h"
#include "befehle.h"
#include "brett.h"
#include "user.h"
#include "io.h"
#include "modem.h"
#include "terminal.h"
#include "tompaths.h"
#include "tomtcl.h"
#include "box_conf.h"
#include "routing.h"

extern int logout_modus;

static BOOLEAN init_done;

Tcl_Interp *ttcl_interp;
static Tcl_Interp *ttcl_hook_interp;
static Tcl_Interp *ttcl_init_interpreter(void);

#define EXITCHECK() \
do { \
	if (logout_modus) \
	{ \
		Tcl_SetResult(interp,"logout_modus set",TCL_STATIC); \
		return TCL_ERROR; \
	} \
} while (0)

extern BOOLEAN ttcl_perl_script(char *name, char **result_string);



int 
ttcl_hook(const char *fn, const char *para, char **newpara)
{
	static BOOLEAN in_hook=FALSE;
	char *buf;
	char *buf2;
	int errorflag;
	static BOOLEAN hook_init_done=FALSE;
	char datei[PATH_MAX];
	char *fn_lower;

	/* Endlosrekursionen vermeiden ... */
	if (in_hook)
		return TTCL_HRES_NOTDONE;

	fn_lower=entumlauten1(fn);
	XNULLCHECK(fn_lower);
	strlwr(fn_lower);

	sprintf(datei,"%shook_%s.tcl",box_conf.tcl_dir,fn_lower);
	if (!fileexist(datei))
	{
		free(fn_lower);
		return TTCL_HRES_NOTDONE;
	}

	if (!hook_init_done)
	{
		hook_init_done=TRUE;
		ttcl_hook_interp=ttcl_init_interpreter();
		if (!ttcl_hook_interp)
		{
			free(fn_lower);
			syslog(LOG_ERR,"Kann TCL Interpreter nicht anlegen");
			return TTCL_HRES_NOTDONE;
		}
	}
	if (!ttcl_hook_interp)
	{
		free(fn_lower);
		return TTCL_HRES_NOTDONE;
	}

	buf=XMALLOC(PATH_MAX
		+ ( fn ? strlen(fn) : 0) 
		+ ( para ? strlen(para) : 0)
		+ 128);
	buf2=XMALLOC(fn ? strlen(fn) : 1);
	if (fn)
	{
		strcpy(buf2,fn);
		strupr(buf2);
	}
	else
		*buf2=0;
	
	sprintf(buf,"source %s\nreturn [hook_%s {%s} {%s}]\n",
		datei, fn_lower, buf2, para ? para : "");
	free(fn_lower);
	free(buf2);

	in_hook=TRUE;
	errorflag=Tcl_Eval(ttcl_hook_interp, buf);
	in_hook=FALSE;

	free(buf);
	if (errorflag!=TCL_RETURN && errorflag!=TCL_OK)
	{
		char *trace;
		mbprintf("%s: %s\r\n",buf,ttcl_hook_interp->result);
		trace=Tcl_GetVar(ttcl_hook_interp,"errorInfo",TCL_GLOBAL_ONLY);
		if (trace && !logout_modus)
		{
			mbprintf("*** TCL TRACE ***\r\n");
			do {
				char *p=strchr(trace,'\n');
				if (p)
					*p='\0';
				mbprintf("%s\r\n",trace);
				if (p)
					trace=p+1;
				else
					trace=NULL;
			} while (trace!=NULL);
		}
		return TTCL_HRES_ERROR;
	}
	*newpara=NULL;
	if (0==strcmp(ttcl_hook_interp->result,"done"))
		return TTCL_HRES_DONE;
	if (0==strncmp(ttcl_hook_interp->result,"modified",8))
	{
		char *p=strpbrk(ttcl_hook_interp->result,"\t ");
		if (p)
		{
			while (isspace(*p))
				p++;
			*newpara=XSTRDUP(p);
		}
		else
			*newpara=XSTRDUP("");

		return TTCL_HRES_MODIFIED;
	}
	else
		return TTCL_HRES_NOTDONE;
}

int 
ttcl_call_func(const char *filename, const char *func, const char *para,
	char **result_string)
{
	char *buf;
	int errorflag;
	const char *dot;
	dot=strrchr(filename,'.');
	if (dot && 0==strcmp(dot,".pl"))
		return tperl_call_func(filename, func,para,result_string);

	if (!init_done)
	{
		ttcl_interp=ttcl_init_interpreter();
		init_done=TRUE;
	}
	if (!ttcl_interp)
	{
		mbprintf("Kein TCL Interpreter vorhanden!\r\n");
		return FALSE;
	}

	if (!fileexist(filename))
		return -1;

	buf=XMALLOC(PATH_MAX
		+ ( func ? strlen(func) : 0) 
		+ ( para ? strlen(para) : 0)
		+ 128);

	if (para)
		sprintf(buf,"source %s\nreturn [%s \"%s\"]\n",
				filename, func, para);
	else
		sprintf(buf,"source %s\nreturn [%s]\n",
				filename, func);

	errorflag=Tcl_Eval(ttcl_interp, buf);
	free(buf);

	if (errorflag!=TCL_RETURN && errorflag!=TCL_OK)
	{
		char *trace;
		mbprintf("%s: %s\r\n",buf,ttcl_interp->result);
		trace=Tcl_GetVar(ttcl_interp,"errorInfo",TCL_GLOBAL_ONLY);
		if (trace && !logout_modus)
		{
			mbprintf("*** TCL TRACE ***\r\n");
			do {
				char *p=strchr(trace,'\n');
				if (p)
					*p='\0';
				mbprintf("%s\r\n",trace);
				if (p)
					trace=p+1;
				else
					trace=NULL;
			} while (trace!=NULL);
		}
		return -1;
	}
	if (result_string) 
		*result_string=ttcl_interp->result;
	if (0==strcmp(ttcl_interp->result,"ok"))
		return 0;
	return 1;
}

extern void boot_tom();

BOOLEAN 
ttcl_script(char *name, char **result_string)
{
	int errorflag;
	const char *dot;
	dot=strrchr(name,'.');
	if (dot && 0==strcmp(dot,".pl"))
		return ttcl_perl_script(name, result_string);
	if (!init_done)
	{
		ttcl_interp=ttcl_init_interpreter();
		init_done=TRUE;
	}
	if (!ttcl_interp)
	{
		mbprintf("Kein TCL Interpreter vorhanden!\r\n");
		return FALSE;
	}
	errorflag=Tcl_EvalFile(ttcl_interp, name);
	if (errorflag!=TCL_OK)
	{
		char *trace;
		mbprintf("%s: %s\r\n",name,ttcl_interp->result);
		trace=Tcl_GetVar(ttcl_interp,"errorInfo",TCL_GLOBAL_ONLY);
		if (trace && !logout_modus)
		{
			mbprintf("*** TCL TRACE ***\r\n");
			do {
				char *p=strchr(trace,'\n');
				if (p)
					*p='\0';
				mbprintf("%s\r\n",trace);
				if (p)
					trace=p+1;
				else
					trace=NULL;
			} while (trace!=NULL);
		}
		return FALSE;
	}
	if (!result_string && *ttcl_interp->result)
		mbprintf("%s\r\n",ttcl_interp->result);
	if (result_string) 
		*result_string=ttcl_interp->result;
	return TRUE;
}

static Tcl_Interp *
ttcl_init_interpreter(void)
{
	Tcl_Interp *interp;

	interp=Tcl_CreateInterp();
	if (!interp)
	{
		syslog(LOG_ERR,"Kann TCL Interpreter nicht anlegen");
		return NULL;
	}
	Tcl_Init(interp);
	Tom_Init(interp);
	Tcl_LinkVar(interp,"tomtcl_dir",(char *) &box_conf.tcl_dir,
		TCL_LINK_STRING | TCL_LINK_READ_ONLY);
	Tcl_LinkVar(interp,"tomconfig_dir",(char *) &box_conf.config_dir,
		TCL_LINK_STRING | TCL_LINK_READ_ONLY);
	Tcl_LinkVar(interp,"logout_modus",(char *) &logout_modus,
		TCL_LINK_INT);


	return interp;
}

